/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/***************************************************************
 * C++ source
 *
 * File : 	MultiPCA.cc
 *
 * Module :	MultiPCA
 *
 * Author : 	A M Baumberg (CoMIR)
 *
 * Creation Date : Wed Jan  3 11:23:02 1996 
 *
 * Comments : 	
 *
 ***************************************************************/


/* include files */

#include <PCAclass.h>
#include <MultiPCA.h>
#include <stdlib.h>
#include <math.h>
#include <stream.h>
#include <string.h>

static int the_data_size = 0;
static int the_no_clusters = 1;
//static const realno TOO_SMALL = 1e-5;

Cluster::Cluster() 
: Covariance(the_data_size, the_data_size, identity_fn),
  Inverse_Covariance(the_data_size,the_data_size,identity_fn),
  mean(the_data_size)
{  
  determinant = 1.0; 
  alpha_wght = 1.0 / the_no_clusters;
}

PCAclass* Cluster::create_PCAclass()      
{
  int& n = Covariance.no_rows();
  PCAclass* res = new PCAclass(n);
  Covariance.copy(res->covariance);
  mean.copy(res->mean);
  res->get_evals();
  return (res);
}

PCAclass* Cluster::create_PCAclass(NagMatrix& Q, NagVector& offset)      
{
  int& n = Q.no_rows();
  PCAclass* res = new PCAclass(n);
  Covariance.ortho_transform(Q, res->covariance, FALSE);
  Q.multiply(mean, res->mean);
  res->mean.add(offset, res->mean);
  res->get_evals();
  return (res);
}
  
MultiPCA::MultiPCA(int datasize, int nclusters) 
{
  epoch_size = 0;
  data_size = datasize;
  no_clusters = nclusters;
  small_variance = pow(MINDOUBLE, 1.0 /((realno)datasize));

  // assign static copies for Cluster constructor to
  // access (see above)
  the_data_size = data_size;
  the_no_clusters = no_clusters;

  cluster_set = new Cluster[no_clusters];
  scalar_constant = pow(2.0 * M_PI, -0.5 *((realno)no_clusters));
}

void MultiPCA::initialise(realno min, realno max, realno min_var,
			  realno max_var)
{
  if (min_var == 0)
    {
      realno sigma = (max - min) / no_clusters;
      min_var = sigma*sigma;
      max_var = 2 * min_var;
    }
  if (min_var < small_variance) min_var = small_variance;
  if (max_var < small_variance) max_var = small_variance;
  for (int i = 0; i < no_clusters; i++)
    {
      cluster_set[i].mean.set_unif_random(min,max);
      realno var = min_var + ((max_var - min_var) * drand48());
      cluster_set[i].Covariance.scale(var, cluster_set[i].Covariance);
      cluster_set[i].Inverse_Covariance.scale(1.0 / var,
		       cluster_set[i].Inverse_Covariance);
      cluster_set[i].determinant = pow(var, data_size);
    }
}
	
void MultiPCA::initialise(NagMatrix& training_data)
{
  epoch_size = training_data.no_columns();
  if (no_clusters > epoch_size)
    {
      fprintf(stderr, " Too many clusters ! in MultiPCA::initialise\n");
	fflush(stderr);
      abort();
    }
  if (training_data.no_rows() != data_size)
    {
      fprintf(stderr, "bad call to MultiPCA::initialise \n");
	fflush(stderr);
      abort();
    }
  
  for (int i = 0; i < data_size; i++)
    {
      realno max_i = -1e12; realno min_i = 1e12;
      realno mean_i = 0; realno sq_sum_i = 0;
      for (int t = 0; t < epoch_size; t++)
	{
	  realno dat = training_data.read(i,t);
	  if (dat > max_i) max_i = dat;
	  if (dat < min_i) min_i = dat;
	  mean_i += dat / ((realno) epoch_size);
	  sq_sum_i += (dat * dat)  / ((realno) epoch_size);
	}
      realno var_i = sq_sum_i - (mean_i * mean_i);
      realno sig_i = sqrt(var_i);
      realno max_var = 2.0 * var_i;
      if (max_var < small_variance) max_var = small_variance;
      realno min_var = var_i / 100.0;
      if (min_var < small_variance) min_var = small_variance;
      fprintf(stderr, "sig_%d %f, max_var %f, min_var %f \n", i, sig_i, max_var, min_var);
      for (int lp =0; lp < no_clusters; lp++)
	{
	  if (i == 0) cluster_set[lp].determinant = 1.0;
	  realno rnd = min_var + ((max_var - min_var) * drand48());
	  //realno rnd = min_var;
	  cluster_set[lp].determinant *= rnd;
	  cluster_set[lp].Covariance.set(i,i,rnd);
	  cluster_set[lp].Inverse_Covariance.set(i,i,1.0/rnd);
	  cluster_set[lp].mean[i] = min_i +
	    ((max_i - min_i) * drand48());
	}
    }
//   int* flags = new int[epoch_size];
//   for (int lp = 0; lp < epoch_size; lp++)
//     flags[lp] = FALSE;

//   for (int i = 0; i < no_clusters; i++)
//     {
//       int j = rand() % epoch_size;
//       while (flags[j]) j = rand() % epoch_size;
//       (training_data.column(j)).copy(cluster_set[i].mean);
//       flags[j] = TRUE;	 // don't pick this
// 			 // point again
//     }

//   delete [] flags;
}


realno MultiPCA::prob_density(NagVector& x, int cluster)
{
  NagVector diff(x - cluster_set[cluster].mean);

  NagVector tmp_v;
  cluster_set[cluster].Inverse_Covariance.multiply(diff, tmp_v);
  realno res = exp(-0.5 * diff.dot(tmp_v));
  res *= (scalar_constant) / sqrt(cluster_set[cluster].determinant);
  //if (res < TOO_SMALL) res = TOO_SMALL;
  return (res);
}


static void get_inverse_determinant(NagMatrix& M, NagMatrix& Inv_M,
				    realno& det_m,
				    realno small_variance)
{
  NagMatrix Q;
  NagVector evals;
  NagMatrix Omega(M.no_rows(), M.no_rows(), identity_fn);
  Omega.scale(small_variance, Omega);

  // Add scalar multiple of identity to ensure 
  // invertibility of M
  
  M.add(Omega, M);
  M.invert(Inv_M);
  M.get_eigenvectors(Q,evals);
  det_m = 1.0;
  for (int i = 0; i < evals.get_size(); i++)
    det_m *= evals[i];
  return;

  NagMatrix M_prime;
  M.add(Omega, M_prime);
  M_prime.get_eigenvectors(Q, evals);

  det_m = 1.0;
  NagMatrix Q_t;
  Q.transpose(Q_t);

  for (i = 0; i < evals.get_size(); i++)
    {
      evals[i] -= small_variance;	// remove effect of Omega
      if (evals[i] < small_variance) evals[i] = small_variance;
      det_m *= evals[i];
      NagVector tmp_v(Q.no_rows(), Q.get(0, i));
      tmp_v.scale(1.0 / evals[i], tmp_v);
    }

  Q.multiply(Q_t, Inv_M);
}

int MultiPCA::classify(NagVector& x_t)
{
  // find the 'closest' cluster to x_t
  realno min_dist = 0;
  int nearest = 0;
  for (int i = 0; i < no_clusters; i++)
    {
      NagVector diff(x_t - cluster_set[i].mean);
      NagVector tmp_v;
      cluster_set[i].Inverse_Covariance.multiply(diff, tmp_v);
      realno d = diff.dot(tmp_v);
      d -= 2 * log(cluster_set[i].alpha_wght);

      if (i == 0) min_dist = d;
      if (d < min_dist)
	{
	  min_dist = d;
	  nearest = i;
	}
    }
  return (nearest);
}

static void scale_clusters(Cluster* cset, int no_clusters,
			   realno factor)
{
  for (int i = 0; i < no_clusters; i++)
    {
      cset[i].Covariance.scale(factor, cset[i].Covariance);
      cset[i].Inverse_Covariance.scale(1.0 / factor,
				       cset[i].Inverse_Covariance);
      cset[i].determinant *= factor;
    }
}
  

				       
  
realno MultiPCA::update(NagMatrix& training_data)
{
  epoch_size = training_data.no_columns();
  NagMatrix h_parms(no_clusters, epoch_size);

  int j,t;
  realno L = 0;
  // Expectation Step
  for (t = 0; t < epoch_size; t++)
    {
      NagVector x_t(training_data.column(t));
      NagVector h_col(h_parms.column(t));
      for (j = 0; j < no_clusters; j++)
	{
	  if (cluster_set[j].alpha_wght > 0)
	    h_col[j] = cluster_set[j].alpha_wght * prob_density(x_t,j);
	  else h_col[j] = 0;
	}
      realno prob_xt = h_col.sum();
      if (prob_xt <= MINDOUBLE)
	{
	  fprintf(stderr, "Can't classify data item! \n");
	  // abort();
	  fprintf(stderr, "reinitialising ... \n");
	  scale_clusters(cluster_set, no_clusters, 2.0);
	  return (0.0);
	  //h_col.clear(1.0 / no_clusters);	// add to all clusters!
	}
      else
	{
	  h_col.scale(1.0 / prob_xt, h_col);
	  L += log(prob_xt);
	}
    }

  fprintf(stderr," Log Likelihood = %f \n", L);
  // Maximization Step

  for (j = 0; j < no_clusters; j++)
    {
      NagVector tmp_row;
      h_parms.get_row(j, tmp_row);
      realno tmp_sum = tmp_row.sum();
      if (tmp_sum <= MINDOUBLE)
	{
	  fprintf(stderr,"Cluster has no data !\n");
	  fprintf(stderr,"reinitialising \n");;
	  scale_clusters(&cluster_set[j], 1, 4.0);
	  return (0.0);
	}
    }

  realno change_in_means = 0;
  for (j = 0; j < no_clusters; j++)
    {
      if (cluster_set[j].alpha_wght == 0) continue;
      realno scalar_sum = 0;
      NagVector vec_sum(data_size);
      vec_sum.clear();
      NagMatrix mat_sum(data_size, data_size);
      mat_sum.clear();
      NagVector& mean_j = cluster_set[j].mean;
      NagVector tmp_vec;
      NagMatrix tmp_m;
      for (t = 0; t < epoch_size; t++)
	{
	  realno& h_j_t = *(h_parms.get(j,t));
	  scalar_sum += h_j_t;
//	  fprintf(stderr, "h_%d_%d = %f", j, t, h_j_t);
	  NagVector x_t(training_data.column(t));
	  x_t.scale(h_j_t, tmp_vec);
	  vec_sum.add(tmp_vec, vec_sum);

	  NagVector diff(x_t - mean_j);
	  diff.mult_trans(diff, tmp_m);
	  tmp_m.scale(h_j_t, tmp_m);
	  mat_sum.add(tmp_m, mat_sum);

	}

      if (scalar_sum <= MINDOUBLE)
	{
	  fprintf(stderr,"Cluster has no data! in MultiPCA::update\n");
	fflush(stderr);
	  abort();
	}
	  
      else 
	{
	  vec_sum.scale(1.0 / scalar_sum, tmp_vec);
	  NagVector diff(tmp_vec - mean_j);
	  change_in_means += diff.length2();
	  tmp_vec.copy(mean_j);
	  mat_sum.scale(1.0 / scalar_sum, cluster_set[j].Covariance);
	  cluster_set[j].alpha_wght = scalar_sum / ((realno) epoch_size);
	  get_inverse_determinant(cluster_set[j].Covariance, 
				  cluster_set[j].Inverse_Covariance, 
				  cluster_set[j].determinant,
				  small_variance);
	}
    }
//  return (change_in_means);
  return (L);

}

static int compare_clusters(const void* c1,
			    const void* c2)
{
  Cluster* cluster1 = (Cluster*) c1;
  Cluster* cluster2 = (Cluster*) c2;

  if (cluster1->alpha_wght <
      cluster2->alpha_wght) return (1);
  if (cluster1->alpha_wght >
      cluster2->alpha_wght) return (-1);
  return (0);
}

  
void MultiPCA::sort_models()
{
  
  qsort((void*) cluster_set, no_clusters,
	sizeof(Cluster), compare_clusters);
}

void MultiPCA::filter_models(realno min_items)
{
  sort_models();
  realno thresh = min_items / epoch_size;
  for (int i = 0; i < no_clusters; i++)
    if (cluster_set[i].alpha_wght < thresh) break;
  no_clusters = i;
}

void MultiPCA::build_distance_table(NagMatrix& training_data)
{
  epoch_size = training_data.no_columns();
  NagMatrix p_j_t(no_clusters, epoch_size);

  distance_matrix.reconstruct(no_clusters, no_clusters);
  
  int i,j,t, t1,t2;
  realno L = 0;

  NagVector sum_p_j(no_clusters);
  sum_p_j.clear();

  int* classification = new int[epoch_size];
  for (t = 0; t < epoch_size; t++)
    {
      NagVector x_t(training_data.column(t));
      for (j = 0; j < no_clusters; j++)
	{
	  p_j_t.set(j,t,prob_density(x_t,j));
	  sum_p_j[j] += p_j_t.read(j,t);
	}
      classification[t] = classify(x_t);

    }

  distance_matrix.clear(1e30);

  for (t1 = 0; t1 < epoch_size; t1++)
    for (t2 = 0; t2 < epoch_size; t2++)
      {
	NagVector x_t1(training_data.column(t1));
	NagVector x_t2(training_data.column(t2));
	int i1 = classification[t1];
	int i2 = classification[t2];
	realno d12 = (x_t1 - x_t2).length();
	if (d12 < distance_matrix.read(i1,i2))
	  distance_matrix.set(i1,i2, d12);
      }
  distance_matrix.output();
  return;
	
	
  NagMatrix point_distances(epoch_size, epoch_size);
  realno epoch2 = (1.0 / ((realno) (epoch_size * epoch_size)));
  for (t1 = 0; t1 < epoch_size; t1++)
    for (t2 = 0; t2 < epoch_size; t2++)
      *point_distances.get(t1,t2) = 
	(training_data.column(t1) -
	 training_data.column(t2)).length2();
  
//  point_distances.output();
//  NagMatrix joint_prob;
//  p_j_t.multiply(trans, joint_prob);
//  joint_prob.scale(1.0 / ((realno) epoch_size), joint_prob);

  point_distances.ortho_transform(p_j_t, distance_matrix, FALSE);
  
  for (i = 0; i < no_clusters; i++)
    for (j = 0; j < no_clusters; j++)
      *distance_matrix.get(i,j) /= (sum_p_j[i] * sum_p_j[j]);

  distance_matrix.output();
}

void MultiPCA::dump_pca_models(char* base_file_name)
{
  char* tmp_str = new
    char[strlen(base_file_name) + 20];
  sprintf(tmp_str,"%s.info",base_file_name);
  ofstream mpca_info(tmp_str);
  mpca_info << "Multiple PCA info" << endl;
  mpca_info << no_clusters << " clusters" << endl;
  mpca_info << "distance matrix" << endl << distance_matrix;
  mpca_info << "Sub Models: " << endl;
  for (int i = 0; i < no_clusters; i++)
    {
      PCAclass* pca_model =
	cluster_set[i].create_PCAclass();

      sprintf(tmp_str,"%s%d",base_file_name,i);
      mpca_info << "Cluster " << i << endl;
      mpca_info << "file: " << tmp_str << endl;
      mpca_info << "alpha weight " <<
	cluster_set[i].alpha_wght << endl;

      ofstream file_out(tmp_str);
      file_out << *pca_model << endl;
      delete (pca_model);
    }
  delete [] tmp_str;
}


void MultiPCA::dump_pca_models(char* base_file_name, NagMatrix& Q,
			       NagVector& offset)
{
  char* tmp_str = new
    char[strlen(base_file_name) + 20];
  sprintf(tmp_str,"%s.info",base_file_name);
  ofstream mpca_info(tmp_str);
  mpca_info << "Multiple PCA info" << endl;
  mpca_info << no_clusters << " clusters" << endl;
  mpca_info << "distance_matrix" << endl << distance_matrix;
  mpca_info << "Sub_Models " << endl;
  for (int i = 0; i < no_clusters; i++)
    {
      PCAclass* pca_model =
	cluster_set[i].create_PCAclass(Q, offset);

      sprintf(tmp_str,"%s%d",base_file_name,i);
      mpca_info << "Cluster " << i << endl;
      mpca_info << "file: " << tmp_str << endl;
      mpca_info << "alpha_weight " <<
	cluster_set[i].alpha_wght << endl;

      ofstream file_out(tmp_str);
      file_out << *pca_model << endl;
      delete (pca_model);
    }
  delete [] tmp_str;
}


