/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/*
 * PCAclass
 * 
 * a class for carrying out Principal Component Analysis on
 * arbitrary data
 *
 */

#include <PCAclass.h>
#include <H_metric.h>
#include <math.h>
#include <string.h>
#include <AmbErrorHandler.h>

realno MIN_COV_ENTRY = 0.0;

PCAclass::PCAclass(char* fname, int max)
{
  ifstream data_in(fname);
  if (!data_in) 
    {
      sprintf(AmbErrorHandler::buffer,"Couldn't open %s", fname);
      AmbErrorHandler::handle_error("in PCAclass:PCAclass\n");
	  fprintf(stderr,"abort in PCAclass::PCAclass\n");
	fflush(stderr);
      abort();	// unrecoverable error so abort
    }

  data_in >> *this;
  max_n = max;
}

PCAclass::PCAclass(int size, int max):
data_size(size),
covariance(size,size,0.0),
change_covariance(size, size, 0.0),
mean(size),
noise_model(),
eigenvalues(),
eigenvectors(),
weights()
{
  no_processed = 0;
  max_n = max;
  mean.clear();
}

void PCAclass::process_cov(NagVector& new_data)
{
  //  new version calculates cov(X Xt) = E(X Xt) - E(X)E(Xt)
  if (new_data.get_size() < data_size)
    {
      fprintf(stderr," bad call to PCAclass::process_cov \n");
	fflush(stderr);
      abort();
    }

  no_processed++;
  if (no_processed > max_n) no_processed = max_n;

  // update E(X)
  NagVector diff(mean - new_data);
  mean -= (diff / no_processed);

  // update E(X Xt)
  realno tmp;
  for (int i = 0; i < data_size; i++)
    for (int j = i; j < data_size; j++)
      {
	tmp = *cov(i,j) - new_data[i] * new_data[j];
	*cov(i,j) -= (tmp / no_processed);
      }
}

void PCAclass::process_cov(NagVector& new_data, NagVector& old_data)
{
  NagVector deriv;
  new_data.subtract(old_data, deriv);
  if (new_data.get_size() < data_size)
    {
      fprintf(stderr," bad call to PCAclass::process_cov \n");
	fflush(stderr);
      abort();
    }
  
  no_processed++;
  if (no_processed > max_n) no_processed = max_n;
  
  // update E(X)
  NagVector diff(mean - new_data);
  mean -= (diff / no_processed);
  
  // update E(X Xt), E(del_X del_X_t)

  realno tmp;
  for (int i = 0; i < data_size; i++)
    for (int j = i; j < data_size; j++)
      {
	tmp = *cov(i,j) - new_data[i] * new_data[j];
	*cov(i,j) -= (tmp / no_processed);
	tmp = change_covariance.read(i,j) - deriv[i] * deriv[j];
	*change_covariance.get(i,j) -= (tmp / no_processed);
      }
  
}
void PCAclass::restore_covariance()
{
  // subtract off E(X)E(Xt) (see above)
  int i,j;
  for (i = 0; i < data_size; i++)
    for (j = i; j < data_size; j++)
      {
	*cov(i,j) -= mean[i] * mean[j];
	*cov(j,i) = *cov(i, j);
      }
  for (i = 0; i < data_size; i++)
    for (j = i; j < data_size; j++)
      {
	realno rnk_cor = ((*cov(i,j)) * (*cov(i,j))) / 
	  ((*cov(i,i)) * (*cov(j,j)));
	if (rnk_cor < MIN_COV_ENTRY)
	  {
	    *cov(i,i) -= *cov(i,j);
	    *cov(j,j) -= *cov(i,j);
	    *cov(i,j) = 0;
	    *cov(j,i) = 0;
	    
	  }
      }
}

void PCAclass::process_mean(NagVector& new_data)
{
  NagVector diff(mean - new_data);
  no_processed++;
  if (no_processed > max_n) no_processed = max_n;
  mean.subtract(diff / no_processed, mean);
}

static void break_down_evals(NagVector& evals)
{
  realno l = evals.sum();
  realno sum = 0;
  int j = 1;
  for (int i = evals.get_size(); i > 0; i--)
    {
      realno indv = evals[i-1] / l;
      sum += indv;
      fprintf(stderr," mode %d indv %f cum %f \n", j++, indv, sum);
    }
}
	      
void PCAclass::get_evals()
{
//  restore_covariance();
//  covariance.get_eigenvectors(eigenvectors, eigenvalues);
//  eigenvalues.output();
//  break_down_evals(eigenvalues);
  setup_H_metric(data_size);

// NagMatrix tmp
// covariance.ortho_transform(*get_H_root(), tmp);

  NagMatrix tmp;
  covariance.get_eigenvectors(tmp,eigenvalues);

  get_H_inv_root()->multiply(tmp, eigenvectors);
  NagVector tmp_v;
  mean.copy(tmp_v);
  get_H_inv_root()->multiply(tmp_v, mean);

//  break_down_evals(eigenvalues);

}

void PCAclass::add_noise(realno noise_variance, int no_spline_pnts)
{

//  for (int i = 0; i < data_size; i++)
//    *covariance.get(i,i) += noise_variance;


  setup_H_metric(data_size);
  NagMatrix tmp1;
  build_covariance_matrix(tmp1, data_size / 2, no_spline_pnts);
  NagMatrix delta_cov;
  tmp1.ortho_transform(*get_H_root(), delta_cov, FALSE);
  delta_cov.scale(noise_variance, delta_cov);
  covariance.add(delta_cov, covariance);

  
  NagMatrix tmp;
  
  covariance.get_eigenvectors(tmp,eigenvalues);

  get_H_inv_root()->multiply(tmp, eigenvectors); 
}

void PCAclass::get_noise_model()
{
  NagMatrix tmp;
  if (get_H_root()->get_data() != NULL)
    get_H_root()->multiply(eigenvectors, tmp);
  else eigenvectors.copy(tmp);
  change_covariance.ortho_transform(tmp, change_covariance);
  change_covariance.read_diagonal(noise_model);
//  noise_model.output();
}

void PCAclass::get_mode(int index, realno lmda, NagVector& result)
{
  NagVector e_i(data_size, get_evec(index));
  mean.add(lmda * e_i, result);
  e_i.reset();
}

void PCAclass::adjust_mode(int index, realno lmda, NagVector& result)
{
  NagVector e_i(data_size, get_evec(index));
  result.add(lmda * e_i, result);
  e_i.reset();
}

realno PCAclass::Hdot(NagVector& v1, NagVector& v2)
{
  setup_H_metric(data_size);
  NagVector tmp;
  get_H()->multiply(v2, tmp);
  return (v1.dot(tmp));
}

void PCAclass::map_to_shape(int no_modes, NagVector& result, realno
			    max_distance)
{
  NagVector diff(result - mean);
  NagVector lambda(no_modes);
  realno tmp, m_dist = 0;
  for (int i = 0; i < no_modes; i++)
    {
      NagVector e_i(data_size, get_evec(i+1));
      lambda[i] = Hdot(e_i, diff);
      m_dist += (lambda[i] * lambda[i]) / get_eval(i+1);
      e_i.reset();
    }
  if (m_dist >= max_distance)
    lambda.scale(sqrt(max_distance / m_dist), lambda);

  get_mode(1,lambda[0], result);
  for (i = 1; i < no_modes; i++)
    adjust_mode(i+1, lambda[i], result);
}
  
void PCAclass::map_to_model(int no_modes, NagVector& shape, NagVector& result)
{
  if (shape.get_size() < data_size)
    {
      // data only containts first 'n' components
      // use pseudo-inverse to find shape parameters

      NagVector diff(shape - mean);
      
      // solve Pb = dx
      NagMatrix P(shape.get_size(), no_modes);
      for (int i = 0; i < no_modes; i++)
	{
	  NagVector e_i(shape.get_size(), get_evec(i+1));
	  P.set_column(i, e_i);
	  e_i.reset();
	}
      
      NagMatrix P_inv;

      P.pseudo_invert(P_inv);
      P_inv.multiply(diff, result);
      return;
    }

  // else 
  NagVector diff(shape - mean); 
  realno tmp, *evec;
  if (result.get_data() == NULL) result.reconstruct(no_modes);
  for (int i = 0; i < no_modes; i++)
    {
      NagVector e_i(data_size, get_evec(i+1));
      result[i] = Hdot(diff, e_i);
      e_i.reset();
    }
}

realno PCAclass::map_to_mode(int index, NagVector& shape)
{
  NagVector e_i(data_size, get_evec(index+1));
  NagVector diff(shape - mean);
  realno res = Hdot(diff, e_i);
  e_i.reset();
  return (res);
}


ostream& operator<<(ostream& strm_out, PCAclass& model)
{
  
  strm_out << "PCA_MODEL" << endl;
  strm_out << " no_processed " << model.no_processed << endl;
  int data_size = model.data_size;
  strm_out << " data_size " << model.data_size << endl;
  strm_out << " Covariance_Matrix " << model.covariance;
  strm_out << " Mean " << model.mean;
  strm_out << " Weights " << model.weights;
  strm_out << " Eigenvalues " << model.eigenvalues;
  strm_out << " Eigenvectors " << model.eigenvectors;
  if (model.noise_model.get_data() != NULL)
    strm_out << " Noise_Model " << model.noise_model;
  return (strm_out);
}

istream& operator>>(istream& strm_in, PCAclass& model)
{
  char dummy[50];
  strm_in >> dummy;
  strm_in >> dummy >> model.no_processed >> dummy >> model.data_size;
  strm_in >> dummy >> model.covariance;
  strm_in >> dummy >> model.mean;
  strm_in >> dummy >> model.weights;
  strm_in >> dummy >> model.eigenvalues;
  strm_in >> dummy >> model.eigenvectors;
  strm_in >> dummy;
  if (strcmp(dummy,"Noise_Model") == 0)

      strm_in >> model.noise_model;
  else model.eigenvalues.copy(model.noise_model);
  return (strm_in);
}

