/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/*
 * BWImage.cc
 * 
 * a Gray8Image whith each pixel containing
 * either MARK or CLEAR
 *
 * and with added functionality
 *
 */

#include <BWImage.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stack.h>

#ifdef gl_interface
#include <gl/gl.h>
#else
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#endif

#include <time.h>

#define TEST(p) if (*(p) == MARK) {PUSH(p); *(p) = CLEAR;}


BWImage* BWImage::compress(BWImage* res)
{
  if (res == NULL) res = new BWImage(width, height);
  unsigned char* idat = data, *odat = res->get_data();
  unsigned char flag;
  int counter;
  while (idat < end_data)
    {
      flag = *idat;
      counter = -1;
      while ((idat < end_data) && (*idat == flag) && counter < 127)
	{
	  counter++;
	  idat++;
	}
      if (flag == MARK) counter = counter | 128;
      *odat++ = counter;
    }
  res->set_enddata(odat);
  return (res);

}


BWImage* BWImage::uncompress(BWImage* res)
{
  if (res == NULL) res = new BWImage(width, height);
  unsigned char* idat = data, *odat = res->get_data();
  unsigned char* edat = res->endofdata();
  unsigned char flag;
  int counter;
  while (odat < edat)
    {
      counter = (*idat & 127) + 1;
      if ((*idat & 128) != 0) flag = MARK;
      else flag = CLEAR;
      while (counter > 0  && odat < edat)
	{
	  counter--;
	  *odat++ = flag;
	}
      idat++;
    }
  return (res);
}

Image* BWImage::neighbour_order(int n, Image* res)
{
  if (res == NULL) 
    {
      res = new BWImage(width, height);
      res->clear_border();
    }
  int n9 = 9 - n;
  unsigned char *upnt, *pnt, *dpnt, *out, *epnt;
  for (int y = 1; y < height-1; y++)
    {
      upnt = get_pixel(0,y-1);
      pnt = get_pixel(0,y);
      epnt = get_pixel(width-2,y);
      dpnt = get_pixel(0,y+1);  
      out = res->get_pixel(1,y);
      while (pnt < epnt)
	{  
	  int c = 0;
	  if (upnt[0] == MARK) c++; if (upnt[1] == MARK) c++; if (upnt[2] == MARK) c++;
	  if (pnt[0] == MARK) c++; if (pnt[1] == MARK) c++; if (pnt[2] == MARK) c++;
	  if (dpnt[0] == MARK) c++; if (dpnt[1] == MARK) c++; if (dpnt[2] == MARK) c++;
	  if (c <= n9) *out = CLEAR;
	  else *out = MARK;
	  upnt++; pnt++; dpnt++;out++;
	}
    }
  return (res);
}

void BWImage::remove_small_regions(int minsize)
{
  Gray8Image* temp_img = (Gray8Image*) copy();
  int* table = new int [minsize+1];
  unsigned char* sptr, *eptr = temp_img->endofdata();
  unsigned char* ptr = temp_img->get_data();
  int offset;
  int count, x, y;
  int os1 = width+1;
  int os2 = width;
  int os3 = width-1;
  int tnx = width-1;
  int tny = height-1;
  for (unsigned char* tptr = ptr; tptr < eptr; tptr++)
    {
      if (*tptr == MARK)
	{
	  PUSH(tptr);
	  *tptr = CLEAR;
	  count = 0;
	  while (!STACKEMPTY)
	    {
	      POP(sptr);
	      offset = sptr - ptr;
	      if (count < minsize) table[count] = offset;
	      count++;
	      y = offset / width;
	      x = offset % width;
	      
	      /* add marked neighbours to stack */
	      if (y > 0) {
		if (x > 0) TEST(sptr-os1);
		TEST(sptr-os2);
		if (x < tnx) TEST(sptr-os3);
	      }
	      if (x > 0) TEST(sptr-1);
	      if (x < tnx) TEST(sptr+1);
	      if (y < tny) {
		if (x > 0) TEST(sptr+os3);
		TEST(sptr+os2);
		if (x < tnx) TEST(sptr+os1);
	      }
		
	    }
	  if (count < minsize)
	    {
	      for (int i = 0; i < count; i++)
		*(data + table[i]) = CLEAR;
	    }
	}
    }
  delete temp_img;
  delete [] table;
}



BWImage* BWImage::fix_holes(int gapsize, BWImage* res)
{
  if (res == NULL) res = new BWImage(width, height);
  BWImage* temp_img1, *temp_img2, *dummy;
  temp_img1 = new BWImage(width, height);
  temp_img2 = new BWImage(width,height);
  temp_img1->clear();
  temp_img2->clear();
  this->maximum(temp_img1);
  for (int i = 1; i < gapsize; i++)
    {
      temp_img1->maximum(temp_img2);
      dummy = temp_img1;
      temp_img1 = temp_img2;
      temp_img2 = dummy;
    }

      
  for (i = 0; i < gapsize; i++)
    {
      temp_img1->minimum(temp_img2);
      dummy = temp_img1;
      temp_img1 = temp_img2;
      temp_img2 = dummy;
    }

  temp_img1->copy(res);
  delete temp_img2;
  delete temp_img1;
  return (res);
}

#ifdef gl_interface
// GL interface

long BWImage::display(long awindow)
{
  if (awindow != NULLWIN) glwin = awindow;
  if (glwin == NULLWIN)
    {
      int zoom = image_class_zoom.value;
      char wintitle [40];
      sprintf(wintitle ,"  BWimage %d by %d ", width, height);
      prefsize(width * zoom, height * zoom);
      foreground();
      if (window_title == NULL) glwin = winopen(wintitle);
      else glwin = winopen(window_title);
      winset(glwin);
      reshapeviewport();
      prefsize(width * zoom, height * zoom);
      if (zoom != 1)
	{
	  rectzoom((float) zoom, (float) zoom);
	  viewport(0, width * zoom - 1, 0, height * zoom - 1);
	  ortho2(-0.5, width + 0.5, -0.5, height + 0.5);
	}
      winconstraints();
      pixmode(PM_SIZE, 8);
      // RGBmode();
      if (access_mode == IA_TOP_TO_BOTTOM)
	{
	  pixmode(PM_TTOB, 1);
	  //ortho2(-0.5, width + 0.5, height + 0.5, -0.5);
	}
      cmode();
      /*mapcolor(255, 255, 255, 255);*/
      gflush();
      gconfig();
    }
  winset(glwin);
  lrectwrite(0,0,width-1,height-1,(unsigned long *) data);
  return(glwin);
}


#else
// X interface

Colormap cmap = 0;
static unsigned char pixel;
static unsigned long plane_masks[8];
static unsigned int no_planes = 1;
static int shift;


static void setup_graymap(Display* mydisplay, int screen, Window mywindow)
{
  cmap = DefaultColormap(mydisplay, screen);
  Status res = 0;
  unsigned long pix;
  no_planes = 1;
  res = XAllocColorCells(mydisplay, cmap, True, plane_masks,
				   no_planes, &pix, 1);
  if (res == 0)
    {
      cmap = XCreateColormap(mydisplay, mywindow,
			     DefaultVisual(mydisplay, screen),
			     AllocNone);
      res = XAllocColorCells(mydisplay, cmap, True, plane_masks,
				(no_planes = 8) , &pix, 1);
      if (res == 0) 
	{ fprintf(stderr," No colour planes available! \n");exit(1);}
    }
  int no_cells = 1 << no_planes;
  XColor* gray = (XColor *) new XColor [no_cells];
  for (shift = 0; plane_masks[0] != (1 << shift); shift++);
  for (int cell = 0; cell < no_cells; cell++)
    {
      int intensity = (int) ((cell * (long) 65280 ) / (no_cells-1));
      gray[cell].red = gray[cell].green = gray[cell].blue =
	intensity;
      gray[cell].pixel = (cell << shift) | pix;
      gray[cell].flags = DoRed | DoGreen | DoBlue;
    }
  XStoreColors(mydisplay, cmap, gray, no_cells);
  delete [no_cells] gray;
  pixel = (unsigned char) pix;
}
      
long BWImage::display(long dummy)
{
  char *image_mode = getenv("IMAGE_DISPLAY");
  if ((image_mode != NULL) && strcasecmp(image_mode,"OFF") == 0)
    {
      virtual_display();
      return (0);
    }
  if (mydisplay == NULL)
    {
      char wintitle [40];
      sprintf(wintitle ," mono image %d by %d ", width, height);      
      XSizeHints myhints;
      myhints.width = width;
      myhints.height = height;
      myhints.flags = PSize;
      mydisplay = XOpenDisplay("");
      int screen = DefaultScreen(mydisplay);
      unsigned long white = WhitePixel(mydisplay, screen);
      unsigned long black = BlackPixel(mydisplay, screen);
      mywindow = XCreateSimpleWindow (mydisplay, DefaultRootWindow(mydisplay)
				      ,100,100,width,height,5,white, black);
      XSetStandardProperties(mydisplay, mywindow, wintitle, "image", None,
			     NULL, 0, &myhints);
      
      XMapRaised(mydisplay, mywindow);
      myimage = XCreateImage( mydisplay, DefaultVisual(mydisplay, screen),
			     1, XYBitmap, 0, None, width, height,
			     8, 0);
      //if (cmap == 0) cmap = DefaultColormap(mydisplay, screen);
      //XSetWindowColormap(mydisplay, mywindow, cmap);
      mygc = XCreateGC (mydisplay, mywindow, 0, 0);
      XSetBackground (mydisplay, mygc, black);
      XSetForeground (mydisplay, mygc, white);
      myimage->data = new char[(width / 8 + 1) * height];
    }
  int x,y;
  unsigned char *idat, *odat = (unsigned char*) myimage->data;
  for (y = height-1; y>=0; y--)
    {
      idat = get_pixel(0,y);
      for (x = 0; x < width; x += 8)
	{
	  *odat = 0;
	  if (*idat++ > 0) *odat += 128;
	  if (*idat++ > 0) *odat += 64;
	  if (*idat++ > 0) *odat += 32;
	  if (*idat++ > 0) *odat += 16;
	  if (*idat++ > 0) *odat += 8;
	  if (*idat++ > 0) *odat += 4;
	  if (*idat++ > 0) *odat += 2;
	  if (*idat++ > 0) *odat += 1;
	  odat++;
	}
    }
  XPutImage(mydisplay, mywindow, mygc, myimage, 0,0,0,0,width,
	    height);
  return (0);
}


#endif
static int set_seed = 0;
static int iset = 0;
static realno gset;

Image* BWImage::add_gaussian_noise(realno sd, realno& noise, Image* res)
{
  if (res == NULL) res = new BWImage(width, height);
  if (set_seed == 0)
    {
      long rseed = time(NULL);
      srand48(rseed);
      set_seed = 1;
      fprintf(stderr,"Adding noise to BINARY image \n");
    }
  noise = 0;
  unsigned char *in_pnt = data;
  unsigned char *out_pnt = res->get_data();
  while (in_pnt < end_data)
    {
      realno fac, rsq, v1, v2, corrupt;
      if (iset == 0)
	{
	  do
	    {
	      v1 = 2.0 * drand48() - 1.0;
	      v2 = 2.0 * drand48() - 1.0;
	      rsq = (v1 * v1) + (v2 * v2);
	    } while ((rsq >= 1.0 )|| (rsq == 0.0));
	  
	  fac = sqrt(-2.0 * log(rsq)/rsq);
	  gset = v1 * fac;
	  iset = 1;
	  corrupt = v2 * fac;
	}
      else
	{
	  iset = 0;
	  corrupt = gset;
	}
	  
      int icorrupt = (int) (sd * corrupt + 0.5);
      int ipix = (int) *in_pnt++;
      int new_ipix;
      if (ipix == MARK) new_ipix = ipix - abs(icorrupt);
      if (ipix == CLEAR) new_ipix = ipix + abs(icorrupt);

      if (new_ipix > 255) new_ipix = 255;
      if (new_ipix < 0) new_ipix = 0;
      *out_pnt++ = new_ipix;
      noise += (new_ipix - ipix) * (new_ipix - ipix);
    }
  return (res);
}
      
