/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/***************************************************************
 * C++ source
 *
 * File : 	PgmSource.cc
 *
 * Module :	PgmSource
 *
 * Author : 	A M Baumberg (CoMIR)
 *
 * Creation Date : Wed Jun 29 20:25:25 1994 
 *
 * Comments : 	
 *
 ***************************************************************/


/* include files */
#include <PgmSource.h>
#include <PnmStream.h>
#include <math.h>

#include <stdio.h>
#include <ctype.h>

Image* read_pnm(FILE *infile, Image* res)
{
  char c1,c2;
  fscanf(infile,"%c%c", &c1, &c2);
  if (c1 != 'P' || ((c2 != '5') && (c2 != '6')))
    {
      // fprintf(stderr," file not ppm / pgm \n");
      // exit(1);
      return (NULL);
    }
  char aline[100];
  fgets(aline, 99, infile);
  while (aline[0] == '#' || aline[0] == '\n') fgets(aline, 99, infile);

  int w,h,byte;
  sscanf(aline,"%d %d\n", &w, &h);
  fgets(aline, 99, infile);
  sscanf(aline,"%d", &byte);
  if (byte != 255)
    {
      fprintf(stderr," Sorry -- unsupported pgm/ ppm image format \n");
      exit(1);
    }
  switch (c2)
    {
    case '6':
      {
	if (res == NULL) res = new RGB32Image(w,h);
	if ((res->imgtype() != RGB32) || (res->get_width() != w) || 
	    (res->get_height() != h))
	  {
	    fprintf(stderr, "portable image format changed \n");
	    exit(1);
	  }
	for (int i = h-1; i  >= 0; i--)
	  {
	    RGB32pixel *dat1 = (RGB32pixel *) res->get_pixel(0,i);
	    for (int j = 0; j < w; j++)
	      {
		dat1->alpha = 0;
		fscanf(infile, "%c%c%c",
		       &(dat1->red),&(dat1->green),&(dat1->blue));
		
		dat1++;
	      }
	  }
	break;
      }

    case '5':
      {
	if (res == NULL) res = new Gray8Image(w, h);
	if ((res->imgtype() != GRAY8) || (res->get_width() != w) || 
	    (res->get_height() != h))
	  {
	    fprintf(stderr, "portable image format changed \n");
	    exit(1);
	  }
	if (res->get_access_mode() == IA_TOP_TO_BOTTOM)
	  fread(res->get_data(),1, w*h, infile); // use easy method !
	else for (int i = h-1; i >= 0; i--)
	  fread(res->get_pixel(0,i),1,w, infile);
	      
	break;
      }
    default:
      {
	fprintf(stderr," Sorry unsupported pgm /ppm \n");
	exit(1);
      }
    }

  return (res);
}

Image* read_pnm(char* filename, Image* res)
{
  FILE* fhandle = fopen(filename, "r");
  if (fhandle == NULL) {
    fprintf(stderr,"can't open %s \n", filename);      
    exit(1);
  }
  return (read_pnm(fhandle, res));
  fclose(fhandle);
}

Image* PnmStream::get_next()
{
  frame_number++;
  if ((input_stream == NULL) || (ferror(input_stream)) 
      || (feof(input_stream))) return (NULL);
  current = read_pnm(input_stream, current);
  return (current);
}


Image* PgmSource::get_next()
{
  if (curr_frame < start_frame)
    {
      frame_number = ++curr_frame;
      return (NULL);
    }
  while (curr_frame > (pow(10, no_digits)-0.5)) 
    {
      no_digits++; 
      sprintf(format_string,"%%s%%.%dd%%s",no_digits);
    }
  char* fname = 
    new char[strlen(file_prefix)+strlen(file_suffix)+no_digits+1];
  sprintf(fname, format_string, file_prefix, curr_frame, file_suffix);
  curr_frame++;
  frame_number = curr_frame;

  FILE *infile;
  if ((infile = fopen(fname,"r")) == NULL)
    return (NULL);
  // check for compressed files
  char* file_ext = strrchr(fname, '.');
  if (file_ext != NULL)
    if ((strcasecmp(file_ext,".gz") == 0) || 
	(strcasecmp(file_ext,".z") == 0))
      {
	// compressed
	fclose(infile);
	char *sys_command = new char[strlen(fname)+10];
	sprintf(sys_command,"zcat %s", fname);
	infile = popen(sys_command,"r");
	delete [] sys_command;
      }
  current = read_pnm(infile, current);
  if (pclose(infile) == -1) fclose(infile);
  delete [] fname;


  return (current); 
}

PgmSource::PgmSource(char* initial_frame) : ImageSource(NULL, 0)
{
  int lpos = strlen(initial_frame) - 1;
  int cpos = lpos;
  // find the last digit of the frame number
  // e.g. "file023.xxx"
  //             ^
  
  while ((cpos >= 0) && !isdigit(initial_frame[cpos]))
    cpos--;
  
  if (cpos < 0) 
	{
	fprintf(stderr,"Abort in PgmSource::PgmSource\n");
	fflush(stderr);
	abort();
	}

  // find the first digit
  // e.g. "file023.xxx"
  //           ^

  int dpos = cpos;
  while ((dpos >= 0) && isdigit(initial_frame[dpos]))
    dpos--;
  dpos++;
  
  file_prefix = new char[dpos+1]; 
  file_suffix = new char[lpos - cpos + 1]; 
  if (dpos > 0)
    strncpy(file_prefix, initial_frame, dpos);
  file_prefix[dpos] = '\0';
  if (cpos < lpos)
    strncpy(file_suffix, &initial_frame[cpos+1], lpos-cpos);
  file_suffix[lpos-cpos] = '\0';
  
  no_digits = cpos - dpos + 1;
  curr_frame = frame_number = start_frame = atoi(&initial_frame[dpos]);

  sprintf(format_string,"%%s%%.%dd%%s",no_digits);

}


int PgmSource::get_no_frames()
{
  char* fname = 
    new char[strlen(file_prefix)+strlen(file_suffix)+no_digits+1];

  int count = 0;
  int tmp_frame = start_frame;
  FILE* infile = NULL;
  do
    {
      sprintf(fname, format_string, file_prefix, tmp_frame, file_suffix);
      infile = NULL;
      if ((infile = fopen(fname,"r")) != NULL)
	{
	  count++;
	  fclose(infile);
	  fprintf(stderr," frame %d ", tmp_frame);
	  tmp_frame++;
	}
    } while (infile != NULL);
  
  delete [] fname;
  return (count);
}

/* goto_frame now uses absolute frame numbering */
void PgmSource::goto_frame(int n)
{
//   int skip = n - frame_number;
//   curr_frame += skip;
//   frame_number += skip;
  frame_number = curr_frame = n;

}

  
