/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/*
 * RGB32Image.cc
 * 24-bit (plus alpha) colour image class derived from generic
 * Image class in image.h/image.cc
 *
 * author : A M Baumberg
 */

#include <stdio.h>
#include <stdlib.h>
#include <stream.h>
#include <string.h>
#include <point2.h>


#include <RGB32Image.h>
#include <Gray8Image.h>
#include <BWImage.h>

#ifdef gl_interface
#include <gl/gl.h>
#else
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#endif

#ifndef MAX
#define MAX(a,b) (( (a) > (b)) ? (a) : (b))
#endif

#ifndef ABS
#define ABS(a)          (((a)<0) ? -(a) : (a))
#endif

#ifndef SGN
#define SGN(a)          (((a)<0) ? -1 : 1)
#endif

Image* RGB32Image::copy(Image* r)
{
  if (r == NULL) r = (Image *) new RGB32Image(width, height);
  RGB32pixel *idat = (RGB32pixel *)  data, *odat = (RGB32pixel *) r->get_data()
    ,*edat = (RGB32pixel *) endofdata();
  for (;idat < edat; idat++)
    { *odat++ = *idat; }
  return (r);
}
 
BasePixel* RGB32Image::get_base_pixel(int x, int y)
{
  RGB32pixel* tmp = (RGB32pixel*) get_pixel(x,y);
  return (new RGBpixel(tmp->red, tmp->green, tmp->blue));
}

// NOTE LACK OF FUNCTIONALITY 

#ifdef gl_interface
// GL interface

long RGB32Image::display(long awindow)
{
  if (awindow != NULLWIN) glwin = awindow;
  if (glwin == NULLWIN)
    {
      int zoom = image_class_zoom.value;
      char wintitle [40];
      sprintf(wintitle ," RGB32 image %d by %d ", width, height);
      prefsize(width * zoom, height * zoom);
      foreground();
      glwin = winopen(wintitle);

      if (access_mode == IA_TOP_TO_BOTTOM) pixmode(PM_TTOB, 1);

      winset(glwin);
      reshapeviewport();
      prefsize(width * zoom, height * zoom);
      if (zoom != 1)
	{
	  rectzoom((float) zoom, (float) zoom);
	  viewport(0, width * zoom - 1, 0, height * zoom - 1);
	  ortho2(-0.5, width + 0.5, -0.5, height + 0.5);
	}
      winconstraints();
      RGBmode();
      pixmode(PM_SIZE, 32);
      cpack(0x00ffffff);
      gflush();
      gconfig();
    }
  winset(glwin);
  lrectwrite(0,0,width-1,height-1,(unsigned long *) data);
  return(glwin);
}


#else
// X interface

static int rshift, gshift, bshift;
XVisualInfo xv_info;
static int matched_depth;

long RGB32Image::display(long dummy)
{
  if (mydisplay == NULL)
    {
      char wintitle [40];
      char* wintitle_pnt = (char*) wintitle;
      sprintf(wintitle ," RGB image %d by %d ", width, height);      
      if (window_title != NULL) wintitle_pnt = window_title;
      XSizeHints myhints;
      myhints.width = width;
      myhints.height = height;
      myhints.flags = PSize;
      mydisplay = XOpenDisplay("");
      int screen = DefaultScreen(mydisplay);
      rshift = bshift = gshift = 0;
      Status res = XMatchVisualInfo(mydisplay, screen, 24, TrueColor,
				    &xv_info);
      matched_depth = 32;
      if (res == False) 
	{
	  Status res = XMatchVisualInfo(mydisplay, screen, 8, TrueColor,
					&xv_info);
	  if (res == False) exit(1);
	  rshift = bshift = gshift = 0;
	  for (int i = xv_info.red_mask; i < 128; i *= 2) rshift++;
	  for (i = xv_info.green_mask; i < 128; i *=2) gshift++;
	  for (i = xv_info.blue_mask; i < 128; i *=2 ) bshift++;
	  matched_depth = 8;
	}
      else
	{
	  for (long i = xv_info.red_mask; i > 256; i /= 256) rshift++;
	  for (i = xv_info.green_mask; i > 256; i /= 256) gshift++;
	  for (i = xv_info.blue_mask; i > 256; i /= 256) bshift++;
	  rshift = 3 - rshift;
	  gshift = 3 - gshift;
	  bshift = 3 - bshift;
	}

      unsigned long white = WhitePixel(mydisplay, screen);
      unsigned long black = BlackPixel(mydisplay, screen);
      Colormap cmap = XCreateColormap(mydisplay,
				      XRootWindow(mydisplay, screen),
				      xv_info.visual, AllocNone);

      mypixmap = XCreatePixmap(mydisplay,
 			       XRootWindow(mydisplay, screen), width,
 			       height, xv_info.depth);

      XSetWindowAttributes swa;
      swa.background_pixel = BlackPixel(mydisplay, screen);
      swa.border_pixel = WhitePixel(mydisplay, screen);
      swa.colormap = cmap;
      swa.background_pixmap = mypixmap;

      mywindow = XCreateWindow (mydisplay, RootWindow(mydisplay, screen)
				      ,100,100,width,height,1,
				xv_info.depth, InputOutput,
				xv_info.visual, CWBackPixmap |CWColormap
				| CWBorderPixel, &swa);

      XSetStandardProperties(mydisplay, mywindow, wintitle_pnt,
 			     "image", None,
 			     NULL, 0, &myhints);
      XMapRaised(mydisplay, mywindow);

      myimage = XCreateImage( mydisplay, xv_info.visual,
			     xv_info.depth, ZPixmap, 0, None, width, height,
			     matched_depth, 0);
#ifdef LITTLE_ENDIAN_ARCHITECTURE
      myimage->byte_order = LSBFirst;
      myimage->bitmap_bit_order = LSBFirst;
#else
      myimage->byte_order = MSBFirst;
      myimage->bitmap_bit_order = MSBFirst;
#endif
      
      //mygc = XCreateGC (mydisplay, mywindow, 0, 0);
      mygc = XCreateGC (mydisplay, mypixmap, 0, 0);
      if ((matched_depth == 32) && (access_mode == IA_TOP_TO_BOTTOM)
	  && (rshift == 3) && (gshift == 2) && (bshift == 1))
	myimage->data = (char *) data;
      else myimage->data = new char[width * height * matched_depth / 8];
    }
  if (xv_info.depth == 8)
    {
      
      int x,y;
      RGB32pixel* idat;
      unsigned char *odat = (unsigned char*) myimage->data;
      unsigned char red_mask = xv_info.red_mask;
      unsigned char green_mask = xv_info.green_mask;
      unsigned char blue_mask = xv_info.blue_mask;

      for (y = height-1; y>=0; y--)
	{
	  idat = (RGB32pixel *)get_pixel(0,y);
	  for (x = 0; x < width; x++)    
	    {
	      *odat++ = (((idat->red) >> rshift) & red_mask)
		+ (((idat->green) >> gshift) & green_mask)
		+ (((idat->blue) >> bshift) & blue_mask);
	      idat++;
	    }
	}
    }
  
  else if (myimage->data != data)
    {
      // xv_info.depth == 24
      int x,y;
      RGB32pixel* idat;
      unsigned char *odat = (unsigned char*) myimage->data;
      for (y = height-1; y>=0; y--)
	{
	  idat = (RGB32pixel *)get_pixel(0,y);
	  for (x = 0; x < width; x++)    
	    {
	      odat[rshift] = idat->red;
	      odat[gshift] = idat->green;
	      odat[bshift] = idat->blue;
	      idat++;
	      odat += 4;
	    }
	}
    }
      
    

  //  XPutImage(mydisplay, mywindow, mygc, myimage, 0,0,0,0,width,
  //	    height);
  XPutImage(mydisplay, mypixmap, mygc, myimage, 0,0,0,0,width,
 	    height);
  XClearArea(mydisplay, mywindow, 0,0,0,0, True);
  return(0);

}

#endif

void RGB32Image::save_pm(char *fname)
{
  ostream* strm_pnt = NULL;
  if (fname == NULL) strm_pnt = &cout;
  else strm_pnt = (ostream*) new ofstream(fname);
  ostream& data_out = *strm_pnt;
  data_out << "P6" << endl << width << " " << height << endl;
  data_out << "255" << endl;
  for (int i = height-1; i >= 0; i--)
    {
      RGB32pixel *dat1 = (RGB32pixel *) get_pixel(0,i);
      for (int j = 0; j < width; j++)
	{
	  data_out << dat1->red << dat1->green << dat1->blue;
	  dat1++;
	}
    }
  if (fname != NULL) delete ((ofstream*) strm_pnt);

}


// RGB32Image* read_ppm(char *fname)
// {
//   ifstream data_in(fname);
//   char line_in[50];
//   data_in >> line_in;
//   if (line_in[0] != 'P') exit(1);
//   if (line_in[1] != '6' ) exit(1);
//   int w,h,byte;
//   data_in >> w >> h >> byte;
//   RGB32Image* img = new RGB32Image(w,h);
//   for (int i = h-1; i >= 0; i--)
//     {
//       RGB32pixel *dat1 = (RGB32pixel *) img->get_pixel(0,i);
//       for (int j = 0; j < w; j++)
// 	{
// 	  dat1->alpha = 0;
// 	  data_in >> (dat1->red) >> (dat1->green) >> (dat1->blue);
// 	  dat1++;
// 	}
//     }
//   return (img);
// }

Image* RGB32Image::simple_difference(Image* image2, realno thresh, 
				     int& no_marked, Image* res)
{
  no_marked = 0;

  // rescale threshold for colour images
  int threshold = (int) (thresh * 3.0);

  if (thresh < 0)
    {
      no_marked = 0;
      return (difference(image2,0,res));
    }

  if (image2->get_width() != width || image2->get_height() != height)
    {
      fprintf(stderr, " can't difference different sized images !");
      exit(1);
    }
  if (res == NULL) res = new BWImage(width, height);
  RGB32pixel *idat1 = (RGB32pixel*) data
    , *idat2 = (RGB32pixel*) image2->get_data()
      , *enddat = (RGB32pixel*) endofdata();
  unsigned char *odat = res->get_data();
  for (;idat1 < enddat; idat1++)
    {
      if ((abs(idat1->red - idat2->red) + 
	  abs(idat1->green - idat2->green) +
	  abs(idat1->blue - idat2->blue)) > threshold ) 
	{
	  no_marked++;
	  *odat = MARK;
	}
      else *odat = CLEAR;
      idat2++;odat++;
    }
  return(res);
}

// quick differencing 
static const int BOX_SIZE = 8;
static inline unsigned char rgb_diff(RGB32pixel& pix1, RGB32pixel& pix2)
{
  int ires = (abs(pix1.red - pix2.red) + abs(pix1.green - pix2.green) +
	      abs(pix1.blue - pix2.blue)) / 3;
  return ((unsigned char) ires);
}

static inline void get_detail(RGB32pixel *g1, RGB32pixel *g2,
			      unsigned char* r, int& width)
{
  int lpx,lpy;
  RGB32pixel* tg1, *tg2;
  unsigned char *tr;
  for (lpy = 0; lpy < BOX_SIZE; lpy++)
    {
      tg1 = g1; tg2 = g2;
      tr = r;
      for (lpx = 0; lpx < BOX_SIZE; lpx++)
	*r++ = rgb_diff(*g1++, *g2++);

      g1 = tg1 + width; g2 = tg2 + width;
      r = tr + width;
    }
}

      
Image* RGB32Image::difference(Image* image2, realno threshold, Image* res)
{
  if (this == NULL || image2 == NULL) return (NULL);
  if (image2->imgtype() != RGB32)
    {
      fprintf(stderr, " can't difference different image types! ");
      exit(1);
    }
  if (res == NULL) res = new Gray8Image(width, height);
  res->clear();
  RGB32pixel *g1 = (RGB32pixel*)  data;
  RGB32pixel *g2 = (RGB32pixel*) image2->get_data();

  RGB32pixel *tg1, *tg2;
  RGB32pixel *edat = (RGB32pixel*) endofdata();
  unsigned char* rdat = res->get_data();
  unsigned char *rdat1, *rdat2, *rdat3, *trdat;

  int up = BOX_SIZE * width;
  int left = BOX_SIZE;

  int upleft = left + up;
  int upleft2 = upleft / 2;
  int upright = up - left;

  g1 += (upleft2) * 3; g2 += (upleft2) * 3;
  rdat += (upleft2) * 3;

  int lx, ly;
  int lymax = height / BOX_SIZE;
  int lxmax = width / BOX_SIZE; 

  int interest_threshold = (threshold);

  int adj1 = upleft2, adj2 = left+adj1, adj3 = up+adj1, adj4 = upleft+adj1
    , adj5 = adj1-left, adj6 = adj1-up, adj7 = adj1-upleft
      , adj8 = adj1-upright, adj9 = adj1+upright;

  //lxmax -=1 ;lymax -=1;
  Image* flag_img = new Image(lxmax, lymax);
  flag_img->clear(FALSE);
  for (ly = 1; ly < lymax; ly++)
    {
      tg1 = g1; tg2 = g2; 
      trdat = rdat;
      for (lx = 1; lx < lxmax; lx++)
	{
	  *rdat = rgb_diff(*g1,*g2);
	  if (*rdat >= interest_threshold)
	    {
	      unsigned char* c;
	      // this square
	      c = flag_img->check_pixel(lx,ly);
	      if ((c != NULL) && (*c == FALSE))
		{
		  get_detail(g1-adj1,g2-adj1, rdat-adj1, width);
		  *c = TRUE;
		}
	      // left
	      c = flag_img->check_pixel(lx-1,ly);
	      if ((c != NULL) && (*c == FALSE))
		{
		  get_detail(g1-adj2,g2-adj2, rdat-adj2, width);
		  *c = TRUE;
		}
	      // up
	      c = flag_img->check_pixel(lx,ly-1);
	      if ((c != NULL) && (*c == FALSE))
		{
		  get_detail(g1-adj3,g2-adj3, rdat-adj3, width);
		  *c = TRUE;
		}
	      // up-left
	      c = flag_img->check_pixel(lx-1,ly-1);
	      if ((c != NULL) && (*c == FALSE))
		{
		  get_detail(g1-adj4,g2-adj4, rdat-adj4, width);
		  *c = TRUE;
		}
	      // right
	      c = flag_img->check_pixel(lx+1,ly);
	      if ((c != NULL) && (*c == FALSE))
		{
		  get_detail(g1-adj5,g2-adj5, rdat-adj5, width);
		  *c = TRUE;
		}
	      // down
	      c = flag_img->check_pixel(lx,ly+1);
	      if ((c != NULL) && (*c == FALSE))
		{
		  get_detail(g1-adj6,g2-adj6, rdat-adj6, width);
		  *c = TRUE;
		}
	      // down-right
	      c = flag_img->check_pixel(lx+1,ly+1);
	      if ((c != NULL) && (*c == FALSE))
		{
		  get_detail(g1-adj7,g2-adj7, rdat-adj7, width);
		  *c = TRUE;
		}
	      // down-left
	      c = flag_img->check_pixel(lx-1,ly+1);
	      if ((c != NULL) && (*c == FALSE))
		{
		  get_detail(g1-adj8,g2-adj8, rdat-adj8, width);
		  *c = TRUE;
		}
	      // up-right
	      c = flag_img->check_pixel(lx+1,ly-1);
	      if ((c != NULL) && (*c == FALSE))
		{
		  get_detail(g1-adj9,g2-adj9, rdat-adj9, width);
		  *c = TRUE;
		}
	    } 
	  g1 += left; g2 += left;
	  rdat += left;
	}
      g1 = tg1 + up; g2 = tg2 + up;
      rdat = trdat + up;

    }
  delete (flag_img);
  return (res);
}

int RGB32Image::measure_contrast(void* p1, void* p2)
{
  RGB32pixel* pix1 = (RGB32pixel*) p1;
  RGB32pixel* pix2 = (RGB32pixel*) p2;
  if ((pix1 == NULL) || (pix2 == NULL)) return (0);
  int r1 = (int) pix1->red; int r2 = (int) pix2->red;
  int g1 = (int) pix1->green; int g2 = (int) pix2->green;
  int b1 = (int) pix1->blue; int b2 = (int) pix2->blue;

  return (abs(r1 - r2) + abs(g1 - g2) + abs(b1 - b2));
}

Image* RGB32Image::resample(int xstep , int ystep , Image* resampled)
{
  int new_width = (width / xstep);
  if (resampled == NULL) resampled = (Image *)
	    new RGB32Image(new_width, height / ystep);
  RGB32pixel* idat;
  RGB32pixel* odat;
  int outy = 0;
  for (int y = 0; y < height; y += ystep)
    {
      idat = (RGB32pixel*) get_pixel(0,y);
      odat = (RGB32pixel*) resampled->get_pixel(0, outy);
      for (int x = 0; x < new_width; x ++)
	{
	  *odat = *idat;
	  odat++;
	  idat = idat + xstep;
	}
      outy++;
    }
  return (resampled);
}
 
Image* RGB32Image::convolve(Kernel k, Image* convolved)
{
  RGB32pixel *upnt, *pnt, *dpnt;
  RGB32pixel *out;
  int temp_total;
  int k11 = k.coeff[0][0], k12 = k.coeff[0][1], k13 = k.coeff[0][2]; 
  int k21 = k.coeff[1][0], k22 = k.coeff[1][1], k23 = k.coeff[1][2]; 
  int k31 = k.coeff[2][0], k32 = k.coeff[2][1], k33 = k.coeff[2][2];
  int ksum = abs(k11)+abs(k12)+abs(k13)
    +abs(k21)+abs(k22)+abs(k23)
      +abs(k31)+abs(k32)+abs(k33);
  if (convolved == NULL) convolved = new RGB32Image(width, height);
  //convolved->clear_border();
  for (int y = 1; y < (height-1); y++)
    {
      upnt = (RGB32pixel*) get_pixel(0,y-1);
      pnt = (RGB32pixel*) get_pixel(0,y);
      dpnt = (RGB32pixel*) get_pixel(0,y+1);
      out = (RGB32pixel*) convolved->get_pixel(1,y);
      for (int x = 1; x < (width-1); x++)
	{
	  out->red = abs
	    (k11 * upnt->red + k12 * upnt[1].red + k13 * upnt[2].red
	     + k21 * pnt->red + k22 * pnt[1].red + k23 * pnt[2].red
	     + k31 * dpnt->red + k32 * dpnt[1].red + k33 *
	     dpnt[2].red) / ksum;
	  out->green = abs
	    (k11 * upnt->green + k12 * upnt[1].green + k13 * upnt[2].green
	     + k21 * pnt->green + k22 * pnt[1].green + k23 * pnt[2].green
	     + k31 * dpnt->green + k32 * dpnt[1].green + k33 *
	     dpnt[2].green) / ksum;  
	  out->blue = abs
	    (k11 * upnt->blue + k12 * upnt[1].blue + k13 * upnt[2].blue
	     + k21 * pnt->blue + k22 * pnt[1].blue + k23 * pnt[2].blue
	     + k31 * dpnt->blue + k32 * dpnt[1].blue + k33 *
	     dpnt[2].blue) / ksum;

	  upnt++; pnt++; dpnt++; out++;
	}
    }
  return (convolved);
}

Image* RGB32Image::sobel(int threshold, Image* res)
{
  int sqthresh3 = 3 * threshold * threshold;

  if (res == NULL) res = new Gray8Image(width, height);

  Image* hsobel = convolve(SOBELH_KER, NULL);
  Image* vsobel = convolve(SOBELV_KER, NULL);
  unsigned char *odat, *enddat;
  RGB32pixel *hdat, *vdat;
  hdat = (RGB32pixel*) hsobel->get_data();
  vdat = (RGB32pixel*) vsobel->get_data();
  enddat = res->endofdata();
  for (odat = res->get_data(); odat < enddat; odat++)
    {
      RGB32pixel& hpix = *hdat;
      RGB32pixel& vpix = *vdat;
      int r1 = hpix.red * hpix.red + vpix.red * vpix.red;
      int g1 = hpix.green * hpix.green + vpix.green * vpix.green;
      int b1 = hpix.blue * hpix.blue + vpix.blue * vpix.blue;

      if ((r1 + g1 + b1) > sqthresh3) *odat = MARK;
      else *odat = CLEAR;
      hdat++; vdat++;
    }
  delete vsobel;
  delete hsobel;
  return (res);
}

Image* RGB32Image::sobel(Image* res)
{
  if (res == NULL) res = new Gray8Image(width, height);

  Image* hsobel = convolve(SOBELH_KER, NULL);
  Image* vsobel = convolve(SOBELV_KER, NULL);
  unsigned char *odat, *enddat;
  RGB32pixel *hdat, *vdat;
  hdat = (RGB32pixel*) hsobel->get_data();
  vdat = (RGB32pixel*) vsobel->get_data();
  enddat = res->endofdata();
  for (odat = res->get_data(); odat < enddat; odat++)
    {
      RGB32pixel& hpix = *hdat;
      RGB32pixel& vpix = *vdat;
      int r1 = hpix.red * hpix.red + vpix.red * vpix.red;
      int g1 = hpix.green * hpix.green + vpix.green * vpix.green;
      int b1 = hpix.blue * hpix.blue + vpix.blue * vpix.blue;
      
      int pix = sqrt((r1+g1+b1) / 6);
      *odat = pix;

      hdat++; vdat++;
    }
  delete vsobel;
  delete hsobel;
  return (res);
}

static int compar_char(const void* c1, const void* c2)
{
  return (int) (*((unsigned char *) c1) - *((unsigned char *) c2));
}


void RGB32Image::median_img_array(Image** img_array, int asize)
{

  if (img_array[0]->imgtype() != RGB32)
    {
      fprintf(stderr," bad call to RGB32Image::median_img_array \n");
	fflush(stderr);
      abort();
    }
  
  unsigned char* rtable = new unsigned char[asize];
  unsigned char* gtable = new unsigned char[asize];
  unsigned char* btable = new unsigned char[asize];
  int cpos = (asize + 1) / 2;
  int i,j,k;
  for (i = 0; i < width; i++)
    for (j = 0; j < height; j++)
      {
	for (k = 0; k < asize; k++)
	  {
	    RGB32pixel* rgb_pix = (RGB32pixel*)
	      img_array[k]->check_pixel(i,j);
	    rtable[k] = rgb_pix->red;
	    gtable[k] = rgb_pix->green;
	    btable[k] = rgb_pix->blue;
	  }

	qsort(rtable, asize, sizeof(unsigned char), &compar_char);
	qsort(gtable, asize, sizeof(unsigned char), &compar_char);
	qsort(btable, asize, sizeof(unsigned char), &compar_char);

	RGB32pixel* rgb_pix = (RGB32pixel*) get_pixel(i,j);
	rgb_pix->red = rtable[cpos];
	rgb_pix->green = gtable[cpos];
	rgb_pix->blue = btable[cpos];
      }
	
  delete [] rtable;
  delete [] gtable;
  delete [] btable;
}


Image* RGB32Image::half_blur(Image* res)
{
  int new_width = width  / 2;
  int new_height = height / 2;
  if (res == NULL) res = new RGB32Image(new_width, new_height);
  RGB32pixel* odat = (RGB32pixel*) res->get_data();
  for (int y = 0; y < new_height; y++)
    {
      RGB32pixel* in_even = (RGB32pixel*) get_pixel(0,y * 2);
      RGB32pixel* in_odd = (RGB32pixel*) get_pixel(0,y * 2 + 1);
      for (int x = 0; x < new_width; x++)
	{
	  int r1 = in_even->red + in_odd->red;
	  int g1 = in_even->green + in_odd->green;
	  int b1 = in_even->blue + in_odd->blue;
	  in_even++; in_odd++;

	  r1 += in_even->red + in_odd->red;
	  g1 += in_even->green + in_odd->green;
	  b1 += in_even->blue + in_odd->blue;

	  odat->red = r1 >> 2;
	  odat->green = g1 >> 2;
	  odat->blue = b1 >> 2;

	  in_even++; in_odd++;
	  odat++;
	}
    }
  return (res);
}

Image* RGB32Image::image_blend(Image* i2, int a, int b, Image* res)
{
  if (res == NULL) res = new RGB32Image(width, height);
  RGB32pixel *idat1 = (RGB32pixel*) data, 
    *idat2 = (RGB32pixel*) i2->get_data(),
    *odat = (RGB32pixel*) res->get_data(),  
    *enddat = (RGB32pixel*) endofdata();
  int pixr, pixg, pixb;
  int aplusb = a + b;
  if (i2->imgtype() == RGB32)
    {
      for (;idat1 < enddat; idat1++)
	{
	  pixr = (a * idat1->red) + (b * idat2->red);
	  pixg = (a * idat1->green) + (b * idat2->green);
	  pixb = (a * idat1->blue) + (b * idat2->blue);
	  odat->red = (pixr / aplusb);
	  odat->green = (pixg / aplusb);
	  odat->blue = (pixb / aplusb);
	  
	  idat2++; odat++;
      
	}
    }
  else if (i2->imgtype() == GRAY8)
    {
      unsigned char* graydat = i2->get_data();
      for (;idat1 < enddat; idat1++)
	{
	  pixr = (a * idat1->red) + (b * *graydat);
	  pixg = (a * idat1->green) + (b * *graydat);
	  pixb = (a * idat1->blue) + (b * *graydat);
	  odat->red = (pixr / aplusb);
	  odat->green = (pixg / aplusb);
	  odat->blue = (pixb / aplusb);
	  
	  graydat++; odat++;
      
	}
    }
  else 
    {
      fprintf(stderr,"bad call to RGB32Image::image_blend \n");
	fflush(stderr);
      abort();
    }
  return (res);
}

// written by ajh
// optimised by amb
Image* RGB32Image::transform(int centre_x, int centre_y, realno angle,
			     realno scale, Image* res)
{
  if (res == NULL) res = copy_type();
  RGB32pixel *src, *dest;
  int x_from, y_from;
  realno cosa = cos(angle) / scale;
  realno sina = sin(angle) / scale;
  int& dw = res->get_width();
  int& dh = res->get_height();
  int dw2 = dw / 2;
  int dh2 = dh / 2;
  for (int y = 0; y < dh; ++y) 
    {
      dest = (RGB32pixel*) res->get_pixel(0,y);
      for (int x = 0; x < dw; ++x) 
	{
	  int x1 = x - dw2;
	  int y1 = y - dh2;

	  x_from = (int) (centre_x + x1 * cosa + y1 * sina);
	  y_from = (int) (centre_y + y1 * cosa - x1 * sina);

	  if ((x_from < 0) || (y_from < 0) || (x_from >= width) 
	      || (y_from >= height))
	    dest->red = dest->green = dest->blue = 0;
	  else
	    {
	      src = (RGB32pixel*) get_pixel(x_from, y_from);
	      dest->red = src->red;
	      dest->green = src->green;
	      dest->blue = src->blue;
	    }
	  dest++;
	}
    }
  return (res);
}


Image* RGB32Image::map_intensities(PntGrayMap gmap, Image* res)
{
  if (res == NULL) res = new RGB32Image(width, height);
  RGB32pixel *idat = (RGB32pixel*) data, 
    *odat = (RGB32pixel*) res->get_data(),
    *enddat = (RGB32pixel*) endofdata();
  for (;idat < enddat; idat++)
    {
      odat->red = gmap[idat->red];
      odat->green = gmap[idat->green];
      odat->blue = gmap[idat->blue];
      odat++;
    }
  return (res);
}

RGB32Image* RGB32Image::rgb_read_image(RGB32Image* res)
{

#ifdef gl_interface
  
  long xsize, ysize;
  winset(glwin);
  getsize(&xsize, &ysize);

  if (res == NULL) res = new RGB32Image(xsize, ysize);

  lrectread(0,0,xsize-1,ysize-1, (unsigned long*) res->get_data());

  return (res);
#else

  // to be implemented !
  fprintf(stderr,"RGB32Image::rgb_read_image not implemented in X !");
	fflush(stderr);
  abort();
  return (NULL);

#endif
}


// Stuff added by Tony Heap (ajh) 12/7/96
// Is this a bit of a hack or what (ed) !

static const int HIST_SHIFT = 2;
static const int HIST_RANGE = 256 >> HIST_SHIFT;
static const int HIST_GROUP = 1 << HIST_SHIFT;
static const int HIST_SHIFT_RED = HIST_SHIFT;
static const int HIST_SHIFT_GREEN = HIST_SHIFT*2;
static const int HIST_SHIFT_BLUE = HIST_SHIFT*3;
static const unsigned long HIST_MASK_RED = (HIST_RANGE-1) << HIST_SHIFT;
static const unsigned long HIST_MASK_GREEN = (HIST_RANGE-1) << 8+HIST_SHIFT;
static const unsigned long HIST_MASK_BLUE = (HIST_RANGE-1) << 16+HIST_SHIFT;

inline unsigned long lookup_index(unsigned long l)
{
/*
  // Easy-to-read version:
  int r = (l & HIST_MASK_RED) >> HIST_SHIFT_RED;
  int g = (l & HIST_MASK_GREEN) >> HIST_SHIFT_GREEN;
  int b = (l & HIST_MASK_BLUE) >> HIST_SHIFT_BLUE;
  unsigned long ind = r | g | b;
  return ind;
*/
  // Optimized version:
  return ((l & HIST_MASK_RED) >> HIST_SHIFT_RED) |
         ((l & HIST_MASK_GREEN) >> HIST_SHIFT_GREEN) |
	 ((l & HIST_MASK_BLUE) >> HIST_SHIFT_BLUE);
}

// RGB32Image::grayscale_using_lookup added by Tony Heap (ajh) 12/7/96
Gray8Image* RGB32Image::grayscale_using_lookup(char* lookup, Gray8Image* dest)
{
  unsigned long *idat = (unsigned long *) data;
  if (dest == NULL) dest = new Gray8Image(width, height);
  char* odat = (char*) dest->get_data();
  for (int i = 0; i < width*height; ++i) {
    unsigned long index = lookup_index(idat[i]);
    char o = lookup[index];
    odat[i] = o;
  }
  return dest;
}

void RGB32Image::set_draw_colour(unsigned char red, unsigned char green,
				       unsigned char blue)
{
  RGB32_draw_color.red = red;
  RGB32_draw_color.green = green;
  RGB32_draw_color.blue = blue;

}

void RGB32Image::set_draw_colour(unsigned char grey_value)
{
  RGB32_draw_color.red = grey_value;
  RGB32_draw_color.green = grey_value;
  RGB32_draw_color.blue = grey_value;
}

void RGB32Image::draw_horizontal(int y, int xl, int xr)
{
  RGB32pixel* pixleft = (RGB32pixel*) check_pixel(xl, y);
  RGB32pixel* pixright = (RGB32pixel*) check_pixel(xr,y);
  if ((pixright == NULL) || (pixleft == NULL)) return;
  while (pixleft <= pixright) *pixleft++ = RGB32_draw_color;
}

void RGB32Image::plot_pixel(int x, int y)
{
  RGB32pixel *pnt = (RGB32pixel*) check_pixel(x,y);
  if (pnt != NULL)
    *pnt = RGB32_draw_color;
}

