/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/***************************************************************
 * C++ source
 *
 * File : 	SourceSplitter.cc
 *
 * Module :	SourceSplitter
 *
 * Author : 	A M Baumberg (CoMIR)
 *
 * Creation Date : Mon Jun 17 12:26:15 1996 
 *
 * Comments : 	
 *
 ***************************************************************/


/* include files */

#include "SourceSplitter.h"

// SSNode -- one output source
class SSNode : public ImageSource
{

int node_id;
SourceSplitter* parent_ss;

protected:

  friend SourceSplitter;

  SSNode() : (NULL,0)  { node_id = 0; parent_ss = NULL; }

  set_node_id(int n) { node_id = n; }
  set_parent(SourceSplitter* parent) { parent_ss = parent; }

  
  ~SSNode() { current = NULL; };

  Image* get_next()
  { 
    frame_number++; 
    return (current = parent_ss->request_new_frame(node_id));
  }

};

SourceSplitter::SourceSplitter(ImageSource* src_in, int n)
{
  update_flags = new bool[n];
  no_outputs = n;
  for (int i = 0; i < n; i++) update_flags[i] = TRUE;
  output_sources = new SSNode[n];
  for (i = 0; i < n; i++)
    {
      output_sources[i].set_node_id(i);
      output_sources[i].set_parent(this);
      output_sources[i].frame_number = src_in->frame_number;
    }

  input_source = src_in;
}

SourceSplitter::~SourceSplitter()
{
  delete [] update_flags;
  delete [] output_sources;
  delete (input_source);
}

Image* SourceSplitter::request_new_frame(int id)
{
  if (update_flags[id])
    {
      // we have recieved two requests from this node
      // so it's time to update the input
      
      if (input_source->get_next() == NULL) return (NULL);

      // reset all flags
      for (int i = 0; i < no_outputs; i++)
	update_flags[i] = FALSE;
      
    }
  update_flags[id] = TRUE;
  return (input_source->get_current());


}

ImageSource* SourceSplitter::get_source(int n)
{ 
  if (n < no_outputs) 
    return((ImageSource*) &output_sources[n]);
  else return (NULL); 
}

Image* BufferSplitter::request_new_frame(int id)
{
  if (update_flags[id])
    {
      // we have recieved two requests from this node
      // so it's time to update the input

      if (batch_mode)
	{
	  // fill the buffer of Images 
	  if (!fill_buffer()) return (NULL);
	  update_flags[id] = TRUE;
	  return (output_sources[id].get_current());
	}

      if (input_source->get_next() == NULL) return (NULL);

      img_buffer[count] =
	input_source->get_current()->copy(img_buffer[count]);
      for (int i = 0; i < no_outputs; i++)
	{
	  Image* tmp_img = 
	    img_buffer[(count-i+no_outputs)%no_outputs];
	  if (tmp_img != NULL)
	    {
	      if (output_sources[i].get_current() == NULL)
		output_sources[i].set_current(tmp_img->virtual_copy());
	      
	      else
		output_sources[i].get_current()->set_data(tmp_img->get_data());
	    }

	  update_flags[i] = FALSE;
	}

      count = (count + 1) % no_outputs;
      
    }
  update_flags[id] = TRUE;
  return (output_sources[id].get_current());
}

bool BufferSplitter::fill_buffer()
{
  if (batch_mode)
    {
      for (int i = 0; i < no_outputs; i++)
	{
	  int j = no_outputs - i - 1;
	  if (input_source->get_next() == NULL) return (FALSE);
	  img_buffer[j] =
	    input_source->get_current()->copy(img_buffer[j]);
	}
      
      for (i = 0; i < no_outputs; i++)
	{
	  if (output_sources[i].get_current() == NULL)
	    output_sources[i].set_current(img_buffer[i]);
	  update_flags[i] = FALSE;
	}
    }
  else
    {
      for (int i = 0; i < no_outputs; i++)	
	request_new_frame(0);
      update_flags[0] = FALSE;
    }

  return (TRUE);
}
      
