/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/*
 * YUVdiff211.cc
 *
 * quick differencing using subsampling
 * to speed things up for video differencing
 *
 * - AMB 6/4/93
 *
 */

#include <math.h>
#include <YUVImage.h>
#include <stdio.h>

static const int BOX_SIZE = 4;
static const int ywidth = 384 - 8;
static const int hwidth = ywidth / 2;
static const int y_up = BOX_SIZE * ywidth;
static const int u_up = (BOX_SIZE) * hwidth;
static const int y_left = BOX_SIZE;
static const int u_left = BOX_SIZE / 2;
static const int y_upleft = y_left + y_up;
static const int u_upleft = u_left + u_up;
static const int y_upright = y_up - y_left;
static const int u_upright = u_up - u_left;

static inline void get_detail(unsigned char* Y1, unsigned char* Y2,
			      unsigned char* U1, unsigned char* U2,
			      unsigned char* V1, unsigned char* V2,
			      unsigned char* R, realno& threshold, int& width)
{
  int lpx,lpy;
  unsigned char* TY1, *TY2, *TU1, *TU2, *TV1, *TV2, *TR;
  Y1 -= y_upleft / 2; Y2 -= y_upleft/ 2; R -= y_upleft / 2;
  U1 -= u_upleft / 2; U2 -= u_upleft / 2;
  V1 -= u_upleft / 2; V2 -= u_upleft / 2;
  for (lpy = 0; lpy < BOX_SIZE; lpy++)
    {
      TY1 = Y1; TY2 = Y2; TU1 = U1; TU2 = U2; TV1 = V1; TV2 = V2;
      TR = R;
      int UDIFF, VDIFF;
      for (lpx = 0; lpx < u_left; lpx++)
	{
	  UDIFF = *U2++ - *U1++;
	  VDIFF = *V2++ - *V1++;
	  *R++ = weight_sum(*Y2++ - *Y1++, UDIFF, VDIFF, threshold);
	  *R++ = weight_sum(*Y2++ - *Y1++, UDIFF, VDIFF, threshold);
	}
      Y1 = TY1 + ywidth; Y2 = TY2 + ywidth;
      U1 = TU1 + hwidth; U2 = TU2 + hwidth;
      V1 = TV1 + hwidth; V2 = TV2 + hwidth;
      R = TR + ywidth;
    }
}

      
Image* YUVImage211::difference(Image* image2, realno threshold, Image* res)
{
  if (this == NULL || image2 == NULL) return (NULL);
  if (image2->imgtype() != YUV211)
    {
      fprintf(stderr, " can't difference different image types! ");
      exit(1);
    }
  if (res == NULL) res = new Gray8Image(width, height);
  res->clear(CLEAR);
  unsigned char* y1 = data;
  unsigned char* u1 = U_image->get_data();
  unsigned char* v1 = V_image->get_data();
  
  unsigned char* y2 = ((YUVImage *)image2)->get_ydata();
  unsigned char* u2 = ((YUVImage *)image2)->get_udata();
  unsigned char* v2 = ((YUVImage *)image2)->get_vdata();

  unsigned char *ty1, *ty2, *tu1, *tu2, *tv1, *tv2, *trdat;
  unsigned char* edat = endofdata();
  unsigned char* rdat = res->get_data();
  unsigned char *rdat1, *rdat2, *rdat3;

  int udiff, vdiff, ydiff;
  y1 += (y_upleft / 2) * 3;; y2 += (y_upleft / 2) * 3;
  rdat += (y_upleft / 2) * 3;
  u1 += ( u_upleft / 2) * 3; u2 += ( u_upleft / 2) * 3;
  v1 += ( u_upleft / 2) * 3; v2 += ( u_upleft / 2) * 3;
  int lx, ly;
  int lymax = height / BOX_SIZE;
  int lxmax = ywidth / BOX_SIZE; 
  Image* flag_img = new Image(lxmax, lymax);
  lxmax--;lymax--; 
  flag_img->clear(FALSE);
  for (ly = 1; ly < lymax; ly++)
    {
      ty1 = y1; ty2 = y2; tu1 = u1; tu2 = u2; tv1 = v1; tv2 = v2;
      trdat = rdat;
      for (lx = 1; lx < lxmax; lx++)
	{
	  udiff = (*u1 - *u2);
	  vdiff = (*v1 - *v2);
	  ydiff = (*y1 - *y2); 
	  *rdat =  weight_sum(ydiff, udiff, vdiff,
			    threshold);
	  if (*rdat >= threshold)
	    {
	      // this square
	      if (!*(flag_img->get_pixel(lx,ly)))
		{
		  get_detail(y1,y2,u1,u2,v1,v2,rdat, threshold, width);
		  *(flag_img->get_pixel(lx,ly)) = TRUE;
		}
	      // left
	      if (!*(flag_img->get_pixel(lx-1,ly)))
		{
		  get_detail(y1-y_left, y2-y_left, u1 - u_left, 
			     u2 - u_left, v1 - u_left,
			     v2 - u_left, rdat-y_left,
			     threshold, width);
		  *(flag_img->get_pixel(lx-1,ly)) = TRUE;
		}
	      // up
	      if (!*(flag_img->get_pixel(lx,ly-1)))
		{
		  get_detail(y1-y_up, y2-y_up, u1-u_up, 
			     u2-u_up, v1-u_up, v2-u_up, 
			     rdat-y_up, threshold, width);
		  *(flag_img->get_pixel(lx,ly-1)) = TRUE;
		}
	      // up-left
	      if (!*(flag_img->get_pixel(lx-1,ly-1)))
		{
		  get_detail(y1-y_upleft, y2-y_upleft, u1-u_upleft,
			     u2-u_upleft, v1-u_upleft, v2-u_upleft,
			     rdat-y_upleft, threshold, width);
		  *(flag_img->get_pixel(lx-1,ly-1)) = TRUE;
		}
	      // right 
	      if (!*(flag_img->get_pixel(lx+1,ly)))
		{
		  get_detail(y1+y_left, y2+y_left, u1+u_left,
			     u2+u_left, v1+u_left, v2+u_left,
			     rdat+y_left, threshold, width);
		  *(flag_img->get_pixel(lx+1,ly)) = TRUE;
		}
	      // down
	      if (!*(flag_img->get_pixel(lx,ly+1)))
		{
		  get_detail(y1+y_up, y2+y_up, u1+u_up,
			     u2+u_up, v1+u_up, v2+u_up,
			     rdat+y_up, threshold, width);
		  *(flag_img->get_pixel(lx,ly+1)) = TRUE;
		}
	      // down-right
	      if (!*(flag_img->get_pixel(lx+1,ly+1)))
		{
		  get_detail(y1+y_upleft, y2+y_upleft, u1+u_upleft,
			     u2+u_upleft, v1+u_upleft, v2+u_upleft,
			     rdat+y_upleft, threshold, width);
		  *(flag_img->get_pixel(lx+1,ly+1)) = TRUE;
		}
	      // down-left
	      if (!*(flag_img->get_pixel(lx-1,ly+1)))
		{
		  get_detail(y1+y_upright, y2+y_upright, u1+u_upright,
			     u2+u_upright, v1+u_upright, v2+u_upright,
			     rdat+y_upright, threshold, width);
		  *(flag_img->get_pixel(lx-1,ly+1)) = TRUE;
		}
	      // up-right
	      if (!*(flag_img->get_pixel(lx+1,ly-1)))
		{
		  get_detail(y1-y_upright, y2-y_upright, u1-u_upright,
			     u2-u_upright, v1-u_upright, v2-u_upright,
			     rdat-y_upright, threshold, width);
		  *(flag_img->get_pixel(lx+1,ly-1)) = TRUE;
		}
	    } 
	  u1 += u_left; u2 += u_left;
	  v1 += u_left; v2 += u_left;
	  y1 += y_left; y2 += y_left;
	  rdat += y_left;
	}
      y1 = ty1 + y_up; y2 = ty2 + y_up;
      u1 = tu1 + u_up; u2 = tu2 + u_up;
      v1 = tv1 + u_up; v2 = tv2 + u_up;
      rdat = trdat + y_up;

    }
  delete (flag_img);
  return (res);
}
