/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/*-*-c-*-*/
/*
 * sgimgsrc.cc
 * sgi videoboard image source class
 *
 * author: A.M.Baumberg (amb)
 * creation date: 8/1/93
 */



#include <stdlib.h>
// #include <CC/osfcn.h>

#include <RGB8Image.h>
#include <Gray8Image.h>
#include <YUVImage.h>
#include <sgimgsrc.h>
#include <svideo.h>
#include <gl/gl.h>

/*
extern "C"
{
  void sginap(int);
}
*/

static struct sv_data
{
  SVhandle V;
  svCaptureInfo svci;
};


static void get_max_size(SVhandle V, int *w, int *h)
{
  long param[2];
  
  /* Determine max window size based on broadcast std and capture format */
  param[0] = SV_BROADCAST;
    if (svGetParam(V, param, 2) < 0) {
        svPerror("getparam");
        exit(1);
    }
    if (param[1] == SV_PAL) {
        *w = SV_PAL_XMAX;
        *h = SV_PAL_YMAX;
    } else {
        *w = SV_NTSC_XMAX;
        *h = SV_NTSC_YMAX;
    }
}

static void doYUVtoY(long* yuv_frame, unsigned char* y_frame,
		     int width, int height)
{
  unsigned char* y_frame_odd = y_frame + width * (height - 1);
  unsigned char* y_frame_even = y_frame + width * (height - 2);
  int width3 = 3 * width;
  for (int y = 0; y < height; y += 2)
    {
      for (int x = 0; x < width; x++)
	{
	  *y_frame_odd++ = (unsigned char) ((*yuv_frame >> 4) & 255);
	  *y_frame_even++ = (unsigned char) (((*yuv_frame++) >> 16) & 255);
	}
      y_frame_odd -= width3;
      y_frame_even -= width3;
    }
}


SGimgsrc::SGimgsrc(int bufsize, int rate, ImageType img_format, int w,
		   int h, int DO_BURST)
{
  extra_data = (void*) new sv_data;
  /* Open and initialize the IndigoVideo board */
  SVhandle& V = ((sv_data*)extra_data)->V;
  svCaptureInfo& svci = ((sv_data*)extra_data)->svci;

  if ((V = svOpenVideo()) == NULL) 
    {
      svPerror("svOpenVideo");
      exit(1);
    }
  format = img_format;
  USE_BURST = DO_BURST;
  
  long param[8];
  param[0] = SV_DITHER; param[1] = FALSE;
  param[2] = SV_COLOR;
  param[4] = SV_INPUT_BYPASS;
  param[6] = SV_FIELDDROP; param[7] = FALSE;
  if (format == GRAY8) 
    {
      param[3] = SV_MONO;
      param[5] = TRUE;
    }
  else 
    {
      param[3] = SV_DEFAULT_COLOR;
      param[5] = FALSE;
    }
  svSetParam(V, param, 8);

  if ((format != RGB8) && (format != GRAY8))
    {
      svci.format = SV_YUV411_FRAMES;
      frame_size = w * h * sizeof(long);
    }
  else 
    {
      svci.format = SV_RGB8_FRAMES;
      frame_size = w * h;
    }
  if (w == 0) get_max_size(V, &w, &h);
  svci.width = w;
  svci.height = h;
  svci.size = bufsize;
  svci.samplingrate = rate;
  frame_number = 0;
  videoframe = NULL;
  if (USE_BURST) 
    {
      int nbytes;
      svQueryCaptureBufferSize(V, &svci, &nbytes);
      videoframe = new unsigned char[nbytes];
    }
  current = NULL;
  glwin = NULLWIN;

}

void SGimgsrc::start()
{
  SVhandle& V = ((sv_data*)extra_data)->V;
  svCaptureInfo& svci = ((sv_data*)extra_data)->svci;
  if (USE_BURST)
    {
      if (svCaptureBurst(V, &svci, (void*) videoframe, NULL) < 0)
	{
	  svPerror("svCaptureBurst");
	  exit(1);
	}
    }
  else
    {
      if (svInitContinuousCapture(V, &svci) < 0) 
	{
	  svPerror("svInitContinuousCapture");
	  exit(1);
	}
    }
}

void SGimgsrc::stop()
{
  SVhandle& V = ((sv_data*)extra_data)->V;
  svCaptureInfo& svci = ((sv_data*)extra_data)->svci;
  if (!USE_BURST)
    svEndContinuousCapture(V);
}
 
Image* SGimgsrc::get_next(void)
{
  SVhandle& V = ((sv_data*)extra_data)->V;
  svCaptureInfo& svci = ((sv_data*)extra_data)->svci;
  long fieldID;
  if (frame_number == 0) 
    {
      start();
      switch (format)
	{
	case RGB8: 
	  current = new RGB8Image(svci.width, svci.height);
	  break;
	case GRAY8:
	  current = new Gray8Image(svci.width, svci.height);
	  break;
	case YUV411:
	  current = new YUVImage411(svci.width - 12, svci.height);
	  break;
	case YUV211:
	  current = new YUVImage211((svci.width >> 1) - 8, svci.height >> 1);
	  break;
	case YUV111:
	  current = new YUVImage111((svci.width >> 2) - 4, svci.height >> 2);
	  break;
	}
    }
  /* Get captured video data */
  unsigned char* video_data;
  if (USE_BURST) video_data = videoframe + 
    ((frame_number % svci.size) * frame_size);
  else 
    {
      do 
	{
	  if (svGetCaptureData(V, (void **)&video_data, &fieldID) < 0) 
	    {
	  svPerror("svGetCaptureData");
	  svEndContinuousCapture(V);
	  exit(1);
	}
	  if (video_data == NULL) 
	    sginap(1);
	} while (video_data == NULL);
    }
  switch (format)
    {
    case RGB8: 
    case GRAY8:
      svInterleaveFields(TRUE, (char *) video_data, 
			 (char*) current->get_data(),
			 svci.width, svci.height);
      break;
//    case GRAY8: 
//      doYUVtoY((long *) video_data, current->get_data(), 
//	       svci.width, svci.height);
//      break;
    case YUV411:
    case YUV211: 
    case YUV111: 
      ((YUVImage*)current)->video_to_YUV((long*) video_data);
      break;

    }
  frame_number++;
  if (USE_BURST)
    {
      if ((frame_number % (svci.size)) == 0)
	start();
    }
  else svUnlockCaptureData(V, video_data);
  return (current);
}

SGimgsrc::~SGimgsrc()
{
  SVhandle& V = ((sv_data*)extra_data)->V;
  svCaptureInfo& svci = ((sv_data*)extra_data)->svci;
  stop();
  svCloseVideo(V);
  delete [] videoframe;
  if (glwin != NULLWIN) winclose(glwin);
  delete (current);   /* delete the image*/
  delete ((sv_data*) extra_data);
}

long SGimgsrc::display(long win)
{
  SVhandle& V = ((sv_data*)extra_data)->V;
  svCaptureInfo& svci = ((sv_data*)extra_data)->svci;
  if (frame_number == 0)
    {
      glwin = win;
      if (win != NULLWIN) svBindGLWindow(V, glwin, SV_IN_REPLACE);
      else
	{
	  prefsize(svci.width, svci.height);
	  foreground();
	  glwin = winopen(" live video input ");
	  if (svBindGLWindow(V, glwin, SV_IN_REPLACE) < 0) {
	    fprintf(stderr, " sgimgsrc: svBindGLWindow error \n");
	    exit(1);
	  }
	}
      return (glwin);
    }
  else return (current->display(win));
}
