/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/*
 * sgmoviestore.cc
 * class stores images in a moviefile format
 *
 * author: A.M.Baumberg (amb)
 * creation date: 5/2/93
 */

#include "sgmoviestore.h"
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>  /* For open/read/close */
#include <stdlib.h>
#include <strings.h>

static int movie_formats[10];

MovieStore::MovieStore(char* filename, int n_frames, bool mcompress)
{
  if (mcompress)
    {
      movie_formats[RGB8] = IMAGE_FORMAT_MVC1;
      movie_formats[GRAY8] = IMAGE_FORMAT_GRAYSCALE | IMAGE_FORMAT_MVC1;
      movie_formats[YUV411] = IMAGE_FORMAT_MVC1;
      movie_formats[YUV211] = IMAGE_FORMAT_MVC1;
      movie_formats[YUV111] = IMAGE_FORMAT_MVC1;
      movie_formats[RGB32] = IMAGE_FORMAT_MVC1;
    }
  else
    {
      movie_formats[RGB8] = IMAGE_FORMAT_8RGB;
      movie_formats[GRAY8] = IMAGE_FORMAT_GRAYSCALE | 
	IMAGE_FORMAT_UNCOMPRESSED;
      movie_formats[YUV411] = movie_formats[YUV211] = movie_formats[YUV111]
	= IMAGE_FORMAT_MVC1;
      movie_formats[RGB32] = IMAGE_FORMAT_MVC1;
    }
      
  if (mcompress) compressionScheme = CL_MVC1;
  else compressionScheme = CL_UNCOMPRESSED;

  mh.frameCount = n_frames;
  mh.framesPerSec = 15;
  if ((outfd = open(filename, O_CREAT|O_RDWR|O_TRUNC, 0644)) < 0) 
    {
      fprintf(stderr, " moviestore: couldn't open file %s", filename);
      exit(1);
    }
  mh.movie_id = MOVIE_ID;
  mh.version = MOVIE_VERSION_2;
  mh.movieType = SILENT;
  mh.showType = MOVIE;
  mh.imageCsize = 4;
  mh.imageDatatype = 2;
  mh.loop_mode = LOOP_CONTINUOUSLY;
  strcpy(mh.title, filename );
  strcpy(mh.comments, " created using MovieStore class ");
  strcpy(mh.field_extensions, "field_extensions is the next area");
  strcpy(&mh.field_extensions[248], "the end");
  
  direntry = new DirectoryEntry[n_frames];
  frame_no = 0;
  if ((write(outfd,(char*) &mh, sizeof(MovieHeader))) != sizeof(MovieHeader)) 
    {
        fprintf(stderr, "moviestore: Can't write file %s", filename);
        exit(1);
    }
  if (write(outfd, (char*) direntry, (unsigned int) mh.frameCount*sizeof(DirectoryEntry))
      < mh.frameCount*sizeof(DirectoryEntry))
    {
      fprintf(stderr, "moviestore: can't write file %s", filename);
      exit(1);
    }
  compressed_data = uncompressed_data = NULL;
} 


int  MovieStore::add_a_frame(Image* new_frame)
{
  if (new_frame == NULL) return (TRUE);
  if (frame_no >= mh.frameCount) return (FALSE);
  if (frame_no == 0)
    {
      static int    paramBuf[][2] = {
	CL_IMAGE_WIDTH, 0, 
	CL_IMAGE_HEIGHT, 0, 
	CL_ORIGINAL_FORMAT, 0, 
	CL_BITS_PER_COMPONENT, 0, 
	CL_FRAME_RATE, 0 
      };
      mh.imageXsize = (long) (new_frame->get_width());
      mh.imageYsize = (long) (new_frame->get_height());
      mh.frameWidth = (short) (mh.imageXsize);
      mh.frameHeight = (short) (mh.imageYsize);
      direntry[frame_no].location = sizeof(MovieHeader) + 
	mh.frameCount*sizeof(DirectoryEntry);

      paramBuf[0][1] = mh.imageXsize;
      paramBuf[1][1] = mh.imageYsize;
      paramBuf[3][1] = 8;
      paramBuf[4][1] = 30;
      
      switch (new_frame->imgtype())
      {
      case RGB8:
	paramBuf[2][1] = CL_RGB332;
	if (compressionScheme == CL_MVC1) 
	  {
	    uncompressed_data = (unsigned char*) 
	      new long[mh.imageXsize * mh.imageYsize];
	    mh.image_format = IMAGE_FORMAT_MVC1;
	  }
	else mh.image_format = IMAGE_FORMAT_8RGB;
	break;

      case GRAY8:
	paramBuf[2][1] = CL_GRAYSCALE;
	if (compressionScheme == CL_MVC1)
	  {
	    uncompressed_data = (unsigned char*) 
	      new long[mh.imageXsize * mh.imageYsize];
	    mh.image_format = IMAGE_FORMAT_MVC1 |
	      IMAGE_FORMAT_GRAYSCALE;
	  }
	else mh.image_format = IMAGE_FORMAT_GRAYSCALE |
	  IMAGE_FORMAT_UNCOMPRESSED;
	break;

      case RGB32:
      case YUV411:
      case YUV211:
      case YUV111:
	paramBuf[2][1] = CL_RGBX;
	/*if (compressionScheme == CL_MVC1)
	  mh.image_format = IMAGE_FORMAT_MVC1;
	else mh.image_format = IMAGE_FORMAT_8RGB;*/
	compressionScheme = CL_MVC1;
	mh.image_format = IMAGE_FORMAT_MVC1;
	break;

      /*case YUV411:
      case YUV211:
      case YUV111:
	paramBuf[2][1] = CL_RGBX;
	compressionScheme = CL_MVC1;
	mh.image_format = IMAGE_FORMAT_MVC1;
	break;*/
      }

      if (compressionScheme != CL_UNCOMPRESSED)
	{
	  clOpenCompressor(compressionScheme, &compressorHdl);  
	  clSetParams(compressorHdl, (int*) paramBuf, 4);
	  compressed_data = (unsigned char*) new long[mh.imageXsize *
						    mh.imageYsize];
	}
    }
  if (isYUV(new_frame->imgtype()))
    new_frame = ((YUVImage*) new_frame)->get_rgb32();

  int videosize;
  if (compressionScheme == CL_UNCOMPRESSED)
    {
      videosize = new_frame->endofdata() -
	new_frame->get_data(); 
      if (write(outfd, (char*) new_frame->get_data(), videosize) < videosize)
	{
	  fprintf(stderr, "moviestore:bad write frame %d in movie %s\n",
	     frame_no, mh.title);
	  exit(1);
	} 
    }
  else
    {
      if (uncompressed_data != NULL)
	{
	  memcpy(uncompressed_data, new_frame->get_data(),
		 new_frame->endofdata() - new_frame->get_data());
	  clCompress(compressorHdl, 1, uncompressed_data, &videosize,
		 compressed_data);
	}
      else
	clCompress(compressorHdl, 1, new_frame->get_data(), &videosize,
		 compressed_data);
      if (write(outfd, (char*) compressed_data, videosize) < videosize)
	{
	  fprintf(stderr, "moviestore:bad write frame %d in movie %s\n",
		  frame_no, mh.title);
	  exit(1);
	}
    }

  direntry[frame_no].asize = 0;
  direntry[frame_no].vsize = videosize;
  
  frame_no++;
  if (frame_no < mh.frameCount)
    {
      direntry[frame_no].location = direntry[frame_no-1].location +
	direntry[frame_no-1].vsize;
      return (TRUE);
    }
  else 
    {
      int j;
      /* Seek back to beginning of file */
      if ((j = (int) lseek(outfd, 0, SEEK_SET)) != 0)
	{
	  fprintf(stderr, "moviestore: lseek failed %s, %d not 0\n", mh.title, j);
	  exit(1);
	}
      /* Rewrite movie header and directory */
      if ((write(outfd, (char*) &mh, sizeof(MovieHeader))) != sizeof(MovieHeader)) 
	{
	  fprintf(stderr, "moviestore: Can't write file %s", mh.title);
	  exit(1);
	}
      if (write(outfd, (char*) direntry, (unsigned) mh.frameCount*sizeof(DirectoryEntry))
	  < mh.frameCount*sizeof(DirectoryEntry)) 
	{
	  fprintf(stderr, "moviestore: Can't write file %s", mh.title);
	  exit(1);
	}
      if (compressionScheme != CL_UNCOMPRESSED)
	clCloseCompressor(compressorHdl);
      close(outfd);
      fprintf(stderr," writing movie\n");
      return (FALSE);
    }
}

MovieStore::~MovieStore()
{
  delete [] direntry;
  delete [] compressed_data;
  close(outfd);
}

