/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/*
 * diffsrc.c++
 * 
 * a source of differenced (thresholded) images taken 
 * from two input image sources
 *
 */

#include <stdio.h>
#include <diffsrc.h>
#include <image.h>
#include <Gray8Image.h>
#include <RGB8Image.h>
#include <YUVImage.h>
#include <math.h>


static const int UNDEFINED = -1;

Diffsrc::Diffsrc(ImageSource* s1, ImageSource* s2, int t )
     : (NULL,0)
{
  source1 = s1;
  source2 = s2;
  threshold1 = threshold2 = threshold3 = t;
  no_deviations = 1;
  recalc();
}

Diffsrc::Diffsrc(ImageSource* s1, ImageSource* s2, int t1, int t2,
		  int t3) : (NULL,0)
{
  source1 = s1;
  source2 = s2;
  threshold1 = t1;
  threshold2 = t2;
  threshold3 = t3;
  no_deviations = 1;
  recalc();
}

Diffsrc::Diffsrc(ImageSource* s1, ImageSource* s2, realno no_devs )
     : (NULL,0)
{
  source1 = s1;
  source2 = s2;
  threshold1 = threshold2 = threshold3 = UNDEFINED;
  no_deviations = no_devs;
  recalc();
}

Image* Diffsrc::recalc()
{
  if ( source1->get_current() == NULL || source2->get_current() == NULL)
    return (NULL);
  if (current == NULL) current = source1->get_current()->copy();
  if (threshold1 == UNDEFINED) return (recalculate_thresholds());
  switch (source1->get_current()->imgtype())
    {
    case GRAY8:
    case RGB8:
      
      current = source1->get_current()
	->difference(source2->get_current(),threshold1, current);
      break;
      
    case YUV411:
    case YUV211:
    case YUV111:
      {
      YUVImage* src1 = (YUVImage*) source1->get_current();
      YUVImage* src2 = (YUVImage*) source2->get_current();
      YUVImage* curr = (YUVImage*) current;

      ((Gray8Image*) (src1->y_image())) 
	->difference(src2->y_image(), threshold1, curr->y_image());
      src1->u_image()->difference(src2->u_image(), threshold2,
				  curr->u_image());
      src1->v_image()->difference(src2->v_image(), threshold2,
				  curr->v_image());
      /*curr->or_fields(curr->y_image());*/

      break;
    }
    }
  return (current);
}
						  
						  
Image* Diffsrc::recalculate_thresholds()
{
  if ( source1->get_current() == NULL || source2->get_current() == NULL)
    return (NULL);
  if (current == NULL) current = source1->get_current()->copy();
  realno mean, var;
  switch (source1->get_current()->imgtype())
    {
    case RGB8:
      fprintf (stderr, " Diffsrc: not yet implemented for RGB8 \n");
      break;

    case GRAY8:
      {
	Gray8Image* gcurr = (Gray8Image*) source1->get_current();
	gcurr->diff_stats(source2->get_current(), mean, var, current);
	
	realno stdev = sqrt(var);
	threshold1 = (int) (no_deviations * stdev);
	current->threshold(threshold1, current);
	break;
      }

    case YUV411:
    case YUV211:
    case YUV111:
      {
	YUVImage* curr = (YUVImage*) current
	  , *src1 = (YUVImage*) source1->get_current()
	    , *src2 = (YUVImage*) source2->get_current();

	src1->y_image()->diff_stats(src2->y_image(), mean, var, 
				    curr->y_image());
	threshold1 = (int) (no_deviations * sqrt(var));
	curr->y_image()->threshold(threshold1, curr->y_image());
	
	src1->u_image()->diff_stats(src2->u_image(), mean, var, 
				    curr->u_image());
	threshold2 = (int) (no_deviations * sqrt(var));
	curr->u_image()->threshold(threshold2, curr->u_image());
	
	src1->v_image()->diff_stats(src2->v_image(), mean, var, 
				    curr->v_image());
	threshold3 = (int) (no_deviations * sqrt(var));
	curr->v_image()->threshold(threshold3, curr->v_image());
	
	/*curr->or_fields(curr->y_image());*/

	break;
      }
    }
  
  return (current);
}
    


Image* Diffsrc::get_next()
{
  Image* im1 = source1->get_next();
  Image* im2 = source2->get_next();
  if (im1 == NULL || im2 == NULL) return (NULL);
  else return (recalc());
}

