/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/*
 * image.h
 * general image class
 * not fully implemented yet
 *
 * author : A M Baumberg
 */

// this base class could be replaced by Gray8Image 

#ifndef __IMAGE_H__
#define __IMAGE_H__

#include <kernel.h>
#include <realno.h>

#define NULL 0
#define gl_interface

#ifdef gl_interface
//#include <gl/gl.h>
#else 

// use X interface if not on sgi 
#include <X11/X.h>
#include <X11/Xlib.h>
// #include <X11/Xutil.h>
#include <X11/Xatom.h>
#endif
#include <base_pixel.h>
#include <GrayMap.h>
#include <realno.h>
#include <EnvParameter.h>

// use setenv IA_B_TO_T ON
// if image data is always stored bottom to top

// use setenv IMAGE_ZOOM 2
// to display images magnified by a factor of 2
extern EnvIntParameter image_class_zoom;

// set color -- GL / VOGL function to set the drawing color
// checks to for rgb mode

void set_color(int col);

class RGB32Image;
class Point2;

const unsigned char MARK = 255;
const unsigned char CLEAR = 0;

const long NULLWIN = 0;


// valid image types
enum ImageType { BASE, GRAY8, RGB8, RGB32, YUV411, YUV211, YUV111};
enum ImageAddressing { IA_TOP_TO_BOTTOM, IA_BOTTOM_TO_TOP, IA_UNDEFINED};

extern ImageAddressing DEFAULT_IMAGE_ADDRESSING;

void output_image_help_string(FILE* stream = stderr);

inline bool isYUV(ImageType tp) { 
  return ((tp == YUV411) || (tp == YUV211) || (tp == YUV111)); }


class Image
{
 protected:

  int width;
  int height;
  int bytes_per_pixel;
  // (x * bytes_per_pixel) should equal (x << bpp_shift)
  int bpp_shift;		
  int line_length;
  
  int new_data;
  unsigned char* data;
  unsigned char* end_data;

  int* ypoints;  // array of offsets to (0,y)
  ImageAddressing access_mode;

  realno DrawLineWidth; // for graphics routines


  char* window_title;
#ifdef gl_interface
  // use GL
  long glwin;
#else
  // use X
  Display* mydisplay;
  Window mywindow;
  XImage* myimage;
  Pixmap mypixmap;
  GC mygc;
#endif
  
public:


  // data access functions 
  inline ImageAddressing get_access_mode() { return (access_mode); }
  inline int get_width() { return width; }
  inline int get_height() { return height; }
  inline int get_bytes_per_pixel() { return bytes_per_pixel; }
  inline unsigned char* get_data() { return data; }

  inline void set_data(unsigned char* pnt)
    {
      if (new_data) delete [] data;
      data = pnt;
      end_data = data + width * height * bytes_per_pixel;
    }
  
  void reset_image_addressing(ImageAddressing ia_mode);
      
  inline unsigned char* endofdata() { return end_data; } 
  inline void set_enddata(unsigned char* e) { end_data = e; }
  void set_height(int new_height) { height = new_height; }
  void set_width(int new_width) { width = new_width; }
  void set_title(char* title) { window_title = title; }

  virtual int measure_contrast(void* p1, void* p2);

  // constructor, destructor 
  Image (int w, int h, int bpp = 1, unsigned char *img_data = NULL);
  Image () { width = height = bytes_per_pixel = 0;
	     data = end_data = NULL; window_title = NULL; }
  virtual ~Image();

  // save
  virtual void save_pm(char* filename = NULL);

  virtual void clear(unsigned char fill_char = CLEAR);

  // get pixel ; no error checking
  inline unsigned char* get_pixel(int x, int y)
    { return (data + ypoints[y] + (x << bpp_shift)); }
  inline unsigned char* nearest_pixel(realno x, realno y)
  { return (check_pixel(real_to_int(x), real_to_int(y))); }
  
  void get_pixel_coords(unsigned char* pnt, int& x, int& y);
  
  bool check_coords(int x, int y)
    { return ( (x >= 0) && (x < width) && (y >= 0) && (y < height)); }

  unsigned char* check_pixel(int x, int y);
  
  void draw_in_image();
  // lrectread the screen pixels
  void read_image() ;


  // VIRTUAL FUNCTIONS 

  // *new* graphics routines

  // virtual graphics routines
  virtual void set_draw_colour(unsigned char /* red */, 
			       unsigned char /* geen */, 
			       unsigned char /* blue */) {}

  virtual void set_draw_colour(unsigned char /* grey */ ) {}

  virtual void draw_horizontal(int /* y */, int /* xl */ , 
			       int /* xr */) {}
  
  virtual void plot_pixel(int /* x */, int /* y */) {}

  // base class graphics routines

  void draw_filled_circle(int x0, int y0, int r);
  
  void set_line_width(realno w) { DrawLineWidth = w; }

  void draw_line(int x0 ,int y0, int x1, int y1);

  void draw_circle(int xc, int yc, int r);

  // draw a filled concave polygon into the image
  void draw_filled_polygon(int nvertices, Point2* vertices);


  // misc virtual functions

  // lrectread the displayed screen pixels to get an *rgb* image
  virtual RGB32Image* rgb_read_image(RGB32Image* res = NULL);

  // get_base_pixel returns a base_pixel class for further
  // type dependent manipulation (e.g. colour subtraction)

  virtual BasePixel* get_base_pixel(int x, int y);
  virtual BasePixel* get_base_pixel(int x, int y, BasePixel* res)
    { BasePixel* tmp = get_base_pixel(x,y);
      if (res == NULL) res = tmp;
      else 
	{ tmp->clone(res); delete (tmp); }
      return (res); }

  // get_value provides a mechanism for calculation on request;
  virtual realno get_value(int x, int y) 
    { if (check_coords(x,y)) return ((realno) *get_pixel(x,y));
      else return (0); }
  virtual long display(long window = NULLWIN);

  virtual void clear_border(); 

  // crude resample 
  virtual Image* resample(int xstep, int ystep, Image* r = NULL);

  // create a physical copy of the image
  virtual Image* copy(Image* r = NULL);
  
  // return a new image of same type and dimensons
  virtual Image* copy_type(int w = 0, int h = 0);

  // virtual_copy -- returns a new wrapper
  // around the same image data
  // (not a physical copy)
  virtual Image* virtual_copy();
       
  virtual ImageType imgtype() { return (BASE); }

  Image* flip_vertically(Image* r= NULL);

  // general image processing functions 
  
  // fcombine - general function of two images
  virtual Image* fcombine(Image* image2, int (*func) (void*, void*),
			  Image* r = NULL);

  // quick difference image and threshold uses grid method
  virtual Image* difference(Image* image2, realno threshold, Image* r = NULL);

  // simple differencing and threshold 
  // threshold is always between 0 and 255
  // (the derived class rescales for different image types)
  virtual Image* simple_difference(Image* image2, realno threshold,
				   int& no_marked, Image* r = NULL);

  // difference image (no threshold) 
  virtual Image* difference(Image* image2, Image* r = NULL);

  // difference image and calculate mean, variance of differences 
  virtual Image* diff_stats(Image* img2, realno& mean, realno& variance,
			    Image* r = NULL);


  // image_blend
  // Result(x,y) = (a * I1(x,y) + b * I2(x,y)) / (a+b)
  virtual Image* image_blend(Image* image2, int a, int b, Image* r = NULL);

  // map_intensities
  // use an array to map intensity values
  virtual Image* map_intensities(PntGrayMap intensity_map, 
				 Image* res = NULL);

  // convolve with 3x3 integer array (& normalise)
  virtual Image* convolve(Kernel k, Image* r = NULL);
  
  virtual Image* blur(Image*r = NULL) { return (convolve(BLUR_KER,r)); }

  // take n'th of ordered local pixels 
  virtual Image* neighbour_order(int n, Image* r = NULL);

  // take median of 9 local pixels 
  virtual Image* middle(Image* r = NULL) { return (neighbour_order(5, r)); }
  
  // take minimum of 9 local pixels 
  virtual Image* minimum(Image* r = NULL) { return (neighbour_order(1, r)); }
  
  // take maximum of 9 local pixels 
  virtual Image* maximum(Image* r = NULL) { return (neighbour_order(9,r)); }

  // fmed - fast middle filter 
  virtual Image* fmed(Image* r = NULL);

  // sobel edge detection       
  virtual Image* sobel(int threshold, Image* r = NULL);

  // sobel filter , don't threshold 
  virtual Image* sobel(Image* r = NULL);

  // threshold image 
  virtual Image* threshold(int threshold, Image* r = NULL, 
			   int* no_marked = NULL);

  // halve width, height and blur 
  virtual Image* half_blur( Image* r = NULL);


  // mask image -- mask out image pixels
  virtual Image* mask(Image* mask, Image* r = NULL);

  
  // extract and clear the subimage  - uses top-left as origin 
  virtual Image* extract_subimage(int xmin, int xmax, int ymin, int ymax,
				  Image* r = NULL);

  virtual void paste_subimage(int xmin, int xmax, int ymin, int ymax,
			      Image* sub);

  Image* get_subimage(int xmin, int xmax, int ymin, int ymax,
		      Image* sub = NULL);

  virtual Image* to_rgb(Image* r = NULL);

  virtual void draw_rectangle(int xmin, int xmax, int ymin, int ymax,
			      int val);

  // Transform image (translate, rotate and scale) into a new image.
  // WARNING: Source & destination images must be of the same type.
  // (Added by T Heap)

  virtual Image* transform(int centre_x, int centre_y, realno angle,
                           realno scale, Image* r = NULL);

  // add noise to image
  // returns noise (square sum)

  virtual Image* add_gaussian_noise(realno sd,  realno& noise, Image*
				    r = NULL);

  virtual realno sum_square_pixels(int zero_value = 0);
  virtual realno sum_pixels();

  // wierd and wonderful ...
  //
  // median_img_array
  // pixelwise "median" over array of images'
  // storing result in *this
  virtual void median_img_array(Image** , int );


  // needs adding to vislib !!!!
  
  // plane project - pass the inverse mapping m00 ..., m22
  Image* plane_project(realno& m00, realno& m01, realno& m02,
		       realno& m10, realno& m11, realno& m12,
		       realno& m20, realno& m21, realno& m22,
		       Image* res = NULL );

};


#endif






