/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/***************************************************************
 * C++ source
 *
 * File : 	imgsrc_utils.cc
 *
 * Module :	imgsrc_utils
 *
 * Author : 	A M Baumberg (CoMIR)
 *
 * Creation Date : Thu Jun 13 13:31:58 1996 
 *
 * Comments : 	
 *
 ***************************************************************/


/* include files */

#include "imgsrc_utils.h"
#include <stdio.h>
#include <stdlib.h>
#include <MpegStream.h>
#include <RAMpegStream.h>
#include <XanimStream.h>
#include <PgmSource.h>
#include <PnmStream.h>
#include <sgimgsrc.h>
#include <sgmoviesrc.h>
#include <miscsrc.h>

static char* name_mangle(const char* filename, const char* new_extension)
{
  if ((filename == NULL) || (new_extension == NULL)) return (NULL);
  char* result =
    new char[strlen(filename) + strlen(new_extension) + 1];

  strcpy(result, filename);

  char* old_extension = strrchr(result, '.');
  if (old_extension == NULL)
    old_extension = &result[strlen(filename)];
  sprintf(old_extension,".%s", new_extension);

  return (result);
}

const char GET_FEED_HELP_STRING[] = "command line arguments :-  \n\n\
-i input_options  \n\n\
input_options is a comma separated list of options :- \n\n\
file=name	: input movie file \n\
		  (guess the type of file from the extension) \n\
mpegfile=name	: input MPEG file \n\
avifile=name	: input AVI file (or any Xanim compatible file) \n\
pnmfile=name	: set of PNM files -- this is the first  \n\
		(use '-' for stdin) \n\
mvfile=name	: input Silicon Graphics Movie file \n\
refimg=name	: PNM file containing a reference image \n \n\
RAmpeg		: Use Random Access MPEG library (caches file offsets) \n\
gray		: force grayscale input for MPEG / live input \n\
GRAY8		: force grayscale input for MPEG / live input \n\
RGB8		: use 8bit RGB input for SG Indigo video \n\
RGB32		: use 32bit RGB input for Indy video (default) \n\
YUV411		: use YUV colour input ratio 4:1:1 for Indigo video \n\
YUV211		: use YUV colour input ratio 2:1:1 for Indigo video \n\
YUV111		: use YUV colour input ratio 1:1:1 for Indigo video \n\n\
width=n		: set live video width \n\
height=n	: set live video height \n\
burst		: use Indigo video burst mode \n\
bufsize=n	: live video buffer size in frames \n\
rate=n		: live video sampling rate (1 - full, 2 - half etc) \n\
skip=n		: use every n'th frame (for movies) \n\
start=n		: skip frames 0 to n-1 \n\n\
help		: output this help message \n";


static Image* global_reference_image = NULL;

Image* get_reference_frame()
{ return (global_reference_image); }

ImageSource* get_live_or_movie_feed(int argc, char** argv)
{
  int c;
  while ((c=getopt(argc, argv, "i:")) != EOF)
    switch(c)
      {
      case 'i':
	return (get_live_or_movie_feed(optarg));
	break;
      default:
      case '?':
	fprintf(stderr,"%s %s \n", argv[0], GET_FEED_HELP_STRING);
	exit(1);
	break;
      }
  return (get_live_or_movie_feed(NULL));
}

char *myopts[] = {
#define FILENAME 0
"file",
#define MPEGFILE 1
"mpegfile",
#define PNMFILE 2
"pnmfile",
#define MVFILE 3
"mvfile",
#define OPTION_GRAY 4
"gray",
#define OPTION_GRAY8 5
"GRAY8",
#define OPTION_RGB8 6
"RGB8",
#define OPTION_RGB32 7
"RGB32",
#define OPTION_YUV411 8
"YUV411",
#define OPTION_YUV211 9
"YUV211",
#define OPTION_YUV111 10
"YUV11",
#define INPUT_WIDTH 11
"width",
#define INPUT_HEIGHT 12
"height",
#define OPTION_BURST 13
"burst",
#define OPTION_BUFSIZE 14
"bufsize",
#define OPTION_RATE 15
"rate",
#define OPTION_SKIP 16
"skip",
#define REF_IMAGE 17
"refimg",
#define AVIFILE 18
"avifile",
#define OPTION_START 19
"start",
#define OPTION_HELP 20
"help",
#define OPTION_RA_MPEG 21
"RAmpeg",
NULL };

ImageSource* get_live_or_movie_feed(char* option_string)
{
  char *ref_imgfile = NULL;
  char *filename = NULL;
  char *mpeg_fname = NULL;
  char *pnm_fname = NULL;
  char *mv_fname = NULL;
  char *avi_fname = NULL;

  ImageType format = RGB32;
  int width = 320, height = 240, bufsize = 4, frate = 1, skip = 1;
  int start_frame = 0;
  bool burst_mode = FALSE;
  bool use_ra_mpeg = FALSE;

  char *value;
  char *options = option_string;
  while ((options != NULL) && (*options != '\0'))
    {
      switch(getsubopt(&options, myopts,&value))
	{
      
	case FILENAME :
	default:
	  if (value == NULL) break;
	  char* file_ext = strrchr(value, '.');
	  if (file_ext == NULL) break;

	  if ((strcasecmp(file_ext,".mpeg") == 0) || 
	      (strcasecmp(file_ext,".mpg") == 0))
	    {
	      filename = mpeg_fname = value;
	      pnm_fname = mv_fname = avi_fname = NULL;
	    }
	  if ((strcasecmp(file_ext,".mv") == 0) ||
	      (strcasecmp(file_ext,".mvc") == 0) || 
	      (strcasecmp(file_ext,".mov") == 0) || // quicktime .mov
						   // is handled by sg library
	      (strcasecmp(file_ext,".movie") == 0))
	    {
	      filename = mv_fname = value;
	      mpeg_fname = pnm_fname = avi_fname = NULL; 
	    }
	  if (strcasecmp(file_ext,".avi") == 0)
	    {
	      filename = avi_fname = value;
	      mpeg_fname = pnm_fname = mv_fname = NULL;
	    }
	      
	  if (filename == NULL)
	    {
	      filename = pnm_fname = value;
	      mpeg_fname = mv_fname = avi_fname = NULL;
	    }
	  break;
	  
	case MPEGFILE:
	  filename = mpeg_fname = value;
	  pnm_fname = mv_fname = avi_fname = NULL;
	  break;
	  
	case PNMFILE:
	  filename = pnm_fname = value;
	  mpeg_fname = mv_fname = avi_fname = NULL;
	  break;
	  
	case MVFILE:
	  filename = mv_fname = value;
	  mpeg_fname = pnm_fname = avi_fname = NULL; 
	  break;

	case AVIFILE:
	  filename = avi_fname = value;
	  mpeg_fname = pnm_fname = mv_fname = NULL;
	  break;
	  
	case REF_IMAGE:
	  ref_imgfile = value;
	  break;
	  
	case OPTION_GRAY:
	case OPTION_GRAY8:
	  format = GRAY8;
	  break;
	  
	case OPTION_RGB8:
	  format = RGB8;
	  break;
	  
	case OPTION_RGB32:
	  format = RGB32;
	  break;
	  
	case OPTION_YUV411:
	  format = YUV411;
	  break;
	  
	  
	case OPTION_YUV211:
	  format = YUV211;
	  break;
	  
	case OPTION_YUV111:
	  format = YUV111;
	  break;
      
	case INPUT_WIDTH:
	  width = atoi(value);
	  height = (3 * width) / 4;
	  break;
	  
	case INPUT_HEIGHT:
	  height = atoi(value);
	  width = (4 * height) / 3;
	  break;
	  
	case OPTION_BURST:
	  burst_mode = TRUE;
	  break;
	  
	case OPTION_BUFSIZE:
	  bufsize = atoi(value);
	  break;
	  
	case OPTION_RATE:
	  frate = atoi(value);
	  break;
	  
	case OPTION_SKIP:
	  skip = atoi(value);
	  break; 

	case OPTION_START:
	  start_frame = atoi(value);
	  break;

	case OPTION_RA_MPEG:
	  use_ra_mpeg = TRUE;
	  break;

	case OPTION_HELP:
	  fprintf(stderr, GET_FEED_HELP_STRING);
	  exit(1);
	}
    }
  

  ImageSource* current_source = NULL;
  if (mv_fname != NULL)
    current_source = (ImageSource*) new SGmoviesrc(mv_fname);
    
  if (mpeg_fname != NULL)
    {
      if (use_ra_mpeg)
	current_source = (ImageSource*) 
	  new RAMpegStream(mpeg_fname, (format != GRAY8));
      else
	current_source = (ImageSource*)
	  new MpegStream(mpeg_fname, (format != GRAY8));
    }

  if (pnm_fname != NULL)
    {
      if (strcmp(pnm_fname, "-") == 0)
	current_source = (ImageSource*) new PnmStream(stdin);
      else
	current_source = (ImageSource*) new PgmSource(pnm_fname);
    }
	
  if (avi_fname != NULL)
    current_source = (ImageSource*) new XanimStream(avi_fname);

  if (filename != NULL)
    {
      if (start_frame > 0)
	current_source->goto_frame(start_frame);
    }

  // live video source
  if (filename == NULL)
    current_source = (ImageSource*)
    new SGimgsrc(bufsize,frate, (ImageType) format, width, height,
		 burst_mode);
  

  // set up reference image if defined
  if ((filename != NULL) && (ref_imgfile == NULL))
    ref_imgfile = name_mangle(filename, "reference");  
  if (ref_imgfile != NULL)
    {
      FILE *pnm_image_file = fopen(ref_imgfile, "r");
      if (pnm_image_file != NULL)
	{
	  global_reference_image = read_pnm(pnm_image_file);
	  fclose(pnm_image_file);
	}
    }
  
  if (skip > 1)
    current_source = new SkipSource(skip, current_source);
  
	       
  return (current_source);
}
