/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/*
 * sgmoviesrc.cc
 *  image source for movie files
 *
 */


#include <stdio.h>  /* For i/o */
#include <fcntl.h>  /* For open/read/close */
#include <stdlib.h> /* For malloc */
#include <unistd.h>
#include <sgmoviesrc.h>

extern "C" 
{ 
  #include "Movie.h" 
}

#define IMAGE_FORMAT_MVC1	    1
#define IMAGE_FORMAT_8RGB	    5
#define IMAGE_FORMAT_JPEG	    6
#define IMAGE_FORMAT_RLE	    7
#define IMAGE_FORMAT_UNCOMPRESSED   8
#define IMAGE_FORMAT_RTR	    9
#define IMAGE_FORMAT_GRAYSCALE	    16
#define IMAGE_FORMAT_YUV	    32

#include <cl.h>
#include <Gray8Image.h>
#include <RGB8Image.h>

static struct mv_extra_data
{
public:

  long decompressionScheme;
  long *rawdata; 
  long frameSize;
  DirectoryEntry *directory;
  CL_Handle decompressorHdl;
  int inFile;
};


static long	debugLevel = 0;
#define	DEBUG(a)    if(debugLevel > 0) a

/*
 * Function for reading compressed video data.
 * 
 * The movie file has a "directory" that stores the location and sizes
 * of the audio and video data. This routine reads the directory entry
 * then the video data for the specified frame. The data is read into
 * the specified buffer along with the size of the frames.
 */

static long GetCompressedVideo(long frameIndex, long size, void *data, 
			DirectoryEntry *directory, int inFile)
{
    long actualSize;
    
    DEBUG(fprintf(stderr, "%4.4ld %9.9ld,  %4.4ld,  %5.5ld  ", frameIndex, 
	directory[frameIndex].location, directory[frameIndex].asize, 
	directory[frameIndex].vsize));

    /* Read the video data which is after the audio data */
    lseek(inFile, directory[frameIndex].location + directory[frameIndex].asize, 
	SEEK_SET);
    actualSize = directory[frameIndex].vsize;
    if(actualSize <= size)
        read(inFile, data, (unsigned) actualSize);
    else
      {
	fprintf(stderr, " Frame too big ");
	exit(FAILURE);
      }
    DEBUG(fprintf(stderr, " %2.2x %2.2x %2.2x %2.2x \n", ((char *)data)[0], 
	((char *)data)[1], ((char *)data)[2], ((char *)data)[3]));
    
    return(actualSize);
}
/*
 *  image source initialisation  
 */

SGmoviesrc::SGmoviesrc(char* mv_name)
{
    MovieHeader	tmvhdr;
    long    grayscale, yuv;
    static long    paramBuf[][2] = {
	CL_IMAGE_WIDTH, 0, 
	CL_IMAGE_HEIGHT, 0, 
	CL_ORIGINAL_FORMAT, 0, 
	CL_BITS_PER_COMPONENT, 0, 
	CL_FRAME_RATE, 0, 
	CL_INTERNAL_FORMAT, 0, 
      };
    hidden = (void*) new (mv_extra_data);

    int& inFile = ((mv_extra_data*)hidden)->inFile;

    if ((inFile = open(mv_name,O_RDONLY)) == -1) {
      fprintf(stderr, ": can't open %s\n", mv_name);
      exit(FAILURE);
    }    
    
    
    /* Read movie file header */
    read(inFile, &tmvhdr, (unsigned) sizeof(MovieHeader));
    ((mv_extra_data*)hidden)->directory = (DirectoryEntry *) 
      malloc((unsigned) (sizeof(DirectoryEntry) * tmvhdr.frameCount));
    read(inFile, ((mv_extra_data*)hidden)->directory, 
	 (unsigned) (sizeof(DirectoryEntry) * tmvhdr.frameCount));
    
    width = tmvhdr.frameWidth;
    height = tmvhdr.frameHeight;
    grayscale = FALSE;
    yuv = FALSE;
    if((tmvhdr.image_format & IMAGE_FORMAT_GRAYSCALE) != 0) {
	grayscale = TRUE;
	tmvhdr.image_format &= ~IMAGE_FORMAT_GRAYSCALE;
      }
    else if((tmvhdr.image_format & IMAGE_FORMAT_YUV) != 0) {
      yuv = TRUE;
      tmvhdr.image_format &= ~IMAGE_FORMAT_YUV;
    }
    
    long &decompressionScheme =
      ((mv_extra_data*)hidden)->decompressionScheme;

    /* Open a Decompressor */
    if(tmvhdr.image_format == IMAGE_FORMAT_MVC1)
      decompressionScheme = CL_MVC1;
    else if((tmvhdr.image_format == IMAGE_FORMAT_MPEG) || 
	    (tmvhdr.image_format == 0))
      decompressionScheme = CL_MPEG_VIDEO;
    else if(tmvhdr.image_format == IMAGE_FORMAT_JPEG)
       decompressionScheme = CL_JPEG;
    else if(tmvhdr.image_format == IMAGE_FORMAT_RLE)
      decompressionScheme = CL_RLE;
    else if(tmvhdr.image_format == IMAGE_FORMAT_UNCOMPRESSED ||
	    tmvhdr.image_format == IMAGE_FORMAT_8RGB)
      decompressionScheme = CL_UNCOMPRESSED;
    else if(tmvhdr.image_format == IMAGE_FORMAT_RTR)
       decompressionScheme = CL_RTR;
    
    clOpenDecompressor(decompressionScheme,
		       &((mv_extra_data*)hidden)->decompressorHdl);
    
    paramBuf[0][1] = tmvhdr.frameWidth;
    paramBuf[1][1] = tmvhdr.frameHeight;
    if(grayscale) {
      paramBuf[2][1] = CL_GRAYSCALE;
    }
    else {
      paramBuf[2][1] = CL_RGBX;
    }
    paramBuf[3][1] = 8;
    paramBuf[4][1] = 30;
    if(yuv) {
      paramBuf[5][1] = CL_YUV422HC;
	clSetParams(((mv_extra_data*)hidden)->decompressorHdl,
		    (long *)paramBuf, sizeof(paramBuf) / sizeof(long));
    }
    else
      clSetParams(((mv_extra_data*)hidden)->decompressorHdl, 
		  (long *)paramBuf, sizeof(paramBuf) / sizeof(long) - 2);
    
    ((mv_extra_data*)hidden)->frameSize = (long) width * height * sizeof(long);
    ((mv_extra_data*)hidden)->rawdata 
      = (long *) malloc((unsigned) ((mv_extra_data*)hidden)->frameSize);
    frame_number = 0;
    no_frames = (int) tmvhdr.frameCount;
    current =  NULL;
    if (grayscale) imgformat = GRAY8;
    else imgformat = RGB32;
} 
    

Image* SGmoviesrc::get_next()
{
  long dataSize;
  frame_number++;
  int& inFile = ((mv_extra_data*)hidden)->inFile;

  if (frame_number > no_frames) return (NULL);
  if (current  == NULL)
    { 
      if (imgformat == GRAY8) current = new Gray8Image(width, height);
      else current = new RGB32Image(width, height);
    }    

  long &decompressionScheme =
      ((mv_extra_data*)hidden)->decompressionScheme;
 
  if (decompressionScheme == CL_UNCOMPRESSED)
    {
      dataSize = GetCompressedVideo(frame_number-1,
				    ((mv_extra_data*)hidden)->frameSize,
				    current->get_data(), 
				    ((mv_extra_data*)hidden)->directory, 
				    inFile);
      current->set_enddata(current->get_data() + dataSize);
    }
  else 
    {
      dataSize = GetCompressedVideo(frame_number-1,
				    ((mv_extra_data*)hidden)->frameSize,
				    ((mv_extra_data*)hidden)->rawdata,
				    ((mv_extra_data*)hidden)->directory,
				    inFile);
      clDecompress(((mv_extra_data*)hidden)->decompressorHdl, 1, dataSize, 
		   ((mv_extra_data*)hidden)->rawdata, 
		   (unsigned char*) (current->get_data()));
    }
  return (current);
}    
   

SGmoviesrc::~SGmoviesrc()
{
  free(((mv_extra_data*)hidden)->rawdata);
  free(((mv_extra_data*)hidden)->directory);
  delete current;
  clCloseDecompressor(((mv_extra_data*)hidden)->decompressorHdl);
  close(((mv_extra_data*)hidden)->inFile);
  delete (hidden);
}

