/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/*
 * sgmoviestore.cc
 * class stores images in a moviefile format
 *
 * author: A.M.Baumberg (amb)
 * creation date: 5/2/93
 */

#include <sgmoviestore.h>
#include <image.h>
#include <YUVImage.h>

#include "Movie.h"
#include "cl.h"
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>  /* For open/read/close */
#include <stdlib.h>
#include <strings.h>

// irix4 movie implementation

struct OldMovieExtras
{
  MovieHeader mh;
  DirectoryEntry *direntry;
  int outfd;
  long compressionScheme;
  CL_Handle compressorHdl;
  unsigned char* compressed_data;
  unsigned char* uncompressed_data;
};


MovieStore::MovieStore(char* filename, int n_frames, bool comp)
{ 
  movie_name = filename; 
  max_frames = n_frames; 
  compress = comp;
  OldMovieExtras* mv_data = new OldMovieExtras;
  hidden = (void*) mv_data;

  if (compress) mv_data->compressionScheme = CL_MVC1;
  else mv_data->compressionScheme = CL_UNCOMPRESSED;

  mv_data->mh.frameCount = max_frames;
  mv_data->mh.framesPerSec = 15;
  if ((mv_data->outfd = open(movie_name, O_CREAT|O_RDWR|O_TRUNC, 0644)) < 0) 
    {
      fprintf(stderr, " moviestore: couldn't open file %s", movie_name);
      exit(1);
    }
  mv_data->mh.movie_id = MOVIE_ID;
  mv_data->mh.version = MOVIE_VERSION_2;
  mv_data->mh.movieType = SILENT;
  mv_data->mh.showType = MOVIE;
  mv_data->mh.imageCsize = 4;
  mv_data->mh.imageDatatype = 2;
  mv_data->mh.loop_mode = LOOP_CONTINUOUSLY;
  strcpy(mv_data->mh.title, movie_name );
  strcpy(mv_data->mh.comments, " created using MovieStore class ");
  strcpy(mv_data->mh.field_extensions, 
	 "field_extensions is the next area");
  strcpy(&mv_data->mh.field_extensions[248], "the end");
  
  mv_data->direntry = new DirectoryEntry[max_frames];
  frame_no = 0;
  if ((write(mv_data->outfd,(char*) &mv_data->mh, sizeof(MovieHeader))) 
      != sizeof(MovieHeader)) 
    {
        fprintf(stderr, "moviestore: Can't write file %s", movie_name);
        exit(1);
    }
  if (write(mv_data->outfd, (char*) mv_data->direntry, 
	    (unsigned int) mv_data->mh.frameCount*sizeof(DirectoryEntry))
      < mv_data->mh.frameCount*sizeof(DirectoryEntry))
    {
      fprintf(stderr, "moviestore: can't write file %s", movie_name);
      exit(1);
    }
  mv_data->compressed_data = mv_data->uncompressed_data = NULL;
} 


static void save_movie_header(OldMovieExtras* mv_data)
{
  int j;
  /* Seek back to beginning of file */
  if ((j = (int) lseek(mv_data->outfd, 0, SEEK_SET)) != 0)
    {
      fprintf(stderr, "moviestore: lseek failed %s, %d not 0\n",
	      mv_data->mh.title, j);
      exit(1);
    }
  /* Rewrite movie header and directory */
  if ((write(mv_data->outfd, (char*) &mv_data->mh, sizeof(MovieHeader))) 
      != sizeof(MovieHeader)) 
    {
      fprintf(stderr, "moviestore: Can't write file %s",
	      mv_data->mh.title);
      exit(1);
    }
  if (write(mv_data->outfd, (char*) mv_data->direntry, 
	    (unsigned) mv_data->mh.frameCount*sizeof(DirectoryEntry))
      < mv_data->mh.frameCount*sizeof(DirectoryEntry)) 
    {
      fprintf(stderr, "moviestore: Can't write file %s", 
	      mv_data->mh.title);
      exit(1);
    }
  if (mv_data->compressionScheme != CL_UNCOMPRESSED)
    clCloseCompressor(mv_data->compressorHdl);
  close(mv_data->outfd);
}

int MovieStore::add_a_frame(Image* new_frame)
{
  OldMovieExtras* mv_data = (OldMovieExtras*) hidden;
  unsigned char*& compressed_data = mv_data->compressed_data;
  unsigned char*& uncompressed_data = mv_data->uncompressed_data;
  int& frame_no = frame_no;

  if (new_frame == NULL) return (TRUE);
  if (frame_no >= mv_data->mh.frameCount) return (FALSE);
  if (frame_no == 0)
    {
      static long    paramBuf[][2] = {
	CL_IMAGE_WIDTH, 0, 
	CL_IMAGE_HEIGHT, 0, 
	CL_ORIGINAL_FORMAT, 0, 
	CL_BITS_PER_COMPONENT, 0, 
	CL_FRAME_RATE, 0 
      };
      mv_data->mh.imageXsize = (long) (new_frame->get_width());
      mv_data->mh.imageYsize = (long) (new_frame->get_height());
      mv_data->mh.frameWidth = (short) (mv_data->mh.imageXsize);
      mv_data->mh.frameHeight = (short) (mv_data->mh.imageYsize);
      mv_data->direntry[frame_no].location = sizeof(MovieHeader) + 
	mv_data->mh.frameCount*sizeof(DirectoryEntry);

      paramBuf[0][1] = mv_data->mh.imageXsize;
      paramBuf[1][1] = mv_data->mh.imageYsize;
      paramBuf[3][1] = 8;
      paramBuf[4][1] = 30;
      
      switch (new_frame->imgtype())
      {
      case RGB8:
	paramBuf[2][1] = CL_RGB332;
	if (mv_data->compressionScheme == CL_MVC1) 
	  {
	    uncompressed_data = (unsigned char*) 
	      new long[mv_data->mh.imageXsize * mv_data->mh.imageYsize];
	    mv_data->mh.image_format = IMAGE_FORMAT_MVC1;
	  }
	else mv_data->mh.image_format = IMAGE_FORMAT_8RGB;
	break;

      case GRAY8:
	paramBuf[2][1] = CL_GRAYSCALE;
	if (mv_data->compressionScheme == CL_MVC1)
	  {
	    uncompressed_data = (unsigned char*) 
	      new long[mv_data->mh.imageXsize * mv_data->mh.imageYsize];
	    mv_data->mh.image_format = IMAGE_FORMAT_MVC1 |
	      IMAGE_FORMAT_GRAYSCALE;
	  }
	else mv_data->mh.image_format = IMAGE_FORMAT_GRAYSCALE |
	  IMAGE_FORMAT_UNCOMPRESSED;
	break;

      case RGB32:
      case YUV411:
      case YUV211:
      case YUV111:
	paramBuf[2][1] = CL_RGBX;
	mv_data->compressionScheme = CL_MVC1;
	mv_data->mh.image_format = IMAGE_FORMAT_MVC1;
	break;

      }

      if (mv_data->compressionScheme != CL_UNCOMPRESSED)
	{
	  clOpenCompressor(mv_data->compressionScheme, 
			   &mv_data->compressorHdl);  
	  clSetParams(mv_data->compressorHdl, (long*) paramBuf, 4);
	  compressed_data = (unsigned char*) 
	    new long[mv_data->mh.imageXsize * mv_data->mh.imageYsize];
	}
    }
  if (isYUV(new_frame->imgtype()))
    new_frame = ((YUVImage*) new_frame)->get_rgb32();

  long videosize;
  if (mv_data->compressionScheme == CL_UNCOMPRESSED)
    {
      videosize = new_frame->endofdata() - new_frame->get_data(); 
      if (write(mv_data->outfd, (char*) new_frame->get_data(),videosize)
	  < videosize)
	{
	  fprintf(stderr, "moviestore:bad write frame %d in movie %s\n",
	     frame_no, mv_data->mh.title);
	  exit(1);
	} 
    }
  else 
    {
      if (uncompressed_data != NULL)
	{
	  memcpy(uncompressed_data, new_frame->get_data(),
		 new_frame->endofdata() - new_frame->get_data());
	  clCompress(mv_data->compressorHdl, 1, uncompressed_data,
		     &videosize, compressed_data);
	}
      else
	clCompress(mv_data->compressorHdl, 1, new_frame->get_data(),
		   &videosize, compressed_data);
      if (write(mv_data->outfd, (char*) compressed_data, videosize) 
	  < videosize)
	{
	  fprintf(stderr, "moviestore:bad write frame %d in movie %s\n",
		  frame_no, mv_data->mh.title);
	  exit(1);
	}
    }

  mv_data->direntry[frame_no].asize = 0;
  mv_data->direntry[frame_no].vsize = videosize;
  
  frame_no++;
  if (frame_no < mv_data->mh.frameCount)
    {
      mv_data->direntry[frame_no].location = 
	mv_data->direntry[frame_no-1].location +
	mv_data->direntry[frame_no-1].vsize;
      return (TRUE);
    }
  else 
    {
      save_movie_header(mv_data);
      return (FALSE);
    }
}

MovieStore::~MovieStore()
{
  OldMovieExtras* mv_data = (OldMovieExtras*) hidden;
  if (frame_no < max_frames)
    {
      for (; frame_no < mv_data->mh.frameCount; frame_no++)
	{
	  mv_data->direntry[frame_no].asize = 0;
	  mv_data->direntry[frame_no].vsize = 
	    mv_data->direntry[frame_no-1].vsize;
	  mv_data->direntry[frame_no].location = 
	    mv_data->direntry[frame_no-1].location;
	}
      save_movie_header(mv_data);
    }
  delete [] mv_data->direntry;
  delete [] mv_data->compressed_data;
  close(mv_data->outfd);
  delete mv_data;
}


