/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/*-*-c-*-*/
/*
 * medsrc.c++
 * 
 * median filter (over time) to extract background
 * 
 * defined as an ImageSource
 */

#include <medsrc.h>
#include <YUVImage.h>

// use byte mask, motion mask image
static void update_median(unsigned char *image, unsigned char *median, unsigned
			  char* nearest, unsigned char *counts, unsigned char*
			  enddat, int rl, unsigned char mask, 
			  unsigned char *motion, int mskip_h,
			  int mskip_v, int mwidth);
// use byte mask 
static void update_median(unsigned char *image, unsigned char *median,
			  unsigned char *nearest, unsigned char *counts,
			  unsigned char *enddat, int rl, unsigned char mask);

// no byte mask, use motion mask
static void update_median(unsigned char *image, unsigned char *median,
			  unsigned char *nearest, unsigned char *counts,
			  unsigned char *enddat, int rl,
			  unsigned char *motion, int mskip_h,
			  int mskip_v, int mwidth);

// no byte mask, no motion mask
static void update_median(unsigned char *image, unsigned char *median,
			  unsigned char *nearest, unsigned char *counts,
			  unsigned char *enddat, int rl);


static const unsigned char ZERO = 128;


static void update_median(unsigned char *image, unsigned char *median, unsigned
			  char* nearest, unsigned char *counts, unsigned char*
			  enddat, int rl, unsigned char mask,
			  unsigned char *motion, int mskip_h,
			  int mskip_v, int mwidth)
{


  if (motion == NULL) 
    {
      (update_median(image, median, nearest, counts, enddat, rl, mask));
      return;
    }

  unsigned char pix, med;
  int mcount1 = mskip_h;
  int mcount2 = mskip_v;
  unsigned char *old_motion = motion;
  unsigned char *new_motion = motion + mwidth;
  for (; image < enddat; image++)
    {
      if (*motion == MARK)
	{ 
	  pix = *image & mask;
	  med = *median & mask;
	  if (pix == *median) (*counts = ZERO);
	  else {
	    if (*counts < ZERO)
	      {
		if (pix < med)
		  {
		    (*counts)++;
		    if (pix > (*nearest & mask)) 
		      *nearest = (*nearest & ~mask) | pix;
		    
		    if (*counts == ZERO)
		      *median = (*median & ~mask) | *nearest;
		  }
		else *counts = ZERO;
	      }
	    else if (*counts > ZERO)
	      {
		if (pix > med)
		  {
		    (*counts)--;
		    if (pix < (*nearest & mask)) 
		      *nearest = (*nearest & ~mask) | pix;
		    if (*counts == ZERO)
		      *median = (*median & ~mask) | *nearest;
		  }
		else  *counts = ZERO;
	      }
	    else 
	      {
		if (pix < med)
		  {
		    *counts = ZERO - rl;
		    *nearest = (*nearest & ~mask) | pix;
		  }
		else if (pix > med)
		  {
		    *counts = rl + ZERO;
		    *nearest = (*nearest & ~mask) | pix;
		  }
	      }
	  }
	}
      counts++;
      median++;
      nearest++;

      if (--mcount1 == 0)
	{
	  mcount1 = mskip_h;
	  if (++motion == new_motion)
	    {
	      if (--mcount2 == 0)
		{
		  mcount2 = mskip_v;
		  old_motion = motion;
		  new_motion = motion + mwidth;
		}
	      else motion = old_motion;
	    }
	}
	  
    }
}
		     


static void update_median(unsigned char *image, unsigned char *median,
		     unsigned char *nearest, unsigned char *counts,
		     unsigned char *enddat, int rl, unsigned char mask)
{
  unsigned char pix, med;
  for (; image < enddat; image++)
    {
      pix = *image & mask;
      med = *median & mask;
      if (pix == *median) (*counts = ZERO);
      else {
	if (*counts < ZERO)
	  {
	    if (pix < med)
	      {
		(*counts)++;
		if (pix > (*nearest & mask)) *nearest = (*nearest & ~mask) | pix;
		if (*counts == ZERO) *median = (*median & ~mask) | *nearest;
	      }
	    else *counts = ZERO;
	  }
	else if (*counts > ZERO)
	  {
	    if (pix > med)
	      {
		(*counts)--;
		if (pix < (*nearest & mask)) *nearest = (*nearest & ~mask) | pix;
		if (*counts == ZERO) *median = (*median & ~mask) | *nearest;
	      }
	    else  *counts = ZERO;
	  }
	else 
	  {
	    if (pix < med)
	      {
		*counts = ZERO - rl;
		*nearest = (*nearest & ~mask) | pix;
	      }
	    else if (pix > med)
	      {
		*counts = rl + ZERO;
		*nearest = (*nearest & ~mask) | pix;
	      }
	  }
      }
      counts++;
      median++;
      nearest++;
    }
}


static void update_median(unsigned char *image, unsigned char *median,
			  unsigned char *nearest, unsigned char *counts,
			  unsigned char *enddat, int rl, 
			  unsigned char *motion, int mskip_h,
			  int mskip_v, int mwidth)
{
  
  if (motion == NULL)
    {
      update_median(image, median, nearest, counts, enddat, rl);
      return;
    }
  unsigned char pix, med;
  int mcount1 = mskip_h;
  int mcount2 = mskip_v;
  unsigned char *old_motion = motion;
  unsigned char *new_motion = motion + mwidth;
  for (; image < enddat; image++)
    {
      if (*motion == MARK)
	{
	  pix = *image;
	  med = *median;
	  if (pix == *median) (*counts = ZERO);
	  else {
	    if (*counts < ZERO)
	      {
		if (pix < med)
		  {
		    (*counts)++;
		    if (pix > *nearest) *nearest = pix;
		    if (*counts == ZERO) *median = *nearest;
		  }
		else *counts = ZERO;
	      }
	    else if (*counts > ZERO)
	      {
		if (pix > med)
		  {
		    (*counts)--;
		    if (pix < *nearest) *nearest = pix;
		    if (*counts == ZERO) *median = *nearest;
		  }
		else  *counts = ZERO;
	      }
	    else 
	      {
		if (pix < med)
		  {
		    *counts = ZERO - rl;
		    *nearest = pix;
		  }
		else if (pix > med)
		  {
		    *counts = rl + ZERO;
		    *nearest = pix;
		  }
	      }
	  }
	}
      counts++;
      median++;
      nearest++;  
      if (--mcount1 == 0)
	{
	  mcount1 = mskip_h;
	  if (++motion == new_motion)
	    {
	      if (--mcount2 == 0)
		{
		  mcount2 = mskip_v;
		  old_motion = motion;
		  new_motion = motion + mwidth;
		}
	      else motion = old_motion;
	    }
	}
    }
}
		     
     
static void update_median(unsigned char *image, unsigned char *median,
		     unsigned char *nearest, unsigned char *counts,
		     unsigned char *enddat, int rl)
{
  unsigned char pix, med;
  for (; image < enddat; image++)
    {
      pix = *image;
      med = *median;
      if (pix == *median) (*counts = ZERO);
      else {
      if (*counts < ZERO)
	{
	  if (pix < med)
	    {
	      (*counts)++;
	      if (pix > *nearest) *nearest = pix;
	      if (*counts == ZERO) *median = *nearest;
	    }
	  else *counts = ZERO;
	}
      else if (*counts > ZERO)
	{
	  if (pix > med)
	    {
	      (*counts)--;
	      if (pix < *nearest) *nearest = pix;
	      if (*counts == ZERO) *median = *nearest;
	    }
	  else  *counts = ZERO;
	}
      else 
	{
	  if (pix < med)
	    {
	      *counts = ZERO - rl;
	      *nearest = pix;
	    }
	  else if (pix > med)
	    {
	      *counts = rl + ZERO;
	      *nearest = pix;
	    }
	}
    }
      counts++;
      median++;
      nearest++;
    }
}
		     

Medsrc::Medsrc(ImageSource* s, int rl, Image* back) : PipeSource(s)
{
  //in_source = s;
  run_length = rl;
  counts_red = counts_green = counts_blue  = nearest = NULL;
  motion_mask = NULL;
  if (back != NULL)
    {
      ImageSource in2(back);
      in = &in2;
      refresh();
      in = s; 
      in2.set_current(NULL);	// prevent back from being deleted
    }
  refresh();
}

void Medsrc::restart()
{

  if (counts_green != NULL) counts_green->clear(ZERO);
  if (counts_blue != NULL) counts_blue->clear(ZERO);
  if (counts_red != NULL) counts_red->clear(ZERO);

  Image* in_curr = in->get_current();
  if (in_curr != NULL)
    current = in_curr->copy(current);
          
}
  
Image* Medsrc::recalc()
{
  Image* in_curr = in->get_current();
  if (in_curr == NULL) return (NULL);
  int w = in_curr->get_width(), h =  in_curr->get_height(); 
  unsigned char *motion = NULL;
  int mskip_h = 0, mskip_v = 0, mwidth = 0;
  if (motion_mask != NULL) 
    {
      motion = motion_mask->get_data();
      mwidth = motion_mask->get_width() * motion_mask->get_bytes_per_pixel();
      mskip_h = (in_curr->get_width() * in_curr->get_bytes_per_pixel())
	/ (mwidth);
      mskip_v = h / motion_mask->get_height();
    }

  switch(in_curr->imgtype())
    {
    case GRAY8:
    case RGB32:
      {
	if (current == NULL)
	  {
	    current = in_curr->copy();
	    counts_red = in_curr->copy_type();
	    nearest = in_curr->copy_type();
	    counts_red->clear(ZERO);	
	  }    
	else
	  {
	    update_median (in_curr->get_data(), current->get_data(), 
			   nearest->get_data(), counts_red->get_data(),
			   in_curr->endofdata(), run_length, motion,
			   mskip_h, mskip_v, mwidth);
	  }
	break;
      }

    case RGB8:
      {
	if (current == NULL)
	  {
	    current = in_curr->copy();
	    counts_red = new Gray8Image(w,h);
	    counts_green = new Gray8Image(w, h);
	    counts_blue = new Gray8Image(w, h);
	    counts_red->clear(ZERO);
	    counts_blue->clear(ZERO);
	    counts_green->clear(ZERO);
	    nearest = new RGB8Image(w,h);
	  }
	else 
	  {
	    update_median (in_curr->get_data(), current->get_data(), 
			   nearest->get_data(), counts_red->get_data(),
			   in_curr->endofdata(), run_length, RED_MASK,
			   motion, mskip_h, mskip_v, mwidth);
	    update_median (in_curr->get_data(), current->get_data(), 
			   nearest->get_data(), counts_green->get_data(),
			   in_curr->endofdata(), run_length, GREEN_MASK, 
			   motion, mskip_h, mskip_v, mwidth);
	    update_median (in_curr->get_data(), current->get_data(), 
			   nearest->get_data(), counts_blue->get_data(),
			   in_curr->endofdata(), run_length, BLUE_MASK,
			   motion, mskip_h, mskip_v, mwidth);
	  }
	break;
      }


    case YUV411:
    case YUV211:
    case YUV111:
      {
	if (current == NULL)
	  {
	    current = in_curr->copy();
	    counts_red = in_curr->copy_type();
	    counts_red->clear(ZERO);
	    nearest = in_curr->copy();
	  }
	else 
	  {
	    YUVImage* in_curr1 = (YUVImage *) in_curr;
	    YUVImage* current1  = (YUVImage *) current;
	    YUVImage* nearest1  = (YUVImage *) nearest;
	    YUVImage* counts1  = (YUVImage *) counts_red;
	    
	    
	    update_median (in_curr1->get_ydata(), current1->get_ydata(), 
			   nearest1->get_ydata(), counts1->get_ydata(),
			   in_curr1->y_image()->endofdata(), run_length,
			   motion, mskip_h, mskip_v, mwidth);
	    update_median (in_curr1->get_udata(), current1->get_udata(), 
			   nearest1->get_udata(), counts1->get_udata(),
			   in_curr1->u_image()->endofdata(), run_length,
			   motion, mskip_h, mskip_v, mwidth);
	    update_median (in_curr1->get_vdata(), current1->get_vdata(), 
			   nearest1->get_vdata(), counts1->get_vdata(),
			   in_curr1->v_image()->endofdata(), run_length,
			   motion, mskip_h, mskip_v, mwidth);
	  }
	break;
      }
    }
  return (current);
}

// Image* Medsrc::get_next()
// {
//   if (in->get_next() != NULL) return (refresh());
//   else return (NULL);
// }
  
Medsrc::~Medsrc()
{
  delete (counts_red);
  if (counts_green != NULL) delete (counts_green);
  if (counts_blue != NULL) delete (counts_blue);
  delete (nearest);
  delete (current);
}
