/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/*
 * miscsrc.h
 *
 * repackages Image processing methods as ImageSources
 *
 */

#ifndef __MISCSRC_H__
#define __MISCSRC_H__

#include <imgsrc.h>

static const int DEF_SOBEL_THRESHOLD = 20;
static const int DEF_DIFF_THRESHOLD = 10;

/****** PipeSource *****/

class PipeSource : public ImageSource
{
 protected:
  ImageSource* in;
  virtual Image* recalc() 
    { fprintf(stderr, " error in PipeSource "); return current; }
 public:
  Image* refresh() { 
    if (in->get_current() == NULL) return (NULL);
    else return(current = recalc());
  }
  Image* get_next() {
    frame_number++;
    if (in->get_next() == NULL) return (NULL);
    else return (refresh());
  }
  ImageSource* get_input_source() { return (in); }
  PipeSource(ImageSource* s1) : (NULL,0) { in = s1; }

  // use clean_up before deleteing so avoid
  // deleting the input ImageSource
  void clean_up()
  { in = NULL; }

  ~PipeSource() { delete (in); }

};
/****** resample *****/
class ResampleSource : public PipeSource
{
 protected:
  int xstep, ystep;
 public:
  Image* recalc()
    { return (in->get_current()->resample(xstep, ystep, current)); }
  ResampleSource(ImageSource* s1, int xs, int ys) : PipeSource(s1)
    { xstep = xs; ystep = ys; refresh();}
};

/****** rotate *****/
class RotateSource : public PipeSource
{
 protected:
  int xcentre, ycentre;
  realno angle, scale;
 public:
  Image* recalc()
    { 
        if ((xcentre == -1) && in->get_current() != NULL)
          xcentre = in->get_current()->get_width() / 2;
        if ((ycentre == -1) && in->get_current() != NULL)
          ycentre = in->get_current()->get_height() / 2;
        return (in->get_current()->transform(xcentre, ycentre, angle,
					     scale, current));
    }
  RotateSource(ImageSource* s1, int xc = -1, int yc =-1, realno ang
                 = 3.14 / 2, realno s = 1.0) : PipeSource(s1)
    { xcentre = xc; ycentre = yc; angle = ang; scale = s; refresh();}

  RotateSource(ImageSource* s1, realno ang, realno s = 1.0) : PipeSource(s1)
  { xcentre = ycentre = -1; angle = ang; scale = s; refresh(); }
  realno get_angle() { return angle; }
  void set_angle(realno theta) { angle = theta; }
};


/****** convolve *****/
class ConvolveSource : public PipeSource
{
 public:
  Kernel kern;
  Image* recalc() 
    { return (in->get_current()->convolve(kern, current)); }
  ConvolveSource(ImageSource* s1, Kernel k) : PipeSource(s1) 
    {kern = k; refresh();}
};
/***** blur *****/
class BlurSource : public PipeSource
{
 public:
  Image* recalc()
    { return (in->get_current()->blur(current)); }
  BlurSource(ImageSource* s1) : PipeSource(s1) { refresh();}
};
/***** half_blur *****/
class HalfBlurSource : public PipeSource
{
 public:
  Image* recalc()
    { return (in->get_current()->half_blur(current)); }
  HalfBlurSource(ImageSource* s1) : PipeSource(s1) { refresh();}
};
  
/***** middle *****/
class MiddleSource : public PipeSource
{
 public:
  Image* recalc() 
    { return (in->get_current()->middle(current)); }
  MiddleSource(ImageSource* s1) : PipeSource(s1) {refresh();}
};

/***** neighbour ordered *****/
class NeighbourSource : public PipeSource
{
  int n;

 public:
  Image* recalc() 
    { return (in->get_current()->neighbour_order(n, current)); }
  NeighbourSource(ImageSource* s1, int pos) : PipeSource(s1)
    { n = pos; refresh();}
};

/***** fmed *****/
class FMedSource : public PipeSource
{
 public:
  Image* recalc() 
    { return (in->get_current()->fmed(current)); }
  FMedSource(ImageSource* s1) : PipeSource(s1) {refresh();}
};

/***** minimum *****/
class MinimumSource : public PipeSource
{
 public:
  Image* recalc() 
    { return (in->get_current()->minimum(current)); }
  MinimumSource(ImageSource* s1) : PipeSource(s1) {refresh();}
};

/***** maximum *****/
class MaximumSource : public PipeSource
{
 public:
  Image* recalc() 
    { return (in->get_current()->maximum(current)); }
  MaximumSource(ImageSource* s1) : PipeSource(s1) {refresh();}
};

/***** sobel (thresholded) *****/
class SobelEdgeSource : public PipeSource
{
 public:
  int threshold;

  Image* recalc() 
    { return (in->get_current()->sobel(threshold, current)); }
  SobelEdgeSource(ImageSource* s1, int t = DEF_SOBEL_THRESHOLD) 
    : PipeSource (s1) 
    { threshold = t; refresh();}
};

/***** sobel (unthresholded) *****/
class SobelSource : public PipeSource
{
 public:
  Image* recalc() 
    { return (in->get_current()->sobel(current)); }
  SobelSource(ImageSource* s1) : PipeSource(s1) { refresh();}
};

/***** threshold *****/
class ThresholdSource : public PipeSource
{
 public:
  unsigned char thresh;
  int no_marked;

  Image* recalc() 
    { return (in->get_current()->threshold(thresh, current, &no_marked)); }
  ThresholdSource(ImageSource* s1, unsigned char t) : PipeSource(s1) 
    { thresh = t; refresh(); }
};

/***** sub_image *****/

class SubImageSource : public PipeSource
{
 protected:
  int xmin, xmax, ymin, ymax;
 public:
  Image* recalc() 
    { return (in->get_current()->extract_subimage(xmin, xmax, ymin, ymax)); }
  SubImageSource(ImageSource* s1, int x1, int x2, int y1, int y2)
    : PipeSource(s1) 
    { xmin = x1; xmax = x2; ymin = y1; ymax = y2; refresh(); }
};


/***** NoisySource *****/

class NoisySource : public PipeSource
{
public:
  realno sd;
  realno noise;
  realno total_noise;
  
  Image* recalc()
    { Image* ret_img = 
	in->get_current()->add_gaussian_noise(sd, noise, current);
      total_noise += noise;
      return (ret_img);
    }
  NoisySource(ImageSource* s1, int sdev) : PipeSource(s1)
    { sd = sdev; total_noise = 0; }

};
    
/***** difference *****/
class DifferenceSource : public ImageSource
{
 protected:
  ImageSource* s1;
  ImageSource* s2;
 public:
  realno threshold;

 protected:
  Image* recalc()
    { return (s1->get_current()->difference(s2->get_current(), threshold, current)); }
 public: 
  Image* refresh()
    { 
      if ((s1->get_current() == NULL) || (s2->get_current() == NULL)) 
	return(NULL);
      return (current = recalc());
    }
  Image* get_next()
  { frame_number++;    
    if ((s1->get_next() != NULL) && (s2->get_next() != NULL))
      return(refresh()); 
    else return (NULL);
  }

  DifferenceSource(ImageSource* is1, ImageSource* is2, realno t = -1)
    : ImageSource(NULL,0)
   { s1 = is1; s2 = is2; threshold = t; }
};


/******************/
class SimpleDiffSource : public ImageSource
{
 protected:
  ImageSource* s1;
  ImageSource* s2;
 public:
  realno threshold;
  int no_marked;

 protected:
  Image* recalc()
    { return (s1->get_current()->
	      simple_difference(s2->get_current(), threshold,
				no_marked, current)); }
 public: 
  Image* refresh()
    { 
      if ((s1->get_current() == NULL) || (s2->get_current() == NULL)) 
	return(NULL);
      else return (current = recalc());
    }
  Image* get_next()
  { 
    frame_number++;
    if ((s1->get_next() != NULL) && (s2->get_next() != NULL))
      return(refresh());
    else return (NULL);
  }

  SimpleDiffSource(ImageSource* is1, ImageSource* is2, realno t = -1)
    : ImageSource(NULL,0)
   { s1 = is1; s2 = is2; threshold = t; }
};

/******************/

class BlendSource : public ImageSource
{
protected:
  ImageSource* s1;
  ImageSource* s2;
  int a, b;

protected:
  Image* recalc()
  { return (s1->get_current()->image_blend(s2->get_current(), 
					   a, b, current)); }
public: 
  Image* refresh()
    { 
      if ((s1->get_current() == NULL) || (s2->get_current() == NULL)) 
	return(NULL);
      else return (current = recalc());
    }
  Image* get_next()
    { 
      frame_number++;
      if ((s1->get_next() != NULL) && (s2->get_next() != NULL))
	return(refresh()); 
      else return (NULL);
    }

  BlendSource(ImageSource* is1, ImageSource* is2, int param_a = 1,
	      int param_b = 1)
    : ImageSource(NULL,0)
  { s1 = is1; s2 = is2; a = param_a; b = param_b; }
};
/******************/


class VFlipSource : public PipeSource
{
public:
  Image* recalc()
    { return (in->get_current()->flip_vertically(current)); }
  VFlipSource(ImageSource* s1) : PipeSource(s1)
    { refresh();}
};

/******************/

// SkipSource
// sample an input ImageSource
// pass on frames 0, n, 2n, ...

class SkipSource : public ImageSource
{
  
protected:

  ImageSource* in;
  int skip;

public:

  SkipSource(int n, ImageSource* s1) : (NULL, 0)
  { in = s1; skip = n; }

  Image* get_next() 
    { 
      in->goto_frame(skip * frame_number);
      current = in->get_next();
      frame_number++;
      return (current);
    }
  
  void clean_up()
  { current = NULL; in = NULL; }

  ~SkipSource() { current = NULL; delete (in); }
  
};

// RepeatSource
// temporally re-sample an input ImageSource
// pass on each frame n times


class RepeatSource : public ImageSource
{
  
protected:

  ImageSource* in;
  int repeat;
  int count;

public:

  RepeatSource(int n, ImageSource* s1) : (NULL, 0)
  { in = s1; repeat = n; count = 0; }

  Image* get_next() 
  { 
    if ((count == 0) && (in->get_next() == NULL)) return (NULL);
    count = (count + 1) % repeat;
    current = in->get_current();
    frame_number++;
    return (current);
  }

  void clean_up()
  { current = NULL; in = NULL; }

  ~RepeatSource() { current = NULL; delete (in); }
  
};

/******************/

// SingleImageSource
// first call to get_next returns an image
// subsequent calls return NULL

class SingleImageSource : public ImageSource
{
public:
  
  SingleImageSource(Image* frame) : (frame, 0) { }
  
  Image* get_next()
  {
    if (frame_number++ == 0) return (current);
    return (NULL);
  }
};


/******************/

class ConcatSource : public ImageSource
{
protected:

  ImageSource** in_array;
  int no_sources;
  int current_source;
public:

  ConcatSource(ImageSource* s1, ImageSource* s2)  : (NULL,0)
  { 
    in_array = new ImageSource*[2]; 
    in_array[0] = s1;
    in_array[1] = s2;
    no_sources = 2;
    current_source = 0;
  }
   
  ConcatSource(ImageSource** ia, int n) : (NULL,0)
  { in_array = ia; no_sources = n; current_source = 0; }
    
  Image* get_next()
  {
    if (current_source >= no_sources) return (NULL);
    Image* curr_img = in_array[current_source]->get_next();
    if ((curr_img == NULL) && (++current_source < no_sources))
      curr_img = in_array[current_source]->get_next();
    
    if (curr_img != NULL)
      {
	if (current == NULL) current = curr_img->virtual_copy();
	if (curr_img->imgtype() != current->imgtype())
	  {
	    fprintf(stderr,"can't concat different image types \n");
	fflush(stderr);
	    abort();
	  }
	if (current->get_data() != curr_img->get_data())
	  current->set_data(curr_img->get_data());
      }
    else return (NULL);
    frame_number++;
    return (current);
  }

  void clean_up()
  { in_array = NULL; current = NULL; }

  ~ConcatSource()
  {
    for (int i = 0; i < no_sources; i++)
      delete in_array[i];
    delete [] in_array;
  }

};

/******************/

class GammaCorrectSource: public PipeSource
{
public:

  GammaCorrectMap gamma_map;
  Image* recalc() 
    { return (in->get_current()->map_intensities(gamma_map,current)); }
  GammaCorrectSource(ImageSource* s1, realno gamma_value) : 
    PipeSource(s1), gamma_map(gamma_value)
  { refresh();}
};

/******************/

#include <medsrc.h>
#endif





