/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/*
 * moviesrc.cc
 * image source for uncompressed movie files
 *
 */

extern "C"
{

#include <stdio.h>  /* For i/o */
#include <fcntl.h>  /* For open/read/close */
#include <stdlib.h> /* For malloc */
#include <unistd.h>

}

#include <moviesrc.h>
#include <Gray8Image.h>
#include <RGB8Image.h>

static long	debugLevel = 0;
#define	DEBUG(a)    if(debugLevel > 0) a

static long imageFormatTable[6] = {IMAGE_FORMAT_UNCOMPRESSED, IMAGE_FORMAT_RLE, 
    IMAGE_FORMAT_JPEG, IMAGE_FORMAT_MPEG, IMAGE_FORMAT_MVC1, IMAGE_FORMAT_RTR};

/*
 * Function for reading compressed video data.
 * 
 * The movie file has a "directory" that stores the location and sizes
 * of the audio and video data. This routine reads the directory entry
 * then the video data for the specified frame. The data is read into
 * the specified buffer along with the size of the frames.
 */

static long GetVideo(long frameIndex, long size, void *data, 
			DirectoryEntry *directory, int inFile)
{
    long actualSize;
    
    DEBUG(fprintf(stderr, "%4.4ld %9.9ld,  %4.4ld,  %5.5ld  ", frameIndex, 
	directory[frameIndex].location, directory[frameIndex].asize, 
	directory[frameIndex].vsize));

    /* Read the video data which is after the audio data */
    lseek(inFile, directory[frameIndex].location + directory[frameIndex].asize, 
	SEEK_SET);
    actualSize = directory[frameIndex].vsize;
    if(actualSize <= size)
        read(inFile, (char*) data, (unsigned) actualSize);
    else
      {
	fprintf(stderr, " Frame too big ");
	exit(1);
      }
    DEBUG(fprintf(stderr, " %2.2x %2.2x %2.2x %2.2x \n", ((char *)data)[0], 
	((char *)data)[1], ((char *)data)[2], ((char *)data)[3]));
    
    return(actualSize);
}
/*
 *  image source initialisation  
 */

MovieSource::MovieSource(char* mv_name)
{
    MovieHeader	tmvhdr;
    if ((inFile = open(mv_name,O_RDONLY)) == -1) {
      fprintf(stderr, ": can't open %s\n", mv_name);
      exit(1);
    }    

    /* Read movie file header */
    read(inFile, (char*) &tmvhdr, (unsigned) sizeof(MovieHeader));
    directory = (DirectoryEntry *) 
      malloc((unsigned) (sizeof(DirectoryEntry) * tmvhdr.frameCount));
    read(inFile, (char*) directory, 
	 (unsigned) (sizeof(DirectoryEntry) * tmvhdr.frameCount));
    
    width = tmvhdr.frameWidth;
    height = tmvhdr.frameHeight;
    int tmp = (int) tmvhdr.image_format;
    if ((tmvhdr.image_format & IMAGE_FORMAT_8RGB) != 0) imgformat = RGB8;
    else if ((tmvhdr.image_format & IMAGE_FORMAT_GRAYSCALE) != 0)
      imgformat = GRAY8;
    else
      {
	fprintf(stderr, "moviesrc: bad image format = %d \n", tmp);
	exit(1);
      }
	     
    frameSize = (long) width * height * sizeof(long);
    rawdata = (long *) malloc((unsigned) frameSize);
    frame_number = 0;
    no_frames = (int) tmvhdr.frameCount;
    current =  NULL;
} 
    

Image* MovieSource::get_next()
{
  long dataSize;
  if (frame_number++ >= no_frames) return (NULL);
  if (current  == NULL)
    { 
      if (imgformat == GRAY8) current = new Gray8Image(width, height);
      else current = new RGB8Image(width, height);
    }    
  dataSize = GetVideo(frame_number-1, frameSize,
				current->get_data(), directory, inFile);
  current->set_enddata(current->get_data() + dataSize);
  return (current);
}    
   

MovieSource::~MovieSource()
{
  free((char*) rawdata);
  free((char*) directory);
  delete current;
  close(inFile);
}









