/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/***************************************************************
 * C++ source
 *
 * File : 	KalmanFilter.cc
 *
 * Module :	KalmanFilter
 *
 * Author : 	A M Baumberg
 *
 * Creation Date : Mon Dec  6 12:53:14 1993 
 *
 * Comments : 	various Kalman Filter classes
 *
 ***************************************************************/


/* include files */

#include <boolean.h>
#include "KalmanFilter.h"
#include <NagMatrix.h>

#include <stdio.h>

//#define USE_OVERWEIGHTING1
#define USE_OVERWEIGHTING2
//#define USE_OVERWEIGHTING3

DynamicKalmanOneD::DynamicKalmanOneD(realno Q11, realno Q12, realno Q22,
			     realno r_0, realno b_0, realno
			     db_0, realno p11_0, realno p12_0, realno
			     p22_0 , realno delta_t, realno eps)
{
  
  q11 = Q11; q12 = Q12; q22 = Q22;
  old_p11 = p11 = p11_0;
  old_p12 = p12 = p12_0;
  old_p22 = p22 = p22_0;
  dt = delta_t;
  old_b = b = b_0;
  old_db = db = db_0;
  def_r = r_0;
  t = 0;
  epsilon = eps;

}


realno DynamicKalmanOneD::get_prediction(realno delta_t)
{
  b += delta_t * db;
  t += delta_t;
  realno new_p11 =  p11 + delta_t * (2 * p12 + delta_t * p22) +
    delta_t * q11;
  realno new_p12 =  p12 + (delta_t * p22) + delta_t * q12;
  realno new_p22 = p22 + (delta_t * q22);
  old_p11 = p11 = new_p11;
  old_p12 = p12 = new_p12;
  old_p22 = p22 = new_p22;
  old_b = b;
  old_db = db;
  return (b);
}

realno DynamicKalmanOneD::add_observation(realno b_obs, realno r)
{
  old_b = b;
  old_db =db;
  old_p11 = p11;
  old_p12 = p12;
  old_p22 = p22;

  b_obs -= b;
  realno denom = p11 + r;

#ifdef USE_OVERWEIGHTING1

  b += ((p11 + epsilon * epsilon * r) / denom) * b_obs;

  b += (p11  / denom) * b_obs;
  db += (p12 / denom) * b_obs;
  realno new_p11 = ((p11 + epsilon * epsilon * r)  * r) / denom;

  
//   realno gain1 = p11 / denom;
//   realno q_pos = 0.5 * q22 * dt * dt * dt;
//   realno gain2 = q_pos / (q_pos + r);
//   realno gain = gain1 + epsilon * gain2;
//   b += gain * b_obs;
//   db += (p12 / denom) * b_obs;

//   realno new_p11 = r * gain;
 
#else

  b += (1.0 + epsilon) * (p11 / denom) * b_obs;
  db += (1.0 + epsilon) * (p12 / denom) * b_obs;
  realno new_p11 = (p11 * r) / denom;

#endif


  realno new_p12 = (p12 * r) / denom;
  realno new_p22 = p22 - ((p12 * p12) / denom);

 
  p11 = new_p11;
  p12 = new_p12;
  p22 = new_p22;
  return (b);
}

realno DynamicKalmanOneD::delta_update(realno delta_b, realno inv_r)
{
  if (inv_r == 0) return (0);
  old_b = b;
  old_db = db;
  old_p11 = p11;
  old_p12 = p12;
  old_p22 = p22;

  realno denom = 1 + p11 * inv_r;

  p22 -= (p12 * p12 * inv_r) / denom;
  p12 /= denom;

#ifdef USE_OVERWEIGHTING1
  p11 = (p11 / denom) + (epsilon * epsilon) / (inv_r * denom);
//   realno method1 = p11  / (1.0 + p11 * inv_r);
//   realno q_pos = 0.5 * q22 * dt * dt * dt;
//   realno method2 = q_pos / (1.0 + q_pos * inv_r);

//   p11 = (1 /* - epsilon */) * method1 + epsilon * method2;

#else
  p11 = (p11 / denom);
#endif

  
  b += p11 * delta_b;
  db += p12 * delta_b;

  // return the Kalman gain for parameter 'b'
  return (p11 * inv_r);

}

void DynamicKalmanOneD::remove_last_observation()
{
  b = old_b;
  db = old_db;
  p11 = old_p11;
  p12 = old_p12;
  p22 = old_p22;
}


void DynamicKalmanOneD::add_noise(realno var)
{
  b += sqrt(var) * normal_random();
  p11 += var;
}

/*****************************************************************************/

StaticKalmanOneD::StaticKalmanOneD(realno Q, realno r_0, realno b_0,
			     realno p11_0, realno eps)
{
  
  q = Q;
  old_p11 = p11 = p11_0;
  old_b = b = b_0;
  def_r = r_0;
  epsilon = eps;

}


realno StaticKalmanOneD::update_state()
{
  realno new_p11 =  p11 + q;
  old_p11 = p11 = new_p11;
  old_b = b;
  return (b);
}

realno StaticKalmanOneD::add_observation(realno b_obs, realno r)
{
  old_b = b; old_p11 = p11;

  b_obs -= b;
  realno denom = p11 + r;

#ifdef USE_OVERWEIGHTING2
 b += ((p11 + epsilon * epsilon * r) / denom) * b_obs;
 p11 = ((p11 + epsilon * epsilon * r)  * r) / denom;

//   realno gain1 = p11 / denom;
//   realno gain2 = q  / (q + r);
//   realno gain = gain1 + epsilon * gain2;
//   b += gain * b_obs;
//   p11 = gain * r;

#else
  b += (1.0 + epsilon) * (p11  / denom) * b_obs;
  p11 = (1.0 + epsilon) * (p11 * r) / denom;
#endif
  return (b);
}

realno StaticKalmanOneD::delta_update(realno delta_b, realno inv_r)
{
  if (inv_r == 0) return (0);

  old_b = b; old_p11 = p11;
  realno denom = 1 + p11 * inv_r;

#ifdef USE_OVERWEIGHTING2
  p11 = ((p11 + (epsilon * epsilon / inv_r))) / denom;
  
//   realno method1 = p11  / (1.0 + p11 * inv_r);
//   realno method2 = q / (1.0 + q * inv_r);

//   p11 = (1 /* - epsilon */) * method1 + epsilon * method2;

#else
  p11 = (p11 / denom);
#endif

  b += p11 * delta_b;
  return (p11 * inv_r);

}

void StaticKalmanOneD::remove_last_observation()
{
  b = old_b;
  p11 = old_p11;
}

void StaticKalmanOneD::add_noise(realno var)
{
  b += sqrt(var) * normal_random();
  p11 += var;
}

/*****************************************************************************/


StaticKalmanTwoD::StaticKalmanTwoD(realno Qxx, realno Qxy, realno Qyy , 
			     realno bx_0, realno by_0,
			     realno pxx_0, realno pxy_0, 
			     realno pyy_0) :
  Q(Qxx, Qxy, Qyy),
  b(bx_0, by_0),
  old_b(bx_0, by_0),
  Cov(pxx_0, pxy_0, pyy_0),
  old_Cov(pxx_0, pxy_0, pyy_0)
{ }


realno StaticKalmanTwoD::delta_update(realno dbx_obs, realno dby_obs,  
			 realno rxx, realno rxy, realno ryy)
{  
  old_Cov = Cov;
  old_b = b;
  
  realno det1 =  Cov.determinant();
  if (det1 < 0) 
	{
		fprintf(stderr,"Abort in StaticKalmanTwoD::delta_update det1 < 0\n");
		fflush(stderr);
		abort(); 
	}
  
  // if det1 too small, measurement has no effect so return
  if (det1 < 1e-30) return (0);
  
  Matrix2by2 R_inv(rxx,rxy,ryy);
  Matrix2by2 tmp_m = Cov.inverse() + R_inv;

  realno det2 = tmp_m.determinant();
  if (det2 < 0) 
	{
		fprintf(stderr,"Abort in StaticKalmanTwoD::delta_update det2 < 0\n");
		fflush(stderr);
		abort(); 
	}

  Cov = tmp_m.inverse();
  Point2 db_obs(dbx_obs, dby_obs);

  b += Cov.multiply(db_obs);

  Matrix2by2 K_gain = Cov * R_inv;
  // return maximum eigenvalue as upper bound on gain
  return (K_gain.get_max_real_eigenvalue());
}

Point2 StaticKalmanTwoD::add_observation(Point2 b_obs, Matrix2by2 R)
{
  Matrix2by2 tmp((Cov + R).inverse());
  Matrix2by2 K = Cov.multiply(tmp);

  Point2 db(b_obs - b);

  old_b = b;

  b += K.multiply(db);

  old_Cov = Cov;

  Cov -= K.multiply(old_Cov);
  
  return (b);
}

void StaticKalmanTwoD::update_state()
{
  Cov = Cov + Q;
  old_Cov = Cov;
}


void StaticKalmanTwoD::remove_last_observation()
{
  Cov = old_Cov;
  b = old_b;
}

void StaticKalmanTwoD::set_polar_noise_cov(realno sv, realno tv)
{
  realno bx2 = b.x * b.x;
  realno by2 = b.y * b.y;
  realno bxy = b.x * b.y;
  Q = Matrix2by2((bx2 * sv + by2 * tv),
		 (bxy * (sv - tv)),
		 (by2 * sv + bx2 * tv));
}


void StaticKalmanTwoD::add_noise(realno var_x, realno var_y)
{
  b.x += sqrt(var_x) * normal_random();
  b.y += sqrt(var_y) * normal_random();
  Cov.val00 += var_x;
  Cov.val11 += var_y;
}

realno StaticKalmanTwoD::get_uncertainty()
{ 
  // return maximum eigenvalue of the covariance matrix
  // which is a bound on the expected errors
  return (Cov.get_max_real_eigenvalue());
  
}

VibrationFilter::VibrationFilter(realno Q11, realno Q12, realno Q22,
				 realno pass_a, realno pass_b,
				 realno r_0, realno x_0, realno dx_0,
				 realno p11_0, realno p12_0, 
				 realno p22_0, realno delta_t)
{
  
  q11 = Q11; q12 = Q12; q22 = Q22;
  a = pass_a; b = pass_b;
  p11 = p11_0;
  p12 = p12_0;
  p22 = p22_0;
  dt = delta_t;
  x = x_0;
  bias = x_p = 0;
  dx = dx_0;
  def_r = r_0;
  t = 0;

}

// assume delta_t is small
// the exact solution exists in terms of exponentials
// too difficult for the moment!

realno VibrationFilter::get_prediction(realno delta_t)
{
//  bias += 0.1 * (x - x_p);
  NagMatrix thi(2,2); 
  NagVector new_x(2);
  NagVector old_x(2);

  thi.set(0,0,1); thi.set(0,1,delta_t);
  thi.set(1,0, a * delta_t); thi.set(1,1,1 + b * delta_t);
  old_x[0] = x; old_x[1] = dx;
  thi.multiply(old_x, new_x);
  x = new_x[0]; dx = new_x[1] - thi.read(1,0) * bias;

  NagMatrix tmp_P(2,2); NagMatrix tmp_Q(2,2);
  NagMatrix new_P;
  
  tmp_P.set(0,0,p11); tmp_P.set(0,1,p12);
  tmp_P.set(1,0,p12); tmp_P.set(1,1,p22);
  tmp_Q.set(0,0,q11); tmp_Q.set(0,1,q12);
  tmp_Q.set(1,0,q12); tmp_Q.set(1,1,q22);

  tmp_P.ortho_transform(thi, new_P, FALSE);
  tmp_Q.scale(delta_t, tmp_Q);
  new_P.add(tmp_Q, new_P);

  p11 = new_P.read(0,0); p12 = new_P.read(0,1);
  p22 = new_P.read(1,1);
  
  t += delta_t;
  x_p = x;
  return (x);
}

realno VibrationFilter::add_observation(realno x_obs, realno r)
{

  x_obs -= x;
  realno denom = p11 + r;

  x += (p11 / denom) * x_obs;
  dx += (p12 / denom) * x_obs;
  realno new_p11 = (p11 * r) / denom;
  realno new_p12 = (p12 * r) / denom;
  realno new_p22 = p22 - ((p12 * p12) / denom);

 
  p11 = new_p11;
  p12 = new_p12;
  p22 = new_p22;
  return (x);
}

realno VibrationFilter::delta_update(realno delta_x, realno inv_r)
{
  if (inv_r == 0) return (0);

  realno denom = 1 + p11 * inv_r;

  p22 -= (p12 * p12 * inv_r) / denom;
  p12 /= denom;
  p11 = (p11 / denom);
  x += p11 * delta_x;
  dx += p12 * delta_x;

  // return the Kalman gain for 'x'
  return (p11 * inv_r);

}


// 2D constant velocity discrete Kalman Filter
// constructor :
// diagonal of system noise passed as (q_x, q_y, q_dx, q_dy)
// diagonal of initial covariance passed as 
//	(var_x, var_y, var_dx, var_dy)
// assume all other covariance terms are zero
// initial position pos_0
// initial velocity vel_0

DynamicKalmanTwoD::DynamicKalmanTwoD(realno q_x, realno q_y,
			       realno q_dx, realno q_dy,
			       Point2 pos_0, Point2 vel_0, 
			       realno var_x, realno var_y, 
			       realno var_dx, realno var_dy,
			       realno del_t) :
  Q00(q_x, q_y), Q11(q_dx, q_dy),
  C00(var_x, var_y), C11(var_dx, var_dy),
  old_C00(var_x, var_y), old_C11(var_dx, var_dy),
  pos(pos_0), velocity(vel_0), old_pos(pos_0), old_velocity(vel_0)
{

  t = 0;
  dt = del_t;
  // nothing left toinitialise

}

void DynamicKalmanTwoD::reset_noise_covariance(realno q_x, realno q_y,
					    realno q_dx, realno q_dy)
{
  Q00 = Matrix2by2(q_x, q_y);
  Q11 = Matrix2by2(q_dx, q_dy);
}

  
  //
  // add an observation  ...
  // pos_obs  	observed value
  // sensor covariance  [ rxx rxy ]
  //			[ rxy ryy ]

Point2 DynamicKalmanTwoD::add_observation(Point2 pos_obs, realno rxx, 
				       realno rxy, realno ryy)
{
  Matrix2by2 R(rxx, rxy, ryy);
  Matrix2by2 tmp((C00 + R).inverse());
  Matrix2by2 K_pos = C00.multiply(tmp);
  Matrix2by2 K_vel = C01.multiply(tmp);

  Point2 dp(pos_obs - pos);

  old_pos = pos;
  old_velocity = velocity;

  pos += K_pos.multiply(dp);
  velocity += K_vel.multiply(dp);

  old_C00 = C00; old_C01 = C01; old_C11 = C11;

  C00 -= K_pos.multiply(old_C00);
  C01 -= K_pos.multiply(old_C01);
  C11 -= K_vel.multiply(old_C01);
}

// delta_update ...
// add an observation 
// with 
//	delta_pos = ( R^{-1} (pos_obs - pos) ) 
// and measurement inverse covariance R^{-1} given by
//
//	R^{-1} =	[ rxx rxy ]
//			[ rxy ryy ]	
// 
// return the Kalman gain

realno DynamicKalmanTwoD::delta_update(Point2 delta_pos, realno rxx, 
				    realno rxy, realno ryy)
{
  
  
  Matrix2by2 R_inv(rxx, rxy, ryy);
  if (fabs(R_inv.determinant()) < 1e-30) return (0.0);

  Matrix2by2 R(R_inv.inverse());

  Point2 dp(R.multiply(delta_pos));

  Matrix2by2 tmp((C00 + R).inverse());
  Matrix2by2 K_pos = C00.multiply(tmp);
  Matrix2by2 K_vel = C01.multiply(tmp);

  old_pos = pos;
  old_velocity = velocity;

  pos += K_pos.multiply(dp);
  velocity += K_vel.multiply(dp);

  old_C00 = C00; old_C01 = C01; old_C11 = C11;

  C00 -= K_pos.multiply(old_C00);
  C01 -= K_pos.multiply(old_C01);
  C11 -= K_vel.multiply(old_C01);
  
  return (K_pos.get_max_real_eigenvalue());
}



//
// get_prediction and update estimates ...
// delta_t 	time elapsed since last update
//

Point2 DynamicKalmanTwoD::get_prediction(realno delta_t)
{
  t += delta_t;

  pos += delta_t * velocity;
  old_pos = pos;
  
  C00 += C01.scale(2 * delta_t) + C11.scale(delta_t * delta_t)
    + Q00.scale(delta_t);
  C01 += C11.scale(delta_t);
  C11 += Q11.scale(delta_t);

  old_C00 = C00; old_C01 = C01; old_C11 = C11;

  return (pos);
}



void DynamicKalmanTwoD::remove_last_observation()
{
  C00 = old_C00; C01 = old_C01; C11 = old_C11;
  pos = old_pos; velocity = old_velocity;
}


realno DynamicKalmanTwoD::get_uncertainty()
{
  return (C00.get_max_real_eigenvalue());
}
