/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/***************************************************************
 * C - C++ Header
 *
 * File : 	KalmanFilter.h
 *
 * Module :	KalmanFilter
 *
 * Author : 	A M Baumberg
 *
 * Creation Date : Mon Dec  6 10:35:28 1993 
 *
 * Comments : 	various Kalman Filter classes
 *
 ***************************************************************/


#ifndef __KALMAN_FILTER_H__
#define __KALMAN_FILTER_H__

#include <Matrix2by2.h>


// base class for 1D filter

enum FilterType { BASE_FILT, KALM_FILT1, KALM_FILT2, VIBE_FILT };


class FilterOneD
{
public:
  virtual realno add_observation(realno, realno ) {}
  virtual realno delta_update(realno, realno ) {}
  virtual realno get_update_gain(realno) {}

  virtual realno add_observation(realno ) {}
  virtual void remove_last_observation() {}
  virtual realno update_state() {}

  virtual realno& get_state() {}
  virtual realno& get_p11 () {}
  virtual FilterType get_type() { return (BASE_FILT); }

  // clone --   returns a dynamically allocated copy of
  //		the derived class

  virtual FilterOneD* clone() {}

  // corrupts the filter state and increases the
  // covariance -- for simluated annealing
  virtual void add_noise(realno variance) {}

};


// 1D constant velocity discrete Kalman Filter
// default initial conditions  
// p11 = infinity (b unknown)
// p12 = 0
// p22 = 0, db/dt = 0 (stationary)

class DynamicKalmanOneD : public FilterOneD
{
  
public:
  
  realno p11, p12, p22; // covariance matrix
  realno old_p11, old_p12, old_p22, old_b, old_db; // temporary p11, b
  realno epsilon;	// set to zero for standard Kalman
  realno b, db; 	// current estimates
  realno q11, q12, q22;	// system noise covariance
  realno def_r; 	// default sensor variance
  realno dt;  		// time interval between measurements
  realno t; 		// time elapsed
  
  FilterOneD* clone() 
    { DynamicKalmanOneD* res = new DynamicKalmanOneD();
      *res = *this; return (res); }

      
  DynamicKalmanOneD() { p11 = p12 = p22 = epsilon = 
		      b = db = q11 = q12 = q22 = 
			def_r = dt = t = 0; }
  
  DynamicKalmanOneD(realno Q11, realno Q12, realno Q22,
		realno r_0, realno b_0 = 0, realno db_0
		= 0, realno p11_0 = 1e12, realno p12_0 = 0, realno p22_0
		= 0, realno dt = (1.0/15.0), realno eps = 0.0);
  
  //
  // add an observation  ...
  // b_obs  	observed value
  // r     	sensor variance
  //

  realno add_observation(realno b_obs, realno r);

  // delta_update ...
  // add an observation 
  // with (b_obs - b_est) * inv_r = delta_b
  // and measurement inverse covariance = inv_r
  // 
  // return the Kalman gain

  realno delta_update(realno delta_b, realno inv_r);
  realno get_update_gain(realno inv_r)
    { return ((p11 * inv_r) / (1 + p11 * inv_r)); }


  realno add_observation(realno b_obs)
    { return (add_observation(b_obs, def_r)); }

  //
  // get_prediction and update estimates ...
  // delta_t 	time elapsed since last update
  //

  realno get_prediction(realno delta_t);
  realno get_prediction()
    { return(get_prediction(dt)); }
  realno update_state() { return (get_prediction(dt)); }

  realno state_change_b() { return (b - old_b); }
  realno state_change_db() { return (db - old_db); }

  void remove_last_observation();

  realno& get_velocity() 
    { return db; }
  realno& get_state()
    { return b; }
  realno& get_p11() { return p11; }
  void add_noise(realno variance);

  FilterType get_type() { return (KALM_FILT1); }
  
};

//
// 1D discrete Kalman Filter
//

class StaticKalmanOneD : public FilterOneD
{
  
public:
  
  realno p11, old_p11;		// covariance matrix
  realno b, old_b;	 	// current estimates
  realno q;		       	// variance of system noise
  realno def_r; 		// default sensor variance
  realno epsilon;		// weight current measurement

  StaticKalmanOneD() { p11 = b = q = def_r = epsilon = 0; }
  StaticKalmanOneD(realno Q, realno r_0, realno b_0 = 0,
		realno p11_0 = 1e12, realno eps = 0);
  realno& get_state() { return b; }
  realno& get_p11() { return p11; }
  realno add_observation(realno b_obs, realno r);
  realno add_observation(realno b_obs)
    { return (add_observation(b_obs, def_r)); }
  realno delta_update(realno delta_b, realno inv_r);
  realno get_update_gain(realno inv_r)
    { return ((p11 * inv_r) / (1 + p11 * inv_r)); }
  realno state_change() { return (b - old_b); }
  realno update_state(); 
  void remove_last_observation();
  FilterType get_type() { return (KALM_FILT2); }
  FilterOneD* clone() 
    { StaticKalmanOneD* res = new StaticKalmanOneD();
      *res = *this; return (res); }
  void add_noise(realno variance);

  void set_variance(realno v) { p11 = old_p11 = v; }
  void set_state(realno x) { b = old_b = x; }


};

// 1D dynamic Kalman filter 
// (state x, dx/dt)
// with 2nd order model
// d  [x ] = [0 1][x - bias ] + noise
// dt [dx]   [a b][dx]
//
// noise has covariance Q

class VibrationFilter : public FilterOneD
{
public:
  
  realno p11, p12, p22; // covariance matrix
  realno x, dx; 	// current estimates
  realno x_p;		// predicted value of x
  realno bias;		// set to x(t-1)
  realno a,b;		// 2nd order model 
  realno q11, q12, q22;	// system noise covariance
  realno def_r; 	// default sensor variance
  realno dt;  		// time interval between measurements
  realno t; 		// time elapsed
  
  VibrationFilter()
    { p11 = p12 = p22 = a = b = 
	x = dx = bias = q11 = q12 = q22 = 
	  def_r = dt = t = 0; }
  
  VibrationFilter(realno Q11, realno Q12, realno Q22,
		  realno a, realno b,
		  realno r_0, realno x_0 = 0, realno dx_0 = 0,
		  realno p11_0 = 1e12, realno p12_0 = 0, 
		  realno p22_0 = 0, realno dt = 1.0) ;
  
  //
  // add an observation  ...
  // x_obs  	observed value
  // r     	sensor variance
  //

  realno add_observation(realno x_obs, realno r);
  realno delta_update(realno delta_x, realno inv_r);

  realno add_observation(realno x_obs)
    { return (add_observation(x_obs, def_r)); }

  //
  // get_prediction and update estimates ...
  // delta_t 	time elapsed since last update
  //

  realno get_prediction(realno delta_t);
  realno get_prediction()
    { return(get_prediction(dt)); }
  realno update_state() 
    { return (get_prediction(dt)); }
  
  realno& get_velocity() 
    { return dx; }
  realno& get_state()
    { return x; }
  realno& get_p11() { return p11; }
  FilterType get_type() { return (VIBE_FILT); }
  FilterOneD* clone() 
    { VibrationFilter* res = new VibrationFilter();
      *res = *this; return (res); }
};


/********************************************************************/
/* 2D filters                                                       */
/********************************************************************/

class FilterTwoD
{
public:

  virtual FilterTwoD* clone() 
	{ 
		fprintf(stderr,"Abort in FilterTwoD* clone\n");
		fflush(stderr);
		abort(); 
	}

  // return 2D static state
  virtual Point2& get_state() 
	{ 
		fprintf(stderr,"Abort in FilterTwoD* get_state\n");
		fflush(stderr);
		abort(); 
	}

  // return 2x2 covaraince of static part of state
  virtual Matrix2by2& get_cov() 
	{ 
		fprintf(stderr,"Abort in FilterTwoD* get_cov\n");
		fflush(stderr);
		abort(); 
	}

  // return the variance in the most uncertain direction
  // (i.e. the principal eigenvalue of the covariance matrix)
  virtual realno get_uncertainty()
	{ 
		fprintf(stderr,"Abort in FilterTwoD* get_uncertainty\n");
		fflush(stderr);
		abort(); 
	}

  // delta_update(dp, M): 
  // add an observation with ...
  // measurement covariance R = M^{-1}
  // and p_observed - p_estimate = R dp
  //
  // returns Kalman gain
  virtual realno delta_update(Point2 /*dp*/, Matrix2by2 /*M*/) { abort();}

  // add an observation 
  // (pass observed,  measurement covariance)
  // returns new state
  virtual Point2 add_observation(Point2 /*pos_obs*/, Matrix2by2 /*R*/)
	{ 
		fprintf(stderr,"Abort in FilterTwoD* add_observation\n");
		fflush(stderr);
		abort(); 
	}

  // the change in state after last observation
  virtual Point2 state_change() 
	{ 
		fprintf(stderr,"Abort in FilterTwoD* state_change\n");
		fflush(stderr);
		abort(); 
	}
};


//
// 2D disrcete Kalman filter 
//


class StaticKalmanTwoD : public FilterTwoD
{
  
public:
  
  Matrix2by2 Cov;	// 2 x 2 covariance
  Matrix2by2 old_Cov;

  Point2 b;				// current estimates
  Point2 old_b;

  Matrix2by2 Q;			       	// system noise matrix

  StaticKalmanTwoD() : Cov(), old_Cov(), b(), old_b(), Q() {}

  StaticKalmanTwoD(realno Qxx, realno Qxy, realno Qyy , 
		realno bx_0 = 0, realno by_0 = 0,
		realno pxx_0 = 1e12, realno pxy_0 = 0, 
		realno pyy_0 = 1e12);

  FilterTwoD* clone()
  { 
    StaticKalmanTwoD* res = new StaticKalmanTwoD();
    *res = *this; return (res);
  }

  realno& get_xstate() { return b.x; }
  realno& get_ystate() { return b.y; }
  Point2& get_state() { return b; }

  realno get_uncertainty();

  realno state_change_x() { return (b.x - old_b.x); }
  realno state_change_y() { return (b.y - old_b.y); }
  Point2 state_change() { return (b - old_b); }

  // update state (as before) and return maximum 
  // eigenvalue of gain matrix
  realno delta_update(realno dbx_obs, realno dby_obs,  
		      realno rxx, realno rxy, realno ryy);

  realno delta_update(Point2 db , Matrix2by2 R_inv)
  { return (delta_update(db.x, db.y, R_inv.val00, R_inv.val01,
			 R_inv.val11)); }

  Point2 add_observation(Point2 b_obs, Matrix2by2 R);

  // set_polar_noise_cov
  // sets a polar noise model covariance by linearising
  // about current estimates
  // -- input variance of relative scale noise term
  // -- and variance of angular noise term
  void set_polar_noise_cov(realno scale_var, realno theta_var);

  void update_state(); 
  void remove_last_observation();

  void add_noise(realno var_x, realno var_y);

  void set_covariance(Matrix2by2 new_cov) 
  { Cov = old_Cov = new_cov; }
  Matrix2by2& get_cov() { return Cov; }
  void set_state(Point2 x)
  { b = old_b = x; }

};


// 2D constant velocity discrete Kalman Filter
class DynamicKalmanTwoD : public FilterTwoD
{
public:
  
  // covariance matrix in 2 by 2 blocks
  Matrix2by2 C00;
  Matrix2by2 C01;
  Matrix2by2 C11;

  // old covariance matrix in 2 by 2 blocks
  Matrix2by2 old_C00;
  Matrix2by2 old_C01;
  Matrix2by2 old_C11;


  Point2 pos;		// current estimate of position
  Point2 velocity;	// current estimate of velocity
  Point2 old_pos;
  Point2 old_velocity;

  // system noise in 2 by 2 blocks 
  Matrix2by2 Q00;
  Matrix2by2 Q01;
  Matrix2by2 Q11;
  
  realno dt;  		// time interval between measurements
  realno t; 		// time elapsed
  

  // null constructor
  DynamicKalmanTwoD() : 
  pos(0,0), velocity(0,0), old_pos(0,0), old_velocity(0,0)
  { dt = t = 0; }
  
  // constructor :
  // diagonal of system noise passed as (q_x, q_y, q_dx, q_dy)
  // diagonal of initial covariance passed as 
  //	(var_x, var_y, var_dx, var_dy)
  // assume all other covariance terms are zero
  // initial position pos_0
  // initial velocity vel_0
  DynamicKalmanTwoD(realno q_x, realno q_y, realno q_dx, realno q_dy,
		 Point2 pos_0, Point2 vel_0,
		 realno var_x, realno var_y, realno var_dx, realno var_dy,
		 realno dt = (1.0/15.0));

  FilterTwoD* clone()
  {
    DynamicKalmanTwoD* res = new DynamicKalmanTwoD();
    *res = *this; return (res); 
  }


  // reset diagonal of system noise
  void reset_noise_covariance(realno q_x, realno q_y, 
			      realno q_dx, realno q_dy);


  //
  // add an observation  ...
  // pos_obs  	observed value
  // sensor covariance  [ rxx rxy ]
  //			[ rxy ryy ]

  Point2 add_observation(Point2 pos_obs, realno rxx, realno rxy,
			 realno ryy);

  Point2 add_observation(Point2 pos_obs, Matrix2by2 R)
  { return (add_observation(pos_obs, R.val00, R.val01, R.val11)); }

  realno delta_update(Point2 delta_pos, realno rxx, realno rxy, 
		      realno ryy);

  realno delta_update(Point2 delta_pos, Matrix2by2 R_i)
  {return (delta_update(delta_pos, R_i.val00, R_i.val01, R_i.val11));}
  //
  // get_prediction and update estimates ...
  // delta_t 	time elapsed since last update
  //

  Point2 get_prediction(realno delta_t);
  Point2 get_prediction()
    { return(get_prediction(dt)); }
  Point2 update_state() { return (get_prediction(dt)); }

  Point2 state_change() { return (pos - old_pos); }

  void remove_last_observation();

  Point2& get_velocity() 
    { return velocity; }
  Point2& get_state()
    { return pos; }

  void get_positional_covariance(realno& pxx, realno& pxy,
				 realno& pyy)
  {pxx = C00.val00; pxy = C00.val01; pyy = C00.val11; }

  realno get_uncertainty();

  Matrix2by2& get_cov() { return C00; }

};

#endif


  
