/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/***************************************************************
 * C - C++ Header
 *
 * File : 	MathRoutines.h
 *
 * Module :	MathRoutines
 *
 * Author : 	A M Baumberg (CoMIR)
 *
 * Creation Date : Mon Sep 26 11:05:21 1994 
 *
 * Comments : 	
 *
 ***************************************************************/



// some general math functions
// implemented using NAG
//

#ifndef __MATH_ROUTINES_H__
#define __MATH_ROUTINES_H__

#include <realno.h>

#ifdef USE_FLOAT
#define e04abf_ e04abe_
#define e04dgf_ e04dge_
#endif

#ifndef MAX
#define MAX(a,b) (( (a) > (b)) ? (a) : (b))
#endif

extern "C"
{
  typedef void (*func_of_x)(realno *x, realno *result);

  void e04abf_(func_of_x, realno* eps, realno* t, realno *a, realno *b,
	       int *maxcal, realno *x, realno *f_x, int *IFAIL);
}

inline realno get_local_minimum(func_of_x fn_x,
				realno a, realno b, realno eps = 0.0,
				realno t = 0.0, int maxcal = 50)
{
  realno x, f_x;
  int ifail = -1;
  e04abf_(fn_x, &eps, &t, &a, &b, &maxcal, &x, &f_x, &ifail);
  return (x);
}

extern "C"
{
  
  typedef void (*FUNGRD) (int* mode, int* n, realno* x, realno* objf, realno*
		g, int* nstate, int* iuser, realno* user);
  void e04dgf_(int* N, FUNGRD, int* ITER, realno* objf, realno*
	       objgrd, realno* x, int* iwork, realno* work, int*
	       iuser, realno* user, int* ifail);
  void e04dkf_(char* string, int len);
}
  

// get_local_minimum of a function of many variables
// using quasi-Newton conjugate gradient method
//
// X[0..N] will be set to parameters at minimum
// function returns minimum value
//
// func is a user suplied function that
// returns function + gradient values
//
// USER, IUSER are passed on to func
//

inline realno get_local_minimum(int N, FUNGRD func, realno* X,
				int* IUSER = NULL,
				realno* USER = NULL)

{
  int ITER;
  realno OBJF;
  realno* OBJGRD = new realno[N];
  int* IWORK = new int[N+1];
  realno* WORK = new realno[13 * N];
  int ifail = -1;
  
  e04dgf_(&N, func, &ITER, &OBJF, OBJGRD, X, IWORK, WORK, IUSER, USER,
	  &ifail);

  delete [] IWORK;
  delete [] WORK;
  delete [] OBJGRD;

  return (OBJF);
}


// general_minimise
//
// minimises a general non-linear constrained problem
// (using NAG routine)

extern "C"
{
  typedef void (*CONJAC)
    (int* mode, int* ncnln, int* n, int* nrowj, int* needc,
     realno* x, realno* c, realno* cjac, int* nstate,
     int* iuser, realno* user);

  void e04uef_(char*, int);
  void e04ucf_(int* n, int* nclin, int* ncnlin, int* nrowa,
	      int* nrowj, int* nrowr, realno* A, realno* BL,
	      realno* BU, CONJAC confun, FUNGRD objfun,
	      int* iter, int* istate, realno* C, realno* CJAC,
	      realno* clamda, realno* objf, realno* objgrd,
	      realno* R, realno* X, int* iwork, int* liwork,
	      realno* work, int* lwork, int* iuser, realno* user,
	      int* ifail);
}

inline realno 
general_minimise(int no_variables, int no_linear_constraints,
		 int no_non_linear_cons, FUNGRD objfn, CONJAC confn,
		 realno* lower_constraints, realno* upper_constraints,
		 realno* X,
		 int* iuser = NULL, realno* user = NULL)
{
  int N = no_variables;
  int NCLIN = no_linear_constraints;
  int NCNLN = no_non_linear_cons;
  int NROWA = MAX(1, NCLIN);
  int NROWJ = MAX(1, NCNLN);
  int NROWR = N;
  realno* A = NULL;
  if (NCLIN > 0) A = new realno[NROWA * N];
  realno* BL = lower_constraints;
  realno* BU = upper_constraints;
  int ITER;
  int* ISTATE = new int[N + NCLIN + NCNLN];
  realno* C = new realno[(MAX(1,NCNLN))];
  realno* CJAC = NULL;
  if (NCNLN > 0) CJAC = new realno[NROWJ * N];
  realno* CLAMDA = new realno[N + NCNLN + NCNLN];
  realno OBJF;
  realno* OBJGRD = new realno[N];
  realno* R = new realno[NROWR * N];
  int LIWORK = (3 * N + NCLIN + 2 * NCNLN);
  int* IWORK = new int[LIWORK];
  
  int LWORK;
  if ((NCLIN + NCNLN) == 0) LWORK = 20 * N;
  if ((NCNLN == 0) && (NCLIN > 0)) LWORK = 2 * N * N + 20 * N + 11 *
    NCLIN;
  if ((NCNLN > 0) && (NCLIN >=0)) LWORK = 2 * N * N
    + N * NCLIN + 2 * N * NCNLN + 20 * N + 11 * NCLIN + 21 * NCNLN;
  realno* WORK = new realno[LWORK];

  int ifail = -1;

  e04ucf_(&N, &NCLIN, &NCNLN, &NROWA, &NROWJ, &NROWR, A, BL, BU,
	  confn, objfn, &ITER, ISTATE, C, CJAC, CLAMDA, &OBJF,
	  OBJGRD, R, X, IWORK, &LIWORK, WORK, &LWORK, iuser, user,
	  &ifail);

  delete [] A;
  delete [] ISTATE;
  delete [] C;
  delete [] CJAC;
  delete [] CLAMDA;
  delete [] OBJGRD;
  delete [] R;
  delete [] IWORK;
  delete [] WORK;

  return (OBJF);
}
			       
#endif




