/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

// 
// NagMatrix.cc
// 
// A class for manipulating (2D) matrices
// (uses nag routines)

#include <stdio.h>
#include <stdlib.h>
#include "NagMatrix.h"

#ifndef MAX
#define MAX(a,b) (( (a) > (b)) ? (a) : (b))
#endif

#ifndef MIN
#define MIN(a,b) (( (a) < (b)) ? (a) : (b))
#endif

#ifdef USE_FLOAT
#define dsyev_ ssyev_
#define dgetrf_ sgetrf_
#define dgetri_ sgetri_
#define dgemv_ sgemv_
#define dgemm_ sgemm_
#define dgeev_ sgeev_
#define f01aaf_ f01aae_
#define f02abf_ f02abe_
#define f04aef_ f04aee_
#define f02agf_ f02age_
#define f01qcf_ f01qce_
#define f01qef_ f01qee_
#define f02wef_ f02wee_
#endif

#ifdef USE_LAPACK
extern "C"
{
  int ilaenv_(int *ispec, char *name, char *opts, int *n1,
		  int *n2, int *n3, int *n4);
  void dsyev_(char *jobz, char *uplo, int *n, realno *a,
	     int *lda, realno *w, realno *work, int *lwork, int *info);

  void dgetrf_(int *m, int *n, realno *a, int * lda, int *ipiv, int
	      *info);
    
  void dgetri_ (int* , realno* , int*, int*, realno*, int*, int*);

  void dgeev_ (char *jobvl, char *jobvr, int *n, realno* a, int *lda,
	       realno* wr, realno* wi, realno* vl, int *ldvl, 
	       realno* vr, int *ldvr, realno* work, int *lwork,
	       int *info);
	       
}
#else
static const int DEF_IFAIL = -1; // show error message but continue
static int IFAIL = DEF_IFAIL;
extern "C"
{
  void f01aaf_ (realno*, int*, int*, realno*, int*, realno*, int* );
  void f02abf_ (realno*, int*, int*, realno*, realno*, int*, 
		realno*, int*);
  void f04aef_(realno*, int*, realno*, int*, int*, int*, realno*,
	       int*, realno*, realno*, int*, realno*, int*, int*);  
  void f02agf_(realno*, int*, int*, realno*, realno*, realno*, int*,
	       realno*, int*, int*, int*);
  void f01qcf_(int*, int*, realno*, int*, realno*, int*);
  void f01qef_(char*, int*, int*, int*, realno* , int*, realno*, realno *,
	  int*);
  void f02wef_(int*, int*, realno*, int*, int*, realno*, int*, int*, 
	       realno*, int*, realno*, int*, realno*, int*,
	       realno* , int*);
}
#endif

extern "C"
{
  dgemv_ (char*, int*, int*, realno*, realno*, int*, realno*, int*, realno*,
	  realno*, int* );
  dgemm_ (char*, char*, int*, int*, int*, realno*, realno*, int*, 
	  realno*, int*, realno*, realno*, int*);
}

static void matrix_error(char * message)
{
  fprintf(stderr,"Error in NagMatrix library routine \n");
  fprintf(stderr, message);
  fprintf(stderr,"\n");
		fflush(stderr);
  abort();
}

NagMatrix::NagMatrix(int n, int m)
{
  rows = n;
  columns = m;
  data = new realno[n*m];
  own_memory = TRUE;
}

void NagMatrix::reconstruct(int n, int m)
{
  if ((data != NULL) && own_memory) delete [] data;
  rows = n;
  columns = m;
  if ((n*m) > 0) data = new realno[n*m];
  else data = NULL;
  own_memory = TRUE;
}


NagMatrix::NagMatrix(int n, int m, realno init)
{
  rows = n;
  columns = m;
  data = new realno[n*m];
  own_memory = TRUE;
  realno* enddata = (&data[n*m]);
  while ((--enddata) >= data) *enddata = init;
}

NagMatrix::NagMatrix(int n, int m, NagVector& v)
{
  rows = n;
  columns = m;
  if ((n * m) == v.size)
    {
      data = v.data;
      v.data = NULL;
      own_memory = v.own_memory;
    }
  else if ((n == m) && (n == v.size))
    {
      data = new realno[n*m];
      clear(0);
      for (int i = 0; i < rows; i++)
	set(i,i, v[i]);
      own_memory = TRUE;
    }
}

void NagMatrix::reconstruct(int n, int m , NagVector& v)
{
  if ((data != NULL) && own_memory) delete [] data;
  rows = n;
  columns = m;
  data = v.data;
  v.data = NULL;
  own_memory = v.own_memory;
}
  
realno identity_fn(int i, int j)
{
  if (i == j) return (1);
  return (0);
}

NagMatrix::NagMatrix(int n, int m, realno (*func) (int,int))
{
  rows = n;
  columns = m;
  data = new realno[n*m];
  own_memory = TRUE;
  for (int i = 0; i < rows; i++)
    for (int j = 0; j < columns; j++)
      set(i,j, (*func) (i,j));
}

NagMatrix::~NagMatrix()
{
  if ((data != NULL) && own_memory)
    delete [] data;
}

void NagMatrix::transpose( NagMatrix& result)
{
  if (result.data == NULL) result.reconstruct(columns, rows);
  NagMatrix tmp;
  copy(tmp);
  for (int i = 0; i < rows; i++)
    for (int j = 0; j < columns; j++)
	result.set(j,i,read(i,j));

}

void NagMatrix::invert( NagMatrix& result)
{
  if (rows != columns) 
    matrix_error(" can not invert non-square matrix");
#ifdef USE_LAPACK
  copy(result);
  int N = columns;
  int LDA = rows;
  int *IPIV = new int[N];
  int INFO;
  int ISPEC = 1; 
  //int NB = ilaenv_(&ISPEC, "dgetri", "", &N, &LDA, &minus1, &minus1);
  int NB = 64;
  int LWORK = (N * NB);
  NagVector WORK(LWORK);
  dgetrf_(&LDA, &N, result.data, &LDA, IPIV, &INFO);
  if (INFO != 0) matrix_error(" can't factorise");
  dgetri_(&N, result.data, &LDA, IPIV, WORK.data, &LWORK, &INFO);
  if (INFO != 0) matrix_error(" can't invert matrix");
#else
  if (result.data == NULL) result.reconstruct(rows, columns);
  int IA = rows;
  int N = rows;
  int IX = result.no_rows();
  NagMatrix tmp;
  copy(tmp);
  NagVector P(rows);
  IFAIL = DEF_IFAIL;
  f01aaf_(tmp.data, &IA, &N, result.get_data(), &IX, P.data, &IFAIL);
  if (IFAIL != 0) matrix_error(" failed to invert");
#endif
}

void NagMatrix::scale(realno s,  NagMatrix& result)
{
  if (result.data == NULL) result.reconstruct(rows, columns);
  if ((rows != result.rows) || (columns != result.columns))
    matrix_error(" illegal matrix in scale");
  realno* data1 = data;
  realno* data2 = result.data;
  for (int i = rows * columns; i > 0 ; i--)
    *data2++ = s * *data1++;
}


void NagMatrix::clear(realno s)
{
  realno* data1 = data;
  for (int i = rows * columns; i > 0 ; i--)
    *data1++ = s;
}

void NagMatrix::add( NagMatrix& m2,  NagMatrix& result)
{
  if (result.data == NULL) result.reconstruct(rows, columns);
  if ((rows != m2.rows) || (columns != m2.columns) ||
      (rows != result.rows) || (columns != result.columns))
    matrix_error(" illegal Matrix to add");
  realno* data1 = data;
  realno* data2 = m2.data;
  realno* data_res = result.data;
  for (int i = rows * columns; i > 0; i--)
    *data_res++ = (*data1++) + (*data2++);
}

void NagMatrix::subtract( NagMatrix& m2,  NagMatrix& result)
{
  if (result.data == NULL) result.reconstruct(rows, columns);
  if ((rows != m2.rows) || (columns != m2.columns) ||
      (rows != result.rows) || (columns != result.columns))
    matrix_error(" illegal Matrix to subtract");
  realno* data1 = data;
  realno* data2 = m2.data;
  realno* data_res = result.data;
  for (int i = rows * columns; i > 0; i--)
    *data_res++ = (*data1++) - (*data2++);
}

void NagMatrix::multiply( NagMatrix& m2,  NagMatrix& result)
{
  if (result.data == NULL) result.reconstruct(rows, m2.columns);
  if ((rows != result.rows) || (m2.columns != result.columns)
      || (columns != m2.rows))
    matrix_error(" error in matrix multiply");
  char transa = 'N';
  char transb = 'N';
  int M = rows;
  int N = m2.columns;
  int K = columns;
  realno alpha = 1.0;
  realno beta = 0.0;

  int LDA = rows;
  int LDB = m2.rows;
  int LDC = result.rows;

  realno* B = m2.data;
  realno* C = result.data;
  realno* A = data;
  dgemm_(&transa, &transb, &M, &N, &K, &alpha, A, &LDA,
	 B, &LDB, &beta, C, &LDC);
}

void NagMatrix::multiply( NagMatrix& m2,  NagMatrix& result, 
			 char transa, char transb, realno alpha)
{
  if (transa == 't') transa = 'T';
  if (transb == 't') transb = 'T';
  int M = rows;
  if (transa == 'T') M = columns;
  int N = m2.columns;
  if (transb == 'T') N = m2.rows;
  int K = columns;
  if (transa == 'T') K = rows;

  if (((transb == 'T') && (K != m2.columns))
      || ((transb != 'T') && (K != m2.rows)))
    matrix_error("bad call to matrix multiply");


  realno beta = 0.0;

  if (result.data == NULL) result.reconstruct(M, N);

  int LDA = rows;
  int LDB = m2.rows;
  int LDC = result.rows;

  realno* B = m2.data;
  realno* C = result.data;
  realno* A = data;
  dgemm_(&transa, &transb, &M, &N, &K, &alpha, A, &LDA,
	 B, &LDB, &beta, C, &LDC);
}

void NagMatrix::multiply(NagVector& m, NagVector& r)
{
  if (r.data == NULL) r.reconstruct(rows);
  if ((columns > m.size) || (rows > r.size))
    matrix_error(" vector too small \n");
  char trans = 'N';
  int M = rows;
  int N = columns;
  realno alpha = 1.0;
  realno beta = 0.0;
  int LDA = rows;
  realno* X = m.data;
  int INCX = 1;
  realno* Y = r.data;
  realno* A = data;
  int INCY = 1;
  dgemv_(&trans,&M, &N, &alpha, A, &LDA, X, &INCX, &beta, Y, &INCY);
}

void NagMatrix::multiply(NagVector &m, NagVector& r, char trans,
			 realno alpha)
{
  if (trans == 't') trans = 'T';
  if (trans == 'n') trans = 'N';
  int new_rows = rows;
  int new_cols = columns;
  
  if (trans == 'T') 
    {
      new_rows = columns; new_cols = rows;
    }
  if (r.data == NULL) r.reconstruct(new_rows);
  if ((new_cols > m.size) || (new_rows > r.size))
    matrix_error(" vector too small \n");

  int M = rows;
  int N = columns;
  realno beta = 0.0;
  int LDA = rows;
  realno* X = m.data;
  int INCX = 1;
  realno* Y = r.data;
  realno* A = data;
  int INCY = 1;
  dgemv_(&trans,&M, &N, &alpha, A, &LDA, X, &INCX, &beta, Y, &INCY);

}

void NagMatrix::output()
{
  printf("--------------\n");
  for (int i = 0; i < rows; i++)
    {
      for (int j = 0; j < columns; j++)
	printf(" %2.5f ", read(i,j));
      printf("\n");
    }
  printf("----------------\n");
}

void NagMatrix::get_column(int c, NagVector& result)
{
  if (result.data == NULL) result.reconstruct(rows);
  if (result.size < rows)
    matrix_error(" bad call to get_column");
  
  realno* pnt1 = get(0,c);
  realno* pnt2 = get(rows-1,c);
  realno* pnt3 = result.data;
  while (pnt1 <= pnt2) *pnt3++ = *pnt1++;
}

void NagMatrix::get_row(int rw, NagVector& result)
{
  if (result.data == NULL) result.reconstruct(columns);
  if (result.size < columns)
    matrix_error(" bad call to get_row");
  realno* pnt1 = get(rw,0);
  realno* pnt2 = get(rw, columns-1);
  realno* pnt3 = result.data;
  for (;pnt1 <= pnt2; pnt1 += rows) *pnt3++ = *pnt1;
} 

void NagMatrix::set_column(int c, NagVector& column)
{
  if ((column.size > rows) || (c >= columns))
    matrix_error(" bad call to set_column");
  
  realno* pnt1 = get(0,c);
  realno* pnt2 = &(column[column.size-1]);
  realno* pnt3 = column.data;
  while (pnt3 <= pnt2) *pnt1++ = *pnt3++;
}

void NagMatrix::set_row(int rw, NagVector& row0)
{
  if (row0.size > columns)
    matrix_error(" bad call to get_row");
  realno* pnt1 = get(rw,0);
  realno* pnt2 = &(row0[row0.size-1]);
  realno* pnt3 = row0.data;
  for (;pnt3 <= pnt2; pnt1 += rows) *pnt1 = *pnt3++;

}

void NagMatrix::set_block(int row_off, int col_off, NagMatrix&
			  block_m)
{
  if (((block_m.rows + row_off) > rows) ||
      (block_m.columns + col_off) > columns)
    matrix_error(" bad call to set_block");

  int i1 = row_off;
  for (int i2 = 0; i2 < block_m.rows; i2++)
    {
      int j1 = col_off;
      for (int j2 = 0; j2 < block_m.columns; j2++)
	set(i1,j1++,block_m.read(i2,j2));

      i1++;
    }
}

void NagMatrix::get_block(int row_top, int row_bottom,
			  int col_left, int col_right,
			  NagMatrix& block)
{
  if (block.data == NULL) block.reconstruct(row_bottom - row_top + 1,
					    col_right - col_left + 1);

  if (((block.rows + row_top) > rows) ||
      ((block.columns + col_left) > columns))
    matrix_error("bad call to get_block");

  for (int i1 = row_top; i1 <= row_bottom; i1++)
    for (int j1 = col_left; j1 <= col_right; j1++)
      block.set(i1 - row_top, j1 - col_left, read(i1,j1));
}


void NagMatrix::copy( NagMatrix& res)
{
  if (res.data == NULL) res.reconstruct(rows, columns);
  if ((res.rows != rows) || (res.columns != columns))
    matrix_error(" bad call to matrix copy");
  realno* pnt1 = data;
  realno* pnt2 = get(rows-1, columns-1);
  realno* pnt3 = res.data;
  while (pnt1 <= pnt2) *pnt3++ = *pnt1++;  
  //memcpy(res.data, data, rows*columns*sizeof(realno));
}


void NagMatrix::operator=( NagMatrix& m2)
{
  if ((data != NULL) && (own_memory))
    delete [] data;
  rows = m2.rows;
  columns = m2.columns;
  data = m2.data;
  m2.data = NULL;
  own_memory = m2.own_memory;
}

void NagMatrix::square_entries( NagMatrix& res)
{
  if (res.data == NULL) res.reconstruct(rows, columns);
  if ((rows != res.rows) || (columns != res.columns))
    matrix_error(" bad call to square_entries");
  realno* pnt1 = data;
  realno* pnt2 = res.data;
  realno* edat = data + (rows * columns);
  for (;pnt1 < edat; pnt1++)
    *pnt2++ = (*pnt1 * *pnt1);
}


void NagMatrix::pseudo_invert( NagMatrix& res)
{
  if (res.data == NULL) res.reconstruct(columns, rows);
  if ((rows != res.columns) || (columns != res.rows))
    matrix_error(" bad call to pseudo_inverse");

  NagMatrix h_t(columns,rows);
  transpose(h_t);

  if (rows > columns)
    {

      NagMatrix hTh(columns, columns);
      
      h_t.multiply(*this, hTh);
      
      NagMatrix inv_hTh(columns, columns);
      hTh.invert(inv_hTh);
      
      inv_hTh.multiply(h_t, res);
    }
  else
    {
      NagMatrix hhT(rows,rows);
      multiply(h_t, hhT);

      NagMatrix inv_hhT(rows,rows);
      hhT.invert(inv_hhT);

      h_t.multiply(inv_hhT, res);
    }

}

// get_eigenvectors(result, evals))
//  calculates eigenvalues, eigenvectors of real symetric matrix
//
//  on entry:
//    (*this) real sysmetric matrix
//  on exit:
//     result = (v_1 ... v_n) eigenvectors
//      evals = (e_1 ... e_n) eigenvalues in ascending order


//static char routine_name[32] = "SSYTRD";
//static char u_str[32] = "U";

void NagMatrix::get_eigenvectors( NagMatrix& result, NagVector& evals)
{
  if (rows != columns) matrix_error("bad call to get_eigenvectors");
  if (evals.data == NULL) evals.reconstruct(rows);
  if (result.data == NULL) result.reconstruct(rows, rows);
  if ((evals.size != rows) || (result.rows != rows) ||
      (result.columns != columns))
    matrix_error("bad call to get_eigenvectors");

#ifdef USE_LAPACK

  copy(result);
  char JOBZ = 'V';
  char UPLO = 'U';
  int N = rows;
  int LDA = rows;
  int ISPEC = 1;
  //int NB = ilaenv_(&ISPEC, "dsytrd", "U",  &N,
  //&LDA, &minus1, &minus1); 
  int NB = 1;
  int LWORK = (NB+2)*N;
  int INFO;
  NagVector WORK(LWORK);
  if (WORK.data == NULL) matrix_error("couldn't assign workspace");
  dsyev_(&JOBZ, &UPLO, &N, result.data, &LDA, evals.data,
	 WORK.data, &LWORK, &INFO);
  if (INFO != 0) matrix_error("couldn't get eigenvectors");

#else
  int N = rows;
  NagMatrix tmp;
  copy(tmp);
  NagVector temp(N);
  IFAIL = DEF_IFAIL;
  f02abf_ (tmp.data, &N, &N, evals.data, result.data, &N, temp.data, &IFAIL);
  if (IFAIL != 0) matrix_error(" failed to get eigensystem");
#endif  

}

// get_eigensystem 
// finds the (complex) eigenvalues, eigenvectors
// of an arbitrary square non-sysmetric real matrix

#ifdef USE_LAPACK
int best_work = 0;
#endif

void NagMatrix::get_eigensystem( NagMatrix& res_r,  NagMatrix& res_i,
				NagVector& evals_r, NagVector& evals_i)
{
  if (rows != columns) matrix_error("bad call to get_eigensystem");
  if (evals_r.data == NULL) evals_r.reconstruct(rows);
  if (evals_i.data == NULL) evals_i.reconstruct(rows);
  if (res_r.data == NULL) res_r.reconstruct(rows, rows);
  if (res_i.data == NULL) res_i.reconstruct(rows, rows);

#ifdef USE_LAPACK
  int N = rows;
  int LWORK = 4 * N;
  if (best_work > LWORK) LWORK = best_work;
  NagMatrix tmp;
  copy(tmp);
  char JOBVL = 'N';
  char JOBVR = 'V';
  int LDA = N;
  int LDVL = 1;
  int LDVR = res_r.no_rows();  
  NagVector WORK(LWORK);
  if (WORK.data == NULL) matrix_error("couldn't assign workspace");
  int INFO;

  dgeev_(&JOBVL, &JOBVR, &N, tmp.data, &LDA, evals_r.data,
	evals_i.data, NULL, &LDVL, res_r.data, &LDVR, WORK.data,
	&LWORK, &INFO);

  best_work = (int) WORK[1];
  if (INFO != 0) matrix_error("couldn't get eigenvectors");

  // unpack the complex eigenvectors
  NagVector zero_v(N); zero_v.clear();
  NagVector tmp_v;
  for (int j = 0; j < N; j++)
    {
      if (evals_i[j] == 0)
	res_i.set_column(j, zero_v);
      else
	{
	  res_r.get_column(j+1, tmp_v);
	  res_i.set_column(j, tmp_v);
	  tmp_v *= -1.0;
	  res_i.set_column(j+1, tmp_v);

	  res_r.get_column(j, tmp_v);
	  res_r.set_column(j+1, tmp_v);

	  j++;
	}
    }

#else
  NagMatrix tmp;
  copy(tmp);
  int IA = rows;
  int N = rows;
  int IVR = res_r.no_rows();
  int IVI = res_i.no_rows();
  int* INTGER = new int[N];
  IFAIL = DEF_IFAIL;
  f02agf_(tmp.data, &IA, &N, evals_r.data, evals_i.data,
	 res_r.data, &IVR, res_i.data, &IVI, INTGER, &IFAIL);
  delete [] INTGER;
#endif  

}

void NagMatrix::QR_factorise(NagMatrix& Q)
{
#ifndef USE_LAPACK
  NagMatrix A;
  copy(A);  
  int M = rows;
  int N = columns;
  if (Q.data == NULL) Q.reconstruct(M,M);
  int LDA = M;
  int K = N;
  NagVector zeta(N);
  NagVector WORK(K);
  IFAIL = DEF_IFAIL;
  f01qcf_(&M, &N, A.data, &LDA, zeta.data, &IFAIL);
  IFAIL = DEF_IFAIL;
  f01qef_("S", &M, &N, &K, A.data, &LDA, zeta.data, WORK.data,
	  &IFAIL);
  A.copy(Q);
#endif

}


// sqrt
// returns the positive definite square root
// of a real symetric matrix

void NagMatrix::msqrt(NagMatrix& res)
{
  if (rows != columns)
    matrix_error("bad call to NagMatrix::sqrt");
      
  NagMatrix evecs;
  NagVector evals;
  
  get_eigenvectors(evecs, evals);
  NagMatrix lambda_sqrt(rows, rows, 0.0);
  for (int i = 0; i < rows; i++)
    *lambda_sqrt.get(i,i) = fsqrt(evals[i]);
  NagMatrix evecs_t;  
  evecs.transpose(evecs_t);

  NagMatrix tmp;
  lambda_sqrt.multiply(evecs_t, tmp);
  evecs.multiply(tmp, res);
}
  
  
// sv_decompose
// Singular Value Decomposition
// A =  Q D P'
// where Q , P are orthogonal matrices

void NagMatrix::sv_decompose(NagMatrix& Q, NagVector& sv,
			     NagMatrix& P_t)
{
#ifndef USE_LAPACK
  int M = rows;
  int N = columns;
  int LDA = M;
  NagMatrix A;
  copy(A);
  int NCOLB = 0;
  int LDB = 0;
  int WANTQ = 1;
  int min_mn = MIN(M,N);
  if (Q.data == NULL) 
    Q.reconstruct(M,min_mn);
  int LDQ = M;
  if (sv.data == NULL) sv.reconstruct(min_mn);
  int WANTP = 1;
  if (P_t.data == NULL) P_t.reconstruct(min_mn,N);
  int LDPT = N;
  NagVector WORK(MAX(N * N + 5 * (N - 1), N + 4));
  
  IFAIL = DEF_IFAIL;
  f02wef_(&M, &N, A.data, &LDA, &NCOLB, NULL, &LDB, &WANTQ, Q.data,
	  &LDQ, sv.data, &WANTP, P_t.data, &LDPT, WORK.data, &IFAIL);
  
  if (M == N) A.copy(Q);
  if (M > N) A.get_block(0,M-1,0, N-1, Q);
  if (M < N) A.get_block(0,M-1,0, N-1, P_t);
#endif

}


// ortho transform
// if flag is TRUE (default)
// 		calculates Qt A Q
// else
//		calculates Q A Qt
//
// (where Qt is Q transform)
void NagMatrix::ortho_transform(NagMatrix& Q, NagMatrix& res, bool flag)
{
  NagMatrix Q_t;
  NagMatrix tmp;
  Q.transpose(Q_t);
  if (flag)
    {
      multiply(Q, tmp);
      Q_t.multiply(tmp, res);
    }
  else
    {
      multiply(Q_t, tmp);
      Q.multiply(tmp, res);
    }
}

ostream& operator<<(ostream& out_s,  NagMatrix& m)
{
  out_s << m.rows << " by " << m.columns << " matrix" << endl;
  /*  NagVector col(m.rows);
      for (int i = 0; i < m.columns; i++)
      {
      m.get_column(i,col);
      out_s << "column " << i << endl << col << endl;
      } */
  NagVector v(m);
  out_s << v;
  m.reconstruct(m.rows, m.columns, v);
  return (out_s);
}

istream& operator>>(istream& in_s,  NagMatrix& m)
{
  char dummy[50];
  int r,c;
  in_s >> r >> dummy >> c >> dummy;
  if (m.data == NULL) m.reconstruct(r,c);
  else if ((r != m.rows) || (c != m.columns)) 
    matrix_error("wrong sized matrix");

  /*
     NagVector col(m.rows);
     for (int i = 0; i < m.columns; i++)
     {
     in_s >> dummy >> dummy >> col;
     m.set_column(i,col);
      }*/

  NagVector v(m);
  in_s >> v;
  m.reconstruct(r,c,v);
  return (in_s);
}

void NagMatrix::pack_to_triangle(NagVector& res)
{
  int N = MIN(rows, columns);
  if (res.data == NULL) res.reconstruct(N * (N+1) / 2);

  if (res.size != (N * (N+1) / 2))
    matrix_error("bad call to pack_to_triangle");

  realno* pnt1 = res.data;
  for (int i = 0; i < N; i++)
    for (int j = 0; j <= i; j++)
      *pnt1++ = read(i,j);

}

void NagMatrix::unpack_triangle(NagVector& dat, bool symetric)
{
  if (data == NULL)
    {
      int N = (int)
	(sqrt(2 * dat.size + 0.25) - 0.49999);
      reconstruct(N,N);
    }

  int N = MIN(rows, columns);
  if (dat.size != (N * (N+1) / 2))
    matrix_error("bad call to unpack_triangle");

  realno* pnt1 = dat.data;
  if (symetric)
    {
      for (int i = 0; i < N; i++)
	for (int j = 0; j <= i; j++)
	  *get(i,j) = *get(j,i) = *pnt1++;
    }
  else 
    {
      for (int i = 0; i < N; i++)
	for (int j = 0; j <= i; j++)
	  {
	    *get(j,i) = 0;
	    *get(i,j) = *pnt1++;
	  }
    }
}

void NagMatrix::lower_to_full(NagMatrix& X_full)
{
  // X_full = X_lower + (X_lower)^T
  NagMatrix X_t;
  transpose(X_t);
  add(X_t, X_full);
}

void NagMatrix::scale_diagonal(realno fac)
{
  if (data == NULL)
    matrix_error("bad call to scale_diagonal");

  int N = MIN(rows, columns);
  for (int i = 0; i < N; i++)
    *get(i,i) *= fac;
}

realno NagMatrix::trace()
{
  realno res = 0;
  int N = MIN(rows, columns);
  for (int i = 0; i < N; i++)
    res += *get(i,i);

  return (res);
}


void NagMatrix::read_diagonal(NagVector& diag)
{
  int N = MIN(rows, columns);
  if (diag.data == NULL) diag.reconstruct(N);
  if (diag.size != N)
    matrix_error("bad call to read_diagonal");

  for (int i = 0; i < N; i++)
    diag[i] = *get(i,i);

}

void NagMatrix::set_diagonal(NagVector& diag)
{
  int N = MIN(rows, columns);
  N = MIN(N, diag.size);
  for (int i = 0; i < N; i++)
    set(i,i,diag[i]);

}
realno trace_A_Bt(NagMatrix& A, NagMatrix& B)
{
  if ((A.rows != B.rows) || (A.columns != B.columns))
    matrix_error("bad call to trace_A_Bt");

  realno res = 0.0;
  for (int i = 0; i < A.rows; i++)
    for (int j = 0; j < A.columns; j++)
      res += A.read(i,j) * B.read(i,j);

  return (res);
}

realno trace_A_B(NagMatrix& A, NagMatrix& B)
{
  if ((A.rows != B.columns) || (A.columns != B.rows))
    matrix_error("bad call to trace_A_Bt");
  
  realno res = 0.0;
  for (int i = 0; i < A.rows; i++)
    for (int j = 0; j < A.columns; j++)
      res += A.read(i,j) * B.read(j,i);
  
  return (res);
}

// solves Ax = b_1 b_2 ... b_m  with A an n x n matrix
// using NAG routine 
void NagMatrix::solve_equations(NagMatrix& B, NagMatrix& res)
{
#ifndef USE_LAPACK
  int n = MIN(rows, columns);
  int m = B.columns;
  if (res.data == NULL) res.reconstruct(n, B.columns);
  if ((res.rows != n) || (res.columns != m) ||
      (B.rows != n))
    matrix_error("bad call to solve_equations");
  

  int IA = rows;
  int IB = B.rows;
  int IC = res.rows;
  
  int IAA, IBB;
  IAA = IBB = n;
  
  NagVector wkspace(n);
  NagMatrix AA(n,n);
  NagMatrix BB(n,m);
  IFAIL = DEF_IFAIL;
  f04aef_(data, &IA, B.data, &IB, &n, &m, res.data, &IC, wkspace.data,
	  AA.data, &IAA, BB.data, &IBB, &IFAIL);
#endif

}


void NagMatrix::flip_horizontally()
{
  NagVector c1, c2;
  int imax = (columns / 2);
  for (int i = 0; i < imax; i++)
    {
      get_column(i, c1);
      get_column(columns - i - 1, c2);
      set_column(i, c2);
      set_column(columns - i - 1, c1);
    }
}
