/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

// 
// NagMatrix.h
// 
// A class for manipulating (2D) realno matrices
// (uses nag routines)

#ifndef __NAG_MATRIX_H__
#define __NAG_MATRIX_H__


#include "NagVector.h"
#include <stdlib.h>
#include <boolean.h>

extern realno identity_fn(int i, int j);
#define MATRIX_DEBUG


class NagMatrix 
{
 protected:
  
  int rows;
  int columns;
  realno* data;		// array of values
  bool own_memory;	// if TRUE then delete data when
			// finished

 public:

  
  // constructors, destructor
  
  NagMatrix(int n, int m, realno* dt) 
    { rows = n; columns = m; data = dt; own_memory = FALSE; }

  NagMatrix(int n, int m);
  NagMatrix(int n, int m, realno initial_value);
  NagMatrix(int n, int m, realno (*func) (int, int));
  NagMatrix() 
    { rows = columns = 0; data = NULL; own_memory = FALSE; }
  NagMatrix( NagMatrix& m) { (*this) = m; }
  NagMatrix(int n, int m,  NagVector& v);

  ~NagMatrix();

  //data access
  int no_rows() { return rows; }
  int no_columns() { return columns; }
  realno* get_data() { return data; }

  void set(int i, int j, realno value)
    { 
#ifdef MATRIX_DEBUG
      if ((i >= rows) || (j >= columns) || (i < 0) || (j < 0))
	{
	  fprintf(stderr,"bad call to Matrix::set \n");
	fflush(stderr);
	  abort();
	}
#endif
      data[i + j * rows] = value;
    }
  realno read(int i, int j)
    { return (data[i + j * rows]); }
  realno* get(int i, int j)
    { 
#ifdef MATRIX_DEBUG
      if ((i >= rows) || (j >= columns) || (i < 0) || (j < 0))
	{
	  fprintf(stderr,"bad call to Matrix::set\n");
	fflush(stderr);
	  abort();
	}
#endif
      return &(data[i + j * rows]);
    }

  realno& operator[](int i)
    { return (data[i]); }

  void operator=( NagMatrix& );
  void reconstruct(int rows, int columns);
  void reconstruct(int rows, int columns, NagVector& v);
  void reset() { rows = columns  = 0; data = NULL; }
  void destruct() { if ((data != NULL) && (own_memory))
		      delete [] data; reset(); }

  // set each element randomly
  void set_unif_random(realno min = -1, realno max = 1.0)
    {
      NagVector tmp1(rows * columns, data);
      tmp1.set_unif_random(min, max);
      tmp1.reset();
    }

  void set_gaus_random(realno mean = 0, realno sd = 1.0)
    {
      NagVector tmp1(rows * columns, data);
      tmp1.set_gaus_random(mean, sd);
      tmp1.reset();
    }

  // mathematical functions
  
  void transpose( NagMatrix& r);
  void flip_horizontally();
  void invert( NagMatrix& r);
  void pseudo_invert( NagMatrix& r);

// get_eigenvectors(result, evals))
//  calculates eigenvalues, eigenvectors of real symetric matrix
//
//  on entry:
//    (*this) real sysmetric matrix
//  on exit:
//     result = (v_1 ... v_n) eigenvectors
//      evals = (e_1 ... e_n) eigenvalues in ascending order


  void get_eigenvectors( NagMatrix& result,  NagVector& evals);

// get_eigensystem
// returns complex eigenvalues, eigenvectors 
// of a real non-symetric matrix

  void get_eigensystem( NagMatrix& res_r,  NagMatrix& res_i,
		        NagVector& evals_r,  NagVector& evals_i);

// QR factorise a real matrix

  void QR_factorise(NagMatrix& Q);
		    
// sqrt
// returns the positive definite square root
// of a real symetric matrix

  void msqrt(NagMatrix& res);

// sv_decompose
// Singular Value Decomposition
// A =  Q D P'
// where Q , P are orthogonal matrices

  void sv_decompose(NagMatrix& Q, NagVector& sv, NagMatrix& P_t);

  void scale(realno s,  NagMatrix& r);

  void add( NagMatrix& m,  NagMatrix& r);
  void subtract( NagMatrix& m,  NagMatrix& r);
  void multiply( NagMatrix& m,  NagMatrix& r);
  void multiply( NagVector& m,  NagVector& r);
  void multiply( NagVector& r)
    {
      NagVector tmp;
      r.copy(tmp);
      multiply(tmp, r);
    }
  void multiply(NagMatrix& m)
    {
      NagMatrix tmp;
      tmp = m;
      multiply(tmp, m);
    }
  void solve_equations(NagMatrix& B, NagMatrix& res);

  // result = alpha * op(A) * op(B)
  // where op(A) = A   when transa = 'N'
  // and   op(A) = A^T when transa = 'T'
  void multiply( NagMatrix& m2,  NagMatrix& result, 
		char transa, char transb, realno alpha = 1.0);
  void multiply(NagVector &m, NagVector& r, char trans,
			 realno alpha = 1.0);
  // misc

  friend ostream& operator<<(ostream&,  NagMatrix&);
  friend istream& operator>>(istream&,  NagMatrix&);
  friend NagVector;

  void clear(realno s = 0);

  void output();
  
  // column --	returns a NagVector referencing the i'th
  //		column
  NagVector column(int i) { return (NagVector(rows, get(0,i))); }

  // get_column -- copy contents of column into result
  void get_column(int col,  NagVector& r);
  void get_row(int row,  NagVector& r);
  void set_column(int col,  NagVector& column);
  void set_row(int row,  NagVector& row_v);
  
  void set_block(int row_off, int col_off, NagMatrix& block_m);
  void get_block(int row_top, int row_bottom,
		 int col_left, int col_right,
		 NagMatrix& block);

  
  // pack a symetric  into an ordered vector
  // (lower triangle used)
  void pack_to_triangle(NagVector& res);
  // unpack a vector into a symetric matrix
  void unpack_triangle(NagVector& data, bool symetric = TRUE);

  void NagMatrix::copy( NagMatrix& res);

  // ortho transform
  // calculates Qt A Q
  // (where Qt is Q transform)
  void ortho_transform(NagMatrix& Q, NagMatrix& res, bool flag = TRUE);
  void square_entries( NagMatrix& r);
  void scale_diagonal(realno fac = 0.5);
  void read_diagonal(NagVector& diag);
  void set_diagonal(NagVector& diag);

  realno trace();

  // calculate sum (i,j) { A_ij B_ij }
  friend realno trace_A_Bt(NagMatrix& A, NagMatrix& B);
  // calculate sum (i,j) { A_ij B_ji }
  friend realno trace_A_B(NagMatrix& A, NagMatrix& B);
  
  // X_full = *this + (*this)^T
  void lower_to_full(NagMatrix& X_full);

};

realno trace_A_Bt(NagMatrix& A, NagMatrix& B);
realno trace_A_B(NagMatrix& A, NagMatrix& B);

ostream& operator<<(ostream&,  NagMatrix&);
istream& operator>>(istream&,  NagMatrix&);  

inline NagMatrix operator*( NagMatrix& m1,  NagMatrix& m2)
{
  NagMatrix res;
  m1.multiply(m2, res);
  return (NagMatrix(res));
}

inline NagMatrix operator+( NagMatrix& m1,  NagMatrix& m2)
{
  NagMatrix res;
  m1.add(m2, res);
  return (NagMatrix(res));
}

inline NagMatrix operator-( NagMatrix& m1,  NagMatrix& m2)
{
  NagMatrix res;
  m1.subtract(m2, res);
  return (NagMatrix(res));
}

inline NagVector operator*( NagMatrix& m,  NagVector& v)
{
  NagVector res;
  m.multiply(v, res);
  return (NagVector(res));
}

inline NagMatrix operator*(double k,  NagMatrix& m)
{
  NagMatrix res;
  m.scale(k, res);
  return (NagMatrix(res));
}

inline NagMatrix operator*( NagMatrix& m, double k)
{
  NagMatrix res;
  m.scale(k, res);
  return (NagMatrix(res));
}

inline NagMatrix operator/( NagMatrix& m, double k)
{
  NagMatrix res;
  m.scale(1.0 / k, res);
  return (NagMatrix(res));
}


#endif

