/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

// 
// NagVector.cc
// 
// A class for manipulating realno n-vectors
// (uses nag routines)

#include <stdio.h>
#include <time.h>
#include <sys/types.h>
#include <unistd.h>


#include "NagMatrix.h"
#define MAX(a,b) (( (a) > (b)) ? (a) : (b))
#define MIN2(a,b) (( (a) < (b)) ? (a) : (b))
#define MIN3(a,b,c) (MIN2((a),MIN2((b),(c))))

#include <EnvParameter.h>

static EnvParameter output_binary_mode("NAG_BINARY_OUTPUT", TRUE);

static void vector_error(char* message)
{
  fprintf(stderr,"Error in NagVector library routine \n");
  fprintf(stderr,message);
  fprintf(stderr,"\n");
		fflush(stderr);
  abort();
}

NagVector::NagVector(int n, realno* dt, int asize)
{
  alloc_size = size = n;
  data = dt;
  if (asize > 0) alloc_size = asize;
  own_memory = FALSE;		// we didn't allocate this memory so
				// don't delete it !
}

void NagVector::reconstruct(int n, int asize)
{
  if ((data != NULL) && (own_memory)) delete [] data;
  alloc_size = size = n;
  if (asize > 0) alloc_size = asize;
  if (alloc_size > 0) data = new realno[alloc_size];
  else data = NULL;
  own_memory = TRUE;
}

NagVector::NagVector(int n, int asize)
{
  alloc_size = size = n;
  if (asize > 0) alloc_size = asize;
  if (alloc_size > 0) data = new realno[alloc_size];
  else data = NULL;
  own_memory = TRUE;
}

// NagVector::NagVector(int n, realno init, int asize )
// {
//   alloc_size = size = n;
//   if (asize > 0) alloc_size = asize;
//   data = new realno[alloc_size];
//   realno* enddata = &data[size];
//   while ((--enddata) >= data) *enddata = init;
// }

NagVector::NagVector(NagVector& v)
{
  alloc_size = v.alloc_size;
  size = v.size;
  data = v.data;
  own_memory = v.own_memory;
  v.data = NULL;
}

NagVector::NagVector(NagMatrix& m)
{
  alloc_size = size = m.rows * m.columns;
  data = m.data;
  own_memory = m.own_memory;
  m.data = NULL;
}

NagVector::NagVector(int n, realno (*func) (int), int asize)
{
  alloc_size = size = n;
  if (asize > 0) alloc_size = asize;
  data = new realno[size];
  own_memory = TRUE;
  for (int i = 0; i < size; i++)
    data[i] = (*func) (i);
}

void NagVector::reset()
{ 
  // this function is obsolete 
  if (own_memory == TRUE)
    vector_error("bad call to NagVector::reset");

  // own_memory is FALSE so data won't be deleted anyway!
  // size = alloc_size = 0; 
  // data = NULL; 
}

NagVector::~NagVector()
{
  if ((data != NULL) && (own_memory))
    delete [] data;
}

void NagVector::copy_data(float* cdata, int n)
{
  if (n == 0) n = size;
  if ((data == NULL) && (n > 0)) reconstruct(n);
  if (n > size) vector_error("too much data to copy");
  realno* dat1 = data;
  realno* dat2 = &data[n];
  while (dat1 < dat2) *dat1++ = (realno) *cdata++;
}

void NagVector::copy_data(double* cdata, int n)
{
  if (n == 0) n = size;
  if ((data == NULL) && (n > 0)) reconstruct(n);
  if (n > size) vector_error("too much data to copy");
  realno* dat1 = data;
  realno* dat2 = &data[n];
  while (dat1 < dat2) *dat1++ = (realno) *cdata++;
}
 

void NagVector::operator=(NagVector& vec2)
{
  if ((data != NULL) && (own_memory))
    delete [] data;
  alloc_size = vec2.alloc_size;
  size = vec2.size;
  data = vec2.data;
  own_memory = vec2.own_memory;
  vec2.data = NULL;
  vec2.size = 0;
  vec2.alloc_size = 0;
}

void NagVector::output()
{
  printf("[\n");
  for (int i = 0; i < size; i++)
    printf(" %3.3f \n", data[i]);
  printf("]\n");
}

void NagVector::clear(realno value)
{
  realno* enddata = &data[size];
  while ((--enddata) >= data) *enddata = value;
}

void NagVector::add(realno item)
{
  if (data == NULL) reconstruct(0,2);
  int size1 = size + 1;
  if (size1 > alloc_size)
    {
      NagVector new_stack(size1, alloc_size * 2);
      this->copy(new_stack);
      new_stack.data[size] = item;
      *this = new_stack;
    }
  else
    {
      data[size] = item;
      size = size1;
    }
}

void NagVector::add(NagVector& vec2, NagVector& res)
{
  if (res.data == NULL) res.reconstruct(MIN2(size, vec2.size));
  int n = MIN3(size, vec2.size, res.size);

  realno* vdat1 = data;
  realno* vdat2 = vec2.data;
  realno* rdat = res.data;
  realno* edat = &(data[n]);
  
  while (vdat1 < edat)
    *rdat++ = (*vdat1++) + (*vdat2++);
}

NagVector& NagVector::operator+=(NagVector& v2)
{
  int n = MIN2(size, v2.size);
  realno* dat1 = data;
  realno* dat2 = v2.data;
  realno* edat = &(data[n]);
  while (dat1 < edat) 
    *dat1++ += *dat2++;
  return (*this);
}

NagVector& NagVector::operator-=(NagVector& v2)
{
  int n = MIN2(size, v2.size);
  realno* dat1 = data;
  realno* dat2 = v2.data;
  realno* edat = &(data[n]);
  while (dat1 < edat) 
    *dat1++ -= *dat2++;
  return (*this);
}

NagVector& NagVector::operator*=(realno a)
{
  realno* dat1 = data;
  realno* edat = &(data[size]);
  while (dat1 < edat) *dat1++ *= a;
  return (*this);
}

NagVector& NagVector::operator/=(realno a)
{
  realno* dat1 = data;
  realno* edat = &(data[size]);
  while (dat1 < edat) *dat1++ /= a;
  return (*this);
}

void NagVector::subtract(NagVector& vec2, NagVector& res)
{
  if (res.data == NULL) res.reconstruct(MIN2(size, vec2.size));
  int n = MIN3(size, vec2.size, res.size);
  realno* vdat1 = data;
  realno* vdat2 = vec2.data;
  realno* rdat = res.data;
  realno* edat = &(data[n]);
  
  while (vdat1 < edat)
    *rdat++ = (*vdat1++) - (*vdat2++);
}

void NagVector::copy(NagVector& res)
{
  if (res.data == NULL) res.reconstruct(size);
  realno* dat1 = data;
  realno* rdat = res.data;
  realno* edat = &(data[MIN2(size,res.size)]);

  while (dat1 < edat) *rdat++ = (*dat1++);

//  memcpy(res.data, data, MIN2(size,res.size) * sizeof(realno));
}


realno NagVector::length2()
{
  realno* vdat = data;
  realno* edat = &(data[size]);
  realno lngth = 0, tmp;
  while (vdat < edat)
    {
      tmp = *vdat++;
      lngth += (tmp * tmp);
    }
  return (lngth);
}

void NagVector::scale(realno lmda, NagVector& res)
{
  if (res.data == NULL) res.reconstruct(size);
  realno *dat = data;
  realno *edat = &(data[MIN2(size,res.size)]);
  realno *rdat = res.data;
  while (dat < edat)
    *rdat++ = lmda * (*dat++);
}

void NagVector::sub_vector(int i, int j, NagVector& res)
{
  if (res.data == NULL) res.reconstruct(j-i+1);
  if ((res.size < (j - i + 1)) || (i < 0) || (j >=  size) || (j < i))
    vector_error(" sub_vector  bad parameters ");
  realno *dat = &(data[i]);
  realno *edat = &(data[j]);
  realno *rdat = res.data;
  while (dat <= edat) *rdat++ = *dat++;
}

realno NagVector::dot(NagVector& vec2)
{
  realno *dat1 = data;
  realno *dat2 = vec2.data;
  realno *edat = &(data[MIN2(size,vec2.size)]);
  realno res = 0;
  while (dat1 < edat)
    res += (*dat1++ * *dat2++);
  return (res);
}

void NagVector::convolve(NagVector& kern, NagVector& res)
{
  if (res.data == NULL) res.reconstruct(size);
  int h = (kern.size -1) / 2;

  for (int i = 0; i < size; i++)
    {
      realno itmp = 0.0;
      int k = 0;
      for (int j = size-h; j <= size+h; j++)
	itmp += data[(i+j) % size] * kern[k++];
      res[i] = itmp;
    }
}


void NagVector::concat(NagVector& vec2, NagVector& res)
{
  if (res.data == NULL) res.reconstruct(size + vec2.size);
  if (res.size < (size + vec2.size))
    vector_error("Not enough storage space in NagVector::concat");
  realno* dat1 = data;
  realno* rdat = res.data;
  realno* edat = &(data[size]);
  
  while (dat1 < edat)
    *rdat++ = (*dat1++);  

  dat1 = vec2.data;
  edat = &(vec2[vec2.size]);

  while (dat1 < edat)
    *rdat++ = (*dat1++);   

}  

void  NagVector::flip(NagVector& res)
{
  if (res.data == NULL) res.reconstruct(size);
  if (res.size > size)
    vector_error("error in NagVector::flip");
  
  realno *fdat = &(data[size]);
  realno *dat1 = res.data;
  realno *dat2 = &(res.data[res.size]);
  
  while (dat1 < dat2)
    *dat1++ = *(--fdat);
}
  

ostream& operator<<(ostream& out_s, NagVector& v)
{
  if ((v.size == 0) || (v.data == NULL))
    return (out_s << "0" << endl);

  if (output_binary_mode.is_on())
    {
#ifdef USE_FLOAT
      out_s << v.size << " f[" << endl;
#else
      out_s << v.size << " d[" << endl;
#endif
      out_s.write((char*) v.data, v.size * sizeof(realno));
      out_s << "]" <<endl;
    }
  else
    {
      out_s << v.size << " [" << endl;
      realno *dat1 = v.data;
      realno *dat2 = &(v[v.size]);
      while (dat1 < dat2) out_s << (*dat1++) << endl;
      out_s << "]" << endl;
    }
  return (out_s);
}
  
istream& operator>>(istream& in_s, NagVector& v)
{
  char dummy[50];
  int sz;
  in_s >> sz ;
  if (v.data == NULL) v.reconstruct(sz);
  else if (v.size < sz)
    vector_error("not enough storage space");
  if (sz == 0) return (in_s);
  in_s >> dummy;
  if (dummy[0] == 'f')
    {
      in_s.ignore(1);
      if (sizeof(realno) == sizeof(float))
	in_s.read((char*) v.data, sz * sizeof(float));
      else
	{
	  float* tmp_data = new float[sz];
	  in_s.read((char*) tmp_data, sz * sizeof(float));
	  v.copy_data(tmp_data, sz);
	  delete [] tmp_data;
	}
    }
  else if (dummy[0] == 'd')
    {
      in_s.ignore(1);
      if (sizeof(realno) == sizeof(double))
	in_s.read((char*) v.data, sz * sizeof(double));
      else
	{
	  double* tmp_data = new double[sz];
	  in_s.read((char*) tmp_data, sz * sizeof(double));
	  v.copy_data(tmp_data, sz);
	  delete [] tmp_data;
	}
    }
  else 
    {
      realno *dat1 = v.data;
      realno *dat2 = &(v[v.size]);
      while (dat1 < dat2) in_s >> (*dat1++);
    }
  in_s >> dummy;
  return (in_s);
}

realno NagVector::sum()
{
  realno res = 0;
  for (int i = 0; i < size; i++)
    res += data[i];

  return (res);
}

// generate a normally distributed random variable

static int iset = 0;
static int set_seed = 0;
static realno gset;

static void check_seed()
{
  if (set_seed == 0)
    {
      long rseed = time(NULL) + getpid();
      srand48(rseed);
      set_seed = 1;
    }
}

realno normal_random()
{
  
  check_seed();
  realno fac,rsq,v1,v2;

  if (iset == 0)
    {
      do
	{
	  v1 = 2.0 * drand48() - 1.0;
	  v2 = 2.0 * drand48() - 1.0;
	  rsq = (v1 * v1) + (v2 * v2);
	} while ((rsq >= 1.0 )|| (rsq == 0.0));

      fac = sqrt(-2.0 * log(rsq)/rsq);
      gset = v1 * fac;
      iset = 1;
      return (v2 * fac);
    }

  iset = 0;
  return (gset);

}

void NagVector::set_unif_random(realno min, realno max)
{
  check_seed();
  realno diff = max - min;
  for (int i = 0; i < size; i++)
    data[i] = min + (diff * drand48());
}  



void NagVector::set_gaus_random(realno mean, realno sd)
{
  for (int i = 0; i < size; i++)
    data[i] = mean + (sd * normal_random());
}


void NagVector::mult_trans(NagVector& vec2, NagMatrix& res)
{
  if (res.data == NULL) res.reconstruct(size, vec2.size);
  if ((res.rows != size) || (res.columns != vec2.size))
    vector_error("bad call to mult_trans");

  int i,j;
  for (i = 0; i < size; i++)
    for (j = 0; j < vec2.size; j++)
      *res.get(i,j) = data[i] * vec2[j];
}


#ifndef SWAP
#define SWAP(a,b) temp=(a);(a)=(b);(b)=temp;
#endif

realno nr_select(unsigned long k, unsigned long n, realno* arr)
{
  
  unsigned long i,ir,j,l,mid;
  realno a, temp;

  l = 1;
  ir = n;
  for (;;) {    
    if (ir <= l+1) {
      if (ir == l+1 && arr[ir] < arr[l]) {
	SWAP(arr[l], arr[ir]);
      }
      return arr[k];
    } else {
      mid = (l+ir) >> 1;
      SWAP(arr[mid],arr[l+1]);
      if (arr[l] > arr[ir]) {
	SWAP(arr[l],arr[ir]);
      }
      if (arr[l+1] > arr[ir]) {
	SWAP(arr[l+1],arr[ir]);
      }
      if (arr[l] > arr[l+1]) {
	SWAP(arr[l],arr[l+1]);
      }

      i=l+1;
      j=ir;
      a=arr[l+1];
      for (;;) {
	do i++; while (arr[i] < a);
	do j--; while (arr[j] > a);
	if (j < i) break;
	SWAP(arr[i],arr[j]);
      }
      arr[l+1]=arr[j];
      arr[j]=a;
      if (j >= k) ir=j-1;
      if (j <= k) l=i;
    }
  }
  return (arr[k]);
}
      
      
realno NagVector::select(realno x)
{
  NagVector tmp_v;
  copy(tmp_v);
  unsigned long k = 0.5 + size * x;
  return (nr_select(k, size, tmp_v.get_data()-1));
}
	  

