/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

// 
// NagVector.h
// 
// A class for manipulating realno n-vectors
// (uses nag routines)

#ifndef __NAG_VECTOR_H__
#define __NAG_VECTOR_H__

#include <realno.h>

#include <stdlib.h>
#include <stream.h>
#include <math.h>
#include <boolean.h>

class NagMatrix;

realno normal_random();

class NagVector
{
 protected:

  int alloc_size;	// size of allocated data
  int size;		// size of vector
  realno* data;		// array of values
  bool own_memory;	// if TRUE then delete data when
			// finished

 public:
  
  // constructors , destructors;
  // this constructor sets own_memory to FALSE
  NagVector(int n, realno* dt, int asize = 0);

  // all other constructors create NagVectors 
  // that dynamically allocate memory
  NagVector(int n, int asize = 0);
  NagVector(int n, realno (*func) (int), int asize = 0);
  NagVector() { size = alloc_size = 0; data = NULL;
		own_memory = TRUE; }
  NagVector(NagVector& v);
  NagVector(NagMatrix& );
  ~NagVector();
  
  void operator=(NagVector& vec2);
  void reconstruct(int n, int asize = 0);

  friend NagMatrix;
  friend istream& operator>>(istream&, NagVector& );
  friend ostream& operator<<(ostream&,  NagVector& );

  // data access
  int get_size() { return size; }
  realno* get_data() { return data; }
//  operator realno*() { return data; }
  
  // reset to null, don't delete data
  void reset();

//   void set(int i, realno value)
//     { data[i] = value; }
//   realno read(int i)
//     { return (data[i]); }
//   realno* get(int i)
//     { return &(data[i]); }

  realno& operator[](int i) 
     { return (data[i]); }

  void output() ;

  // utility functions

  void clear(realno value = ((realno) 0.0));
  void add(realno itm);
  void copy_data(float* cdata, int n = 0);
  void copy_data(double* cdata, int n = 0);

  // more functions to come ...
  void add( NagVector& vec2, NagVector& res) ;
  void subtract( NagVector& vec2, NagVector& res) ;
  void copy(NagVector& res) ;
  void scale(realno lmda, NagVector& res) ;
  void sub_vector(int i, int j, NagVector& res) ;
  void convolve( NagVector& kern, NagVector& res) ;
  void concat( NagVector& vec2, NagVector& res) ;
  void flip(NagVector& res);

  // x.mult_trans(y, M) sets M = x y^T
  void mult_trans(NagVector& vec2, NagMatrix& res);

  // set each element randomly
  void set_unif_random(realno min = -1, realno max = 1.0);
  void set_gaus_random(realno mean = 0, realno sd = 1.0);


  // select(x)
  // pick the (n * x)'th item from the sorted data
  // (see Numerical Recipies)
  realno select(realno x = 0.5);


  NagVector sub_vector(int i, int j) 
    {
      NagVector res;
      sub_vector(i,j,res);
      return (NagVector(res));
    }
  realno length2() ;
  realno length() { return (sqrt(length2())); }
  realno sum();

  realno dot( NagVector& vec2) ;
  
  NagVector& operator+=( NagVector& v2);
  NagVector& operator-=( NagVector& v2);
  NagVector& operator*=(realno a);
  NagVector& operator/=(realno a);

  // other functions
  //friend void complex_1D_ft(NagVector& real, NagVector& imag);
  //  friend void real_1D_ft(NagVector& x);

};

istream& operator>>(istream&,  NagVector& );
ostream& operator<<(ostream&,  NagVector& );

inline NagVector operator+( NagVector& v1,  NagVector& v2)
{
  NagVector res;
  v1.add(v2,res);
  return (NagVector(res));
}

inline NagVector operator-( NagVector& v1,  NagVector& v2)
{
  NagVector res;
  v1.subtract(v2,res);
  return (NagVector(res));
}

inline NagVector operator*(realno k,  NagVector& v)
{
  NagVector res;
  v.scale(k,res);
  return (NagVector(res));
}

inline NagVector operator*( NagVector& v, realno k)
{
  NagVector res;
  v.scale(k,res);
  return (NagVector(res));
}

inline NagVector operator/( NagVector& v, realno k)
{
  NagVector res;
  v.scale(1.0 / k,res);
  return (NagVector(res));
}

#endif

