/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/***************************************************************
 * C++ source
 *
 * File : 	PolyFilter.cc
 *
 * Module :	PolyFilter
 *
 * Author : 	A M Baumberg
 *
 * Creation Date : Mon Nov  8 13:39:32 1993 
 *
 * Comments : 	An object for prediting the next value in a sequence
 *	        of equally spaced data values using a polynomial fit.
 *		(Each data value is a vector.)
 *
 ***************************************************************/


/* include files */

#include "PolyFilter.h"
#include <math.h>
#include <stdlib.h>
#include <stdio.h>


PolyFilter::PolyFilter(int order, int neqns, int vsize)
: past_data(neqns, vsize),
  poly_coefficients(order, vsize),
  extra_row(1, order),
  pseudo_inverse(order, neqns)
{
  curve_order = order;
  no_eqns = neqns;
  vector_size = vsize;
  t_stamp = 0;
  int i,j,k;
  NagMatrix A(no_eqns, curve_order);
  for (i = 0; i < no_eqns; i++)
    for (j = 0; j < curve_order; j++)
      A.set(i,j,pow(i,j));

  for (k = 0; k < curve_order; k++)
    extra_row.set(0,k, pow(no_eqns, k));

  NagMatrix A_trans(curve_order, no_eqns);
  A.transpose(A_trans);

  NagMatrix At_A(curve_order, curve_order);
  A_trans.multiply(A, At_A);

  NagMatrix At_A_inv (curve_order, curve_order);
  At_A.invert(At_A_inv);

  At_A_inv.multiply(A_trans, pseudo_inverse);

}


void PolyFilter::add_observation(NagVector& ob)
{
  int i,j;
  if (t_stamp++ == 0)
    {
      for (i = 0; i < no_eqns; i++)
	for (j = 0; j < vector_size; j++)
	  past_data.set(i,j,ob[j]);
    }

  else
    {
      int tmp = no_eqns - 1;
      for (i = 0; i < tmp; i++)
	for (j = 0; j < vector_size; j++)
	  past_data.set(i,j ,past_data.read(i+1,j));
      for (j = 0; j < vector_size; j++)
	past_data.set(tmp, j,ob[j]);
    }
}
	  
void PolyFilter::get_prediction(NagVector& predicted)
{
  if (t_stamp == 0) 
    {
      fprintf(stderr, " PolyFilter: can not predict with no data \n");
		fflush(stderr);
      abort();
    }
  NagMatrix res(1,vector_size);
  pseudo_inverse.multiply(past_data, poly_coefficients);
  extra_row.multiply(poly_coefficients,res);
  res.get_row(0, predicted);
}

