/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/* test program */

#include <image.h>
#include <sgimgsrc.h>
#include <miscsrc.h>
#include <stdio.h>
#include <gl/gl.h>
#include <getopt.h>
#include <RegionSet.h>
#include <ProfileSet.h>
#include <BWImage.h>
#include <sgmoviesrc.h>
#include <MpegStream.h>
#include <EnvParameter.h>
#include <utils.h>
#include <imgsrc_utils.h>

extern void setup_spline_matrices();
#define USAGE "usage: data_extraction [options] \n\
-q quiet mode (no image display)\n\
-l run Length for median filter \n\
-t differencing Threshold \n\
-n No. of control points \n\
-i input options (-i help for details) \n\
-d don't align/ normalise shapes \n\
-g gaps size for merging regions \n\
-v vertical flip images \n\
-m minimum blob size as proportion of image area (e.g. 0.06) \n\
-p parameter_file (use file to specify parameters)\n"

/* fixed globals */

bool quiet_mode = 
AmbGlobals.register_bool("QUIET_MODE", FALSE,
			 &quiet_mode, TRUE,
			 "main", 
			 "If set ON then don't display any images");

int& default_ncp = 
AmbGlobals.register_int("NO_CONTROL_POINTS", 32,
			&NO_CONTROL_POINTS, TRUE,
			"main", 
			"The number of spline control points used to \
represent each shape");

int time_skip =
AmbGlobals.register_int("BACKGROUND_UPDATE_SKIP", 4,
			&time_skip, TRUE,
			"main",
			"Temporal resampling for background filtering");

int default_window_length=
AmbGlobals.register_int("BACKGROUND_RUN_LENGTH", 50,
			&default_window_length, TRUE,
			"main",
			"Medain window length (in frames) \
for background image update");

bool default_do_blur = 
AmbGlobals.register_bool("BLUR_MOTION",
			 TRUE, &default_do_blur,
			 TRUE, "main",
			 "Blur images in motion detection (slow) ?");


bool do_normalisation = 
AmbGlobals.register_bool("DO_NORMALISE",
			 TRUE, &do_normalisation,
			 TRUE, "main",
			 "Normalise profiles ? (i.e. recenter using \
centroid and rotate principal axis to vertical");


bool flip_images = 
AmbGlobals.register_bool("FLIP_VERTICALLY",
			 FALSE, &flip_images,
			 TRUE, "main",
			 "Vertically flip input images ?");

bool reduce_noise = 
AmbGlobals.register_bool("REDUCE_NOISE",
			 TRUE, &reduce_noise,
			 TRUE, "main",
			 "Use bluring and dilation to reduce noise \
(slow) ?");

/* variable globals */

int gap_size = 
AmbGlobals.register_int("GAP_SIZE", 1,
			&gap_size, FALSE, "main",
			"The dilation size in pixels when filtering \
binary motion image. The effect is to join blobs that are this many \
pixels apart.");

			
realno camera_moved_threshold = 
AmbGlobals.register_real("CAMERA_MOVED_THRESH", 0.5,
			 0.0, 1.0,
			 &camera_moved_threshold,
			 FALSE, "main", 
			 "If the number of pixels where there is \
movement over the total number of pixels \
is greater than this threshold, then the camera has moved.");

realno diff_threshold = 
AmbGlobals.register_real("DETECT_DIFF_THRESH", 0.1,
			 0.0, 1.0, &diff_threshold, 
			 FALSE, "main",
			 "Differencing threshold for object detection");


realno max_region_ratio =
AmbGlobals.register_real("MAX_REGION_SIZE", 0.5,
			 &max_region_ratio,
			 FALSE, "main",
			 "Maximum size of a motion blob over \
size of image");

realno min_region_ratio = 
AmbGlobals.register_real("MIN_REGION_SIZE", 0.001,
			 &min_region_ratio,
			 FALSE, "main",
			 "Minimum size of a motion blob over \
size of image");
 

main(int argc, char** argv)
{
  int c;
  ImageSource* input_source = NULL;


  Profile::draw_boxes = TRUE;
  char *parameter_file = NULL;

  while ((c = getopt(argc, argv, "qp:m:i:dg:vl:t:n:?")) != EOF)
    switch(c)
      {
      case 'q':
	quiet_mode = TRUE;
	break;

      case 'p':
	parameter_file = optarg;
	break;

      case 'm':
	if (optarg != NULL) min_region_ratio = atof(optarg);
	break;

      case 'i':
	input_source = get_live_or_movie_feed(optarg);
	break;

      case 'd':
	do_normalisation = FALSE;
	break;
      case 'g':
	reduce_noise = TRUE;
	if (optarg != NULL) gap_size = atoi(optarg);
	break;
      case 'v':
	flip_images = TRUE;
	break;
      case 'l':
	if (optarg != NULL) default_window_length = atoi(optarg);
	break;
      case 't':
	if (optarg != NULL) diff_threshold = atof(optarg);
	break;
      case 'n':
	if (optarg != NULL) NO_CONTROL_POINTS = atoi(optarg);
	break;
      case '?':
      default:
	fprintf(stderr, USAGE);
	exit(1);
      } 
  AmbGlobals.parse_parameter_file(parameter_file);
  if (!quiet_mode) AmbGlobals.create_interface(argc, argv);

  Image* reference_image = get_reference_frame();
  setup_spline_matrices();
  cout << NO_CONTROL_POINTS << "control points" << endl;
  
  if (flip_images)
    {
      ImageSource* old_in = input_source;
      input_source = new VFlipSource(old_in);
      if (reference_image != NULL)
	reference_image->flip_vertically(reference_image);
    }
  
  ImageSource& vidsrc = *input_source;
  int rlength = default_window_length / time_skip;
  Medsrc background(&vidsrc, rlength, reference_image);
  DifferenceSource foreground1(&vidsrc, &background, 128.0 * diff_threshold);
  BlurSource blur_fore(&foreground1);
  
  background.get_next();
  ImageType data_format = vidsrc.get_current()->imgtype();
  
  ThresholdSource* fore2_pnt;
  if (reduce_noise) fore2_pnt = 
      new ThresholdSource(&blur_fore, 255.0 * diff_threshold);
  else fore2_pnt =
	 new ThresholdSource(&foreground1, 255.0 * diff_threshold);
  
  
  ThresholdSource& foreground2 = *fore2_pnt;

  RegionSet curr_regions;
  ProfileSet curr_profiles;
  
  background.get_next();
  int vwidth = vidsrc.get_current()->get_width();
  int vheight = vidsrc.get_current()->get_height();
  int vsize = vwidth * vheight;

  if (isYUV((ImageType) data_format))
    {
      for (int y = 0; y < 5; y++)
	{
	  background.get_next();
	}
      recalculate_weights(vidsrc.get_current(),
			  background.get_current());
    }
  fprintf(stderr, " and ACTION! \n");

  bool found_some = FALSE;
  Image* pnt_image;
  Point2 object_center;
  bool flag_center = FALSE;
  for (int x=0; x >=0 ; x++)
    {
      int min_region_size = vsize * min_region_ratio;
      int max_region_size = vsize * max_region_ratio; 
      int max_marked = vsize * camera_moved_threshold;
      foreground1.threshold = 128.0 * diff_threshold;
      foreground2.thresh = 255.0 * diff_threshold;

      if ((x % time_skip) == 0) 
	pnt_image = background.get_next();
      else  pnt_image = vidsrc.get_next();
      if (pnt_image == NULL) break;
      foreground2.get_next();
      if (foreground2.no_marked > max_marked)
	{
	  cout << "list_length = 0" << endl << "***********" << endl;
	  while (foreground2.no_marked > max_marked)
	    {
	      max_marked = vsize * camera_moved_threshold;
	      foreground1.threshold = 255.0 * diff_threshold;
	      foreground2.thresh = 255.0 * diff_threshold;
	      background.get_next();
	      foreground2.get_next();
	    }
	}
      if (reduce_noise == TRUE)
	{
	  BWImage* fore_bw = (BWImage*) foreground2.get_current();
	  if (gap_size > 0)
	    fore_bw->fix_holes(gap_size, fore_bw);
	}

      if (!quiet_mode)
	{
	  background.display();
	  foreground1.display();
	  foreground2.display();      
	  vidsrc.display();
	  vidsrc.get_current()->draw_in_image();
	}
      curr_regions.grab_regions(foreground2.get_current(),
				min_region_size, max_region_size, TRUE);


      curr_regions.trace_regions();
      curr_profiles = curr_regions.to_profiles();
      if (!quiet_mode) curr_profiles.draw();      
      if (do_normalisation) curr_profiles.normalise();
      else if (curr_profiles.no_items > 0)
	curr_profiles.add_origins();
      if (curr_profiles.no_items > 0 || found_some)
	{
	  cout << curr_profiles << "***********" << endl;;
	  if (curr_profiles.no_items > 0) found_some = TRUE;
	  else found_some = FALSE;
	}
      curr_profiles.destroy_all();
      curr_regions.destroy_all();
    }
}


// Local Variables:
// compile-command: "make data_extraction |& grep -v warning"
// End:
