/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/* test program */

#include <image.h>
#include <imgsrc.h>
#include <stdio.h>
#include <getopt.h>
#include <ProfileSet.h>
#include <sgmoviestore.h>
#include <RGB8Image.h>
#include <utils.h>
#include <imgsrc_utils.h>
#include <GroundPeopleTracker.h>
#include <ReadingExternalData.h>


#define USAGE "ground_track -i options [-p file] [-l n] [-q] \
[-o file][-z file] [-c] \n\
-i input image options (-i help for details) \n\
-p parameter file \n\
-l last frame number to process \n\
-q quick mode -- display no images \n\
-o optional output movie file \n\
-z optional output image file  \n\
-c input car data (files specified in parameter file)\n"

 
void do_some_visualisation(GroundPeopleTracker*,ImageSource&, ProfileSet& );
void output_results(GroundPeopleTracker*, ImageSource&,
		    ProfileSet&, MovieStore*, char* );


int MAX_OBJECTS = 
AmbGlobals.register_int("MAX_OBJECTS", 8,
			&MAX_OBJECTS,  FALSE,
			"main", 
			"The maximum number of silhouettes to track");

char* default_car_images =
AmbGlobals.register_string("DEFAULT_CAR_IMAGES", 
"/vol/COMMON/ftp/scs/people/amb/READING_IN/images/silouhette_279.pgm.gz",
			     &default_car_images, TRUE, "main",
			     "Optional car silouhette images");

char* default_car_boxes = 
AmbGlobals.register_string("DEFAULT_CAR_BOXES",
"/vol/COMMON/ftp/scs/people/amb/READING_IN/boxes/gpbbox_279.dat",
			   &default_car_boxes, TRUE, "main",
			   "Optional car bounding box data file");


main(int argc, char** argv)
{
  Profile::draw_colour = TRUE;
  Profile::draw_boxes = FALSE;

  bool show_images = TRUE;
  char *output_movie = NULL;
  char *parameter_file = NULL;
  char* output_image = NULL;
  ImageSource* input_source = NULL;
  MovieStore* mv_out = NULL; 

  bool use_car_data = 
    AmbGlobals.register_bool("USE_CAR_DATA", FALSE,
			     &use_car_data, TRUE, "main",
			     "If this option is set then Reading car \
tracking data is used");
			     
  char *car_images = NULL;
  char *car_boxes = NULL;

  int last_frame = 0;
  int c;
  while ((c = getopt(argc, argv, "i:p:l:qo:z:c")) != EOF)
    switch(c)
      {
      case 'i':
	input_source = get_live_or_movie_feed(optarg);
	break;

      case 'p':
	parameter_file = optarg;
	break;

      case 'l':
	if (optarg != NULL) last_frame = atoi(optarg);
	break;

      case 'q':
	show_images = FALSE;
	break;

      case 'z':
	if (optarg != NULL) 
	  output_image = optarg;
	break;

      case 'o':
	if (optarg != NULL) output_movie = optarg;
	break;
	
      case 'c':
	use_car_data = TRUE;
	break;
		  
      case '?':
      default:
	fprintf(stderr, USAGE);
	exit(1);
      }

  AmbGlobals.parse_parameter_file(parameter_file);
  AmbGlobals.create_interface(argc, argv);
  if (use_car_data)
    {
      car_images = default_car_images;
      car_boxes = default_car_boxes;
    }

  ImageSource& vidsrc = *input_source;
  ProfileSet curr_profiles; // pedestrian silhouettes

  // optional car tracking input
  ReadingExternalData* reading_data = NULL;
  if ((car_images != NULL) || (car_boxes != NULL))
    reading_data = new ReadingExternalData(car_images, car_boxes,
					   vidsrc.get_frame_no());

  GroundPeopleTracker* people_tracker = new
    GroundPeopleTracker(&vidsrc, NULL, &curr_profiles,
			reading_data);


  if (show_images) vidsrc.display();

  if (output_movie != NULL)
    mv_out = new MovieStore(output_movie, 0, TRUE);

  /********************************************************************/
  /*                                                                  */
  /*                      Main Loop                                   */
  /*                                                                  */
  /********************************************************************/

  while (vidsrc.get_next() != NULL) 
    {
      if ((last_frame != 0) && (vidsrc.get_frame_no() >= last_frame))
	break;
      people_tracker->MainLoop(MAX_OBJECTS);
      if (show_images)  
	do_some_visualisation(people_tracker, vidsrc, curr_profiles);
      else fprintf(stderr," frame %d \n", vidsrc.get_frame_no());

      output_results(people_tracker, vidsrc, curr_profiles, mv_out,
		     output_image);
    }
  delete (mv_out);  

  /********************************************************************/
}


// do_some_visualisation
// display images etc
void do_some_visualisation(GroundPeopleTracker* people_tracker,
			   ImageSource& vidsrc,
			   ProfileSet& curr_profiles)
{
  static RGB8Image* rgb_viz_image = NULL;
  vidsrc.set_title("Input image");
  // display video image
  vidsrc.display();
  // display profiles over current video image on screen
  curr_profiles.draw();
  
  Image* viz_image = people_tracker->get_viz_image();
  if (rgb_viz_image == NULL)
    {
      // wrap an RGB8 Image around the 8-bit grayscale data
      rgb_viz_image = new RGB8Image(viz_image->get_width(),
				    viz_image->get_height(),
				    viz_image->get_data());
    }

  rgb_viz_image->set_title("Camera view of computer model");
  rgb_viz_image->display();

  // display the current background image
  people_tracker->detect_motion->background->set_title("Background \
image");

  people_tracker->detect_motion->background->display();
}


// output_results
// if required output movie visualisation.
// output silhouette images and data files
void output_results(GroundPeopleTracker* people_tracker,
		    ImageSource& vidsrc,
		    ProfileSet& curr_profiles,
		    MovieStore* mv_out,
		    char* output_filename)
{
  // dump a movie frame showing input image with superimposed contours
  if (mv_out != NULL)
    {
      people_tracker->get_tracked_profiles()->
	draw_in_image(vidsrc.get_current(), 
		      GroundPeopleTracker::SMALL_NO_FRAMES);
      mv_out->add_a_frame(vidsrc.get_current());
    }

  // dump the silhouette image and the ground plane data file
  if (output_filename != NULL)
    {
      char str_buffer[200];
      int frame_no = vidsrc.get_frame_no() - 1;
/*
      sprintf(str_buffer,"%s%.5d.pgm", output_filename, frame_no);
      people_tracker->get_viz_image()->save_pm(str_buffer);
*/

      sprintf(str_buffer,"%s%.5d.dat", output_filename, frame_no);
      ofstream data_out(str_buffer);
      for (curr_profiles.start(); curr_profiles.current_ok();
	   curr_profiles.forward())
	{
	  if (curr_profiles.get_current().get_type() == TRACKED_PROFILE)
	    {
	      TrackedProfile* curr_prf = (TrackedProfile*)
		&curr_profiles.get_current();
	      if (curr_prf->time_alive > GroundPeopleTracker::SMALL_NO_FRAMES)
		curr_prf->dump_reading_info(data_out);
	    }
	}
      data_out.close();
/*
      sprintf(str_buffer,"gzip -f %s%.5d.pgm %s%.5d.dat", output_filename,
	      frame_no, output_filename, frame_no);
      system(str_buffer);	// compress image, data	  
*/
    }
}


