/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/* test for ProfileSequence class */

#include <ProfileSequence.h>
#include <gl.h>
#include <getopt.h>
#include <math.h>
#include <stdlib.h>
#include <utils.h>

#define USAGE " process_sequence -a alignfile -t trackfile -o outputfile \
-p pca_file -s show_file \n\
-r reflect_file \n\
-m model_input_file -n look_at_mode_n \n\
-c center_file \n\
-x use_extended_model \n\
-q quiet mode (no GUI) \n"


static bool use_extended_model =
AmbGlobals.register_bool("EXTENDED_MODEL", FALSE,
			 &Profile::draw_arrow, TRUE,
			 "main", 
			 "Use extended model ? The extended model \
includes the tracked displacement of each object as part of the \
shape vector");

static int mode_index = 
AmbGlobals.register_int("MODE_NUMBER", 1,
			&mode_index, FALSE,
			"main",
			"The shape mode to display (minimum is 1)");

static char *inputfile = 
AmbGlobals.register_filename("INPUT_SEQUENCE_FILE", "",
			     &inputfile, 
			     "main",
			     "The input file containing extracted \
silouhette (profile) data");
			     
static char *model_file =
AmbGlobals.register_filename("INPUT_MODEL_FILE", "",
			     &model_file,
			     "main",
			     "The input linear model file");

static char *outfile =
AmbGlobals.register_filename("OUTPUT_FILE", "",
			     &outfile,
			     "main",
			     "Filename for output");

static char *functions[] = { "align", "center", "reflect",		     
			     "track", "show shapes", 
			     "principal component analysis",
			     "display modes",
			     NULL };

static char *function_mode = 
AmbGlobals.register_string("FUNCTION", "display_modes",
			   &function_mode, TRUE,
			   "main", 
			   "Pick a function ... \n\
If no output file has been specified, use the input file with \
an additional extension described below.\
\n\n\
align: Align the shapes in the data file. Output extension is '.align'\
\n\n\
center: Recenter the shapes (i.e. align without rotating). \
Output extension is '.centered' \
\n\n\
reflect: Reflect each input shape. Output extension is '.reflect'\
\n\n\
track: Label and track the input shapes between frames. Output \
extension is '.track' \
\n\n\
show_shapes: Display the input shapes. If a model file has been \
specified project each shape into the model subspace. If an \
output file has been specified then typing 'n' at the terminal will \
throw away the displayed shape. Any other keys accepts the shape\
\n\n\
principal component analysis: Do PCA on input data. Output a model \
file for tracking etc\
\n\n\
display modes: Visualize the shape modes of variation for the input \
model file\n\n", functions);



#include <termios.h>
#include <osfcn.h>

int setterm (void)
{
  termios term ;
  ioctl (0, TCGETS, &term) ;
  term.c_lflag = 0 ;
  term.c_cc [VEOF] = 1 ;
  term.c_cc [VEOL] = 0 ;
  if (ioctl (0, TCSETS, &term))
    return 0 ;
  return 1 ;
}

main(int argc, char** argv)
{
  Profile::draw_colour = TRUE;
  Profile::draw_boxes = TRUE;
  Profile::draw_linewidth = 8;
  Profile::draw_box_size = 2.0;

  bool no_interface = FALSE;
  char mode = 'd';
  int c;
  bool tmp_sequence = FALSE;
  while ((c = getopt(argc, argv, "qxn:m:a:c:r:t:o:p:s:")) != EOF)
    switch(c)
      {
      case 'n':
	if (optarg != NULL) mode_index = atoi(optarg);
	break;
      case 'm':
	if (optarg != NULL) model_file = optarg;
	break;
      case 'a':
      case 'c':
      case 'r':
      case 't':
      case 'p':
      case 's':
	inputfile = optarg;
	mode = c;
	break;
      case 'o':
	outfile = optarg;
	break;	

      case 'x':
	Profile::draw_arrow = TRUE;
	break;

      case 'q':
	no_interface = TRUE;
	break;

      case '?':
      default:	
	fprintf(stderr," usage - %s", USAGE);
	exit(1);
      }
  for (int i = 0; functions[i] != NULL; i++)
    if (functions[i][0] == mode) function_mode = functions[i];      

  if (!no_interface)
    AmbGlobals.create_interface(argc, argv);

  mode = function_mode[0];

  char* in_ext = strrchr(inputfile, '.');
  if (in_ext != NULL)
    if ((strncmp(in_ext,".Z",2) == 0) ||
	(strncmp(in_ext, ".gz", 2) == 0))
      {
	char sys_command[200];
	sprintf(sys_command," zcat %s > /tmp/tmp_sequence.txt",
		inputfile);
	system(sys_command);
	inputfile = "/tmp/tmp_sequence.txt";
	tmp_sequence = TRUE;
      }
  
  if (strlen(outfile) == 0) outfile = NULL;
  if (strlen(inputfile) == 0) inputfile = NULL;
  if (strlen(model_file) == 0) model_file = NULL;

  ProfileSequence data(inputfile, model_file);

  switch (mode)
    {
    case 'a':
      data.align(outfile);
      break;
    case 'c':
      data.center(outfile);
      break;
    case 'r':
      data.reflect(outfile);
      break;
    case 't':
      data.track(outfile);
      break;
    case 's':
      {
	prefsize(300 ,400);
#ifndef VOGL
	foreground();
#endif
	winopen(" profiles ");
	color(BLACK);
	clear();  
	Point2 centre(150, 200);
	if (outfile == NULL) data.display_shapes(centre);
	else 
	  {
	    setterm();
	    data.edit_shapes(centre, outfile);
	  }
	break;
      }


    case 'p':
      data.do_analysis();
      if (outfile != NULL)
	{
	  ofstream out_file(outfile);
	  out_file << *(data.pca) << endl;
	  out_file.close();
	  break;
	}
      outfile = NULL;

    case 'd':
      {
	if (data.pca  == NULL) break;
	int indx = 0;
	if (outfile != NULL)
	  {
	    ofstream data_out(outfile);
	    data_out << *(data.pca);
	    data_out.close();
	  }
	data.display_mode(mode_index);
	gexit();
      }
    }    
  if (tmp_sequence) system("rm /tmp/tmp_sequence.txt");

}

// Local Variables:
// compile-command: "make sequencetest"
// End:
