/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/*
 * BoundaryPoints.cc
 *
 * the ordered set of points on the boundary of an object
 * 
 */

#include <BoundaryPoints.h>
#include <gl.h>



static const realno LARGE =  1e6; // a large number
static const realno HEIGHT = 100; // fixed height of person in pixels

void BoundaryPoints::gauss_smooth(realno sd, int window_size)
{
  realno* coeff = new realno[2 * window_size + 1];
  realno csum = 0;
  realno denom = 2 * sd * sd;
  int n,j;
  int win2 = window_size << 1;
  for (n = -window_size; n <= window_size; n++)
      csum += (coeff[n + window_size] = exp(-((realno) n * (realno) n)/denom));
  for (n = 0; n <= win2; n++)
    coeff[n] /= csum;

  PointVector result(no_points);
  Point2 temp;
  for (int i = 0; i < no_points; i++)
    {
      temp = Point2(0,0);
      j = (i - window_size) + no_points;
      for (n = 0; n <= win2; n++)
	temp = temp + (coeff[n] * point_data()[j++ % no_points]);
	
      result.point_data()[i] = temp;
    }
  delete [] coeff;
  this->PointVector::operator=(result);
}
 
void BoundaryPoints::find_best_line(realno& inv_grad, realno& ox, realno& oy)
{
  realno sy2, sx2, sx, sy, sxy, tx, ty;
  sx = sy = sxy = sy2 = sx2 = 0.0;
  Point2* enddata = end_data();
  for (Point2* curr = point_data(); curr < enddata; curr++)
    {
      tx = curr->x;
      ty = curr->y;
      sx += tx;
      sy += ty;
      sx2 += (tx * tx);
      sy2 += (ty * ty);
      sxy += (tx * ty);
    }

  sx2 -= (sx * sx) / ((realno) no_points);
  sy2 -= (sy * sy) / ((realno) no_points);
  sxy -= (sx * sy) / ((realno) no_points);

  realno k = 2 * sxy / (sx2 - sy2);
  realno im1 = k / (-1 - sqrt(1 + k*k));
  realno im2 =  k  / (sqrt(1 + k*k) -1);
  realno err1 = (im1*im1*sy2 - im1*2*sxy + sx2) / (im1 * im1 + 1);
  realno err2 = (im2*im2*sy2 - im2*2*sxy + sx2) / (im2 * im2 + 1);
  if (err1 <= err2) inv_grad = im1;
  else inv_grad = im2;

  ox = sx / no_points;
  oy = sy / no_points;

}

void BoundaryPoints::find_robust_best_line(realno& inv_grad, realno&
					   ox, realno& oy)
{
  // first robust center
  find_best_line(inv_grad, ox, oy);
  BoundaryPoints trimmed_pnts(no_points);
  copy(trimmed_pnts);
  for (int lp = 0; lp < 40; lp++)
    {
      realno sum_dist = 0;
      realno sum_sq_dist = 0;
      Point2 n(inv_grad, 1);
      int np = trimmed_pnts.get_no_points();
      fprintf(stderr,"%d %d (%f, %f) \n",lp, np , ox, oy);

      for (Point2* curr = trimmed_pnts.point_data(); 
	   curr < trimmed_pnts.end_data(); curr++)     
	{
	  realno d = fabs((*curr - Point2(ox,oy)).length());
	  sum_dist += d;
	  sum_sq_dist += d * d;
	}
      realno mean_dist = sum_dist / np;
      realno sd_dist = 
	sqrt((sum_sq_dist / np) - (mean_dist * mean_dist));

      realno thresh_d = mean_dist + (1.6 * sd_dist);
      BoundaryPoints new_trimmed;
      for (curr = trimmed_pnts.point_data(); 
	   curr < trimmed_pnts.end_data(); curr++)     
	{
	  realno d = fabs((*curr - Point2(ox,oy)).length());
	  if (d < thresh_d) new_trimmed.add_point(*curr);
	}
      trimmed_pnts = new_trimmed;
      if (trimmed_pnts.no_points < 4) break;
      trimmed_pnts.find_best_line(inv_grad, ox, oy);
    }
  inv_grad = 0;
}

      
void BoundaryPoints::find_end(Point2& n, Point2& origin)
{
  int closest, closest2;
  realno best_dist,  dist;
  realno best_dist2;
  best_dist = 1e6;
  best_dist2 = 1e6;

  for (int i = 0; i < no_points; i++)
    {
      Point2 p =  point_data()[i] - origin;
      dist = fabs(p^n);
      realno side = p*n;

      if (side >= 0)
	{
	  if (dist < best_dist)
	    {
	      closest = i;
	      best_dist = dist;
	    }
	}
      else
	{
	  if (dist < best_dist2)
	    {
	      closest2 = i;
	      best_dist2 = dist;
	    }
	}
    }
  PointVector result(no_points);
  closest2 = (closest2 - closest + no_points) % no_points;
  for (i = 0; i < no_points; i++)
    result.point_data()[i] = point_data()[(closest++ % no_points)] - origin;
  this->PointVector::operator=(result);
  anchor_point = closest2;
}
  
void BoundaryPoints::find_ends(realno inv_grad, realno ox, realno oy)
{
  int closest, closest2;
  realno best_dist,  dist;
  realno best_dist2;
  best_dist = 1e6;
  best_dist2 = 1e6;

  for (int i = 0; i < no_points; i++)
    {
      dist = fabs((point_data()[i].x - ox) - 
		  inv_grad * (point_data()[i].y - oy));
      if (point_data()[i].y > oy)
	{
	  dist = (20 * dist - point_data()[i].y);
	  if (dist < best_dist)
	    {
	      closest = i;
	      best_dist = dist;
	    }
	}
      else
	{
	  dist = (20 * dist + point_data()[i].y);
	  if (dist < best_dist2)
	    {
	      closest2 = i;
	      best_dist2 = dist;
	    }
	}
    }
  PointVector result(no_points);
  Point2 centre(ox,oy);
  closest2 = (closest2 - closest + no_points) % no_points;
  for (i = 0; i < no_points; i++)
    result.point_data()[i] = point_data()[(closest++ % no_points)] - centre;
  this->PointVector::operator=(result);
  anchor_point = closest2;
}

void BoundaryPoints::recenter(realno ox, realno oy)
{
  Point2 centre(ox,oy);
  Point2* lastpnt = end_data();
  while ((--lastpnt) >= point_data()) *lastpnt = *lastpnt - centre;
}
      
void BoundaryPoints::draw_points(int xlo, int ylo)
{
  float vert[2];
  //color(RED);
  bgnpoint();
  Point2* enddata = end_data();
  for (Point2* curr = point_data(); curr < enddata; curr++)
    {
      vert[0] = curr->x + xlo;
      vert[1] = curr->y + ylo;
      v2f(vert);
    }
  endpoint();
  /*color(GREEN);
  bgnline();
  while (curr != NULL)
    {
      vert[0] = curr->dat->x + xlo;
      vert[1] = curr->dat->y + ylo;
      v2f(vert);
      curr = curr->next;
    }
    endline();*/
}

extern Matrix bsplinematrix;

void BoundaryPoints::draw_curve(int ox, int oy, int step)
{
  defbasis(1,bsplinematrix);
  curveprecision(15);
  linewidth(2);
  Coord geom1[4][3];
  int i,j;
  geom1[0][2] = geom1[1][2] = geom1[2][2] = geom1[3][2] = 0;
  //char test_str[50];
  for (i = 0; i < no_points; i += step)
    {
      geom1[0][0] = point_data()[i].x + ox;
      geom1[0][1] = point_data()[i].y + oy;
      j = (i+step) % no_points;
      geom1[1][0] = point_data()[j].x + ox;
      geom1[1][1] = point_data()[j].y + oy;
      j = (j+step) % no_points;
      geom1[2][0] = point_data()[j].x + ox;
      geom1[2][1] = point_data()[j].y + oy;
      j = (j+step) % no_points;
      geom1[3][0] = point_data()[j].x + ox;
      geom1[3][1] = point_data()[j].y + oy;
      crv(geom1);
    }
}
