/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/*
 * RegionSet.cc
 * 
 * functions for RegionSet class
 * (declared in region.h)
 *
 */

#include <RegionSet.h>
#include <ProfileSet.h>

#include <list.h>
#include <stack.h>
#include <stdio.h>
#include <BWImage.h>
#include <gl.h>

#define BORDER_WIDTH 8

#define MIN(a,b) (( (a) < (b)) ? (a) : (b))
#define MAX(a,b) (( (a) > (b)) ? (a) : (b))

#define TEST(p) if (*(p) == MARK) {PUSH(p); *(p) = CLEAR;}

listimplement(Region);

void RegionSet::grab_regions(Image* img, int min_size, int max_size,
			     bool keep_edge_regions)
{
  Image*  scene = new BWImage(img->get_width(), img->get_height());
  scene->clear();
  /*if (gl_win == NULLWIN)
    {
      prefsize(img->get_width(), img->get_height());
      gl_win = winopen("silhouettes");
      color(BLACK);
      clear();
      }*/
  Image* temp_image = NULL;
  unsigned char* sptr, *eptr, *tptr;
  unsigned char* ptr = img->get_data();
  unsigned char* scene_data = scene->get_data();
  int offset;
  int width = scene->get_width();
  int height = scene->get_height();
  int count, x, y;
  int os1 = width+1;
  int os2 = width;
  int os3 = width-1;
  int tnx = width-BORDER_WIDTH;
  int tny = height-BORDER_WIDTH;
  bool on_edge = FALSE;
  for (y = BORDER_WIDTH; y < tny; y++)
    {
      tptr = img->get_pixel(BORDER_WIDTH,y);
      eptr = img->get_pixel(tnx,y);
      for (;tptr < eptr; tptr++)
	{
	  if (*tptr == MARK)
	    {
	      PUSH(tptr);
	      *tptr = CLEAR;
	      int xhi = 1;
	      int yhi = 1;
	      int xlo = tnx;
	      int ylo = tny;
	      count = 0;
	      on_edge = FALSE;
	      while (!STACKEMPTY)
		{
		  POP(sptr);
		  img->get_pixel_coords(sptr, x, y);
		  offset = sptr - ptr;
		  *(scene_data + offset) = MARK;
		  count++;
		  
		  if (x < xlo) xlo = x;
		  if (x > xhi) xhi = x;
		  if (y < ylo) ylo = y;
		  if (y > yhi) yhi = y;
		  
		  /* add marked neighbours to stack */
		  if (y > BORDER_WIDTH) {
		    if (x > BORDER_WIDTH) TEST(sptr-os1);
		    TEST(sptr-os2);
		    if (x < tnx) TEST(sptr-os3);
		  } 
		  else on_edge = TRUE;
		  if (x > BORDER_WIDTH) { TEST(sptr-1) }
		  else on_edge = TRUE;
		  if (x < tnx) { TEST(sptr+1)}
		  else on_edge = TRUE;
		  if (y < tny) {
		    if (x > BORDER_WIDTH) TEST(sptr+os3);
		    TEST(sptr+os2);
		    if (x < tnx) TEST(sptr+os1);
		  }
		  else on_edge = TRUE;
		  
		}
	      if (keep_edge_regions) on_edge = FALSE;
	      xlo -= 2; xhi += 2;
	      ylo -= 2; yhi += 2;
	      xhi = ((((xhi - xlo) >> 3)+1) << 3) + xlo - 1;
	      yhi = ((((yhi - ylo) >> 3)+1) << 3) + ylo - 1;
	      if (xlo < 0) xlo = 0;
	      if (xhi >=  width ) xhi = width -1;
	      if (ylo < 0) ylo = 0;
	      if (yhi >= height) yhi = height -1;
	      temp_image = scene->extract_subimage(xlo,xhi,ylo,yhi);
	      if (!on_edge && (count <= max_size) && (count >= min_size))
		add(new Region(xlo,xhi,ylo,yhi,temp_image));
	      else if (temp_image != NULL) delete(temp_image);
	    }
	}
    }
  delete (scene);
}


void RegionSet::grab_all_regions(Image* img, int min_size, int max_size)
{
  unsigned char* sptr, *eptr, *tptr;
  unsigned char* ptr = img->get_data();
  int offset;
  int width = img->get_width();
  int height = img->get_height();
  int count, x, y;
  int os1 = width+1;
  int os2 = width;
  int os3 = width-1;
  int tnx = width-BORDER_WIDTH;
  int tny = height-BORDER_WIDTH;
  for (y = BORDER_WIDTH; y < tny; y++)
    {
      tptr = img->get_pixel(BORDER_WIDTH,y);
      eptr = img->get_pixel(tnx,y);
      for (;tptr < eptr; tptr++)
	{
	  if (*tptr == MARK)
	    {
	      PUSH(tptr);
	      *tptr = CLEAR;
	      int xhi = 1;
	      int yhi = 1;
	      int xlo = tnx;
	      int ylo = tny;
	      count = 0;
	      while (!STACKEMPTY)
		{
		  POP(sptr);
		  img->get_pixel_coords(sptr,x,y);
		  offset = sptr - ptr;
		  count++;
		  
		  if (x < xlo) xlo = x;
		  if (x > xhi) xhi = x;
		  if (y < ylo) ylo = y;
		  if (y > yhi) yhi = y;
		  
		  /* add marked neighbours to stack */
		  if (y > BORDER_WIDTH) {
		    if (x > BORDER_WIDTH) TEST(sptr-os1);
		    TEST(sptr-os2);
		    if (x < tnx) TEST(sptr-os3);
		  } 
		  if (x > BORDER_WIDTH) { TEST(sptr-1) }
		  if (x < tnx) { TEST(sptr+1)}
		  if (y < tny) {
		    if (x > BORDER_WIDTH) TEST(sptr+os3);
		    TEST(sptr+os2);
		    if (x < tnx) TEST(sptr+os1);
		  }
		}

	      //if (!((xlo < (0.75 * width)) && 
	      //(xhi > (0.25 * width)) &&
	      //(ylo < (0.75 *height)) &&
	      //(yhi > (0.25 * height))) &&
	      
	      if ((xlo > BORDER_WIDTH) &&
		  (xhi < (width - BORDER_WIDTH)) &&
		  (ylo > BORDER_WIDTH) &&
		  (yhi < (height - BORDER_WIDTH)) &&
		  ((count <= max_size) && (count >= min_size)))
		add(new Region(xlo,xhi,ylo,yhi,NULL));
	    }
	}
    }
}

void RegionSet::label_regions(int& next_ref)
{
  for (item(Region)* curr = first; curr != NULL; curr = curr->next)
    if (curr->dat->ref_no == 0) curr->dat->ref_no = next_ref++;
}

void RegionSet::draw_regions()
{
  // winset(gl_win);
  for (item(Region)* curr = first; curr != NULL; curr = curr->next)
    curr->dat->draw_points();
}


void RegionSet::draw_boxes()
{
  for (item(Region)* curr = first; curr != NULL; curr = curr->next)
     curr->dat->draw_box();
}


void RegionSet::clear_window()
{
  //winset(gl_win);
  color(BLACK);
  clear();
}

void RegionSet::trace_regions()
{
  for (item(Region)* curr = first; curr != NULL; curr = curr->next)
      curr->dat->trace();
}

void RegionSet::smooth(realno sd, int window_size)
{
  for (item(Region)* curr = first; curr != NULL; curr = curr->next)
    curr->dat->smooth_boundary(sd, window_size);
}

ProfileSet RegionSet::to_profiles()
{
  ProfileSet profiles = ProfileSet();
  Profile* tmp;
  for (item(Region)* curr = first; curr != NULL; curr = curr->next)
      {
	 tmp = curr->dat->to_profile();
	 if (tmp != NULL) profiles.add(tmp);
       }
  return (profiles);
}      

static realno bbdistance(Region* r, Region* t)
{
    int d1, d2, dx, dy;
    d1 = r->xhi - t->xlo;
    d2 = t->xhi - r->xlo;
    if (d1>0) dx=d1;
    else 
      {
	if (d2>0) dx=d2;
	else dx=0;
      }
    d1 = r->yhi - t->ylo;
    d2 = t->yhi - r->ylo;
    if (d1>0) dy=d1;
    else
      {
	if (d2>0) dy=d2; 
	else dy=0;
      }
    d1 = MAX(r->xhi - r->xlo, r->yhi - r->ylo) + 1;
    d2 = MAX(t->xhi - t->xlo, t->yhi - t->ylo) + 1;
    
    realno dd = (dx  + dy) / ((realno) (d1 + d2));
    return (dd);
}

void RegionSet::merge_regions(realno dthld)
{
  int flag=1;
  //dthld = dthld*dthld;    /* bbdistance returns square of distance */
  item(Region)* reg_r, *reg_t, *reg_t2;
  
  while (flag==1) 
    {
      flag = 0;
      for (reg_r=first; reg_r != NULL; reg_r = (reg_r->next))
	{
	  for (reg_t = (reg_r->next); reg_t != NULL; reg_t = (reg_t->next))
	    {
	      if (bbdistance(reg_r->dat, reg_t->dat) < dthld)
		{
		  reg_r->dat->merge(reg_t->dat);
		  flag = 1;
		  reg_t2 = (reg_t->prev);
		  destroy(reg_t);
		  reg_t = reg_t2;
		}
	    }
	}
    }
}

void RegionSet::filter(realno min_region_size = 0.0)
{
  for (item(Region)* regn = first; regn != NULL; )
    {
      if (fabs((regn->dat->xhi - regn->dat->xlo) * ( regn->dat->yhi - regn->dat->ylo))
	  < min_region_size) 
	{
	  item(Region)* rgn_tmp = regn;
	  regn = regn->next;
	  destroy(rgn_tmp);
	}
      else regn = regn->next;

    }
}
	  


