/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/*
 * Tracker.cc
 * 
 * a module for tracking Regions from frame to frame
 *
 */ 

#include <RegionTrack.h>
#include <list.h>


static const realno LARGE = 1e10;
static const realno DEFAULT_NEAR_THRESHOLD = 4.0;

#ifndef MAX
#define MAX(a,b) (( (a) > (b)) ? (a) : (b))
#endif


realno track_distance(Region* reg1, Region* reg2)
{
  Point2 diff = Point2(reg1->ox - reg2->ox, reg1->oy - reg2->oy);
  realno w = (reg1->width + reg2->width) / 2;
  realno h = (reg1->height + reg2->height) / 2;

  realno dw = fabs(reg1->width - reg2->width) / w;
  realno dh = fabs(reg1->height - reg2->height) / h;
  realno dl = diff.length() / MAX(w,h);
  if (dw > 1.0 || dh > 1.0 || dl > 1.0) return (LARGE);  
  return (dw + dh + dl);
}
 

RegionTracker::RegionTracker(DistFunc* dfn, realno near_thresh)
{
  last_frame = NULL;
  new_frame = NULL;
  if (dist_fn != NULL) dist_fn = dfn;
  else dist_fn = (&track_distance);
  near_threshold = near_thresh;
  last_no = new_no = 0;
}

RegionTracker::RegionTracker()
{
  last_frame = NULL;
  new_frame = NULL;
  dist_fn = (&track_distance);
  near_threshold = DEFAULT_NEAR_THRESHOLD;
  last_no = new_no = 0;
}


void RegionTracker::calculate_distances()
{
  if (last_frame == NULL || new_frame == NULL)
    {
      fprintf(stderr, " RegionTracker: no data in calculate_distances \n");
      exit(1);
    }
  last_no = last_frame->no_items;
  new_no = new_frame->no_items;
  if (new_no > MAX_TRACK_OBJECTS)
    {
      fprintf(stderr, " Sorry too many objects to track \n");
      new_no = MAX_TRACK_OBJECTS;
    }
  if (last_no > MAX_TRACK_OBJECTS) last_no = MAX_TRACK_OBJECTS;
 
  item(Region)* curr1 = last_frame->first;
  item(Region)* curr2;
  int i,j;
  for (i = 1; i <= last_no ; i++)
    {
      curr2 = new_frame->first;
      for (j = 1; j <= new_no ; j++)
	{
	  far[i][j] = dist_fn(curr1->dat, curr2->dat);
	  curr2 = curr2->next;
	}
      curr1 = curr1->next;
    }
  for (i = 1; i <= last_no; i++)
    far[i][0] = 0;
  for (j = 1; j <=  new_no; j++)
    far[0][j] = 0;
}

realno RegionTracker::find_best_match(int& from_indx, int& to_indx)
{
  realno best_so_far = LARGE;
  from_indx = to_indx = -1;
  for (int j,i = 1; i <= last_no; i++)
    {
      if (far[i][0] == 0)
	for  (j = 1; j <= new_no; j++)
	  if ((far[0][j] == 0) && (far[i][j] < best_so_far))
	    {
	      from_indx = i;
	      to_indx = j;
	      best_so_far = far[i][j];
	    }
    }
  if (from_indx > 0)
    {
      far[from_indx][0] = 1;
      far[0][to_indx] = 1;
      from_indx--;to_indx--;
    }
  return (best_so_far);
}

void RegionTracker::process_frame(RegionSet* next_frame)
{
  if ((next_frame != NULL) && (next_frame->no_items > MAX_TRACK_OBJECTS))
    return;
  last_frame = new_frame;
  new_frame = next_frame;
  if ((last_frame == NULL) || (last_frame->first == NULL)) return;
  if ((new_frame == NULL) || (new_frame->first == NULL)) return;
  calculate_distances();
  
  int from_indx, to_indx;
  realno best_dist = find_best_match(from_indx, to_indx);
  while ((best_dist < near_threshold) && (from_indx != -1))
    {
      Region* r2 = (*new_frame)[to_indx]->dat;
      Region* r1 = (*last_frame)[from_indx]->dat;
      r2->ref_no = r1->ref_no;
      r2->direction = Point2(r2->ox - r1->ox,
			     r2->oy - r1->oy);

      best_dist = find_best_match(from_indx, to_indx);
    }
}



