/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

//
// bspline.cc
// 
// function to convert a RegionBoundary
// to spline form.
//
// (added reparametrisation optimisation method
// 14/12/94)
//
//

#include <BoundaryPoints.h>
#include <math.h>
#include <boolean.h>
#include <NagMatrix.h>
#include <stdlib.h>
#include <stdio.h>
#include <SplineWeights.h>

#ifndef MIN
#define MIN(a,b) (( (a) < (b)) ? (a) : (b))
#endif

#include <shape_constants.h>
#include <profile.h>

class SplineMatrix : public NagMatrix
{
public:

  realno* vdata;
  
  SplineMatrix(int nsubs);
  ~SplineMatrix() { delete [] vdata; }

};

SplineMatrix::SplineMatrix(int nsubs)
: NagMatrix(NO_CONTROL_POINTS, NO_CONTROL_POINTS)
{

  fprintf(stderr, " building matrix \n ");
  // calculate the matrix 
  const int N = NO_CONTROL_POINTS;
  NagVector tempdata(N);
  realno temp;
  realno Ui;
  int i,j,l;
  for (l = 0; l < N; l++)
    {
      temp = 0;
      for (i = 0; i < (N*nsubs); i++)
	{
	  Ui = i / (realno) nsubs;
	  temp += B_func(l, Ui) * B_func(0, Ui);
	}
      tempdata[l] = temp;
    }
  
  NagMatrix tmp_matrix(N, N);
  for (i = 0; i < N; i++)
    for (j = 0; j < N; j++) 
      tmp_matrix.set(i,j, tempdata[(i - j + N) %  N]);
  tmp_matrix.invert(*this);

  vdata = new realno[4 * nsubs + 1];
  for (j = 0; j <= 4 * nsubs; j++)
    vdata[j] = B_func(0,(j - (2 * nsubs)) / (realno) nsubs);
} 

static SplineMatrix* s_matrices[6] = { NULL, NULL, NULL, NULL, NULL, NULL};

void setup_spline_matrices()
{
  fprintf(stderr, " ... setting up matrices \n");
  s_matrices[0] = new SplineMatrix(32);
  s_matrices[1] = new SplineMatrix(16);
  s_matrices[2] = new SplineMatrix(8);
  s_matrices[3] = new SplineMatrix(4);
  s_matrices[4] = new SplineMatrix(2);
  s_matrices[5] = new SplineMatrix(1);

}


bool BoundaryPoints::filter()
// select data points by arc-length division 
{
  int SUB_N = 32;		
  int indx = 0;

  // choose a precomputed points-to-spline matrix
  // so that
  // # boundary points > # new sampled boundary points
  // but only just !
  // 
  // Hence start with RHS large and decrease until
  // condition satisfied 

  while ((no_points < (NO_CONTROL_POINTS * SUB_N)) && (SUB_N >= 2))
    {
      SUB_N /= 2;
      indx++;
    }

  // if the setup function has not been called yet, initialise
  // the required matrix
  if (s_matrices[indx] == NULL) s_matrices[indx] = new SplineMatrix(SUB_N);
  

  // pass1 - store all lengths 
  realno* arc = new realno[no_points+1];
  arc[0] = 0;
  Point2 diff, tmp;
  Point2* input_pnts = point_data();
  for (int i = 0; i < no_points; i++)
    {
     diff = input_pnts[(i+1) % no_points] - input_pnts[i];
     arc[i+1] = arc[i] + diff.length();
    }

  realno total_arclength = arc[no_points];

  // pass2 - select relevant points

  // set the spacing interval for parametric values
  realno interval = arc[no_points] / ((realno)(NO_CONTROL_POINTS * SUB_N));

  int curr_pos = 1;
  int new_no_points = NO_CONTROL_POINTS * SUB_N;

  BoundaryPoints result(new_no_points);
  result.point_data()[0] = point_data()[0];
  realno l1,l2, iarc;
  for (i = 1; i < new_no_points; i++)
    {
      iarc = ((realno)i) * interval;

      // skip points if necessary
      while (arc[curr_pos] < iarc)
	curr_pos++;
      if (curr_pos > no_points) 
	{
	fprintf(stderr,"BoundaryPoints::filter ABORT\n");
		fflush(stderr);
	abort();
	}

      // use linear interpolation of lengths to define new point
      l1 = arc[curr_pos-1]; 
      l2 = arc[(curr_pos) % no_points];
      // check for gone full circle
      if (curr_pos == no_points) l2 += total_arclength;

      realno lmda = (iarc - l1) / (l2 - l1);
      
      result.point_data()[i] = point_data()[curr_pos-1] * (1.0 - lmda) 
	+ point_data()[(curr_pos) % no_points] * lmda;

    }
  delete [] arc;
  *this = result;
  return (TRUE);
}

bool BoundaryPoints::convert_to_spline(PointVector* result)
{
  if (filter() == FALSE)  return (FALSE);
  to_spline(result);
  return (TRUE);
}

void BoundaryPoints::to_spline(PointVector* result)
{
  if (result->get_no_points() < NO_CONTROL_POINTS)
    {
      fprintf(stderr, " error in BoundaryPoints::to_spline \n");
		fflush(stderr);
      abort();
    }

  // get the relevant points-to-spline matrix 
  // (again !!)
  int SUB_N = no_points / NO_CONTROL_POINTS;
  int temp_n = SUB_N; int indx = 0;
  while (temp_n < 32) { temp_n *= 2; indx++; }
  if (s_matrices[indx] == NULL) s_matrices[indx] = new SplineMatrix(SUB_N);
  SplineMatrix* s_matrix = s_matrices[indx];

  // calculate the RHS of the simultaneous equations
  realno* vdat = s_matrix->vdata;
  int i,j,l, i2; 
  NagVector rhsx(NO_CONTROL_POINTS);
  NagVector rhsy(NO_CONTROL_POINTS);
  int ntotal = SUB_N * NO_CONTROL_POINTS;
  realno temp; 
  // get the corresponding vectors 
  for (l = 0; l < NO_CONTROL_POINTS; l++)
    {
      rhsx.get_data()[l] = 0;
      rhsy.get_data()[l] = 0;
      j = 0;
      for (i = (l - 2) * SUB_N;i <= (l + 2) * SUB_N; i++)
	{
	  i2 = (i + ntotal) % ntotal;
	  temp = vdat[j++];
	  rhsx.get_data()[l] += temp * (point_data()[i2].x);
	  rhsy.get_data()[l] += temp * (point_data()[i2].y);
	}
    }
  NagVector resx(NO_CONTROL_POINTS);
  NagVector resy(NO_CONTROL_POINTS);
  s_matrix->multiply(rhsx,resx);
  s_matrix->multiply(rhsy,resy);
  Point2* spln = result->point_data();
  for (l = 0; l < NO_CONTROL_POINTS; l++)
    *spln++ = Point2(resx[l],resy[l]);
}

//////////////////////////////////////////////////////////////////////////
// alternative methods ....						//
//									//
//////////////////////////////////////////////////////////////////////////


// calculate a spline from an arbitrary set of Boundary points
// without worrying about speed / efficiency
void calculate_spline(NagVector& u_values, PointVector* data,
			     PointVector* result)
{
  NagMatrix M(NO_CONTROL_POINTS, NO_CONTROL_POINTS);
  M.clear(0);
  int n = u_values.get_size();
  int i;

  for (int j = 0; j < NO_CONTROL_POINTS; j++)
    for (int k = 0; k < NO_CONTROL_POINTS; k++)
      for (i = 0; i < n; i++)
	*M.get(j,k) += B_func(j, u_values[i]) *
	  B_func(k,u_values[i]);


  NagMatrix B(NO_CONTROL_POINTS, 2);
  NagVector rhs_x(NO_CONTROL_POINTS, B.get(0,0));
  NagVector rhs_y(NO_CONTROL_POINTS, B.get(0,1));

  rhs_x.clear(0);
  rhs_y.clear(0);

  
  for (j = 0; j < NO_CONTROL_POINTS; j++)
    for (i = 0; i < n; i++)
      {
	realno tmp = B_func(j, u_values[i]);
	rhs_x[j] += tmp * data->point_data()[i].x;
	rhs_y[j] += tmp * data->point_data()[i].y;
      }
  rhs_x.reset();
  rhs_y.reset();

  NagMatrix C;
  M.solve_equations(B, C);
  
  for (j = 0; j < NO_CONTROL_POINTS; j++)
    result->point_data()[j] = Point2(C.read(j,0), C.read(j, 1));
}
  

// locally optimise the parameter values associated
// with each data point to obtain the best fit
// -- this method is not guaranteed to work in general !
bool BoundaryPoints::find_optimal_spline(Profile* result)
{
  // setup initial parameter values
  filter();
  bool have_filtered = TRUE;
  NagVector u_values(no_points);
  u_values[0] = 0.0;
  realno u_n = no_points;
  int i;
  if (!have_filtered)
    for (i = 1; i < no_points; i++)
      {
	u_values[i] = u_values[i-1]
	  + (point_data()[i] - point_data()[i-1]).length();
	u_n = u_values[no_points-1]
	  + (point_data()[0] - point_data()[no_points-1]).length();
      }
  else
    for (i = 1; i < no_points; i++)
      u_values[i] = i;

  u_values.scale(NO_CONTROL_POINTS / u_n, u_values);

  // find  corresponding spline

  realno rms_change = 1e6;
  realno old_rms_change;
  realno curve_err = 1e6;
  realno old_curve_err;
  NagVector del_u(no_points);
  NagVector del_v(no_points);
  NagVector old_u(no_points);
  do 
    {
      u_values.copy(old_u);
      realno u_change = 0.0;
      old_curve_err = curve_err;
      curve_err = 0;
      calculate_spline(u_values, this, result);
//      result->draw(TRUE);
      for (i = 0; i < no_points; i++)
	{
	  Point2 p, dp, ddp;
	  p.x = p.y = dp.x = dp.y = ddp.x = ddp.y = 0.0;
	  for (int j = 0; j < NO_CONTROL_POINTS; j++)
	    {
	      p +=(B_func(j, u_values[i]) * result->point_data()[j]);
	      dp += (dB_func(j, u_values[i]) *
		     result->point_data()[j]);
	      ddp += (ddB_func(j, u_values[i]) *
		      result->point_data()[j]);
	    }

// check max error too ?

	  Point2 err_u = point_data()[i] - p;
	  curve_err += err_u.length2();

	  realno du_max ;
	  realno du_min ;
	  du_max = 1.0 / NO_CONTROL_POINTS;
	  du_min = - du_max;

	  del_u[i]  = - 1.0 * (err_u * dp) / ((err_u * ddp)  - (dp * dp));
	  
	  if  (del_u[i] >= du_max) del_u[i] = du_max;
	  if (del_u[i] <= du_min) del_u[i] =  du_min;

	  //if (du_max < du_min) abort();

	}
      //u_change = del_u.length2();
      u_values.add(del_u, u_values);
      // make sure u_0 = 0
      for (i = 0 ; i < no_points; i++)
	u_values[i] -= (u_values[0]);
      
      for (i = 1; i < no_points; i++)
	{
	  del_v[i-1] = u_values[i] - u_values[i-1];
	  if (del_v[i-1] < 0) del_v[i-1] = 0;
	}
      
      del_v[no_points-1] = u_values[0] + NO_CONTROL_POINTS -
	u_values[no_points-1];

      del_v.scale(NO_CONTROL_POINTS / del_v.sum(), del_v);

      for (i = 1; i < no_points; i++)
	u_values[i] = u_values[i-1] + del_v[i-1];

      old_u.subtract(u_values,del_u);
      
      old_rms_change = rms_change;
      u_change = del_u.length2();
      rms_change = sqrt(u_change / no_points);
//      fprintf(stderr,"rms_change %f err %f \n", rms_change, curve_err);
//      result->draw(FALSE);
      
    } while ((fabs(rms_change - old_rms_change) > 1e-3) &&
	     (curve_err < old_curve_err));
  if (old_curve_err < curve_err) 
    calculate_spline(old_u, this, result);
      

  return (TRUE);
}
