/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/*
 * profile.cc
 * 
 * a record class for storing the profile data
 * of an object
 *
 * AMB 27/4/93
 */

#include <profile.h>
#include <gl.h>

#include <math.h>
#include <stdlib.h>
#include <SplineWeights.h>
#include <EnvParameter.h>
#ifndef VOGL
#include <gl/get.h>
#endif
#include <image.h>

extern void set_color (int col);

#define SQUARE(x) ((x)*(x))

int NO_CONTROL_POINTS = 32;

bool Profile::draw_boxes = TRUE;
bool Profile::draw_arrow = FALSE;
bool Profile::draw_colour = TRUE;
bool Profile::draw_rectangle = FALSE;
bool Profile::draw_spline = TRUE;
bool Profile::draw_fill = FALSE;
bool Profile::output_points = TRUE;

int Profile::draw_linewidth = 4;
realno Profile::draw_box_size = 1.5;

static const realno LARGE = 1e10;

Profile::Profile(Point2& p_o, realno& p_w, realno& p_h, realno&
		 p_grad, PointVector& pnts, Point2& d) :
		 PointVector(NO_CONTROL_POINTS+1),
		 direction(point_data()[NO_CONTROL_POINTS])
{
  origin = p_o;
  width = p_w;
  height = p_h;
  inv_grad = p_grad;
  if (pnts.get_data() != NULL)
    pnts.NagVector::copy(*this);
  else this->NagVector::clear(0.0);
  direction = d;
  ref_no = 0;
}

Profile::Profile(realno ox, realno oy, realno wdth, realno hgt) :
PointVector(NO_CONTROL_POINTS+1),
direction(point_data()[NO_CONTROL_POINTS]),
origin(ox,oy)
{
  width = wdth;
  height = hgt;
  inv_grad = 0;
  direction.x = direction.y = 0.0;
  clear(0);
  ref_no = 0;
}

Profile::Profile(Profile& original) :
PointVector(NO_CONTROL_POINTS+1),
direction(point_data()[NO_CONTROL_POINTS])
{   
  origin = original.origin;
  width = original.width;
  height = original.height;
  direction = original.direction;
  for (int i = 0 ; i < NO_CONTROL_POINTS; i++)
    spline[i] = original.spline[i];
  ref_no = 0;
  inv_grad = 0;
}

Profile& Profile::operator=(Profile& original)
{ 
  origin = original.origin;
  width = original.width;
  height = original.height;
  direction = original.direction;
  ref_no = original.ref_no;
  inv_grad = 0;
  for (int i = 0 ; i < NO_CONTROL_POINTS; i++)
    spline[i] = original.spline[i];
  return (*this);
}
  
Profile::Profile() : PointVector(NO_CONTROL_POINTS+1),
direction(point_data()[NO_CONTROL_POINTS])
{
  origin.x = origin.y = width = height = inv_grad = 0; 
  direction.x = direction.y = 0;
}

Profile::~Profile()
{
  delete [] data;
  data = NULL;
}

void Profile::grab_data(Profile& original)
{
  origin = original.origin;
  width = original.width;
  height = original.height;
  direction = original.direction;
  for (int i = 0 ; i < NO_CONTROL_POINTS; i++)
    spline[i] = original.spline[i];
}

ostream& operator<<(ostream& out_strm, Profile& pf)
{
  pf.output_to_stream(out_strm);
  return (out_strm);
}

void Profile::output_to_stream(ostream& out_strm)
{
  if (ref_no != 0) out_strm << "label " << ref_no << endl; 
  out_strm << "origin " << origin << endl;
  out_strm << "width " << width << endl;
  out_strm << "height " << height << endl;
  if (direction != Point2(0,0)) out_strm << "direction " <<
				  direction << endl;  
  if (Profile::output_points)
    {
      out_strm << "control_points ";
      for (int i = 0; i < NO_CONTROL_POINTS; i++)
	out_strm << spline[i] << " ";
    }
}

static inline void putback_string(istream& in_strm, char *str1)
{
  while ((*str1) != 0)
    in_strm.putback(*str1++);
}
      
// new version of operator>> allows more flexibility
istream& operator>>(istream& in_strm, Profile& pf)
{
  char dummytag[32];
  bool finished = FALSE;
  
  // keep going until we reach a 'control_points' tag
  // OR a second 'label' or 'origin' tag

  bool found_label = FALSE;
  bool found_origin = FALSE;
  pf.ref_no = 0; pf.inv_grad = 0;
  pf.direction = Point2(0,0);
  while (!finished)
    {
      in_strm >> dummytag;
      if (strcmp(dummytag,"label") == 0)
	{
	  if (found_label) 
	    {
	      finished = TRUE;
	      putback_string(in_strm, dummytag);
	      break;
	    }
	  found_label = TRUE;
	  in_strm >> pf.ref_no;
	  continue;
	}
      else if (strstr(dummytag,"origin") != NULL) // matches *origin*
	{
	  if (found_origin)
	    {
	      finished = TRUE;
	      putback_string(in_strm,dummytag);
	      break;
	    }
	  found_origin = TRUE;
	  in_strm >> pf.origin;
	}
      else if (strstr(dummytag,"width") != NULL) // matches *width*
	in_strm >> pf.width;
      else if (strstr(dummytag,"height") != NULL) // matches *height*
	in_strm >> pf.height;
      else if (strstr(dummytag,"direction") != NULL) // matches *direction*
	in_strm >> pf.direction;
      else if (strstr(dummytag,"control_points") != NULL)
	{
	  for (int i = 0; i < NO_CONTROL_POINTS; i++)
	    in_strm >> pf.spline[i];
	  finished = TRUE;
	}

      // ignore up to end of line 
      in_strm.ignore(999,'\n');
    }
}
Matrix bsplinematrix = {
  { -1.0/6.0, 3.0/6.0, -3.0/6.0, 1.0/6.0 },
  { 3.0/6.0, -6.0/6.0, 3.0/6.0, 0 },
  { -3.0/6.0, 0, 3.0/6.0, 0 },
  { 1.0/6.0, 4.0/6.0, 1.0/6.0, 0 },
};

void Profile::draw(bool flag)
{ 
  if (flag) draw(BLUE,GREEN, YELLOW);
  else draw(BLACK,BLACK, BLACK);
}

typedef Coord gl_vec3[3];

void Profile::draw(int col1, int col2, int col3)
{
  int i,j;
  if (Profile::draw_colour) set_color(col1);
  defbasis(1,bsplinematrix);
  curveprecision(15);
  curvebasis(1);
  linewidth(Profile::draw_linewidth);
  if (Profile::draw_spline && (!Profile::draw_fill))
    {
      gl_vec3* geom1 = new gl_vec3[NO_CONTROL_POINTS + 4];
      for (i = 0; i < NO_CONTROL_POINTS + 4; i++)
	{
	  geom1[i][0] = spline[i % NO_CONTROL_POINTS].x + origin.x;
	  geom1[i][1] = spline[i % NO_CONTROL_POINTS].y + origin.y;
	  geom1[i][2] = 0.0;
	}
      crvn(NO_CONTROL_POINTS+4,geom1);
      delete [] geom1;
    }
  if (Profile::draw_fill) draw_filled();

#ifndef VOGL

  if (Profile::draw_rectangle)
    {
      //get_size();
      sbox((float) (origin.x - width / 2),
	   (float) (origin.y - height/ 2),
	   (float) (origin.x + width / 2),
	   (float) (origin.y + height / 2));
    }

#endif
	    
  float vert[2];
  vert[0] = spline[0].x + origin.x;
  vert[1] = spline[0].y + origin.y;
  /*bgnpoint();
  v2f(vert);
  endpoint();*/
  if (Profile::draw_boxes)
    {
      for (j = 0; j < NO_CONTROL_POINTS; j++)
	{
	  if (Profile::draw_colour) 
	    {
	      if (j == 0) set_color(col3);
	      else set_color(col2);
	    }
	  vert[0] = spline[j].x + origin.x;
	  vert[1] = spline[j].y + origin.y;
	  rectf((float) (vert[0] - Profile::draw_box_size), 
		(float) (vert[1] - Profile::draw_box_size),
		(float) (vert[0] + Profile::draw_box_size), 
		(float) (vert[1] + Profile::draw_box_size));
	  /* if ((j % 4) == 0)
	    {
	      cmov2(spline[j].x + origin.x, spline[j].y + origin.y);
	      char nstr[20];
	      sprintf(nstr, "%d", j);
	      charstr(nstr);
	    }*/
	}
    }
  
  if (Profile::draw_arrow)
    {
      if (Profile::draw_colour) set_color(col3);
      if (height == 0) get_size();
      Point2 arr_start(origin.x, origin.y + 0.2 * height);
      double arr0 = 0.2 * height;
      double arr1 = 5 * arr0;
      double arr2 = 4 * arr0;
      linewidth(Profile::draw_linewidth);
      bgnline();
      vert[0] = arr_start.x;
      vert[1] = arr_start.y;
      v2f(vert);
      vert[0] = arr_start.x + arr1 * direction.x;
      vert[1] = arr_start.y + arr1 * direction.y;
      v2f(vert);
      vert[0] = arr_start.x + arr2 * direction.x - arr0 * direction.y;
      vert[1] = arr_start.y + arr2 * direction.y + arr0 * direction.x;
      v2f(vert);
      vert[0] = arr_start.x + arr1 * direction.x;
      vert[1] = arr_start.y + arr1 * direction.y;
      v2f(vert);
      vert[0] = arr_start.x + arr2 * direction.x + arr0 * direction.y;
      vert[1] = arr_start.y + arr2 * direction.y - arr0 * direction.x;
      v2f(vert);  
      endline();
      linewidth(Profile::draw_linewidth);
    }
}

void Profile::scale(realno scale_fac)
{
  for (int i = 0; i < NO_CONTROL_POINTS; i++)
    spline[i] = scale_fac * spline[i];
}

void Profile::scale_coords(realno scale_fac)
{
  scale(scale_fac);
  origin *= scale_fac;
  width *= scale_fac;
  height *= scale_fac;
  direction *= scale_fac;
}


realno Profile::get_size()
{
  realno ymin = LARGE;
  realno ymax = -LARGE;
  realno xmin = LARGE;
  realno xmax = - LARGE;
  for (int  i = 0; i < NO_CONTROL_POINTS; i++)
    {
      if (spline[i].y < ymin) ymin = spline[i].y;
      if (spline[i].y > ymax) ymax = spline[i].y;
      if (spline[i].x < xmin) xmin = spline[i].x;
      if (spline[i].x > xmax) xmax = spline[i].x;
    }
  width = (xmax - xmin);
  height = (ymax - ymin);
  return (0.5 * (height + width));
}

void Profile::get_y_bounds(realno& ymin, realno& ymax)
{
  ymin = LARGE;
  ymax = -LARGE;
  for (int  i = 0; i < NO_CONTROL_POINTS; i++)
    {
      if (spline[i].y < ymin) ymin = spline[i].y;
      if (spline[i].y > ymax) ymax = spline[i].y;
    }
}

void Profile::get_box(int& ixmin, int& ixmax, int& iymin, int& iymax)
{
  realno ymin = LARGE;
  realno ymax = -LARGE;
  realno xmin = LARGE;
  realno xmax = - LARGE;
  for (int  i = 0; i < NO_CONTROL_POINTS; i++)
    {
      if (spline[i].y < ymin) ymin = spline[i].y;
      if (spline[i].y > ymax) ymax = spline[i].y;
      if (spline[i].x < xmin) xmin = spline[i].x;
      if (spline[i].x > xmax) xmax = spline[i].x;
    }
  ixmin = ((int) (xmin - 0.5)) + origin.x;
  ixmax = ((int) (xmax + 0.5)) + origin.x;
  iymin = ((int) (ymin - 0.5)) + origin.y;
  iymax = ((int) (ymax + 0.5)) + origin.y;

}
					    
void Profile::normalise(bool flag)
{
  recenter();
  if (!flag) scale (get_size() / 200.0);
  // recenters points, rotates and scales
  realno sqsin = 1 / (1 + inv_grad * inv_grad);
  realno sqcos = (1 - sqsin);
  realno sinang, cosang;
  sinang = sqrt(sqsin);
  if (inv_grad >= 0) cosang = sqrt(sqcos);
  else cosang = - sqrt(sqcos);
  if (!flag) cosang = - cosang;
  Point2 newx(sinang,-cosang);
  Point2 newy(cosang,sinang);
  for (int i = 0; i < NO_CONTROL_POINTS; i++)
    spline[i] = Point2(newx * spline[i], newy * spline[i]);
  get_size();
  if (flag) scale(200.0/get_size());
}

void Profile::recenter(realno* w)
{
  Point2 sum(0,0);
  realno wt = (1.0 / NO_CONTROL_POINTS);
  if (w != NULL)
    {
      realno wsum = 0;
      for (int i = 0;i < NO_CONTROL_POINTS; i++)
	{
	  sum += w[i] * spline[i];
	  wsum += w[i];
	}
      sum = (1.0 / wsum) * sum;
    }
  else
    {
      realno ymin = LARGE;
      realno ymax = -LARGE;
      realno xmin = LARGE;
      realno xmax = - LARGE;
      for (int  i = 0; i < NO_CONTROL_POINTS; i++)
	{
	  if (spline[i].y < ymin) ymin = spline[i].y;
	  if (spline[i].y > ymax) ymax = spline[i].y;
	  if (spline[i].x < xmin) xmin = spline[i].x;
	  if (spline[i].x > xmax) xmax = spline[i].x;
	}
      sum = Point2((xmax + xmin) / 2.0, (ymax + ymin) / 2.0);
    }
  origin += sum;
  for (int i = 0; i < NO_CONTROL_POINTS; i++)
    spline[i] -= sum;
}

void Profile::map(realno& ax, realno& ay, Profile* result)
{
  if (result == NULL) result = this;
  PointVector::map(ax,ay,*result);
}

void Profile::align_to(Profile* mean, Profile* result,
		       realno& s, realno& th,
		       realno* weights)
{
  
  //mean->recenter(weights);
  recenter(weights);
  realno C1, C2, Z;
  C1 = C2 = Z = 0;
  Point2* prf1 = &(mean->spline[0]);
  Point2* prf2 = &(spline[0]);
  realno tmp = (1.0 / NO_CONTROL_POINTS);
  for (int k = 0; k < NO_CONTROL_POINTS; k++)
    {
      if (weights != NULL) tmp = weights[k];
      Z += tmp * (prf2[k].x * prf2[k].x + prf2[k].y * prf2[k].y);
      C1 += tmp * (prf1[k].x * prf2[k].x + prf1[k].y * prf2[k].y);
      C2 += tmp * (prf1[k].y * prf2[k].x - prf1[k].x * prf2[k].y);
    }
  
  realno ax = C1 / Z;
  realno ay = C2 / Z;
  
  s = sqrt(SQUARE(ax) + SQUARE(ay));
  th = (atan2(ay,ax));

  map(ax,ay,result);
}

void Profile::display_shape(realno ox, realno oy, bool flag)
{
  Point2 tmp = origin;
  origin = Point2(ox,oy);
  draw(flag);
  origin = tmp;
}

bool Profile::is_ok()
{
  if (ALL_OK) return TRUE;
  realno ratio = (height / ((realno) width));
  if (ratio > MAX_HEIGHT_TO_WIDTH || ratio < MIN_HEIGHT_TO_WIDTH)
    return (FALSE);
  return (TRUE);
}

Profile* Profile::reflect(Profile* result)
{
  if (result == NULL) result = new Profile();
  result->origin = origin;
  result->width = width;
  result->height = height;
  result->direction = Point2(-direction.x, direction.y);
  result->spline[0] = Point2(-spline[0].x, spline[0].y);
  for (int i = 1; i < NO_CONTROL_POINTS; i++)
    result->spline[NO_CONTROL_POINTS - i]
      = Point2(- spline[i].x, spline[i].y);
  return (result);
}

realno* Profile::get_shape(int i)
{
  // if (i == (2 * NO_CONTROL_POINTS)) return (&(direction.x));
  // if (i == (2 * NO_CONTROL_POINTS + 1)) return (&(direction.y));
  if ((i % 2) == 0) return &(spline[i / 2].x);
   else return &(spline[i / 2].y);
}

realno Profile::get_data(int i)
{
  if (i < (2*NO_CONTROL_POINTS)) return (*get_shape(i));
  if (i == (2*NO_CONTROL_POINTS)) return direction.x;
  if (i == (2 * NO_CONTROL_POINTS + 1)) return direction.y;
}

void Profile::set_data(int i, realno val)
{
  if (i < (2*NO_CONTROL_POINTS)) *get_shape(i) = val;
  if (i == (2 * NO_CONTROL_POINTS)) direction.x = val;
  if (i == (2 * NO_CONTROL_POINTS + 1)) direction.y = val;
}

int Profile::get_data_size()
{
  if (Profile::draw_arrow)
    return (2 * NO_CONTROL_POINTS+2);
  return (2 * NO_CONTROL_POINTS);
}

void Profile::draw_filled()
{
  if (spline_weights != NULL) delete (spline_weights);
  int no_divisions = 128 / NO_CONTROL_POINTS;
  spline_weights = new SplineWeights(no_divisions);
  BoundaryPoints* boundary = to_points(NULL);
  float vert[2];
  concave(TRUE);
  polymode(PYM_FILL);
  bgnpolygon();
  for (int i = 0; i < boundary->get_no_points(); i++)
    {
      vert[0] = boundary->point_data()[i].x;
      vert[1] = boundary->point_data()[i].y;
      v2f(vert);
    }
  endpolygon();
  delete boundary;
}
      
BoundaryPoints* Profile::to_points(BoundaryPoints* res)
{
  setup_spline_weights();

  if (res == NULL) res = 
    new BoundaryPoints(NO_CONTROL_POINTS * spline_weights->nsubs);
  else if (res->get_no_points() < (NO_CONTROL_POINTS * spline_weights->nsubs))
    {
      fprintf(stderr," error in Profile::ro_points \n");
		fflush(stderr);
      abort();
    }
  int k;
  Point2 p0,p1,p2,p3, tmp;
  Point2* curr = res->point_data();
  for (int i = 0; i < NO_CONTROL_POINTS; i++)
    {
      k = (i-1); 
      if (k < 0) k += NO_CONTROL_POINTS;
      p0 = spline[k];
      p1 = spline[(k+1) % NO_CONTROL_POINTS];
      p2 = spline[(k+2) % NO_CONTROL_POINTS];
      p3 = spline[(k+3) % NO_CONTROL_POINTS];
      for (int j = 0; j < spline_weights->nsubs; j++)
	{
	  *curr++ =
	    (spline_weights->w0[j] * p0) + (spline_weights->w1[j] * p1) +
	    (spline_weights->w2[j] * p2) + (spline_weights->w3[j] * p3) +
	      origin;
	}
    }
  return (res);
}

void interpolate(NagVector& coeff, NagVector& res)
{
  setup_spline_weights();
  if ((coeff.get_size() != NO_CONTROL_POINTS)
      || (res.get_size() != (NO_CONTROL_POINTS * spline_weights->nsubs)))
    {
      fprintf(stderr," error in Profile::interpolate \n");
		fflush(stderr);
      abort();
    }
  int k;
  realno p0,p1,p2,p3, tmp;
  realno* curr = res.get_data();
  for (int i = 0; i < NO_CONTROL_POINTS; i++)
    {
      k = (i-1); 
      if (k < 0) k += NO_CONTROL_POINTS;
      p0 = coeff[k];
      p1 = coeff[(k+1) % NO_CONTROL_POINTS];
      p2 = coeff[(k+2) % NO_CONTROL_POINTS];
      p3 = coeff[(k+3) % NO_CONTROL_POINTS];
      for (int j = 0; j < spline_weights->nsubs; j++)
	{
	  *curr++ = 
	    (spline_weights->w0[j] * p0) + (spline_weights->w1[j] * p1) +
	    (spline_weights->w2[j] * p2) + (spline_weights->w3[j] * p3);
	}
    }
}

void Profile::get_normals(Profile& res)
{ 
  int k;
  Point2 p0,p1,p2,p3, tmp;
  for (int i = 0; i < NO_CONTROL_POINTS; i++)
    {
      k = (i-1); 
      if (k < 0) k += NO_CONTROL_POINTS;
      p0 = spline[k];
      p2 = spline[(k+2) % NO_CONTROL_POINTS];
      tmp = 0.5 * (p2 - p0);
      tmp.normalise();
      res.spline[i] = Point2(-tmp.y, tmp.x);
    }
}
  
BoundaryPoints* Profile::sample_normals(BoundaryPoints* res)
{ 
  setup_spline_weights();

  if (res == NULL) res = 
    new BoundaryPoints(NO_CONTROL_POINTS * spline_weights->nsubs);
  else if (res->get_no_points() < (NO_CONTROL_POINTS * spline_weights->nsubs))
    {
      fprintf(stderr," error in Profile::sample_normals \n");
		fflush(stderr);
      abort();
    }
  int k;
  Point2 p0,p1,p2,p3, tmp;
  Point2* curr = res->point_data();
  for (int i = 0; i < NO_CONTROL_POINTS; i++)
    {
      k = (i-1); 
      if (k < 0) k += NO_CONTROL_POINTS;
      p0 = spline[k];
      p1 = spline[(k+1) % NO_CONTROL_POINTS];
      p2 = spline[(k+2) % NO_CONTROL_POINTS];
      p3 = spline[(k+3) % NO_CONTROL_POINTS];
      for (int j = 0; j < spline_weights->nsubs; j++)
	{
	  *curr =
	    (spline_weights->dw0[j] * p0) + (spline_weights->dw1[j] * p1) +
	    (spline_weights->dw2[j] * p2) + (spline_weights->dw3[j] * p3);
	  curr->normalise();
	  *curr = Point2(-curr->y, curr->x);
	  curr++;
	}
      
    }
  return (res);
}

void Profile::points_to_spline()
{
  // convert a set of points to a spline 
  BoundaryPoints* pnts = new BoundaryPoints(NO_CONTROL_POINTS);
  copy(*pnts);
  pnts->to_spline(this);
  delete (pnts);
}

void Profile::draw_in_image(Image* canvas, Point2 offset, bool filled)
{
  if (spline_weights == NULL) spline_weights = new SplineWeights(8);
  Point2 keep_origin = origin;
  origin = offset;

  BoundaryPoints* prf_outline = to_points();
  int np = prf_outline->get_no_points();
  Point2* pdata = prf_outline->point_data();
  if (filled)
    {
      canvas->draw_filled_polygon(np, pdata);
    }
  for (int i = 0; i < np; i++)
    {
      Point2 p1 = pdata[i];
      Point2 p2 = pdata[(i+1)%np];
      canvas->set_line_width(3.0);
      canvas->draw_line(p1.x+0.5,p1.y+0.5,p2.x+0.5,p2.y+0.5);
    }
  delete (prf_outline);
  origin = keep_origin;
}

