/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/*
 * Region.cc
 *
 * a region is a rectangular subimage and
 * in general will consist of one connected blob
 *
 * class created 2/3/93
 * author: Adam Baumberg
 *
 */

#include <region.h>
#include <profile.h>
#include <boolean.h>
#include <shape_constants.h>
#include <BWImage.h>
#include <gl.h>
//#define REGION_USE_DIRECTION
#define REGION_USE_PRINCIPAL_AXIS

Region::~Region()
{
  delete (region_img);
  delete (region_boundary);
}

int start_list[8] = {5,6,7,0,1,2,3,4};
int end_list[8] = {4,5,6,7,0,1,2,3};

Point2 directions[8] = {Point2(-1,-1) , Point2(-1,0) ,Point2(-1,1), 
			Point2(0,1), Point2(1,1), 
			Point2(1,0), Point2(1,-1), Point2(0,-1)};

void Region::trace()
{
  if (region_boundary == NULL) region_boundary = new BoundaryPoints();
  int imgwidth = region_img->get_width();
  int imgheight = region_img->get_height();
  bool found =  FALSE;
  /*int i = imgwidth / 2;
  int j = 0;
  for (; j < imgheight && !found; j++)
      if (*(region_img->get_pixel(i,j)) != 0) found = TRUE; */
  int i = 0;
  int j = 0;
  while (j < imgheight && !found)
    {
      if (*(region_img->get_pixel(i,j)) != 0) found = TRUE;
      else if ((++i) >= imgwidth)
	{
	  i = 0;
	  j++;
	}
    }
  if (!found) 
    {
      fprintf(stderr, " Region::trace: can't find starting point in region \n");
      exit(1);
    }

  region_boundary->add_point( Point2(i,j));
  int forward = 1;
  int old_tmp = forward;
  int tmp;
  bool flag;
  int loop1;
  unsigned char* lpix;
  do
    {
      flag = FALSE;
/*      forward = (forward+1) & 7;*/
      loop1 = start_list[forward];
      while (loop1 != end_list[forward] && flag == FALSE)
	{
	  Point2 look_dir = directions[(loop1+1)&7];
	  lpix = region_img->check_pixel(i +  look_dir.x, j + look_dir.y);

	  if ((lpix == NULL) || (*lpix == 0))
	    loop1 = (loop1+1) & 7;
	  else 
	    {
	      i += (int) look_dir.x;
	      j += (int) look_dir.y;
	      region_boundary->add_point(Point2(i,j));
	      flag = TRUE;
	      forward = loop1;
	      tmp = (loop1+1) & 7;
	    }
	}
      
      if (flag == FALSE)
	{
	  Point2 look_dir = directions[(loop1+1)&7];
	  lpix = (region_img->check_pixel(i + look_dir.x, j + look_dir.y));
	  if (((lpix == NULL) || (*lpix == 0)))
	    {
	      region_boundary->pop();
	      forward = old_tmp;
	      Point2* endpoint = region_boundary->last();
	      i = endpoint->x;
	      j = endpoint->y;
	    }
	  else
	    {
	      i += look_dir.x;
	      j += look_dir.y;
	      region_boundary->add_point(Point2(i,j));
	      flag = TRUE;
	      forward = loop1;
	      tmp = (loop1+1) & 7;
	    }
	}
      old_tmp = tmp;
    } while ((*(region_boundary->last()) !=
	      (*(region_boundary->first()))) &&
	     (region_boundary->get_no_points() < 800));
  region_boundary->pop();
  region_boundary->find_best_line(inv_grad, ox, oy);
  ox += xlo; oy += ylo;
}

bool Region::is_ok() 
{
  if (ALL_OK) return TRUE;

  realno ratio = ((realno) (yhi - ylo)) / ((realno) (xhi - xlo));
  if (ratio > MAX_HEIGHT_TO_WIDTH || ratio < MIN_HEIGHT_TO_WIDTH)
    return (FALSE);
  return (TRUE);
}

istream& operator>>(istream& strmin, Region& r) 
{ 
  char dummy[100];
  strmin >> dummy >> r.ref_no;
  strmin >> dummy >> r.xlo >> r.xhi >> r.ylo >> r.yhi;
  strmin >> dummy >> r.direction;
  return (strmin);
}
  
ostream& operator<<(ostream& strmout, Region& r) 
{ 
  strmout << "label " << r.ref_no << endl;
  strmout << "bb " << r.xlo << " " << r.xhi << " " << r.ylo << " " <<
    r.yhi << endl;
  strmout << "direction " << r.direction << endl;
  return (strmout); 
}


Profile* Region::to_profile()
{
  Profile* rval = NULL;
  Point2 tmp_pnt;
  PointVector tmp_pnts(0);
  if (region_boundary == NULL)
    {
      // region not traced -- use bounding box;
      rval = new Profile(Point2(ox, oy), width, height, 0.0,
			 tmp_pnts, direction);
    }
  else
    {
      Point2 new_origin;
      region_boundary->find_best_line(inv_grad, new_origin.x, new_origin.y);
#ifdef REGION_USE_PRINCIPAL_AXIS
      region_boundary->find_end(Point2(inv_grad, 1), new_origin);
#elif defined REGION_USE_DIRECTION
      if (direction.length2() > 1) {
	region_boundary->find_end(direction, new_origin);
#else
      region_boundary->find_end(Point2(0,1), new_origin);
      inv_grad = 0.0; 
#endif
	  
      if (is_ok())
	{ 
	  rval = new Profile(new_origin + Point2(xlo, ylo), 
			     width, height,
			     inv_grad, tmp_pnts, direction);
	  //region_boundary->find_optimal_spline(rval);
	  region_boundary->convert_to_spline(rval);
	}
      
#ifdef REGION_USE_DIRECTION
     }
#endif

    }
  if (rval != NULL) rval->ref_no = ref_no;
  return rval;
}
  
static int combine(void* pix1, void* pix2)
{
  return (int) ((*((unsigned char*) pix1) * 10  + *((unsigned char*) pix2))
		/ 11);
}
 
static int and_pixels (void* pix1, void* pix2)
{
  return (int) (*((unsigned char*) pix1) & *((unsigned char*) pix2));
}

static int or_pixels (void* pix1, void* pix2)
{
  return (int) (*((unsigned char*) pix1) | *((unsigned char*) pix2));
}

void Region::draw_points()
{ 
  get_best_line();
  region_boundary->draw_points(xlo,ylo);
  color(BLACK);
  move2((float) (-45 * inv_grad + ox + xlo), (float) (ylo + oy - 45));
  draw2((float) (65 * inv_grad + ox + xlo), (float) (oy + 65 + ylo));

}

void Region::merge(Region* reg2)
{
  int nxlo, nxhi, nylo, nyhi;
  nxlo = xlo; nxhi = xhi; nylo = ylo; nyhi = yhi;
  if (nxlo > (reg2->xlo)) nxlo = reg2->xlo;
  if (nxhi < (reg2->xhi)) nxhi = reg2->xhi;
  if (nylo > (reg2->ylo)) nylo = reg2->ylo;
  if (nyhi < (reg2->yhi)) nyhi = reg2->yhi;
  
  int nwidth = nxhi - nxlo + 1;
  int nheight = nyhi - nylo + 1;
  if (region_img != NULL)
    {
      Image* newimage = (Image*) new Gray8Image(nwidth,nheight);
      newimage->paste_subimage(xlo - nxlo, xhi - nxlo, ylo - nylo,
			       yhi - nylo, region_img);
      newimage->paste_subimage(reg2->xlo - nxlo, reg2->xhi - nxlo,
			       reg2->ylo - nylo, reg2->yhi - nylo,
			       reg2->region_img);
      delete (region_img);
      region_img = newimage;
    }
  ylo = nylo; yhi = nyhi;
  xlo = nxlo; xhi = nxhi;
  ox = (xlo + xhi) / 2.0;
  oy = (ylo + yhi) / 2.0;
  width = nwidth;
  height = nheight;
}

void Region::draw_box()
{
  set_color((ref_no & 7) + 1);
  recti(xlo,ylo,xhi, yhi);
}
