/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/*
 * region.h
 *
 * a region is a rectangular subimage and 
 * in general will consist of one connected blob
 * 
 * class created 2/3/93
 * author: Adam Baumberg
 * 
 */

#ifndef __REGION_H__
#define __REGION_H__

#include <image.h>
#include <BoundaryPoints.h>

class RegionSet;
class Profile;
class RegionTracker;

class Region
{
 public:
  
  // bounding box
  int xlo, xhi, ylo, yhi;

  // reference number
  int ref_no;

  // origin (ox, oy)
  realno ox, oy;

  // direction of motion of blob 
  Point2 direction;

  // width and height
  realno width, height;

  // shape of region
  Image* region_img;
  
  // misc
  realno inv_grad;
  BoundaryPoints* region_boundary;
  
 public:

  friend RegionSet;
  friend RegionTracker;

  Region(int x1, int x2, int y1, int y2, Image* img) 
    { 
      xlo = x1; xhi = x2; ylo = y1; yhi = y2; region_img = img;
      width =  xhi - xlo + 1;
      height = yhi - ylo + 1;
      ox = (xhi + xlo) / 2.0;
      oy = (yhi + ylo) / 2.0;
      region_boundary = NULL;
      ref_no = 0;
      direction = Point2(0,0);
    }
  Region() 
    { 
      xlo = xhi = ylo = yhi = 0; 
      ref_no = 0;
      region_img = NULL;
      region_boundary = NULL;
      direction = Point2(0,0);
    }
  ~Region();

  // merge two regions
  void merge(Region* reg2);

  void draw_points();
  void draw_box();

  void smooth_boundary(realno sd = 1.0, int win_size = 8) 
    { region_boundary->gauss_smooth(sd, win_size); }

  
  void get_best_line() 
    { region_boundary->find_best_line(inv_grad, ox, oy); }

  void trace();

  void find_ends() 
    { region_boundary->find_ends(inv_grad, ox, oy); }

  bool to_spline(PointVector* res)
    { return(region_boundary->convert_to_spline(res)); }

  void recenter()
    { 
      region_boundary->recenter(ox, oy);
    }

  bool is_ok(); // checks that region is reasonable
  Profile* to_profile();

};

istream& operator>>(istream& strmin, Region& dummy);
ostream& operator<<(ostream& strmout, Region& dummy);



#endif
