/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/***************************************************************
 * C++ source
 *
 * File : 	ActiveModel.cc
 *
 * Module :	ActiveModel
 *
 * Author : 	Adam
 *
 * Creation Date : Mon May 16 15:25:51 1994 
 *
 * Comments : 	
 *
 ***************************************************************/


/* include files */
#include "ActiveModel.h"

#include <FeatureDetector.h>
#include <NagMatrix.h>
#include <KalmanFilter.h>
#include <unistd.h>
#include <SplineWeights.h>
#include <profile.h>
#include <EnvParameter.h>
#include <RGB32Image.h>
#include <utils.h>
#include <default_pca_model.h>


#ifdef sgi
#include <gl.h>
#endif

realno ActiveModel::cloud_size = 
AmbGlobals.register_real("PCA_CLOUD_SIZE", 2.0,
			 &ActiveModel::cloud_size,
			 FALSE, "ActiveModel", 
			 "The maximum Mahalanobis distance to the \
mean shape (in std deviations)");

realno ActiveModel::noise_scalar = 
AmbGlobals.register_real("SHAPE_NOISE_SD", sqrt(0.2),
			 &ActiveModel::noise_scalar,
			 FALSE, "ActiveModel",
			 "The noise term for each shape parameter \
in std deviations per frame");

realno ActiveModel::initial_noise =
AmbGlobals.register_real("INIT_SHAPE_SD", 1.0,
			 &ActiveModel::initial_noise,
			 FALSE, "ActiveModel",
			 "The initial uncertainty of each shape \
parameter in std deviations");

realno ActiveModel::measure_fac_sd =
AmbGlobals.register_real("MEASURE_FAC_SD", 0.1,
			 &ActiveModel::measure_fac_sd,
			 FALSE, "ActiveModel",
			 "The standard deviation of measurements \
proportional to the search window size");

realno ActiveModel::measure_const_sd =
AmbGlobals.register_real("MEASURE_CONST_SD", 0.5,
			 &ActiveModel::measure_const_sd,
			 FALSE, "ActiveModel",
			 "An additional constant value added to the \
std deviation of the measurement process (in pixels)");


realno ActiveModel::epsilon = 0.0;


static int default_subdivisions =
AmbGlobals.register_int("SPLINE_SUBDIVISIONS", 8,
			&default_subdivisions,
			TRUE, "ActiveModel",
			"The number of sample points between spline \
control points");

static int default_model_depth = 
AmbGlobals.register_int("MODEL_DEPTH", 10,
			&default_model_depth,
			TRUE, "ActiveModel",
			"The number of shape parameters used");

int ActiveModel::bunch_size =
AmbGlobals.register_int("BUNCH_SIZE", 2,
			&ActiveModel::bunch_size,
			FALSE, "ActiveModel",
			"The number of measurement made \
per free parameter");


bool ActiveModel::use_heirarchical_search = 
AmbGlobals.register_bool("HEIRARCHICAL", TRUE,
			 &ActiveModel::use_heirarchical_search,
			 FALSE, "ActiveModel",
			 "Vary the number of shape paramaters and \
measurements as the fitting process converges");

bool ActiveModel::do_m_search =
AmbGlobals.register_bool("MAHALANOBIS_DIRECTION", TRUE,
			 &ActiveModel::do_m_search,
			 FALSE, "ActiveModel",
			 "Search along the Mahalanobis optimal search \
direction");

realno ActiveModel::minimum_best_gain = 
AmbGlobals.register_real("MIN_GAIN", 0.5,
			 0.0, 1.0,
			 &ActiveModel::minimum_best_gain,
			 FALSE, "ActiveModel",
			 "The desired initial gain and the minimum \
subsequent gain for parameter fitting");


#ifdef DEBUG
#ifdef sgi
#include <sgmoviestore.h> // for outputing demo's
MovieStore* demo_movie = NULL;
EnvStringParameter demo_movie_name("DEMO_MOVIE_NAME");
RGB32Image* demo_image = NULL;
#endif
#endif


static EnvIntParameter start_ref_no("START_REF_NO", 0);

#ifdef DEBUG
int ActiveModel::debug_level =
AmbGlobals.register_int("DEBUG_LEVEL", 0,
			&ActiveModel::debug_level,
			FALSE, "ActiveModel",
			"Debug level for model fitting");

#endif


#define MAX_BUNCH_SIZE 500
#define M_SQRT1_2       0.70710678118654752440

#ifdef sgi
// GL drawing for debuging
static inline void move2(Point2& p)
{
  move2(p.x, p.y);
}

static inline void draw2(Point2& p)
{
  setlinestyle(0);
  draw2(p.x, p.y);
}

static inline void dotted_draw2(Point2& p)
{
  deflinestyle(1, 0xff00);
  setlinestyle(1);
  draw2(p.x, p.y);
  setlinestyle(0);
}

static inline void rect(Point2& p1, Point2& p2)
{
  rectf(p1.x, p1.y, p2.x, p2.y);
}
static inline void circ(Point2& p, realno r)
{
  circf(p.x, p.y, r);
}
#endif


ActiveModel::ActiveModel(char* pca_file)  : 
P(), H(), N(), evals(), mean_points(0), mean_normals(0)
{
  model_file = pca_file;
  model_depth = default_model_depth;
  no_subdivisions = default_subdivisions;
  if (pca_file != NULL)
    {
      ifstream tmpfile(pca_file);
      if (!tmpfile)
	{
	  sprintf(AmbErrorHandler::buffer,"Can't open %s -- using default \
data", pca_file);
	  AmbErrorHandler::handle_warning("in ActiveModel::ActiveModel\n");
	}
    }
}



// apart from shape there's ox, oy, ax, ay
static const int no_of_non_shape_params = 4;

void ActiveModel::setup()
{
  read_shape_model();
  db.reconstruct(model_depth);
  measure_error.reconstruct(model_depth);

  int old_no_subdivisions = no_subdivisions;

  int max_bunch = (model_depth + no_of_non_shape_params) * bunch_size;
  if (max_bunch > MAX_BUNCH_SIZE)
    {
      max_bunch = MAX_BUNCH_SIZE;
      sprintf(AmbErrorHandler::buffer,"%d", MAX_BUNCH_SIZE);
      AmbErrorHandler::handle_warning("in ActiveModel bunch size may \
be restricted to a maximum of ");
    }
  
  while ((no_sample_points = NO_CONTROL_POINTS * no_subdivisions) 
	 < max_bunch) no_subdivisions *=2;  

  if (no_subdivisions != old_no_subdivisions)
    {
      sprintf(AmbErrorHandler::buffer,"using %d subdivisions \n",
	      no_subdivisions);
      AmbErrorHandler::handle_warning("in ActiveModel::setup() ");
    }
  

  H_no_rows = 2 * NO_CONTROL_POINTS * no_subdivisions;

    
  // setup full model to sub_point transformation
  SplineWeights sweights(no_subdivisions);

  int k;

  NagMatrix G(2 * no_sample_points, 2 * NO_CONTROL_POINTS,0.0);
  NagMatrix dG(2 * no_sample_points, 2 * NO_CONTROL_POINTS,0.0);
  for (int i = 0; i < no_sample_points; i++)
    {
      k = (i / no_subdivisions);
      G.set(2 * i, 2 * ((k-1+NO_CONTROL_POINTS) % NO_CONTROL_POINTS),
	    sweights.w0[i % no_subdivisions]);
      G.set(2 * i, 2 * ((k+NO_CONTROL_POINTS) % NO_CONTROL_POINTS),
	    sweights.w1[i % no_subdivisions]);
      G.set(2 * i, 2 * ((k+1+NO_CONTROL_POINTS) % NO_CONTROL_POINTS),
	    sweights.w2[i % no_subdivisions]);
      G.set(2 * i, 2 * ((k+2+NO_CONTROL_POINTS) % NO_CONTROL_POINTS),
	    sweights.w3[i % no_subdivisions]);

      G.set(1 + 2 * i, 1 + 2 * ((k-1+NO_CONTROL_POINTS) % NO_CONTROL_POINTS),
	    sweights.w0[i % no_subdivisions]);
      G.set(1 + 2 * i, 1 + 2 * ((k+NO_CONTROL_POINTS) % NO_CONTROL_POINTS),
	    sweights.w1[i % no_subdivisions]);
      G.set(1 + 2 * i, 1 + 2 * ((k+1+NO_CONTROL_POINTS) % NO_CONTROL_POINTS),
	    sweights.w2[i % no_subdivisions]);
      G.set(1 + 2 * i, 1 + 2 * ((k+2+NO_CONTROL_POINTS) % NO_CONTROL_POINTS),
	    sweights.w3[i % no_subdivisions]);


      dG.set(2 * i, 2 * ((k-1+NO_CONTROL_POINTS) % NO_CONTROL_POINTS),
	    sweights.dw0[i % no_subdivisions]);
      dG.set(2 * i, 2 * ((k+NO_CONTROL_POINTS) % NO_CONTROL_POINTS),
	    sweights.dw1[i % no_subdivisions]);
      dG.set(2 * i, 2 * ((k+1+NO_CONTROL_POINTS) % NO_CONTROL_POINTS),
	    sweights.dw2[i % no_subdivisions]);
      dG.set(2 * i, 2 * ((k+2+NO_CONTROL_POINTS) % NO_CONTROL_POINTS),
	    sweights.dw3[i % no_subdivisions]);

      dG.set(1 + 2 * i, 1 + 2 * ((k-1+NO_CONTROL_POINTS) % NO_CONTROL_POINTS),
	    sweights.dw0[i % no_subdivisions]);
      dG.set(1 + 2 * i, 1 + 2 * ((k+NO_CONTROL_POINTS) % NO_CONTROL_POINTS),
	    sweights.dw1[i % no_subdivisions]);
      dG.set(1 + 2 * i, 1 + 2 * ((k+1+NO_CONTROL_POINTS) % NO_CONTROL_POINTS),
	    sweights.dw2[i % no_subdivisions]);
      dG.set(1 + 2 * i, 1 + 2 * ((k+2+NO_CONTROL_POINTS) % NO_CONTROL_POINTS),
	    sweights.dw3[i % no_subdivisions]);
    }
  G.multiply(P,H);
  G.multiply(*mean, mean_points);
  mean_sq_distance = mean_points.length2() / no_sample_points;
 
  dG.multiply(P,N);
  dG.multiply(*mean, mean_normals);
  Point2* prf2 = mean->spline;
  mean->get_size();

}

void ActiveModel::read_shape_model()
{
  if (model_file == NULL)
    {
      // use the default 5 mode shape model

      NO_CONTROL_POINTS = default_no_control_points;
      int n2 = 2 * NO_CONTROL_POINTS;
      mean = new Profile();
      mean->NagVector::copy_data(default_mean_shape,n2);
      if (model_depth > n2) model_depth = n2;
      if (model_depth > default_model_modes) 
	model_depth = default_model_modes;
      int depth = MIN(model_depth, default_model_modes);

      P.reconstruct(n2, model_depth);
      P.clear();
      memcpy(P.get_data(), default_eigenvector_data,
	     depth * n2 * sizeof(realno));
      evals.reconstruct(model_depth);
      evals.clear();
      evals.copy_data(default_eigenvalue_data, depth);
      return;
    }

  PCAclass* pca_model = new PCAclass(model_file);
  DB_PRINTF("reading shape model\n");
  NO_CONTROL_POINTS = pca_model->data_size / 2;
  if (Profile::draw_arrow) NO_CONTROL_POINTS--;

  if (model_depth > pca_model->data_size) model_depth = pca_model->data_size;

  P.reconstruct(pca_model->data_size, model_depth);
  evals.reconstruct(model_depth);

  mean = new Profile();
  mean->NagVector::copy_data(pca_model->mean.get_data(), 2*NO_CONTROL_POINTS);
 
  int i;

  // setup model to spline point transformation
  for (int j = 0; j < model_depth; j++)
    {
      realno* evec = pca_model->get_evec(j+1);
      for (i = 0; i < pca_model->data_size; i++)
	P.set(i,j,evec[i]);
      evals[j] = pca_model->get_eval(j+1);
    }
  delete (pca_model);
}

void ActiveModel::setup_shape_filters()
{
  if (currx->filters == NULL)
    currx->filters = new FilterOneD*[model_depth];
  for (int lp = 0; lp < model_depth; lp++)
	currx->filters[lp] = NULL;
  currx->no_shape_parameters = model_depth;

  for (int i = 0; i < model_depth; i++)
    {
      if (currx->filters[i] != NULL) delete (currx->filters[i]);
      currx->filters[i] = (FilterOneD*) new
	StaticKalmanOneD(SQUARE(noise_scalar) *  evals[i]	// q^2 	
		      ,4			// r^2  
		      ,0 			// b0 	
		      ,SQUARE(initial_noise) * evals[i]		// p11 	
		      ,epsilon);
    }
}

void ActiveModel::get_estimate(int id, Point2& p, Point2& n, Point2& q)
{
  int id_x = 2 * id;
  int id_y = id_x + 1;
  q = Point2(mean_points[id_x], mean_points[id_y]);
  Point2 tmp(mean_normals[id_x], mean_normals[id_y]);
  int offset = 0;
  for (int i = 0; i < model_depth; i++)
    {
      realno b_i = currx->filters[i]->get_state();
      q.x += H[id_x+offset] * b_i;
      q.y += H[id_y+offset] * b_i;
      tmp.x += N[id_x+offset] * b_i;
      tmp.y += N[id_y+offset] * b_i;
      offset += H_no_rows; 
    }
  n = tmp.map(currx->a_x, currx->a_y);
  n = Point2(-n.y, n.x);
  p = q.map(currx->a_x, currx->a_y);
}
  
void ActiveModel::b_to_x()
{
  NagVector b(model_depth);
  for (int i = 0; i < model_depth; i++)
    b[i] = currx->filters[i]->get_state();
  P.multiply(b, *currx);
  currx->add(*mean, *currx);
  currx->map(currx->a_x, currx->a_y, currx);
}



void ActiveModel::quick_observe_point(Point2& n, Point2& p, Point2& q, int& id,
				      realno& mxx, realno& mxy, realno& myy,
				      realno& inv_var,
				      realno& m_density, Point2& dp)
{
  Point2 pos;
  int delta_id = no_sample_points / current_bunch_size;
  get_estimate(id, p, n, q);  
  realno len_n = n.length();
  m_density = len_n;
  n = n / len_n;
  pos = p + currx->origin;  
  inv_var = 0;
  mxx = 0; mxy = 0; myy = 0;
  if ((occ_handler->is_occlusion(pos.x, pos.y)) || 
      (occ_handler->is_off_image(pos.x, pos.y)))
    return;

  // if no occlusion then should find feature

  realno cov_xx = 0, cov_yy = 0, cov_xy = 0;
  int id_x = id * 2;
  int id_y = id_x + 1;
  int offset = 0;
  realno curr_ax = currx->a_x; realno curr_ay = currx->a_y;

  for (int i = 0; i < current_depth; i++)
    {
      realno b_i = currx->filters[i]->get_p11();
      realno tmp_x = curr_ax * H[id_x+offset] - curr_ay * H[id_y+offset];
      realno tmp_y = curr_ay * H[id_x+offset] + curr_ax * H[id_y+offset];
      cov_xx += (tmp_x * tmp_x) * b_i;
      cov_xy += (tmp_x * tmp_y) * b_i;
      cov_yy += (tmp_y * tmp_y) * b_i;
      offset += H_no_rows;
	  
    }

  Matrix2by2 point_covariance(cov_xx, cov_xy, cov_yy);
  point_covariance += currx->pos_filter->get_cov();


  // add term due to a_x, a_y 
  realno py2 = q.y * q.y;
  realno px2 = q.x * q.x;
  realno pxy = 2.0 * q.x * q.y;

  Matrix2by2 Cov_a;
  if (currx->a_filter != NULL)
    Cov_a = currx->a_filter->get_cov();
  else
    {
      realno cth = 1.0;
      realno sth = 0.0;
      realno s_hat = currx->s_filter->get_state();
      realno sq_s_hat = s_hat * s_hat;
      realno var_s = currx->s_filter->get_p11();
      realno var_th = 0.0;
      if (currx->th_filter != NULL)
	{
	  realno theta = currx->th_filter->get_state();
	  var_th = currx->th_filter->get_p11();
	  cth = cos(theta);
	  sth = sin(theta);
	}
      
      Cov_a = 
	Matrix2by2(var_s * cth * cth + var_th * sth * sth,
		   (var_s - var_th) * cth * sth,
		   var_s * sth * sth + var_th * cth * cth);
    }
      

  point_covariance += (Matrix2by2(q.x, -q.y, q.y, q.x) * Cov_a) * 
    Matrix2by2(q.x, q.y, -q.y, q.x);

  Point2 edge_orientation = n;

  if (do_m_search)
    {
      // use Mahalanobis search direction
      n = point_covariance.multiply(n);
      n.normalise();
      
    }
  
  mxx = n.x * n.x;
  mxy = n.x * n.y;
  myy = n.y * n.y;

  realno scale = 
    2.0 / sqrt(n * (point_covariance.inverse()).multiply(n));


  realno sq_scale = scale * scale;
  if (sq_scale < 1.0) return;
  

  Point2 direction = scale * n;

  Point2 pos_r = pos + direction;
  Point2 pos_l = pos - direction;
  Point2 pos_m = pos;

  realno measure_var;
  Point2 p_obs;  

  measure_var = frame_step;
  bool ret_val =
    edge_finder->find_feature(id / ((realno) no_sample_points),
			      pos_m, n,
			      edge_orientation, scale,
			      occ_handler, currx, p_obs,
			      measure_var);
  if (ret_val != FD_RETURN_OCCLUSION) no_looks++;
  if (ret_val == FD_RETURN_GOOD)
    {
      dp = p_obs - pos;
      measure_var +=  (var_fac * scale * scale) + var_const;
      inv_var = (frame_step) / 
	(current_bunch_size * measure_var);
    }
  else
    {
      dp = Point2(0,0);
      inv_var = 0;
    }


  Point2 tmp_n(-n.y, n.x);
  realno unit1 = scale / 5.0; 
 
#ifdef DEBUG
#ifdef sgi
  if (debug_level == 1)
    {
      video_image->draw_in_image();
      linewidth(2);
      set_color(YELLOW);
      move2(pos_r + unit1 * tmp_n); draw2(pos_r - unit1 * tmp_n);
      move2(pos_r); draw2(pos_r - unit1 * n);
      move2(pos_l + unit1 * tmp_n); draw2(pos_l - unit1 * tmp_n);
      move2(pos_l); draw2(pos_l + unit1 * n);

      move2(pos_r);
      dotted_draw2(pos_l);

      set_color(WHITE);
      circ(pos, 1);
    }
#endif
#endif
  if (inv_var  < 1e-20) return;
      
#ifdef DEBUG
#ifdef sgi
  if (debug_level == 1)
    {
      set_color(RED);
      circ(p_obs, 1);
    }
#endif
#endif
  
  no_observed++;
}

void ActiveModel::get_new_depth()
{
//   realno sq_pos_scale = currx->x_filter->get_p11() +
//     currx->y_filter->get_p11();

//   realno sq_pos_scale = 
//     MIN(currx->x_filter->get_p11(),currx->y_filter->get_p11());

  realno sq_pos_scale =
   currx->pos_filter->get_cov().get_min_real_eigenvalue();

  sq_pos_scale = (var_fac * 4.0 * sq_pos_scale) + var_const;

  realno frame_sq_scale =  sq_pos_scale / 
    (SQUARE(currx->a_x) + SQUARE(currx->a_y));

  realno m_var = NO_CONTROL_POINTS * frame_sq_scale / frame_step;

  while (current_depth < model_depth)
    {
      realno shape_var = currx->filters[current_depth]->get_p11();
      realno gain = shape_var / 
	(shape_var + m_var);
      if (gain > 0.01) current_depth++;
      else break;
    }
  
  DB_PRINTF("current depth %d \n", current_depth);
}



realno ActiveModel::adjust_state()
{  
#ifdef DEBUG
  if (debug_level == 1)
    {
      video_image->display();
      b_to_x();
      set_color(WHITE);
      currx->draw();
    }
#endif

  realno scale = 0;
  realno size = currx->a_x * currx->a_x + currx->a_y * currx->a_y;

  Point2 n[MAX_BUNCH_SIZE], p[MAX_BUNCH_SIZE], q[MAX_BUNCH_SIZE], 
    dp[MAX_BUNCH_SIZE];
  int ids[MAX_BUNCH_SIZE];
  realno inv_var[MAX_BUNCH_SIZE], mxx[MAX_BUNCH_SIZE]
    , mxy[MAX_BUNCH_SIZE], myy[MAX_BUNCH_SIZE];


  // added new twist --
  // dynamically vary the model depth 
  // and the bunch size
  // based on the positional search scale !

  get_new_depth();
  current_bunch_size = (current_depth + no_of_non_shape_params) 
    * bunch_size;

  if (current_bunch_size > MAX_BUNCH_SIZE) 
    current_bunch_size = MAX_BUNCH_SIZE;

  int id = rand() % no_sample_points;

  NagVector pos_weights(current_bunch_size);
  realno sum_nxx = 0, sum_nxy = 0, sum_nyy = 0;
  
  for (int lp = 0; lp < current_bunch_size; lp++)
    {

      ids[lp] = (id + (lp * no_sample_points) / current_bunch_size) %
	no_sample_points;
      quick_observe_point(n[lp], p[lp], q[lp], ids[lp], mxx[lp], mxy[lp],
		      myy[lp], inv_var[lp], pos_weights[lp], dp[lp]);
    }


  realno ox_err= 0, oy_err = 0, oxy_err = 0, dox = 0, doy = 0;
  
  // weight measurements according to sampling density
  pos_weights.scale(current_bunch_size / (pos_weights.sum()),
		    pos_weights);
  for (int lp1 = 0; lp1 < current_bunch_size; lp1++)
    inv_var[lp1] *= pos_weights[lp1];
  

  // if there was occlusion then make the
  // other obervations more significant

  // but make sure we are tracking a significant section of contour
  //   realno min_no_looks = 
  //     MAX(MIN_SIGNIFICANCE * current_bunch_size, no_looks);
  //   realno look_fac = current_bunch_size / min_no_looks;

  // commented out as it don't work very well

  realno look_fac = 1.0;
  for (lp = 0; lp < current_bunch_size; lp++)
    {
      inv_var[lp] *= look_fac;
      ox_err += inv_var[lp] * mxx[lp];
      dox += inv_var[lp] * (dp[lp].x * mxx[lp] + dp[lp].y * mxy[lp]);
      oy_err += inv_var[lp] * myy[lp];
      oxy_err += inv_var[lp] * mxy[lp];
      doy += inv_var[lp] * (dp[lp].y * myy[lp] + dp[lp].x * mxy[lp]);
    }

  realno gain_pos = 
    currx->pos_filter->delta_update(Point2(dox, doy),
				    Matrix2by2(ox_err, oxy_err,oy_err));

  currx->origin = currx->pos_filter->get_state();


  Point2 delta_origin = currx->pos_filter->state_change();
  
  realno axx_err = 0, axy_err = 0, ayy_err = 0, dax = 0, day = 0;
  for (lp = 0; lp < current_bunch_size; lp++)
    {
      dp[lp] -= delta_origin;
      p[lp] += dp[lp];
      realno& qx = q[lp].x; realno& qy = q[lp].y;
      realno qxx = qx * qx; realno qxy = qx * qy; 
      realno qyy = qy * qy;

      axx_err +=
	inv_var[lp] * (qxx * mxx[lp] + 2 * qxy * mxy[lp] + 
		       qyy * myy[lp]);
      axy_err += 
	inv_var[lp] * ((qxx - qyy) * mxy[lp] + 
		       qxy * (myy[lp] - mxx[lp]));

      ayy_err +=
	inv_var[lp] * (qyy * mxx[lp] - 2 * qxy * mxy[lp] + 
		       qxx * myy[lp]);
  
      realno& px = dp[lp].x; realno& py = dp[lp].y;
      realno pxqy = px * qy; realno pyqx = py * qx;
      realno pxqx = px * qx; realno pyqy = py * qy;

      dax += inv_var[lp] *
	(mxy[lp] * (pxqy + pyqx) + mxx[lp] * pxqx + myy[lp] * pyqy);
      day += inv_var[lp] *
	(mxy[lp] * (pxqx - pyqy) - mxx[lp] * pxqy + myy[lp] * pyqx);
      
    }

  realno gain_s = 0;
  realno gain_th = 0;
  realno gain_a = 0;

  realno old_ax = currx->a_x;
  realno old_ay = currx->a_y;

  if (currx->a_filter == NULL)
    {
      // use polar (s,th) representation
      realno th = 0.0, sth = 0.0, cth = 1.0;
      if (currx->th_filter != NULL)
	{
	  th = currx->th_filter->get_state();
	  sth = sin(th); cth = cos(th);
	}
      realno s = currx->s_filter->get_state();      

      // map measurements to polar representation
      Matrix2by2 H(cth, -s * sth, sth, s * cth);
      Matrix2by2 R_inv(axx_err, axy_err, ayy_err);
      Matrix2by2 tmp_M = H.transpose() * R_inv * H;
      Point2 tmp_P = H.transpose() * Point2(dax, day);

      // filter s and th independently
      gain_s = currx->s_filter->delta_update(tmp_P.x, tmp_M.val00);
      s = currx->s_filter->get_state();
      if (currx->th_filter != NULL)
	{
	  gain_th = 
	    currx->th_filter->delta_update(tmp_P.y, tmp_M.val11);
	  th = currx->th_filter->get_state();
	  cth = cos(th); sth = sin(th);
	}

      currx->a_x = s * cth;
      currx->a_y = s * sth;      
    }
  else
    {

      gain_a = 
	currx->a_filter->delta_update(dax, day, axx_err, axy_err,
				      ayy_err);
      currx->a_x = currx->a_filter->get_xstate();
      currx->a_y = currx->a_filter->get_ystate();
      
    }

  measure_error.clear();
  db.clear();
  for (lp = 0; lp < current_bunch_size; lp++)
    {
  
      Point2 map_n = n[lp].map(old_ax , -old_ay);

      realno nxx = (map_n.x * map_n.x);
      realno nxy = (map_n.x * map_n.y);
      realno nyy = (map_n.y * map_n.y);
      realno mnxx = (map_n.x * n[lp].x);
      realno mnxy = (map_n.x * n[lp].y);
      realno mnyx = (map_n.y * n[lp].x);
      realno mnyy = (map_n.y * n[lp].y);
      
      int id_x = ids[lp] * 2;
      int id_y = id_x + 1;
      int offset = 0;
      for (int i = 0 ; i < current_depth; i++)
	{
	  realno tmp_Hx = H[id_x+offset];
	  realno tmp_Hy = H[id_y+offset];
	  Point2 tmp_h(tmp_Hx, tmp_Hy);
	  tmp_h = tmp_h.map(old_ax, old_ay);
	  
	  offset += H_no_rows;
	  measure_error[i] += inv_var[lp] * 
	    ( tmp_Hx * tmp_Hx * nxx +
	     2 * tmp_Hy * tmp_Hx * nxy +
	     tmp_Hy * tmp_Hy * nyy);
	  db[i] += inv_var[lp] *
	    (dp[lp].x * (tmp_Hx * mnxx +
			tmp_Hy * mnyx) +
	     dp[lp].y * (tmp_Hx * mnxy +
			tmp_Hy * mnyy));	  
	}
    }
  
  realno max_gain = gain_pos;
  max_gain = MAX(max_gain, gain_a);
  max_gain = MAX(max_gain, gain_s);
  max_gain = MAX(max_gain, gain_th);

  for (int i = 0; i < current_depth; i ++)
    {
      realno gain_b_i =
	currx->filters[i]->delta_update(db[i], measure_error[i]);
      max_gain = MAX(max_gain, gain_b_i);
    }
  
  return (max_gain);
}

      
static realno get_M_distance(FilterOneD** filters, int& depth,
			     NagVector& evals)
{
  realno dist = 0;
   for (int i = 0; i < depth; i++)
    {
      realno b = filters[i]->get_state();
      realno sq_b  = b * b;
      dist += sq_b / evals[i];
    }
  return (dist);
}

void ActiveModel::constrain_b()
{
  realno blength = get_M_distance(currx->filters, current_depth, evals);
  if (blength >= (cloud_size * cloud_size))
    {
      realno mfac = cloud_size / sqrt(blength);
      for (int i = 0; i < current_depth; i++)
	currx->filters[i]->get_state() *= mfac;
    }
}



void ActiveModel::get_initial_frame_step()
{
  realno initial_gain = minimum_best_gain;
  realno v_pos = currx->pos_filter->get_cov().get_max_real_eigenvalue();
  realno vb0 = currx->filters[0]->get_p11();
  
  realno va;
  if (currx->a_filter != NULL)
    va = currx->a_filter->get_uncertainty();
  else va = currx->s_filter->get_p11();

  // estimate frame_step for desired gain ... assuming
  // istropic measurements with variance proportional
  // to uncertainty
  
  realno fstep_lower_bound = initial_gain * 4.0 * var_fac / 
    (1.0 - initial_gain);

  
  // estimate frame_step for desired gain ... assuming isotropic
  // measurements with fixed measurement variance
  
  // (gain for position parameter)
  realno fstep_pos = (initial_gain * var_const) / 
    (v_pos * (1.0 - initial_gain));

  // (gain for principal shape parameter)
  realno s2 = SQUARE(currx->a_x) + SQUARE(currx->a_y);
  realno fstep_b0 = (initial_gain * (NO_CONTROL_POINTS * var_const / s2))
    / (vb0 * (1.0 - initial_gain));
  // gain for a_x and a_y
  realno fstep_a = (initial_gain * (var_const / mean_sq_distance))
    / (va * (1.0 - initial_gain));

       
  realno min_frame_step = fstep_pos;
  min_frame_step = MIN(min_frame_step, fstep_b0);
  min_frame_step = MIN(min_frame_step, fstep_a);
  min_frame_step = MAX(min_frame_step, fstep_lower_bound);

  frame_step = min_frame_step;
}

realno ActiveModel::track_profile()
{
  var_fac = SQUARE(measure_fac_sd);
  var_const = SQUARE(measure_const_sd);

  if (edge_finder == NULL)
    {
      AmbErrorHandler::handle_error("in ActiveModel::track_profile\n\
No feature detector has been setup");
      AmbErrorHandler::handle_warning("in ActiveModel::track_profile\n\
resorting to sobel edge detector");
      edge_finder = 
	new SobelDetector(video_image, 0.05, MAXIMUM_SM);
    }

  occ_handler->setup_occlusion_map(currx);
  
  get_initial_frame_step();

  frame_time = 0;

  int no_iterations = 0;
  no_observed = no_looks = 0;
  no_sampled = 0;

  int total_no_looks = 0;

  realno dblength, blength, temp;
  if (use_heirarchical_search)
    current_depth = 0;
  else current_depth = model_depth;
  get_new_depth();

  realno current_scale_sd;
  realno expansion_scale;

  edge_finder->setup_batch(currx);

  while (frame_time < 1.0)
    {
      no_looks = 0;
      realno max_gain = adjust_state();
      no_sampled += current_bunch_size;


#ifdef DEBUG
#ifdef sgi
      if ((debug_level == 1) && (demo_movie_name.value[0] != '\0'))
	{
	  if (demo_movie == NULL)
	    demo_movie = new MovieStore(demo_movie_name.value);
	  demo_image = video_image->rgb_read_image(demo_image);
	  if (!demo_movie->add_a_frame(demo_image))
	    demo_movie_name.value[0] = '\0';
	}
#endif
#endif

   
      frame_step = MIN(frame_step, 1.0 - frame_time);
      DB_PRINTF(" max gain = %f frame_step %f \n", max_gain, frame_step);
      frame_time += frame_step;
      if (max_gain < minimum_best_gain) 
	  frame_step *= 2;
      total_no_looks += no_looks;
      no_iterations++;

      constrain_b();
      currx->origin = currx->pos_filter->get_state();

#ifdef DEBUG
if (debug_level == 2) 
  {
    b_to_x();
    video_image->display();
    currx->draw();
  }
#endif

    }
  b_to_x();

  no_looks = total_no_looks;
  return (get_fitness());
}


void ActiveModel::apply_virtual_shape()
{
  // add a virtual observation 
  for (int i = 0; i < model_depth; i++)
    currx->filters[i]->add_observation(0, 1.0 * evals[i]);
}

void ActiveModel::hold_shape()
{
  for (int i = 0; i < model_depth; i++)
    delete currx->filters[i];
  setup_shape_filters();
}  

