/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/***************************************************************
 * C - C++ Header
 *
 * File : 	ActiveModel.h
 *
 * Module :	ActiveModel
 *
 * Author : 	Adam
 *
 * Creation Date : Mon May 16 15:07:55 1994 
 *
 * Comments : 	module for tracked a TrackedProfile
 *
 ***************************************************************/


#ifndef __ACTIVE_MODEL_H__
#define __ACTIVE_MODEL_H__

/* include files */

#include <ProfileSet.h>
#include <image.h>
#include <PCAclass.h>
#include <NagMatrix.h>
#include <TrackedProfile.h>
#include <OcclusionHandler.h>
#include <SplineWeights.h>
#include <hash_defines.h>

class FeatureDetector;


class ActiveModel
{
protected:
  
  Image* video_image;	//  for debugging, visualisation

  OcclusionHandler* occ_handler;

  char* model_file;	// shape model data file

  NagMatrix P;		// model to control point transformation
  NagMatrix H;		// model to sample point transformation
  NagMatrix N;		// model to sample normal tramsformation
  NagVector evals;	// copy of shape eigenvalues
  int H_no_rows;

  PointVector mean_points;
  PointVector mean_normals;
  // mean square distance from origin to sample points
  realno mean_sq_distance;	

  NagVector measure_error;
  NagVector db;  
  
#ifdef DEBUG
  static int debug_level;
#endif

  // the current tracked profile
  TrackedProfile* currx; 

  int no_sample_points;		// in fact max sample points

  // constants for optimisation scheme
  // maximum M-distance to mean in shape space
  static realno cloud_size;

  // shape model noise parameters
  static realno noise_scalar;
  static realno initial_noise;


  // measurement parameters
  static realno measure_fac_sd;
  static realno measure_const_sd;
  realno var_fac, var_const;

  // overweighting constant
  static realno epsilon;

  // no sample points between spline control points
  int no_subdivisions;

  // number of shape parameters to use
  int model_depth;
  // current number of shape parameters used
  int current_depth;

  static int bunch_size;	// the no of observations per free parameter
  static int min_bunch_size;	// minimum no of observations
  int current_bunch_size;

  realno frame_time;	// notional 'time' between image frames (0 to 1)
  realno frame_step;	// notional time step 

  virtual void constrain_b();

  // use a heirarchical fitting scheme
  static bool use_heirarchical_search;
  // use the mahalonbis search direction
  static bool do_m_search;

public:
  
  Profile* mean;
  FeatureDetector* edge_finder;	// mechanism for making observations

  // no of unoccluded feature points
  int no_looks;
  // no of significant features found
  int no_observed;
  // total no of sample points
  int no_sampled;

  // try and fix the maximum gain around this value or above
  static realno minimum_best_gain;


  void get_estimate(int i, Point2& p, Point2& n, Point2& q);
  realno adjust_state();
  void b_to_x();

  void quick_observe_point(Point2& n, Point2& p, Point2& q, int& id,
			   realno& mxx, realno& mxy, realno& myy,
			   realno& inv_var,
			   realno& m_density, Point2& dp);

  void get_new_depth();
  void get_initial_frame_step();

  ActiveModel(char* pca_model = NULL); // use globals to construct
  
  void setup(); // load in the PCA shape model

  void initialise();
  void set_video_image(Image* video) { video_image = video; }
  
  void set_profile(TrackedProfile* initial_guess)
    { currx = initial_guess; }

  void set_feature_detector(FeatureDetector* fd)
  { edge_finder = fd; }
  void set_occlusion_handler(OcclusionHandler* handler)
  { occ_handler = handler; }

  int get_depth() { return (model_depth); }

  // do the iterative fitting
  realno track_profile();

  // reset to mean shape
  void hold_shape();
  virtual void apply_virtual_shape();
  // initialise shape filters
  virtual void setup_shape_filters();
  // get the shape model data from the model_file
  virtual void read_shape_model();	


  realno get_fitness()
  {
    realno fit;
    if (no_looks > 0) fit = no_observed / ((realno) no_looks);
    else fit = 1.0;
    return (fit);
  }

  realno get_significance()
  {
    realno res = no_looks / ((realno) no_sampled);
    return (res);
  }

};

#endif

