/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/***************************************************************
 * C++ source
 *
 * File : 	FeatureDetector.cc
 *
 * Module :	FeatureDetector
 *
 * Author : 	A M Baumberg (CoMIR)
 *
 * Creation Date : Fri May 17 14:17:32 1996 
 *
 * Comments : 	
 *
 ***************************************************************/


/* include files */

#include <OcclusionHandler.h>
#include <FeatureDetector.h>
#include <image.h>
#include <Gray8Image.h>
#include <RGB32Image.h>
#include <utils.h>


#include <gl.h>

#ifndef ZSGN
#define ZSGN(a)		(((a)<0) ? -1 : (a)>0 ? 1 : 0)	
#endif
#ifndef ABS
#define ABS(a)		(((a)<0) ? -(a) : (a))
#endif


// static inline int real_to_int(realno x)
// {
//   if (x < 0)
//     return ((int) (x-0.5));
//   else
//     return ((int) (x+0.5));
// }

realno FeatureDetector::max_scale_height_ratio = 0.05;
realno FeatureDetector::max_scale_width_ratio = 1.0;

static const realno MAX_RGB_DISTANCE = sqrt(3.0 * 256.0 * 256.0);

// pixelwise edge / contrast detection functions ...

static bool gray_simple(int x, int y, int step, Point2& n, 
			realno& edge, ExtraData userdata)
{
  int dx = real_to_int(n.x * step);
  int dy = real_to_int(n.y * step);

  int xr = x + dx;
  int yr = y + dy;
  int xl = x - dx;
  int yl = y - dy;
  Image *img = (Image*) userdata;

  if (!((img->check_coords(xr,yr))  && (img->check_coords(xl,yl))))
    return (FALSE);

  int c1 = *img->get_pixel(xr,yr);
  int c2 = *img->get_pixel(xl,yl);

  edge = (realno) abs(c1 - c2);
  return (TRUE);
}

static bool gray_sobel(int x, int y, int step, Point2& n, realno& edge,
		       ExtraData userdata)
{

  int xr = x + step;
  int xl = x - step;
  int yr = y + step;
  int yl = y - step;
  Image* img = (Image*) userdata;

  if (!((img->check_coords(xr,yr))  && (img->check_coords(xl,yl))))
    return (FALSE);
  
  int c1 = *img->get_pixel(xl,yl);
  int c2 = *img->get_pixel(x,yl);
  int c3 = *img->get_pixel(xr,yl);
  int c4 = *img->get_pixel(xl,y);
  //  int c5 = *img->get_pixel(x,y);
  int c6 = *img->get_pixel(xr,y);
  int c7 = *img->get_pixel(xl,yr);
  int c8 = *img->get_pixel(x,yr);
  int c9 = *img->get_pixel(xr,yr);

  int v_response = c1 - c7 + (2 * (c2 - c8)) + c3 - c9;
  int h_response = c1 - c3 + (2 * (c4 - c6)) + c7 - c9;
  
  //  edge = fabs((h_response * n.x) + (v_response * n.y));
  edge = sqrt((h_response * h_response) + (v_response * v_response));

  return (TRUE);
}

static bool rgb_simple(int x, int y, int step, 
		       Point2& n, realno& edge, ExtraData userdata)
{
  int dx = real_to_int(n.x * step);
  int dy = real_to_int(n.y * step);

  if ((dx == 0) && (dy == 0))
    {
      fprintf(stderr,"This shouldn't happen! \n");
      fprintf(stderr,"n(%f,%f) step = %d \n", n.x, n.y, step);
    }
  int xr = x + dx;
  int yr = y + dy;
  int xl = x - dx;
  int yl = y - dy;
  Image* img = (Image*) userdata;

  if (!((img->check_coords(xr,yr))  && (img->check_coords(xl,yl))))
    return (FALSE);

  RGB32pixel* p1 = (RGB32pixel*) img->get_pixel(xr,yr);
  RGB32pixel* p2 = (RGB32pixel*) img->get_pixel(xl,yl);

  int rdiff = p1->red - p2->red;
  int gdiff = p1->green - p2->green;
  int bdiff = p1->blue - p2->blue;

  edge = sqrt(rdiff * rdiff + gdiff * gdiff + bdiff * bdiff);

  return (TRUE);
}

static bool rgb_sobel(int x, int y, int step, Point2& n, realno& edge,
		      ExtraData userdata)
{
  int xr = x + step;
  int xl = x - step;
  int yr = y + step;
  int yl = y - step;
  Image* img = (Image*) userdata;
  if (!((img->check_coords(xr,yr))  && (img->check_coords(xl,yl))))
    return (FALSE);
  
  RGB32pixel* c1 = (RGB32pixel*) img->get_pixel(xl,yl);
  RGB32pixel* c2 = (RGB32pixel*) img->get_pixel(x,yl);
  RGB32pixel* c3 = (RGB32pixel*) img->get_pixel(xr,yl);
  RGB32pixel* c4 = (RGB32pixel*) img->get_pixel(xl,y);
  //  RGB32pixel* c5 = (RGB32pixel*) img->get_pixel(x,y);
  RGB32pixel* c6 = (RGB32pixel*) img->get_pixel(xr,y);
  RGB32pixel* c7 = (RGB32pixel*) img->get_pixel(xl,yr);
  RGB32pixel* c8 = (RGB32pixel*) img->get_pixel(x,yr);
  RGB32pixel* c9 = (RGB32pixel*) img->get_pixel(xr,yr);

  int r_v_response = 
    c1->red - c7->red + (2 * (c2->red - c8->red)) + c3->red - c9->red;
  int g_v_response = 
    c1->green - c7->green + (2 * (c2->green - c8->green)) + 
    c3->green - c9->green;
  int b_v_response = c1->blue - c7->blue + (2 * (c2->blue - c8->blue))
    + c3->blue - c9->blue;

  int r_h_response = 
    c1->red - c3->red + (2 * (c4->red - c6->red)) + c7->red - c9->red;
  int g_h_response = c1->green - c3->green + 
    (2 * (c4->green - c6->green)) + 
    c7->green - c9->green;
  int b_h_response = c1->blue - c3->blue + 
    (2 * (c4->blue - c6->blue)) + c7->blue - c9->blue;

  realno r_edge = (r_h_response * n.x) + (r_v_response * n.y);
  realno g_edge = (r_h_response * n.x) + (r_v_response * n.y);
  realno b_edge = (r_h_response * n.x) + (r_v_response * n.y);

  edge = (sqrt(r_edge * r_edge + g_edge * g_edge + b_edge * b_edge));
  return (TRUE);


}

// typedef Image[2] twoIm;
 
static bool gray_moving_edge(int x, int y, int step, Point2& n, realno& edge,
			     ExtraData userdata)
{
  
  int dx = real_to_int(n.x * step);
  int dy = real_to_int(n.y * step);

  int xr = x + dx;
  int yr = y + dy;
  int xl = x - dx;
  int yl = y - dy;
  Image *prev = ((Image**) userdata)[0];
  Image *curr = ((Image**) userdata)[1];
  // Image *prev = ((Image[2]) userdata)[0]; //RJM Took [2] out
  // Image *curr = ((Image[2]) userdata)[1]; // and replaced with **
  

  if (!((curr->check_coords(xr,yr))  && (curr->check_coords(xl,yl))))
    return (FALSE);

  int c1 = *curr->get_pixel(xr,yr);
  int c2 = *curr->get_pixel(xl,yl);
  int c3 = *curr->get_pixel(x,y);
  int c4 = *prev->get_pixel(x,y);

  edge = (realno) (abs(c1 - c2) + abs(c3 - c4));
  return (TRUE);
}

static bool rgb_moving_edge(int x, int y, int step, Point2& n, realno& edge,
			     ExtraData userdata)
{

  int dx = real_to_int(n.x * step);
  int dy = real_to_int(n.y * step);

  int xr = x + dx;
  int yr = y + dy;
  int xl = x - dx;
  int yl = y - dy;
  Image *prev = ((Image**) userdata)[0];
  Image *curr = ((Image**) userdata)[1];
  // Image *prev = ((Image[2]) userdata)[0]; //RJM Took [2] out
  // Image *curr = ((Image[2]) userdata)[1]; // and replaced with **

  if (!((curr->check_coords(xr,yr))  && (curr->check_coords(xl,yl))))
    return (FALSE);

  RGB32pixel* p1 = (RGB32pixel*) curr->get_pixel(xr,yr);
  RGB32pixel* p2 = (RGB32pixel*) curr->get_pixel(xl,yl);
  RGB32pixel* p3 = (RGB32pixel*) curr->get_pixel(x,y);
  RGB32pixel* p4 = (RGB32pixel*) prev->get_pixel(x,y);

  int rdiff = p1->red - p2->red;
  int gdiff = p1->green - p2->green;
  int bdiff = p1->blue - p2->blue;
  
  realno spatial_edge = 
    sqrt(rdiff * rdiff + gdiff * gdiff + bdiff * bdiff);
  
  rdiff = p3->red - p4->red;
  gdiff = p3->green - p4->green;
  bdiff = p3->blue - p4->blue;
  realno temporal_edge = 
    sqrt(rdiff * rdiff + gdiff * gdiff + bdiff * bdiff);
  
  edge = spatial_edge + temporal_edge;

  return (TRUE);
}

static const int max_gray_diff = 256;

static bool gray_foreground_edge(int x, int y, int step, Point2& n, 
				 realno& edge, ExtraData userdata)
{
  ForegroundEdgeDetector* parent = (ForegroundEdgeDetector*) userdata;
  int dx = real_to_int(n.x * step);
  int dy = real_to_int(n.y * step);

  int xr = x + dx;
  int yr = y + dy;
  int xl = x - dx;
  int yl = y - dy;
  Image *back = parent->refimage;
  Image *front = parent->vidimage;
  
  if (!((front->check_coords(xr,yr))  && (front->check_coords(xl,yl))))
    return (FALSE);

  int c1 = *front->get_pixel(xr,yr);
  int c2 = *back->get_pixel(xr,yr);
  int c3 = *front->get_pixel(xl,yl);
  int c4 = *back->get_pixel(xl,yl);
  const int max_diff = 256;

  int d1 = abs(c1 - c2);
  int d2 = abs(c3 - c4);

  if ((d2 <= parent->thresh_foreground) 
      || (d1 >= parent->thresh_background)) 
    edge = 0;
  else edge = (realno) (d2 - d1);
  return (TRUE);
} 


static bool rgb_foreground_edge(int x, int y, int step, Point2& n, 
				 realno& edge, ExtraData userdata)
{
  ForegroundEdgeDetector* parent = (ForegroundEdgeDetector*) userdata;
  int dx = real_to_int(n.x * step);
  int dy = real_to_int(n.y * step);

  int xr = x + dx;
  int yr = y + dy;
  int xl = x - dx;
  int yl = y - dy;
  Image* back = parent->refimage;
  Image* front = parent->vidimage;

  if (!((front->check_coords(xr,yr))  && (front->check_coords(xl,yl))))
    return (FALSE);

  RGB32pixel* p1 = (RGB32pixel*) front->get_pixel(xr,yr);
  RGB32pixel* p2 = (RGB32pixel*) back->get_pixel(xr,yr);
  RGB32pixel* p3 = (RGB32pixel*) front->get_pixel(xl,yl);
  RGB32pixel* p4 = (RGB32pixel*) back->get_pixel(xl,yl);

  int rdiff = p1->red - p2->red;
  int gdiff = p1->green - p2->green;
  int bdiff = p1->blue - p2->blue;
  
  realno d1 = 
    sqrt(rdiff * rdiff + gdiff * gdiff + bdiff * bdiff);
  
  rdiff = p3->red - p4->red;
  gdiff = p3->green - p4->green;
  bdiff = p3->blue - p4->blue;
  realno d2 = 
    sqrt(rdiff * rdiff + gdiff * gdiff + bdiff * bdiff);
  
  if ((d2 <= parent->thresh_foreground) || 
      (d1 >= parent->thresh_background))
    edge = 0;
  else edge = d2 - d1;
  
  return (TRUE);
}


bool GenericFeatureDetector::find_feature(realno u, Point2& pos, 
					Point2& search_line,
					Point2& normal, realno window_size, 
					OcclusionHandler* occlusion_map,
					TrackedProfile* currx,
					Point2& p_obs, realno& var)
{
  curr_u_val = u;

  // ugly switch can't easily be avoided !
  switch (search_mthd)
    {
    case STATISTICAL_SM:
      return (stat_ff(pos,search_line, normal, window_size,
		      occlusion_map, currx, p_obs, var));
    case MAXIMUM_SM:
      return (max_ff(pos,search_line, normal, window_size,
		      occlusion_map, currx, p_obs, var));

    case NEAREST_SM:
      return (nearest_ff(pos,search_line, normal, window_size,
		      occlusion_map, currx, p_obs, var));
    }
  return (FALSE);
}


// mechanisms for combining pixel edge information 

bool GenericFeatureDetector::stat_ff(Point2& pos, Point2& search_line,
				   Point2& normal, realno window_size, 
				   OcclusionHandler* occlusion_map,
				   TrackedProfile* currx,
				   Point2& p_obs, realno& var)
{
  int nsubs = 1; // sampling size
  if (window_size > max_scale)
    nsubs = 0.5 + (window_size / max_scale);
  realno feature_scale = window_size / ((realno) nsubs);
  int grid_scale = feature_scale + 0.5;
  var = 0;
  realno sum_e = 0;
  realno sum_d = 0;
  realno sq_sum = 0;
  realno max_e = -1e30, best = 0;
  Point2 pos_base = pos + Point2(0.5, 0.5);
  Point2 line_offset = feature_scale * search_line;

  for (int i = -nsubs;i <= nsubs; i++)
    {
      realno d_i = i * feature_scale;
      
      realno e_i = significance_threshold-1.0;
      Point2 pos_i = pos_base + (d_i * search_line);
      if (occlusion_map->is_occlusion(pos_i.x, pos_i.y))
	return (FD_RETURN_OCCLUSION);

//       if (!(occlusion_map->is_occlusion(pos_i.x, pos_i.y) ||
// 	    occlusion_map->is_occlusion(pos_i.x + line_offset.x,
// 					pos_i.y + line_offset.y) ||
// 	    occlusion_map->is_occlusion(pos_i.x - line_offset.x,
// 					pos_i.y - line_offset.y)))
      if (!(*edge_response)(pos_i.x, pos_i.y, grid_scale, normal, 
			    e_i, more_data))
	return (FD_RETURN_OCCLUSION);
      realno new_e_i = 0;
      if (e_i > significance_threshold) 
	{
	  new_e_i = e_i;
	  if (e_i > max_e) 
	    { 
	      max_e = e_i;
	      best = d_i;
	    }
	}
	      
      sum_e += new_e_i;
      sum_d += d_i * new_e_i;
      sq_sum += d_i * d_i * new_e_i;
      
    }
 
  if (sum_e == 0) return (FD_RETURN_BAD);
  realno mean = sum_d / sum_e;
  //realno mean = best;
  var = 0.25 * ((sq_sum / sum_e) - (mean * mean));
  p_obs = pos_base + mean * search_line;
  // var += 0.25 * feature_scale * feature_scale;
  return (FD_RETURN_GOOD);
}

bool GenericFeatureDetector::max_ff(Point2& pos, Point2& search_line,
				  Point2& normal, realno window_size, 
				  OcclusionHandler* occlusion_map,
				    TrackedProfile* currx,
				    Point2& p_obs, realno& var)
{
  int nsubs = 1; // sampling size
  if (window_size > max_scale)
    nsubs = 0.5 + (window_size / max_scale);
  realno feature_scale = window_size / ((realno) nsubs);
  int grid_scale = feature_scale + 0.5;
  var = 0;
  realno max_e = -significance_threshold;
  
  Point2 pos_base = pos + Point2(0.5, 0.5);
  for (int i = -nsubs;i <= nsubs; i++)
    {
      Point2 offset =  (i * feature_scale) * search_line;;
      
      realno e_i;
      Point2 pos_i = pos_base + offset;
      if (occlusion_map->is_occlusion(pos_i.x, pos_i.y))
	return (FD_RETURN_OCCLUSION);
      
      if (!(*edge_response)(pos_i.x, pos_i.y, grid_scale, normal, 
			    e_i, more_data))
	return (FD_RETURN_OCCLUSION);
      
      if (e_i > max_e)
	{
	  max_e = e_i;
	  p_obs = pos_i;
	}

    }
  
 if (max_e <= significance_threshold) return (FD_RETURN_BAD);
 // var = 0.25 * feature_scale * feature_scale;
 return (FD_RETURN_GOOD);
}

bool GenericFeatureDetector::nearest_ff(Point2& pos, Point2& search_line,
				      Point2& normal, realno window_size, 
				      OcclusionHandler* occlusion_map,
				      TrackedProfile* currx,
				      Point2& p_obs, realno& var)
{ 
  int nsubs = 1; // sampling size
  if (window_size > max_scale)
    nsubs = 0.5 + (window_size / max_scale);
  realno feature_scale = window_size / ((realno) nsubs);
  int grid_scale = feature_scale + 0.5;
  var = 0;
  realno max_e = -significance_threshold;
  realno e_i;

  Point2 pos_base = pos + Point2(0.5, 0.5);
  p_obs = pos_base;
  if (occlusion_map->is_occlusion(pos_base.x, pos_base.y))
    return (FALSE);
  if (!(*edge_response)(pos_base.x, pos_base.y, grid_scale, normal, 
			e_i, more_data))
    return (FALSE);      
  if (e_i > significance_threshold) return (TRUE);

  for (int i = 1;i <= nsubs; i++)
    {
      Point2 offset =  (i * feature_scale) * search_line;;
      Point2 pos_i = pos_base + offset;

      if (occlusion_map->is_occlusion(pos_i.x, pos_i.y))
	return (FD_RETURN_OCCLUSION);
      
      if (!(*edge_response)(pos_i.x, pos_i.y, grid_scale, normal, 
			    e_i, more_data))
	return (FD_RETURN_OCCLUSION);
      
      if (e_i > significance_threshold)
	{
	  p_obs = pos_i;
	  return (FD_RETURN_GOOD);
	}

      pos_i = pos_base - offset;
      if (occlusion_map->is_occlusion(pos_i.x, pos_i.y))
	return (FD_RETURN_OCCLUSION);
      
      if (!(*edge_response)(pos_i.x, pos_i.y, grid_scale, normal, 
			    e_i, more_data))
	return (FD_RETURN_OCCLUSION);
      if (e_i > significance_threshold)
	{
	  p_obs = pos_i;
	  return (FD_RETURN_GOOD);
	}
    }
  
 return (FD_RETURN_BAD);

} 

// C++ wrappers

SobelDetector::SobelDetector(Image* video_image, realno thresh,
			     EdgeSearchMethod mthd) :
  GenericFeatureDetector(mthd)
{ 
  input_image = video_image;
  more_data = (ExtraData) input_image;
  set_significance_threshold(thresh);
  if (input_image->imgtype() == RGB32)
    edge_response = &rgb_sobel;
  else 
    edge_response = &gray_sobel;
}

realno SobelDetector::get_maximum_response()
{
  if (input_image->imgtype() == GRAY8)
    return (4.0 * 256);
  else if (input_image->imgtype() == RGB32)
    return (2.0 * MAX_RGB_DISTANCE);
    
}

SimpleEdgeDetector::SimpleEdgeDetector(Image* video_image, realno
				       thresh, EdgeSearchMethod mthd)
  : GenericFeatureDetector(mthd)
{
  input_image = video_image;
  more_data = (ExtraData) input_image;
  if (input_image->imgtype() == RGB32)
    edge_response = &rgb_simple;
  else 
    edge_response = &gray_simple;
  set_significance_threshold(thresh);

}

realno SimpleEdgeDetector::get_maximum_response()
{
  if (input_image->imgtype() == GRAY8)
    return (256);
  else if (input_image->imgtype() == RGB32)
    return (MAX_RGB_DISTANCE);
}


MovingEdgeDetector::MovingEdgeDetector(Image* i1, Image* i2, realno
				       thresh, EdgeSearchMethod mthd)
  : GenericFeatureDetector(mthd)
{
  short_sequence[0] = i1;
  short_sequence[1] = i2;
  more_data = (ExtraData) short_sequence;
  if (i1->imgtype() == RGB32)
    edge_response = &rgb_moving_edge;
  else 
    edge_response = &gray_moving_edge;
  set_significance_threshold(thresh);
}

realno MovingEdgeDetector::get_maximum_response()
{
  if (short_sequence[0]->imgtype() == GRAY8)
    return (2.0 * 256);
  else if (short_sequence[1]->imgtype() == RGB32)
    return (2.0 * MAX_RGB_DISTANCE);
}


ForegroundEdgeDetector::ForegroundEdgeDetector(Image* back,
					       Image* front,
					       realno thresh_lower,
					       realno thresh_upper,
					       EdgeSearchMethod mthd)
  : GenericFeatureDetector(mthd)
{
  refimage = back;
  vidimage = front;
  more_data = (ExtraData) this;

  if (front->imgtype() == RGB32)
    edge_response = &rgb_foreground_edge;
  else 
    edge_response = &gray_foreground_edge;

  set_significance_threshold(0.0);
  thresh_background = thresh_lower * get_maximum_response();
  thresh_foreground = thresh_upper * get_maximum_response();
}

realno ForegroundEdgeDetector::get_maximum_response()
{
  if (refimage->imgtype() == GRAY8)
    return (max_gray_diff);
  else if (refimage->imgtype() == RGB32)
    return (MAX_RGB_DISTANCE);
}


realno ColourBlob::colour_distance(RGB32pixel* col)
{
  if (theColour == NULL) return (0);
  int rdiff = col->red - theColour->red;
  int gdiff = col->green - theColour->green;
  int bdiff = col->blue - theColour->blue;

  return (sqrt(rdiff * rdiff + gdiff * gdiff + bdiff * bdiff));

}
  

realno ColourBlob::norm_colour_distance(RGB32pixel* col)
{
  if (theColour == NULL) return (0);
  realno sum1 = 1 + theColour->red + theColour->green + theColour->blue;
  realno r1 = theColour->red / sum1;
  realno g1 = theColour->green / sum1;
  realno b1 = theColour->blue / sum1;

  realno sum2 = 1 + col->red + col->green + col->blue;
  realno r2 = col->red / sum2;
  realno g2 = col->green / sum2;
  realno b2 = col->blue / sum2;

  r1 -= r2; g1 -= g2; b1 -= b2;
  return (sqrt(r1*r1 + g1*g1 + b1*b1));
}

void ColourBlob::update()
{
  if (newColour != NULL)
    {
      if (theColour == NULL) 
	{ 
	  theColour = new RGB32pixel;
	  *theColour = *newColour;
	}
      else 
	{ 
	  int rdiff = newColour->red - theColour->red;
	  if (ABS(rdiff) > 4) rdiff = 4 * ZSGN(rdiff);
	  else rdiff = ZSGN(rdiff);

	  int gdiff = newColour->green - theColour->green;
	  if (ABS(gdiff) > 4) gdiff = 4 * ZSGN(gdiff);
	  else gdiff = ZSGN(gdiff);

	  int bdiff = newColour->blue - theColour->blue;
	  if (ABS(bdiff) > 4) bdiff = 4 * ZSGN(bdiff);
	  else bdiff = ZSGN(bdiff);
	  
	  theColour->red += rdiff;
	  theColour->green += gdiff;
	  theColour->blue += bdiff;
	}
    }
}

static cblob_glwin = NULLWIN;

void ColourBlob::visualise(realno u)
{
  if (theColour == NULL) return;
  if (cblob_glwin == NULLWIN)
    {
      prefsize(200,200);
      foreground();
      cblob_glwin = winopen("Colour Blobs");
      winset(cblob_glwin);
      reshapeviewport();
      prefsize(200,200);
      winconstraints();
      RGBmode();
      pixmode(PM_SIZE, 32);
      cpack(0x00ffffff);
      gflush();
      gconfig();
      RGBcolor(0,0,0);
      clear();
    }

  winset(cblob_glwin);
  RGBcolor((short) theColour->red, (short) theColour->green, (short)
	   theColour->blue);
  circf(100+(50*sin(2 * 3.142 * u)), 100+(50*cos(2 * 3.142 * u)), 10.0);
}

			   

bool 
ColouredFeatureDetector::find_feature(realno u, Point2& pos, 
				      Point2& search_line,
				      Point2& normal, realno window_size, 
				      OcclusionHandler* occlusion_map,
				      TrackedProfile* currx,
				      Point2& p_obs, realno& var)
{
  int indx = u * no_cblobs;
  if (indx >= no_cblobs) 
	{
		fprintf(stderr,"ColouredFeatureDetector::find_feature ABORT\n");
		fflush(stderr);
		abort(); }	// shouldn't happen
  current_blob = &cblobs[indx];

  //current_blob->visualise(indx / (realno) no_cblobs);
  
  int nsubs = 1; // sampling size

  if (window_size > max_scale)
    nsubs = 0.5 + (window_size / max_scale);
  realno feature_scale = window_size / ((realno) nsubs);
  int grid_scale = feature_scale + 0.5;
  var = 0;
  realno max_e = -significance_threshold;
  
  Point2 pos_base = pos + Point2(0.5, 0.5);
  Point2 best_offset(0,0);

  RGB32pixel* best_pix = NULL;
//   realno sum_e = 0;
//   realno sum_d = 0;
//   realno sq_sum = 0;
  for (int i = -nsubs;i <= nsubs; i++)
    {
      realno d_i = i * feature_scale;
      Point2 offset =  (i * feature_scale) * search_line;;
      
      realno e_i;
      Point2 pos_i = pos_base + offset;
      if (occlusion_map->is_occlusion(pos_i.x, pos_i.y))
	return (FD_RETURN_OCCLUSION);
      
      RGB32pixel* inside_pix = 
	edge_response(pos_i.x, pos_i.y, grid_scale, normal, e_i);
      //if (inside_pix == NULL) return (FD_RETURN_OCCLUSION);

      realno new_e_i = 0;
      if ((inside_pix != NULL) && (e_i > significance_threshold))
	{
	  new_e_i = e_i;
	  if (e_i > max_e)
	    {
	      max_e = e_i;
	      best_pix = inside_pix;
	      best_offset = offset;
	    }
	}
//       sum_e += new_e_i;
//       sum_d += d_i * new_e_i;
//       sq_sum += d_i * d_i * new_e_i;
      
    }
  
  if (best_pix == NULL) return (FD_RETURN_BAD);

  current_blob->add_observation(best_pix);
  
//   realno mean = sum_d / sum_e;
//   p_obs = pos + mean * search_line;
//   var = 0.25 * ((sq_sum / sum_e) - (mean * mean));

  p_obs = pos + best_offset;

  return (FD_RETURN_GOOD);
} 

void ColouredFeatureDetector::setup_batch(TrackedProfile* curr)
{
  if (curr->colour_info == NULL)
    curr->colour_info = new ColourBlob[no_cblobs];
  else
    {
      for (int i = 0; i < no_cblobs; i++)
	curr->colour_info[i].update();
    }

  cblobs = curr->colour_info;
  this->FeatureDetector::setup_batch(curr);
}

	  
	  

ColouredForegroundEdgeDetector::
ColouredForegroundEdgeDetector(Image* back, Image* front, 
		     realno thresh_lower,
		     realno thresh_upper,
		     int no_cblobs)

  : ColouredFeatureDetector(no_cblobs)
{
  if (front->imgtype() != RGB32)
    {
      AmbErrorHandler::handle_error("ColouredForegroundEdgeDetector::\
can't construct -- only implemented for RGB images");
	fprintf(stderr,"ColouredForegroundEdgeDetector:: ABORT\n");
		fflush(stderr);
      abort();
    }
  refimage = back; vidimage = front; 
  set_significance_threshold(0.0);
  thresh_background = thresh_lower * get_maximum_response();
  thresh_foreground = thresh_upper * get_maximum_response();
}


RGB32pixel* 
ColouredForegroundEdgeDetector::edge_response(realno x, realno y, 
					      int step, Point2 n,
					      realno& edge)
{
  int dx = real_to_int(n.x * step);
  int dy = real_to_int(n.y * step);


  int xr = x + dx;
  int yr = y + dy;
  int xl = x - dx;
  int yl = y - dy;

  if (!((vidimage->check_coords(xr,yr))  && (vidimage->check_coords(xl,yl))))
    return (NULL);

  RGB32pixel* p1 = (RGB32pixel*) vidimage->get_pixel(xr,yr);
  RGB32pixel* p2 = (RGB32pixel*) refimage->get_pixel(xr,yr);
  RGB32pixel* p3 = (RGB32pixel*) vidimage->get_pixel(xl,yl);
  RGB32pixel* p4 = (RGB32pixel*) refimage->get_pixel(xl,yl);

  int rdiff = p1->red - p2->red;
  int gdiff = p1->green - p2->green;
  int bdiff = p1->blue - p2->blue;
  realno d1 = 
    sqrt(rdiff * rdiff + gdiff * gdiff + bdiff * bdiff);
  
  rdiff = p3->red - p4->red;
  gdiff = p3->green - p4->green;
  bdiff = p3->blue - p4->blue;
  realno d2 = 
    sqrt(rdiff * rdiff + gdiff * gdiff + bdiff * bdiff);
  
  if ((d2 <= thresh_foreground) || (d1 >= thresh_background))
    edge = 0;
  else 
    {
      edge = d2 - d1;
      edge += MAX_RGB_DISTANCE - current_blob->colour_distance(p3);
      
    }

  return (p3);
}


realno ColouredForegroundEdgeDetector::get_maximum_response()
{
  return (2.0 * MAX_RGB_DISTANCE);
}


NormForeColourDetect::
NormForeColourDetect(Image* back, Image* front, 
		     realno thresh_lower,
		     realno thresh_upper,
		     int no_cblobs)
  
  : ColouredFeatureDetector(no_cblobs)
{
  if (front->imgtype() != RGB32)
    {
      AmbErrorHandler::handle_error("NormForeColourDetect::\
can't construct -- only implemented for RGB images");
	fprintf(stderr,"NormForeColourDetect:: ABORT\n");
		fflush(stderr);
      abort();
    }
  refimage = back; vidimage = front; 
  set_significance_threshold(0.0);
  thresh_background = thresh_lower * get_maximum_response();
  thresh_foreground = thresh_upper * get_maximum_response();
}

static const realno MAX_NORM_RGB_DIST = sqrt(2.0);

RGB32pixel* 
NormForeColourDetect::edge_response(realno x, realno y, 
				    int step, Point2 n,
				    realno& edge)
{
  int dx = real_to_int(n.x * step);
  int dy = real_to_int(n.y * step);


  int xr = x + dx;
  int yr = y + dy;
  int xl = x - dx;
  int yl = y - dy;

  if (!((vidimage->check_coords(xr,yr))  && (vidimage->check_coords(xl,yl))))
    return (NULL);

  RGB32pixel* p1 = (RGB32pixel*) vidimage->get_pixel(xr,yr);
  RGB32pixel* p2 = (RGB32pixel*) refimage->get_pixel(xr,yr);
  RGB32pixel* p3 = (RGB32pixel*) vidimage->get_pixel(xl,yl);
  RGB32pixel* p4 = (RGB32pixel*) refimage->get_pixel(xl,yl);

  realno sum1 = 1 + p1->red + p1->green + p1->blue;
  realno sum2 = 1 + p2->red + p2->green + p2->blue;
  realno sum3 = 1 + p3->red + p3->green + p3->blue;
  realno sum4 = 1 + p4->red + p4->green + p4->blue;

  realno rdiff = (p1->red / sum1) - (p2->red / sum2);
  realno gdiff = (p1->green / sum1) - (p2->green / sum2);
  realno bdiff = (p1->blue / sum1) - (p2->blue / sum2);
  realno d1 = 
    sqrt(rdiff * rdiff + gdiff * gdiff + bdiff * bdiff);

  rdiff = (p3->red / sum3) - (p4->red / sum4);
  gdiff = (p3->green / sum3) - (p4->green / sum4);
  bdiff = (p3->blue / sum3) - (p4->blue / sum4);
  realno d2 = 
    sqrt(rdiff * rdiff + gdiff * gdiff + bdiff * bdiff);
  
  if ((d2 <= thresh_foreground) || (d1 >= thresh_background))
    edge = 0;
  else 
    {
      edge = d2 - d1;
      edge += MAX_NORM_RGB_DIST - current_blob->norm_colour_distance(p3);
      
    }

  return (p3);
}


realno NormForeColourDetect::get_maximum_response()
{
  return (2.0 * MAX_NORM_RGB_DIST);
}

