/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/***************************************************************
 * C - C++ Header
 *
 * File : 	FeatureDetector.h
 *
 * Module :	FeatureDetector
 *
 * Author : 	A M Baumberg (CoMIR)
 *
 * Creation Date : Thu May 16 14:01:03 1996 
 *
 * Comments : 	general edge/ contrast etc detection
 *
 ***************************************************************/


/* include files */
#ifndef __FEATURE_DETECTOR_H__
#define __FEATURE_DETECTOR_H__

#include <TrackedProfile.h>
class OcclusionHandler;
#include <RGB32Image.h>


#include <point2.h>
#ifndef MIN
#define MIN(a,b) (( (a) < (b)) ? (a) : (b))
#endif

class Image;

#define FD_RETURN_GOOD 1
#define FD_RETURN_BAD 0
#define FD_RETURN_OCCLUSION -1

// base class 
class FeatureDetector
{
protected:
  
  realno max_scale;			// maximum scale for subsampling

  realno significance_threshold;	// minimum significant response

  virtual realno get_maximum_response() { return (0); }

  static realno max_scale_height_ratio;
  static realno max_scale_width_ratio;

public:

  void set_significance_threshold(realno frac = 0.0)
  { significance_threshold = frac * get_maximum_response(); }

  FeatureDetector() {} 

  // find a feature in a window;
  // look on a line r = 'pos' + k 'search_line'
  // with |k| < 'window_size'
  // the contour normal is 'normal'
  // the parametric value for the curve point is u
  //
  // return the measurement 'p_obs' 
  // and the discrete measurement variance 'var'
  // return TRUE if a significant feature was found
  
  //   bool find_feature(realno u, Point2& pos, Point2& search_line,
  // 	       Point2& normal, realno window_size, 
  //	       OcclusionHandler* occlusion_map, TrackedProfile* curr,
  // 	       Point2& p_obs, realno& var);

  virtual bool find_feature(realno, Point2& , Point2&, Point2&, realno,
			    OcclusionHandler*, TrackedProfile*,
			    Point2& , realno& ) { return (TRUE); }
  

  // get object dependencies before performing a batch
  // of feature detections
  // e.g. need the object size to determine 
  // maximum feature size ( maximum subsample scale)

  //   setup_batch(TrackedProfile* object);
  
  virtual void setup_batch(TrackedProfile* curr) 
  { 
    max_scale = MIN(max_scale_width_ratio * curr->width,
		    max_scale_height_ratio * curr->height);
    if (max_scale < 1.0) max_scale = 1.0;
  }

};

typedef void *ExtraData;
typedef void *ImageData;

enum EdgeSearchMethod { STATISTICAL_SM, MAXIMUM_SM, NEAREST_SM };

class GenericFeatureDetector : public FeatureDetector
{

protected:

  ExtraData more_data;
  realno curr_u_val;

  bool (*edge_response) (int, int, int, Point2&, realno &, ExtraData );
 
  EdgeSearchMethod search_mthd;

public:

  GenericFeatureDetector(EdgeSearchMethod srch = STATISTICAL_SM,
			 ExtraData misc = NULL) 
  { more_data = misc; search_mthd = srch; }

  bool find_feature(realno, Point2& , Point2&, Point2&, realno,
		    OcclusionHandler*, TrackedProfile*, Point2& ,
		    realno& );
  
  // statistical method:
  // look along the profile and obtain an edge response at each point.
  // combine to obtain the mean and variance for the measurement
  // process
  bool stat_ff(Point2& , Point2&, Point2&, realno,
		      OcclusionHandler*, TrackedProfile*, Point2& ,
		      realno& );

  // maximum method:
  // look along the profile and obtain an edge response at each point.
  // find the maximum response
  bool max_ff(Point2& , Point2&, Point2&, realno,
		      OcclusionHandler*, TrackedProfile*, Point2& ,
		      realno& );

  // nearest method:
  // find the nearest significant feature
  bool nearest_ff(Point2& , Point2&, Point2&, realno,
			 OcclusionHandler*, TrackedProfile*, Point2& ,
			 realno& );
};


class SobelDetector : public GenericFeatureDetector
{
protected:

  Image* input_image;

public:
  
  SobelDetector(Image* video_in, realno thresh = 0.0,
		EdgeSearchMethod mthd = STATISTICAL_SM);

  realno get_maximum_response();
};

// Edge (1D)

class SimpleEdgeDetector : public GenericFeatureDetector
{
protected:

  Image* input_image;

public:
  
  SimpleEdgeDetector(Image* video_in, realno thresh = 0.0, 
		     EdgeSearchMethod mthd = STATISTICAL_SM);
  realno get_maximum_response();
  
};

// Moving Edge
//
// looking at spatial and temporal derivative
// but don't assume background image is available

class MovingEdgeDetector : public GenericFeatureDetector
{
protected:

  Image* short_sequence[2];
  // short_sequence[0] == previous image frame
  // short_sequence[1] == current image frame

public:
  
  MovingEdgeDetector(Image* prev, Image* curr, realno thresh = 0.0,
		     EdgeSearchMethod mthd = STATISTICAL_SM);

  realno get_maximum_response();
  
};

// ForegroundEdge (signed edge)


class ForegroundEdgeDetector : public GenericFeatureDetector
{
public:

  realno thresh_background;
  realno thresh_foreground;

  Image* refimage;	// background (reference) image
  Image* vidimage;	// foreground image

  
  // a foreground edge is detected if
  // d1 < thresh_lower, d2 > thresh_upper
  // where
  // d1 = movement detected to the left of the edge
  // d2 = movement detected to the right of the edge
  // scaled such that 0 <= d_i <= 1
  ForegroundEdgeDetector(Image* back, Image* front, 
			 realno thresh_lower = 0.5,
			 realno thresh_upper = 0.5,
			 EdgeSearchMethod mthd = STATISTICAL_SM);

  realno get_maximum_response();
  
};

class ColourBlob
{
public:

  //NagVector mean_colour;
  //NagVector new_mean;

  RGB32pixel* theColour;
  
  RGB32pixel* newColour;
  
  //NagMatrix E_xxT;

  //NagMatrix inv_covariance;
  
  ColourBlob() 
  {
    newColour = NULL;
    theColour = NULL;
  }

  ~ColourBlob()
  {
    delete (newColour);
    delete (theColour);
  }

  void add_observation(RGB32pixel* observed_colour)
  { 
    if (newColour == NULL) newColour = new RGB32pixel;
    *newColour = *observed_colour; 
  }
  void update();
  realno colour_distance(RGB32pixel* col);
  // normalised (hue) colour distance
  realno norm_colour_distance(RGB32pixel* col);

  void visualise(realno u = 0); 
 
};

class ColouredFeatureDetector : public FeatureDetector
{
public:

  ColourBlob* cblobs;
  ColourBlob* current_blob;
  //realno col_gain;

  int no_cblobs;

  realno edge_threshold;
  
  ColouredFeatureDetector(int no_colour_blobs = 20)
  { no_cblobs = no_colour_blobs; }

  bool find_feature(realno, Point2& , Point2&, Point2&, realno,
		    OcclusionHandler*, TrackedProfile*, Point2& ,
		    realno& );
  
  virtual RGB32pixel* 
  edge_response(realno /* x */, realno /* y */, 
		int /* grid_scale */, Point2 /* n */,
		realno& /* response */) {}
  
  void setup_batch(TrackedProfile* curr);
};


// Use a combination of colour image subtraction and
// inside colour (from previous frames)
class ColouredForegroundEdgeDetector : public ColouredFeatureDetector
{

public:

  realno thresh_background;
  realno thresh_foreground;

  Image* refimage;	// background (reference) image
  Image* vidimage;	// foreground image

  
  ColouredForegroundEdgeDetector(Image* back, Image* front, 
				 realno thresh_lower = 0.5,
				 realno thresh_upper = 0.5,
				 int no_cblobs = 20);
  RGB32pixel* edge_response(realno  x, realno y, int grid_scale,
			    Point2 n, realno& repsonse); 

  realno get_maximum_response();

};

// Uses a combination of *normalised* colour subtraction
// and inside normalised colour (from previous frames)
class NormForeColourDetect : public ColouredFeatureDetector
{

public:

  realno thresh_background;
  realno thresh_foreground;

  Image* refimage;	// background (reference) image
  Image* vidimage;	// foreground image

  
  NormForeColourDetect(Image* back, Image* front, 
		       realno thresh_lower = 0.5,
		       realno thresh_upper = 0.5,
		       int no_cblobs = 20);
  RGB32pixel* edge_response(realno  x, realno y, int grid_scale,
			    Point2 n, realno& repsonse); 

  realno get_maximum_response();

};
#endif
