/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/***************************************************************
 * C++ source
 *
 * File : 	GPfilter.cc
 *
 * Module :	GPfilter
 *
 * Author : 	A M Baumberg (CoMIR)
 *
 * Creation Date : Fri Aug  2 15:26:37 1996 
 *
 * Comments : 	filter for tracking Ground Plane position, 
 *		real world height
 *
 ***************************************************************/

#include <TrackedProfile.h>
#include <GPfilter.h>
#include <GroundPlaneToImagePlane.h>
#include <hash_defines.h>
#include <utils.h>


// average height of a person is 1.8m
realno GPfilter::Average_World_Height = 1800;


#define GPtoIP GroundPlaneToImagePlane
#define IPtoGP ImagePlaneToGroundPlane

void GPfilter::setup_image_tracking()
{
  // map ground plane parameters into image plane parameters
  // + covariance etc

  NagVector x_I(3);	 // contact point (image coords)
  NagVector x_W(3);	 // contact point (world coords)
  
  x_W[0] = gp_pos_tracker->get_state().x;
  x_W[1] = gp_pos_tracker->get_state().y;
  x_W[2] = 1;
  GPtoIP.multiply(x_W, x_I);
  Matrix2by2 
    Jacobian(x_I[2] * GPtoIP.read(0,0) - x_I[0] * GPtoIP.read(2,0),
	     x_I[2] * GPtoIP.read(0,1) - x_I[0] * GPtoIP.read(2,1),
	     x_I[2] * GPtoIP.read(1,0) - x_I[1] * GPtoIP.read(2,0),
	     x_I[2] * GPtoIP.read(1,1) - x_I[1] * GPtoIP.read(2,1));
  Jacobian *= 1.0 / SQUARE(x_I[2]);
  Matrix2by2 image_contact_cov =  Jacobian * 
    (gp_pos_tracker->C00 + model_error_cov) *  Jacobian.transpose();

  // (i) etsimate image height = h / Z_i = h * Z_w
  // realno img_height = height_tracker->get_state() / x_I[2];
  realno world_z = 
    world_depth_at_image_coords(x_I[0] / x_I[2], x_I[1] / x_I[2]);
  realno var_world_z =
    Point2(ImagePlaneToDepth[0], ImagePlaneToDepth[1]) *
    (image_contact_cov * Point2(ImagePlaneToDepth[0],
				ImagePlaneToDepth[1]));
  realno img_height = height_tracker->get_state() * world_z;
  realno var_img_height = SQUARE(height_tracker->get_state()) *
    var_world_z + (SQUARE(world_z) * height_tracker->get_p11());

  start_height = img_height;

  // (ii) now estimate scale 's'
  realno y_max, y_min;
  parent->get_y_bounds(y_min, y_max);
  realno old_scale = parent->s_filter->get_state();
  realno old_height = (realno) (y_max - y_min);  
  realno new_scale = old_scale * (img_height / old_height);
  

  realno scale_var = SQUARE(old_scale / old_height) * var_img_height;

  start_scale = new_scale;
  start_scale_var = scale_var;
  parent->s_filter->set_state(new_scale);
  parent->s_filter->set_variance(scale_var);

  // (iii) now estimate origin from image contact point
  Point2 image_contact(x_I[0] / x_I[2], (x_I[1] / x_I[2]));
  Point2 old_offset(0, y_min);
  Point2 new_offset = old_offset * (new_scale / old_scale);
  Point2 new_origin = image_contact - new_offset;

  // (iv) calculate covariance of estimate
  // using extended Kalman approach

  Matrix2by2 origin_cov = image_contact_cov;
  origin_cov.val11 += scale_var * SQUARE(old_offset.y / old_scale);
  start_origin_cov = origin_cov;
  start_origin = new_origin;

  parent->pos_filter->get_state() = new_origin;
  parent->pos_filter->get_cov() = start_origin_cov;
}

// return the "variance" of the combined observations
// on the ground plane 
realno GPfilter::add_image_observations()
{
  // calculate combined "virtual" observations 
  // from converged image frame filters 
  
  Matrix2by2 P_inv = parent->pos_filter->get_cov().inverse();
  Matrix2by2 R_inv = P_inv - start_origin_cov.inverse();
  // if no significant observations return
  if (fabs(R_inv.determinant() < 1e-30)) return (1e30); 
  Matrix2by2 R = R_inv.inverse();
  realno image_frame_error = R.get_max_real_eigenvalue();

  Matrix2by2 K_inv = R * P_inv;
  Point2 z = start_origin + 
    K_inv.multiply(parent->pos_filter->get_state() - start_origin);

  
  // add offset to observed origin
  realno y_max, y_min;
  parent->get_y_bounds(y_min, y_max);
  
  // calculate virtual observation of height 
  // from scale filter
  realno curr_scale = parent->s_filter->get_state();
  realno h_s = (y_max - y_min);
  realno r_s = (start_scale_var * parent->s_filter->get_p11()) /
    (start_scale_var - parent->s_filter->get_p11());
  realno k_s = parent->s_filter->get_p11() / r_s;
  realno z_s = start_scale + ((curr_scale - start_scale)
    / k_s);
  realno observed_offset = y_min * (z_s / curr_scale);
  z += Point2(0, observed_offset);
  // z += Point2(0, y_max - start_height);

  // update image contact point measurement covariance
  R.val11 += r_s *  SQUARE(y_min / curr_scale);
  // R += model_error_cov;
  
  
  realno image_height_obs = h_s * z_s / curr_scale;
  realno var_image_height = r_s * SQUARE(h_s / curr_scale);

  // project image contact point to ground plane coordinates
  NagVector x_I(3);
  NagVector x_W(3);
  x_I[0] = z.x; x_I[1] = z.y; x_I[2] = 1;
  IPtoGP.multiply(x_I, x_W);
  Matrix2by2 
    Jacobian(x_W[2] * IPtoGP.read(0,0) - x_W[0] * IPtoGP.read(2,0),
	     x_W[2] * IPtoGP.read(0,1) - x_W[0] * IPtoGP.read(2,1),
	     x_W[2] * IPtoGP.read(1,0) - x_W[1] * IPtoGP.read(2,0),
	     x_W[2] * IPtoGP.read(1,1) - x_W[1] * IPtoGP.read(2,1));
  Jacobian *= 1.0 / SQUARE(x_W[2]);
  Matrix2by2 gp_pos_cov = Jacobian * R * Jacobian.transpose();
  Point2 gp_pos(x_W[0] / x_W[2], x_W[1] / x_W[2]);

  realno world_depth = world_depth_at_image_coords(x_I[0], x_I[1]);
  realno observed_height = image_height_obs / world_depth;
  realno height_var = var_image_height / SQUARE(world_depth);

  // add observations to GP filters
  height_tracker->add_observation(observed_height, height_var);
  gp_pos_tracker->add_observation(gp_pos, gp_pos_cov  +
				  model_error_cov);

  // the maximum eigenvalue is a good bound on the measurement error
  //return (gp_pos_cov.get_max_real_eigenvalue());
  return (image_frame_error);

}


GPfilter::GPfilter
(
 TrackedProfile* prf,	// object that the GP filter refers to
 Point2 image_contact,	// contact point in image
 realno image_height,	// height in image
 realno init_hght_var,   // initial variance of image height (relative
			// to initial height )
 realno rel_hght_var,	// noise variance (relative to initial height)
 realno init_pos_var,	// initial GP position variance
 realno gp_acc_var,	// GP random acceleration variance
 realno model_err_var	// modeling error term (GP coords)
 )
{
  NagVector x_I(3);
  NagVector x_W(3);

  parent = prf;
  if (IPtoGP.get_data() == NULL)
    setup_default_projection();
      
  // note that we assume the parent profile contains
  // the initial shape control points scaled appropriately

  // (i) map ICP to GCP
    
  x_I[0] = image_contact.x;
  x_I[1] = image_contact.y;
  x_I[2] = 1;

  IPtoGP.multiply(x_I, x_W);
  Point2 GPCP(x_W[0] / x_W[2], x_W[1] / x_W[2]);
  
  gp_pos_tracker = new DynamicKalmanTwoD(gp_acc_var, gp_acc_var, 0, 0,
				      GPCP, Point2(0,0), 
				      init_pos_var, init_pos_var,
				      init_pos_var, init_pos_var,
				      1.0);

  model_error_cov = Matrix2by2(model_err_var, model_err_var);

  // (ii) map image height to world height

//   realno h_w = image_height / x_W[2];
//   realno h_var = rel_hght_var * h_w * h_w;

  // use default fixed initial height

  realno h_w = Average_World_Height;
  realno h_var = init_hght_var * h_w * h_w;
  realno h_noise = rel_hght_var * h_w * h_w;

  height_tracker = new StaticKalmanOneD(h_noise, 1.0, h_w, h_var);

  // (iii) if IP filters haven't been initialised then do it

  
  if (parent->s_filter == NULL)
    parent->s_filter = new StaticKalmanOneD(0,0,parent->a_x, 0);

  if (parent->pos_filter == NULL)
    parent->pos_filter = (FilterTwoD*) new StaticKalmanTwoD;

}      

Point2 GPfilter::IP_velocity()
{
  // project GP velocity into image plane
  NagVector x_I(3);	 // contact point (image coords)
  NagVector x_W(3);	 // contact point (world coords)
  
  x_W[0] = gp_pos_tracker->get_state().x;
  x_W[1] = gp_pos_tracker->get_state().y;
  x_W[2] = 1;

  GPtoIP.multiply(x_W, x_I);

  Matrix2by2 
    Jacobian(x_I[2] * GPtoIP.read(0,0) - x_I[0] * GPtoIP.read(2,0),
	     x_I[2] * GPtoIP.read(1,0) - x_I[1] * GPtoIP.read(2,0),
	     x_I[2] * GPtoIP.read(0,1) - x_I[0] * GPtoIP.read(2,1),
	     x_I[2] * GPtoIP.read(1,1) - x_I[1] * GPtoIP.read(2,1));
  Jacobian *= 1.0 / SQUARE(x_I[2]);
  

 
  return (Jacobian.multiply(gp_pos_tracker->get_velocity()));
}

GPfilter::GPfilter(GPfilter& original, TrackedProfile* prf)
{
  parent = prf;
  
  gp_pos_tracker = (DynamicKalmanTwoD*) original.gp_pos_tracker->clone();
  height_tracker = (StaticKalmanOneD*) original.height_tracker->clone();

  model_error_cov = original.model_error_cov;
  start_origin_cov = original.start_origin_cov;
  start_origin = original.start_origin;
  start_scale = original.start_scale;
  start_scale_var = original.start_scale_var;
}

  
void GPfilter::split_profile(Profile* curr, TrackedProfile* res1,
			     TrackedProfile* res2)
{
  // hypothesis 1 : bottom point corresponds
  // to contact point (feet on ground)
  Point2 contact1(curr->origin - Point2(0, curr->height / 2));
  realno img_height1 = Average_World_Height *
    (ImagePlaneToDepth[0] * contact1.x + ImagePlaneToDepth[1] * contact1.y
     + ImagePlaneToDepth[2]);
  res1->origin = contact1 + Point2(0,img_height1 / 2);
  res1->height = img_height1;
  res1->width = curr->width;

  // hypothesis 2 : top point corresponds to
  // top of head (fixed world height)

  Point2 headpnt(curr->origin + Point2(0, curr->height / 2));

  realno img_height2 = 
    (ImagePlaneToDepth[0] * headpnt.x + ImagePlaneToDepth[1] * headpnt.y 
     + ImagePlaneToDepth[2]) * Average_World_Height
    / (ImagePlaneToDepth[1] * Average_World_Height + 1.0);

  Point2 contact2(headpnt - Point2(0,img_height2));
  res2->origin = contact2 + Point2(0,img_height2 / 2);
  res2->height = img_height2;
  res2->width = curr->width;
}

    
// average height of a person is 1.8m
int GPfilter::register_globals()
{

  Average_World_Height = 
    AmbGlobals.register_real("AVERAGE_WORLD_HEIGHT", 1800,
			     &Average_World_Height,
			     FALSE, "GPfilter",
			     "The average height of a person in ground \
plane units (e.g 1.8m)");

  return (TRUE);
}



