/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/***************************************************************
 * C - C++ Header
 *
 * File : 	GPfilter.h
 *
 * Module :	GPfilter
 *
 * Author : 	A M Baumberg (CoMIR)
 *
 * Creation Date : Thu Aug  1 16:54:36 1996 
 *
 * Comments : 	filter for tracking Ground Plane position, 
 *		real world height
 *
 ***************************************************************/


#ifndef __GP_FILTER_H__
#define __GP_FILTER_H__

#include <NagMatrix.h>
#include <Matrix2by2.h>
#include <KalmanFilter.h>

class Profile;
class TrackedProfile;


class GPfilter
{
public:

  static realno Average_World_Height;  // the typical world height of
				       // a person

  TrackedProfile* parent;	// for accessing profile parameters

  DynamicKalmanTwoD*  gp_pos_tracker; // constant velocity 2D filter

  // additional error covariance to account for modeling errors
  Matrix2by2 model_error_cov;

  // store the origin covariance before measurements were made
  Matrix2by2 start_origin_cov;
  // store the origin coordinates before measurements were made
  Point2 start_origin;
  // store the scale + variance before measurements were made
  realno start_scale, start_scale_var;
  realno start_height;
  static realno y_correct;

  StaticKalmanOneD* height_tracker;	// real world height
					// (arbitrary scale)

  // constructor

  GPfilter(GPfilter &original, TrackedProfile* parent);

  GPfilter( TrackedProfile* prf, Point2 image_contact, 
	    realno image_height, realno init_hght_var,
	    realno rel_hght_var, 
	    realno init_pos_var, realno gp_acc_var,
	    realno model_err_var = 0.0);
  
  ~GPfilter() 
  { 
    delete (gp_pos_tracker);
    delete (height_tracker);
  }

  void setup_image_tracking();
  realno add_image_observations();
  void next_frame() 
  { 
    gp_pos_tracker->get_prediction(); 
    height_tracker->update_state();
  }
  Point2 IP_velocity();
  Point2 GP_velocity() { return (gp_pos_tracker->get_velocity()); }
  Point2 GP_position() { return (gp_pos_tracker->get_state()); }

  realno gp_variance() 
  { return (gp_pos_tracker->get_uncertainty()); }

  // split_profile
  // generate two hypothesis
  // assuming fixed height and possible occlusion
  // (note this is a static member function)
  static void split_profile(Profile* original, TrackedProfile* res1,
			    TrackedProfile* res2);

  static int register_globals();
};

#endif
   

  
  
  
