/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/***************************************************************
 * C++ source
 *
 * File : 	GroundPeopleTracker.cc
 *
 * Module :	GroundPeopleTracker
 *
 * Author : 	A M Baumberg (CoMIR)
 *
 * Creation Date : Mon Oct 21 11:45:28 1996 
 *
 * Comments : 	
 *
 ***************************************************************/


/* include files */

#include "GroundPeopleTracker.h"
#include <point2.h>
#include <GPfilter.h>
#include <utils.h>
#include <FeatureDetector.h>
#include <imgsrc_utils.h>


static int doit1 = gp_to_ip_register_globals();
static int doit2 = GPfilter::register_globals();

static EnvIntParameter start_ref_no("START_REF_NO", 0);

realno GroundPeopleTracker::FITNESS_THRESHOLD  = 
AmbGlobals.register_real("START_FITNESS_THRESHOLD", 0.85,
			 0.0, 1.0, 
			 &GroundPeopleTracker::FITNESS_THRESHOLD,
			 FALSE, "GroundPeopleTracker",
			 "Minimum fitness score to initialise track");

realno GroundPeopleTracker::VERY_POOR_FIT =
AmbGlobals.register_real("TRACK_FITNESS_THRESHOLD", 0.4,
			 0.0, 1.0,
			 &GroundPeopleTracker::VERY_POOR_FIT,
			 FALSE, "GroundPeopleTracker",
			 "Minimum fitness score to maintain track");

int GroundPeopleTracker::SMALL_NO_FRAMES =
AmbGlobals.register_int("SMALL_NO_FRAMES", 3,
			&GroundPeopleTracker::SMALL_NO_FRAMES,
			FALSE, "GroundPeopleTracker",
			"Number of frames before track is accepted");

			
realno GroundPeopleTracker::MIN_SIGNIFICANCE = 
AmbGlobals.register_real("MIN_SIGNIFICANCE", 0.3,
			 0.0, 1.0, 
			 &GroundPeopleTracker::MIN_SIGNIFICANCE,
			 FALSE, "GroundPeopleTracker",
			 "Fraction of curve that must \
be visible for tracking");

realno GroundPeopleTracker::terminate_thresh =
AmbGlobals.register_real("TERMINATION_THRESHOLD", 1600,
			 &GroundPeopleTracker::terminate_thresh,
			 FALSE, "GroundPeopleTracker",
			 "Maximum std deviation of ground plane position");

realno GroundPeopleTracker::init_pos_sd =
AmbGlobals.register_real("GP_INIT_POS_SD", 200,
			 &GroundPeopleTracker::init_pos_sd,
			 FALSE, "GroundPeopleTracker",
			 "Initial standard deviation for \
each ground plane coordinate");


realno GroundPeopleTracker::acc_sd =
AmbGlobals.register_real("GP_POS_ACC_SD", 200,
			 &GroundPeopleTracker::acc_sd,
			 FALSE, "GroundPeopleTracker",
			 "Standard deviation of random \
acceleration term for each ground plane coordinate");


realno GroundPeopleTracker::model_err_sd =
AmbGlobals.register_real("GP_MODEL_ERR_SD", 0,
			 &GroundPeopleTracker::model_err_sd,
			 FALSE, "GroundPeopleTracker",
			 "Standard deviation of additional error\
 term to account for modeling errors");

realno GroundPeopleTracker::init_scale_sd =
AmbGlobals.register_real("INIT_SCALE_SD", 0.015,
			 &GroundPeopleTracker::init_scale_sd,
			 FALSE, "GroundPeopleTracker",
			 "Standard deviation of initial estimate of\
 object size relative to initial estimate");

realno GroundPeopleTracker::rel_scale_sd =
AmbGlobals.register_real("REL_SCALE_SD", 0.0,
			 &GroundPeopleTracker::rel_scale_sd,
			 FALSE, "GroundPeopleTracker",
			 "Standard deviation of scale noise term \
relative to initial scale");

int GroundPeopleTracker::TEMPORAL_SKIP = 
AmbGlobals.register_int("FRAME_SKIP", 4,
			 &GroundPeopleTracker::TEMPORAL_SKIP,
			 TRUE, "GroundPeopleTracker",
			 "Temporal resampling for object detection");



static char* detect_methods[] =
{ "COLOUR_FOREGROUND", "COLOUR_EDGE",
  "MOVING_EDGE", "SIMPLE_EDGE",
  "SOBEL_EDGE", "FOREGROUND_EDGE", 
  "NORMALISED_COLOUR_FOREGROUND", NULL };

static char* detection_method = 
AmbGlobals.register_string("FEATURE_DETECTION", "COLOUR_FOREGROUND",
			   &detection_method, TRUE,
			   "GroundPeopleTracker",
			   "The type of features used for tracking",
			   detect_methods);

static realno feature_threshold = 
AmbGlobals.register_real("FEATURE_THRESH", 0.05,
			 0.0, 1.0, &feature_threshold,
			 TRUE, "GroundPeopleTracker",
			 "A significance threshold for feauture detection");

static char* the_pca_file = 
AmbGlobals.register_string("PCA_MODEL_FILE",
			   NULL, &the_pca_file, TRUE,
			   "GroundPeopleTracker",
			   "An optional eigenshape model file. If \
not set, the default model is used");

GroundPeopleTracker::GroundPeopleTracker(WorldOcclusionHandler* ohandler,   
					 ActiveModel* amodel,
					 MotionDetector* mdetect,
					 ProfileSet* curr_profiles)
{

  setup_default_projection();	// setup the Ground Plane
  if (curr_profiles == NULL) curr_profiles = new ProfileSet;
  profiles = curr_profiles;
  occ_handler = ohandler;
  local_optimiser = amodel;
  detect_motion = mdetect;
  previous_frame = NULL;
}


GroundPeopleTracker::GroundPeopleTracker(ImageSource* img_in, char* pca_file,
					 ProfileSet* profiles_out,
					 ExternalData* tracking_data,
					 FeatureDetector* features)
{
  setup_default_projection();	// setup the Ground Plane
  previous_frame = NULL;
  profiles = profiles_out;
  if (profiles == NULL) profiles = new ProfileSet();

  FeatureDetector* tracking_force = NULL;
  bool use_background_image = FALSE;
  // do we need a full size background image for feature detection?
  if (strstr(detection_method, "FOREGROUND")) 
    use_background_image = TRUE;

  // construct motion detection object
  detect_motion = new MotionDetector(img_in, get_reference_frame(),
				     !use_background_image);

  detect_motion->process_next_frame();
  if (strstr(detection_method, "COLOUR") && 
      (img_in->get_current()->imgtype() != RGB32))
    {
      AmbErrorHandler::handle_warning("ignoring colour feature \
detection");
      detection_method = "FOREGROUND_EDGE";
    }  
  // construct a feature detection object
  if (features != NULL) tracking_force = features;
  else if (strcmp(detection_method, "COLOUR_FOREGROUND") == 0)
    {
      tracking_force = new 
	ColouredForegroundEdgeDetector(detect_motion->background->get_current()
				       ,img_in->get_current(),
				       feature_threshold,
				       feature_threshold);
    }
  else if (strcmp(detection_method,"MOVING_EDGE") == 0)
    {
      previous_frame = img_in->get_current()->copy();
      tracking_force = new
	MovingEdgeDetector(previous_frame, img_in->get_current(),
			   feature_threshold, STATISTICAL_SM);
    }
  else if (strcmp(detection_method,"SIMPLE_EDGE") == 0)
    {
      tracking_force = new
	SimpleEdgeDetector(img_in->get_current(), feature_threshold,
			   STATISTICAL_SM);
    }
  else if (strcmp(detection_method,"SOBEL_EDGE") == 0)
    {
      tracking_force = new
	SobelDetector(img_in->get_current(), feature_threshold,
		      STATISTICAL_SM);
    }
  else if (strcmp(detection_method,"FOREGROUND_EDGE") == 0)
    {
      tracking_force = new
	ForegroundEdgeDetector(detect_motion->background->get_current(),
			       img_in->get_current(),
			       feature_threshold, feature_threshold, 
			       MAXIMUM_SM);
    }
  else if (strcmp(detection_method,"NORMALISED_COLOUR_FOREGROUND") == 0)
    {
      tracking_force = new
	NormForeColourDetect(detect_motion->background->get_current(),
			     img_in->get_current(),
			     feature_threshold,
			     feature_threshold);
    }
  if (tracking_force == NULL)
    {
      AmbErrorHandler::handle_warning("GroundPeopleTracker:: \
Unknown feature detector\n Using FOREGROUND_EDGE");
       tracking_force = new
	 ForegroundEdgeDetector(detect_motion->background->get_current(),
				img_in->get_current(),
				feature_threshold, feature_threshold, 
				MAXIMUM_SM);
    }

  
  // construct an active model object
  if (pca_file == NULL) pca_file = the_pca_file;
  local_optimiser = new ActiveModel(pca_file);
  local_optimiser->setup();
  local_optimiser->set_feature_detector(tracking_force);
  local_optimiser->set_video_image(img_in->get_current());

  // construct a ground plane occlusion handler

  occ_handler = new 
    WorldOcclusionHandler(img_in, detect_motion, 
			  profiles, tracking_data);
  local_optimiser->set_occlusion_handler(occ_handler);
}


void GroundPeopleTracker::predict_profiles()
{
  // update predictions

  // setup local_optimiser if not done yet
  if (local_optimiser->mean == NULL) local_optimiser->setup();

  TrackedProfile* curr_prf;

  // go through list
  for (profiles->start(); profiles->current_ok(); profiles->forward())
    {      
      if (profiles->get_current().get_type() == BASE_PROFILE)
	{
	  // not a TrackedProfile so convert
	  TrackedProfile* tracked =
	    new TrackedProfile(profiles->get_current(),
			       ++(start_ref_no.value));
	  profiles->insert_before(profiles->current, tracked);
	  profiles->current  = profiles->current->prev;
	  profiles->destroy(profiles->current->next);
	}
      curr_prf = (TrackedProfile*) &profiles->get_current();
      local_optimiser->set_profile(curr_prf);
      if (curr_prf->pos_filter == NULL) 
	AmbErrorHandler::handle_error("in \
GroundPeopleTracker::predict_profiles, can't predict unitialised\
profile");
      else 
	{
	  curr_prf->update_shape_filters();
	  local_optimiser->apply_virtual_shape();	  
	  curr_prf->gp_filter->next_frame();
	  occ_handler->constrain_motion(curr_prf);
	}     
      local_optimiser->b_to_x();
      curr_prf->gp_filter->setup_image_tracking();
      local_optimiser->b_to_x();
    }   
}

void GroundPeopleTracker::track_all()
{
  TrackedProfile* curr_prf;
  for (item(Profile)* curr = profiles->first; curr != NULL; )
    {
      curr_prf = (TrackedProfile*) curr->dat;
      local_optimiser->set_profile(curr_prf);
      //occ_handler->setup_occlusion_map(curr_prf);

      realno fitness = local_optimiser->track_profile();
      curr_prf->gp_filter->add_image_observations();

      occ_handler->constrain_motion(curr_prf);

      realno track_error = sqrt(curr_prf->gp_filter->gp_variance());
      curr_prf->frame_error = track_error;

      DB_PRINTF("positional variance %f\n", track_error);
      DB_PRINTF("observed %d out of unoccluded %d out of %d",\
		local_optimiser->no_observed,\
		local_optimiser->no_looks,\
		local_optimiser->no_sampled);	    
      DB_PRINTF("fitness %f \n", fitness);

      curr->dat = curr_prf;
      bool is_ok = TRUE;
      ++(curr_prf->time_alive);
      if ((curr_prf->ref_no == 0) && 
          (curr_prf->time_alive >= SMALL_NO_FRAMES))
        curr_prf->ref_no = ++(start_ref_no.value);

      // if track is fairly new and the fit is poor then kill it
      if ((curr_prf->time_alive < SMALL_NO_FRAMES) && 
          (fitness < FITNESS_THRESHOLD)) is_ok = FALSE;
      // if the positional uncertainty has grown too large, kill it
      if (track_error > terminate_thresh) is_ok = FALSE;

      // if the fit is really bad and a significant part of the
      // contour is observable (unoccluded), kill it
      if ((fitness < VERY_POOR_FIT) &&
	  (local_optimiser->get_significance() > MIN_SIGNIFICANCE)) 
	is_ok = FALSE;

      if (!is_ok)
	{
	  DB_PRINTF("Killing object %d \n", curr_prf->ref_no);
	  DB_PRINTF( "fit = %f\n", track_error);
	  DB_PRINTF( "observed %d out of %d\n",\
		     local_optimiser->no_observed,\
		     local_optimiser->no_looks);
	  //curr = curr->next;
	  item(Profile)* tmp = curr->next;
	  profiles->destroy(curr);
	  curr = tmp;
	}
      
      else curr = curr->next;
    }
}


void GroundPeopleTracker::track_new_objects()
{
  if (local_optimiser->mean == NULL) local_optimiser->setup();

  for (item(Profile)* curr = profiles->first; curr != NULL;)       
    {
      item(Profile)* next_node = curr->next;

      if (curr->dat->get_type() == BASE_PROFILE)
	{
	  TrackedProfile* new_prf1 = new
	    TrackedProfile(start_ref_no.value+1);

	  TrackedProfile* new_prf2 = new
	    TrackedProfile(start_ref_no.value+1);

	  GPfilter::split_profile(curr->dat, new_prf1, new_prf2);

	  local_optimiser->set_profile(new_prf1);
	  initialise_track(new_prf1);
	  if (occ_handler != NULL)
	    occ_handler->constrain_position(new_prf1);
	  local_optimiser->b_to_x();
	  new_prf1->gp_filter->setup_image_tracking();
	  realno fit1 = local_optimiser->track_profile();
	  realno err1 = sqrt(new_prf1->gp_filter->gp_variance());
	  if (local_optimiser->get_significance() < MIN_SIGNIFICANCE)
	    fit1 = 0.0;

	  local_optimiser->set_profile(new_prf2);
	  initialise_track(new_prf2);
	  if (occ_handler != NULL)
	    occ_handler->constrain_position(new_prf2);
	  local_optimiser->b_to_x();
	  new_prf2->gp_filter->setup_image_tracking();
	  realno fit2 = local_optimiser->track_profile();
	  realno err2 = sqrt(new_prf2->gp_filter->gp_variance());
	  if (local_optimiser->get_significance() < MIN_SIGNIFICANCE) 
	    fit2 = 0.0;

	  DB_PRINTF("fit1 %f fit2 %f \n", fit1, fit2);
	  if (fit1 > fit2)
	    delete (new_prf2);
	  else
	    {
	      delete (new_prf1);
	      new_prf1 = new_prf2;
	      err1 = err2;
	      fit1 = fit2;
	    }
	  
	  if ((err1 < terminate_thresh) && (fit1 > FITNESS_THRESHOLD))
	    {
	      delete (curr->dat);
	      curr->dat = new_prf1;
	      (start_ref_no.value)++;
	    }
	  else 
	    {
	      delete (new_prf1);
	      profiles->destroy(curr);
	    }
	      
	}
      curr = next_node;
    }
}


void GroundPeopleTracker::initialise_track(TrackedProfile* curr_prf)
{

  // setup everything that hasn't been initialised

  realno init_scale =  curr_prf->height /
    local_optimiser->mean->height;
  realno init_theta = 0.0;
  curr_prf->a_x = init_scale * cos(init_theta);
  curr_prf->a_y = init_scale * sin(init_theta);

  // IPCP - Image Plane Contact Point
  Point2 IPCP(curr_prf->origin - Point2(0, curr_prf->height / 2));
  
  curr_prf->gp_filter = new GPfilter(curr_prf, IPCP, curr_prf->height,
				     SQUARE(init_scale_sd), 
				     SQUARE(rel_scale_sd),
				     SQUARE(init_pos_sd),
				     SQUARE(acc_sd),
				     SQUARE(model_err_sd));
  // setup shape filters

  local_optimiser->set_profile(curr_prf);
  local_optimiser->setup_shape_filters();
  local_optimiser->b_to_x();

}

static Image* viz_image = NULL;

// tracking loop for each new image
void GroundPeopleTracker::MainLoop(int max_objects)
{
  ImageSource* vidsrc = detect_motion->input_frames;

  occ_handler->update_frame();
  predict_profiles();
  occ_handler->update_profiles();
  track_all();
  detect_motion->process_current_frame();
      
  
  bool detect_new = TRUE;

  if (profiles->no_items >= max_objects)  detect_new = FALSE;
  if (vidsrc != NULL)
    if ((vidsrc->get_frame_no() % TEMPORAL_SKIP) != 0)
      detect_new = FALSE;

  if (detect_new)
    {
      occ_handler->setup_motion_mask();
      ProfileSet tmp_set;
      detect_motion->extract_new_objects(&tmp_set);
      tmp_set.keep_ok();
      profiles->concat(tmp_set);
      track_new_objects();
    }

  if (previous_frame != NULL)
    vidsrc->get_current()->copy(previous_frame);

  // reset the visualisation image to NULL
  viz_image = NULL;
}

Image* GroundPeopleTracker::get_viz_image()
{
  if (viz_image == NULL) 
    viz_image = occ_handler->setup_viz_image();
  return (viz_image);
}

