/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/***************************************************************
 * C++ source
 *
 * File : 	MotionDetector.cc
 *
 * Module :	MotionDetector
 *
 * Author : 	A M Baumberg
 *
 * Creation Date : Mon Jun 13 13:02:58 1994 
 *
 * Comments : 	
 *
 ***************************************************************/


/* include files */

#include <MotionDetector.h>
#include <RegionSet.h>
#include <ProfileSet.h>
#include <utils.h>

realno MotionDetector::camera_moved_threshold = 
AmbGlobals.register_real("CAMERA_MOVED_THRESH", 0.5,
			 0.0, 1.0,
			 &MotionDetector::camera_moved_threshold,
			 TRUE, "MotionDetector", 
			 "If the number of pixels where there is \
movement over the total number of pixels \
is greater than this threshold, then the camera has moved.");

static int resample_shift =
AmbGlobals.register_int("MOTION_RESAMPLE_SHIFT", 2,
			&resample_shift,
			TRUE, "MotionDetector",
			"Resample image dimensions by 2^n \
in motion detection");

static int default_time_skip =
AmbGlobals.register_int("BACKGROUND_UPDATE_SKIP", 4,
			&default_time_skip, TRUE,
			"MotionDetector",
			"Temporal resampling for background filtering");

realno MotionDetector::diff_threshold = 
AmbGlobals.register_real("DETECT_DIFF_THRESH", 0.05,
			 0.0, 1.0, &MotionDetector::diff_threshold, 
			 FALSE, "MotionDetector",
			 "Differencing threshold for object detection");

static int default_window_length=
AmbGlobals.register_int("BACKGROUND_RUN_LENGTH", 50,
			&default_window_length, TRUE,
			"MotionDetector",
			"Medain window length (in frames) \
for background image update");

realno MotionDetector::max_region_ratio =
AmbGlobals.register_real("MAX_REGION_SIZE", 0.5,
			 &MotionDetector::max_region_ratio,
			 FALSE, "MotionDetector",
			 "Maximum size of a motion blob over \
size of image");

realno MotionDetector::min_region_ratio = 
AmbGlobals.register_real("MIN_REGION_SIZE", 0.001,
			 &MotionDetector::min_region_ratio,
			 FALSE, "MotionDetector",
			 "Minimum size of a motion blob over \
size of image");
 
			
realno MotionDetector::merge_threshold = 
AmbGlobals.register_real("REGION_MERGE_THRESH", 0.5,
			 &MotionDetector::merge_threshold,
			 FALSE, "MotionDetector",
			 "Distance threshold for merging blobs \
relative to size of blobs");

static bool default_merge_regions =
AmbGlobals.register_bool("DO_MERGE_REGIONS",
			 TRUE, &default_merge_regions,
			 TRUE, "MotionDetector",
			 "Merge motion blobs ?");

static bool default_do_blur = 
AmbGlobals.register_bool("BLUR_MOTION",
			 FALSE, &default_do_blur,
			 TRUE, "MotionDetector",
			 "Blur images in motion detection (slow) ?");
bool MotionDetector::show_images = 
AmbGlobals.register_bool("SHOW_MOTION_IMAGES",
			 FALSE, &MotionDetector::show_images,
			 FALSE, "MotionDetector",
			 "Show the subtracted motion-detection images ?");


MotionDetector::MotionDetector(ImageSource* in, Image* ref_img,
			       bool use_small_background)
  : input_frames(in)
{
  sample_skip = (1 << resample_shift);

  time_skip = default_time_skip;

  run_length = default_window_length / time_skip;

  do_blur = default_do_blur;
  merge_regions = default_merge_regions;

  if (use_small_background)
    {
      small_input = (PipeSource*) 
	new ResampleSource(in, sample_skip, sample_skip);
      
      Image* tmp_img = NULL;
      if (ref_img != NULL)
	tmp_img = ref_img->resample(sample_skip, sample_skip);
      
      background = new Medsrc(small_input, run_length, tmp_img);
      delete (tmp_img);

      small_background = (PipeSource*) background;
      
      large_background = NULL;
    }
  else
    {
      small_input = (PipeSource*) 
	new ResampleSource(in, sample_skip, sample_skip);

      background = new Medsrc(in, run_length, ref_img);
      large_background = (ImageSource*) background;
      small_background = (PipeSource*)
	new ResampleSource(large_background, sample_skip,
			   sample_skip);
    }

  if (do_blur)
    {
      blur1 = new BlurSource(small_input);
      blur2 = new BlurSource(small_background);
      small_differenced = new SimpleDiffSource(blur1, blur2,
					       256.0 * diff_threshold); 
    } 
  else 
    {
      small_differenced = new SimpleDiffSource(small_input, small_background,
					       256.0 * diff_threshold);
    }

  small_foreground = new NeighbourSource(small_differenced, 5);
  
  frame_count = 0;
  tracked_mask = NULL;
  small_differenced->set_title("Movement Detected");
  small_foreground->set_title("Thresholded new movement");
  have_differenced = FALSE;
  finished = FALSE;
}


MotionDetector::MotionDetector(ImageSource* in, int skip, int length,
			       realno threshold, realno max_ratio, 
			       realno min_ratio, Image* ref_img,
			       bool use_small_background,
			       bool merge_fragments, bool blur_images)
  : input_frames(in)
{
  do_blur = blur_images;
  merge_regions = merge_fragments;

  sample_skip = skip;
  time_skip = default_time_skip;
  run_length = length / time_skip;
  diff_threshold = threshold / 256.0;
  max_region_ratio = max_ratio;
  min_region_ratio = min_ratio;
  if (use_small_background)
    {
      small_input = (PipeSource*) 
	new ResampleSource(in, sample_skip, sample_skip);
      
      Image* tmp_img = NULL;
      if (ref_img != NULL)
	tmp_img = ref_img->resample(sample_skip, sample_skip);
      
      background = new Medsrc(small_input, run_length, tmp_img);
      delete (tmp_img);

      small_background = (PipeSource*) background;
      
      large_background = NULL;
    }
  else
    {
      small_input = (PipeSource*) 
	new ResampleSource(in, sample_skip, sample_skip);

      background = new Medsrc(in, run_length, ref_img);
      large_background = (ImageSource*) background;
      small_background = (PipeSource*)
	new ResampleSource(large_background, sample_skip,
			   sample_skip);
    }

  if (do_blur)
    {
      blur1 = new BlurSource(small_input);
      blur2 = new BlurSource(small_background);
      small_differenced = new SimpleDiffSource(blur1, blur2,
					       diff_threshold); 
    } 
  else 
    {
      small_differenced = new SimpleDiffSource(small_input, small_background,
					       diff_threshold);
    }

  small_foreground = new NeighbourSource(small_differenced, 5);
  
  frame_count = 0;
  tracked_mask = NULL;
  merge_regions = merge_fragments;
  small_differenced->set_title("Movement Detected");
  small_foreground->set_title("Thresholded new movement");
  have_differenced = FALSE;
  finished = FALSE;
}

void MotionDetector::process_next_frame()
{
  if ((frame_count % time_skip) == 0)  
    {
      finished = (small_background->get_next() == NULL);
      if (do_blur) blur2->refresh();
      if (large_background != NULL) small_input->refresh();
    }
  else finished = (small_input->get_next() == NULL);

  if (finished) return;

  if (do_blur) blur1->refresh();
  
  small_differenced->threshold = 256.0 * diff_threshold;
  small_differenced->refresh();
  small_foreground->refresh();
  if (show_images)
      small_foreground->display();

  if (frame_count++ == 0) 	// first frame ?
    {
      total_image_size = small_foreground->get_current()->get_width() *
	small_foreground->get_current()->get_height();
      if (isYUV(input_frames->get_current()->imgtype()))	// YUV image ?
	{
	  // setup Y, U, V, weights
	  for (int y = 0; y < 5; y++)
	    background->get_next();
	  recalculate_weights(background->get_input_source()->get_current(),
			      background->get_current());
	}
    }

  have_differenced = TRUE;
}

void MotionDetector::process_current_frame()
{
  if ((frame_count++ % time_skip) == 0)  
    {
      if (large_background != NULL)
	{
	  background->refresh();
	  small_background->refresh();
	  if (do_blur) blur2->refresh();	
	}
      else
	{
	  small_background->refresh();
	  if (do_blur) blur2->refresh();
	}
    }

  have_differenced = FALSE;
}

void MotionDetector::extract_new_regions(RegionSet* region_set,
					 bool grab_boxes)
{
  if (!have_differenced)
    {
      // Only do image differencing if looking for new objects !
      small_input->refresh();
      if (do_blur) blur1->refresh();
      small_differenced->threshold = 256.0 * diff_threshold;
      small_differenced->refresh();
      small_foreground->refresh();
    }

  // check for camera moved
  if (small_differenced->no_marked > 
      (camera_moved_threshold * total_image_size))
    {
      background->restart();
      return;
    }

  // mask out tracked objects
  small_foreground->get_current()->mask(tracked_mask,
				  small_foreground->get_current());
  if (show_images)
    small_foreground->display();

  RegionSet* res = region_set;
  
  int max_region_size = total_image_size * max_region_ratio;
  int min_region_size = total_image_size * min_region_ratio;

  int min_size = (merge_regions) ? 2 : min_region_size;
  if (grab_boxes)
    res->grab_all_regions(small_foreground->get_current(),
			  min_size, max_region_size);
  else
    res->grab_regions(small_foreground->get_current(),
		      min_size, max_region_size);

  if (merge_regions)
    {
      res->merge_regions(merge_threshold);
      res->filter(min_region_size);
    }
}

void MotionDetector::extract_new_objects(ProfileSet* profile_set)
{
  RegionSet* res = new RegionSet();
  extract_new_regions(res);
  ProfileSet new_profiles = res->to_profiles();
  new_profiles.scale_coords(sample_skip);
  profile_set->concat(new_profiles);

//   for (item(Region)* curr = res->first; curr != NULL; curr =
// 	 curr->next)
//     profile_set->add(new Profile(curr->dat->ox * sample_skip,
// 				 curr->dat->oy * sample_skip,
// 				 curr->dat->width * sample_skip,
// 				 curr->dat->height * sample_skip));

  res->destroy_all();
  delete (res);
}


void MotionDetector::draw_old_objects(ProfileSet* old_objects)
{
  if (tracked_mask == NULL)
    {
      tracked_mask = new
	Gray8Image(small_foreground->get_current()->get_width(), 
		   small_foreground->get_current()->get_height());

      // background->motion_mask = tracked_mask;
    }

  tracked_mask->clear(255);
  for (item(Profile)* curr = old_objects->first; curr != NULL;
       curr = curr->next)
    {
      int xmin, xmax, ymin, ymax;
      curr->dat->get_enclosing_box(xmin,xmax, ymin, ymax);
      xmin /= sample_skip;
      xmax /= sample_skip;
      ymin /= sample_skip;
      ymax /= sample_skip;
      tracked_mask->draw_rectangle(xmin - 1, xmax + 1, 
				     ymin - 1, ymax + 1, 0);
    }
}
      



