/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/***************************************************************
 * C++ source
 *
 * File : 	ProActiveModel.cc
 *
 * Module :	ProActiveModel
 *
 * Author : 	A M Baumberg (CoMIR)
 *
 * Creation Date : Mon Aug 28 16:40:07 1995 
 *
 * Comments : 	
 *
 ***************************************************************/


/* include files */

#include "ProActiveModel.h"
#include <NagMatrix.h>
#include <KalmanFilter.h>
#include <unistd.h>
#include <stdio.h>
#include <MeanProfile.h>
#include <utils.h>

/* global functions */

bool ProActiveModel::overwrite_damping = 
AmbGlobals.register_bool("OVERWRITE_DAMPING", TRUE,
			 &ProActiveModel::overwrite_damping,
			 TRUE,"ProActiveModel",
			 "Don't use learned damping parameters");

realno ProActiveModel::damping_b0 = 
AmbGlobals.register_real("DAMPING_B0", 0.0,
			 &ProActiveModel::damping_b0,
			 TRUE, "ProActiveModel",
			 "Damping parameter b0 for dyanmic tracker");

realno ProActiveModel::damping_b1 = 
AmbGlobals.register_real("DAMPING_B1", 1.0,
			 &ProActiveModel::damping_b1,
			 TRUE, "ProActiveModel",
			 "Damping parameter b1 for dynamic tracker");



void ProActiveModel::read_shape_model()
{
  ifstream data_in(model_file);
  NagMatrix physical_modes;
  int data_size, no_dynamic_modes;
  NagMatrix Cov;
  char dummy[50];
  data_in >> NO_CONTROL_POINTS >> dummy >> dummy >> dummy;
  if (strcmp(dummy, "Dynamic_Model") == 0)
    {
      data_in >> dummy;
      mean = new MeanProfile();
      data_in >> *mean; 
      data_in >> dummy >> data_size >> dummy >> no_dynamic_modes;
      data_in >> dummy >> a_vals >> dummy >> b_vals;

      if (overwrite_damping)
	{
	  cerr << "Using damping parameters " << endl;
	  cerr << "b0 " << damping_b0 << " b1 " << damping_b1 << endl;
	  b_vals.clear(damping_b0);
	  NagVector tmp_b;
	  a_vals.scale(damping_b1, tmp_b);
	  b_vals.add(tmp_b, b_vals);
	}
      if (model_depth > a_vals.get_size()) model_depth =
	a_vals.get_size();
      data_in >> dummy;
      if (strcmp(dummy,"r_vals") == 0)
	{
	  data_in >> r_vals;
	  data_in >> dummy;
	}
      data_in >> physical_modes;
      data_in >> dummy ;
      if (strcmp(dummy,"eigenvalues") == 0)
	data_in >> evals;
      else if (strcmp(dummy,"Covariance") == 0)
	{
	  int no_modes = a_vals.get_size();
	  data_in >> Cov;
	  NagMatrix tmp_sqrt_Cov;
	  NagMatrix tmp_sqrt_inv_Cov;
	  Cov.msqrt(tmp_sqrt_Cov);
	  tmp_sqrt_Cov.invert(tmp_sqrt_inv_Cov);
	  tmp_sqrt_inv_Cov.get_block(0,no_modes-1,
				     0,model_depth-1, sqrt_inv_Cov);
	  tmp_sqrt_Cov.get_block(0,model_depth-1,
				 0,no_modes-1, sqrt_Cov);
	  
	}
    }
  else
    { fprintf(stderr,"NOT a ST model \n");
      data_in.close();
//      return (this->ActiveModel::read_shape_model());      
       this->ActiveModel::read_shape_model();      
	return;
    }

  P.reconstruct(data_size, model_depth);
  H.reconstruct(data_size * no_subdivisions, model_depth);
  N.reconstruct(data_size * no_subdivisions, model_depth);
  mean_points.reconstruct(data_size * no_subdivisions);
  mean_normals.reconstruct(data_size * no_subdivisions);

  // setup model to spline point transformation

  physical_modes.get_block(0, data_size-1, 0, model_depth -1, P);

  if (evals.get_data() == NULL)
    {
      evals.reconstruct(model_depth);
      fprintf(stderr,
	      " no eigenvalues -- generating from frequencies\n");
      for (int j = model_depth -1; j >= 0; j--)
	{
	  if (a_vals[j] != 0) evals[j] = -1.0 / a_vals[j];
	  else evals[j] = evals[j+1];
	}
    }  
}

void ProActiveModel::setup_shape_filters()
{
  if (a_vals.get_data() == NULL)
	{
    // return (this->ActiveModel::setup_shape_filters());
     		this->ActiveModel::setup_shape_filters();
		return;
	}

 for (int i = 0; i < model_depth; i++)
    {
      realno noise_term = noise_scalar * evals[i];
      if (r_vals.get_data() != NULL)
	noise_term = noise_scalar * r_vals[i];
      //if (a_vals[i] != 0)
      if (TRUE)
	{
	  currx->filters[i] = (FilterOneD*) new
	    VibrationFilter(0,0, noise_term,	
			    a_vals[i], b_vals[i],
			    4.0, 0, 0,
			    initial_noise * evals[i], 0,
			    0, 1.0);
	  ((VibrationFilter*)currx->filters[i])->get_prediction(1.0);
	}
      else
	{
	  currx->filters[i] = (FilterOneD*) new
	    StaticKalmanOneD(noise_term
			  ,4			// r^2  
			  ,0 			// b0 	
			  ,initial_noise * evals[i]		// p11 	
			  ,epsilon);
	}
    }

}

void ProActiveModel::apply_virtual_shape()
{
  if (a_vals.get_data() == NULL)
	{
    //return (this->ActiveModel::apply_virtual_shape());
    		this->ActiveModel::apply_virtual_shape();
		return;
	}
  
  for (int i = 0; i < model_depth; i++)
    {
      // add a virtual observation 
      //currx->filters[i]->add_observation(0, 200.0 * noise_scalar * evals[i]);
    }
}


void ProActiveModel::constrain_b()
{
  if (sqrt_Cov.get_data() == NULL) 
	{
	// return (this->ActiveModel::constrain_b());
		this->ActiveModel::constrain_b();
		return;
	}

  NagVector b(model_depth);
  for (int i = 0; i < model_depth; i++)
    b[i] = currx->filters[i]->get_state();

  NagVector new_b;
  sqrt_inv_Cov.multiply(b, new_b);
  realno blength2 = new_b.length2();
  if (blength2 >= cloud_size)
    {
      realno mfac = sqrt(cloud_size / blength2);
      new_b.scale(mfac, new_b);
      sqrt_Cov.multiply(new_b, b);
      for (i = 0; i < model_depth; i++)
	currx->filters[i]->get_state() = b[i];
    }
}  

