/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/***************************************************************
 * C++ source
 *
 * File : 	SimplePeopleTracker.cc
 *
 * Module :	SimplePeopleTracker
 *
 * Author : 	A M Baumberg (CoMIR)
 *
 * Creation Date : Tue Oct 22 15:06:09 1996 
 *
 * Comments : 	High level people tracker without ground plane info
 *
 ***************************************************************/


/* include files */

#include "SimplePeopleTracker.h"
#include <KalmanFilter.h>
#include <utils.h>
#include <FeatureDetector.h>
#include <imgsrc_utils.h>

static EnvIntParameter start_ref_no("START_REF_NO", 0);

realno SimplePeopleTracker::FITNESS_THRESHOLD  = 
AmbGlobals.register_real("START_FITNESS_THRESHOLD", 0.85,
			 0.0, 1.0, 
			 &SimplePeopleTracker::FITNESS_THRESHOLD,
			 FALSE, "SimplePeopleTracker",
			 "Minimum fitness score to initialise track");

realno SimplePeopleTracker::VERY_POOR_FIT =
AmbGlobals.register_real("TRACK_FITNESS_THRESHOLD", 0.4,
			 0.0, 1.0,
			 &SimplePeopleTracker::VERY_POOR_FIT,
			 FALSE, "SimplePeopleTracker",
			 "Minimum fitness score to maintain track");

int SimplePeopleTracker::SMALL_NO_FRAMES =
AmbGlobals.register_int("SMALL_NO_FRAMES", 3,
			&SimplePeopleTracker::SMALL_NO_FRAMES,
			FALSE, "SimplePeopleTracker",
			"Number of frames before track is accepted");

			
realno SimplePeopleTracker::MIN_SIGNIFICANCE = 
AmbGlobals.register_real("MIN_SIGNIFICANCE", 0.3,
			 0.0, 1.0, 
			 &SimplePeopleTracker::MIN_SIGNIFICANCE,
			 FALSE, "SimplePeopleTracker",
			 "Fraction of curve that must \
be visible for tracking");

realno SimplePeopleTracker::terminate_thresh =
AmbGlobals.register_real("TERMINATION_THRESHOLD", 20,
			 &SimplePeopleTracker::terminate_thresh,
			 FALSE, "SimplePeopleTracker",
			 "Maximum std deviation of image plane position");

realno SimplePeopleTracker::init_pos_sd =
AmbGlobals.register_real("INIT_POS_SD", 0.1,
			 &SimplePeopleTracker::init_pos_sd,
			 FALSE, "SimplePeopleTracker",
			 "Initial standard deviation of position\
 relative to object height");

realno SimplePeopleTracker::init_vel_sd =
AmbGlobals.register_real("INIT_VEL_SD", 0.1,
			 &SimplePeopleTracker::init_vel_sd,
			 FALSE, "SimplePeopleTracker",
			 "Initial standard deviation of velocity\
 relative to object height");

realno SimplePeopleTracker::noise_vel_sd =
AmbGlobals.register_real("REL_VEL_SD", 0.1,
			 &SimplePeopleTracker::noise_vel_sd,
			 FALSE, "SimplePeopleTracker",
			 "standard deviation of random velocity\
 term relative to object height");

realno SimplePeopleTracker::noise_acc_sd =
AmbGlobals.register_real("REL_ACC_SD", 0.1,
			 &SimplePeopleTracker::noise_acc_sd,
			 FALSE, "SimplePeopleTracker",
			 "standard deviation of random acceleration\
 term relative to object height");


realno SimplePeopleTracker::init_scale_sd =
AmbGlobals.register_real("INIT_SCALE_SD", 0.1,
			 &SimplePeopleTracker::init_scale_sd,
			 FALSE, "SimplePeopleTracker",
			 "Standard deviation of initial estimate of\
 object size relative to initial estimate");

realno SimplePeopleTracker::noise_scale_sd =
AmbGlobals.register_real("REL_SCALE_SD", 0.05,
			 &SimplePeopleTracker::noise_scale_sd,
			 FALSE, "SimplePeopleTracker",
			 "Standard deviation of scale noise term \
relative to initial scale");


realno SimplePeopleTracker::init_theta_sd =
AmbGlobals.register_real("INIT_THETA_SD", 0.015,
			 &SimplePeopleTracker::init_theta_sd,
			 FALSE, "SimplePeopleTracker",
			 "Standard deviation of initial estimate of\
 rotation angle");

realno SimplePeopleTracker::noise_theta_sd =
AmbGlobals.register_real("NOISE_THETA_SD", 0.0,
			 &SimplePeopleTracker::noise_theta_sd,
			 FALSE, "SimplePeopleTracker",
			 "Standard deviation of theta noise term");


int SimplePeopleTracker::TEMPORAL_SKIP = 
AmbGlobals.register_int("FRAME_SKIP", 4,
			 &SimplePeopleTracker::TEMPORAL_SKIP,
			 TRUE, "SimplePeopleTracker",
			 "Temporal resampling for object detection");

static char* detect_methods[] =
{ "COLOUR_FOREGROUND", "COLOUR_EDGE",
  "MOVING_EDGE", "SIMPLE_EDGE",
  "SOBEL_EDGE", "FOREGROUND_EDGE", 
  "NORMALISED_COLOUR_FOREGROUND", NULL };

static char* detection_method = 
AmbGlobals.register_string("FEATURE_DETECTION", "COLOUR_FOREGROUND",
			   &detection_method, TRUE,
			   "SimplePeopleTracker",
			   "The type of features used for tracking",
			   detect_methods);

static realno feature_threshold = 
AmbGlobals.register_real("FEATURE_THRESH", 0.05,
			 0.0, 1.0, &feature_threshold,
			 TRUE, "SimplePeopleTracker",
			 "A significance threshold for feauture detection");

static char* the_pca_file = 
AmbGlobals.register_string("PCA_MODEL_FILE",
			   NULL, &the_pca_file, TRUE,
			   "SimplePeopleTracker",
			   "An optional eigenshape model file. If \
not set, the default model is used");


SimplePeopleTracker::SimplePeopleTracker(OcclusionImage* ohandler,
					 ActiveModel* loptimise,
					 MotionDetector* mdetect,
					 ProfileSet* curr_profiles) 
 
{
  if (curr_profiles == NULL) curr_profiles = new ProfileSet;
  profiles = curr_profiles;
  occ_handler = ohandler;
  local_optimiser = loptimise;
  detect_motion = mdetect;
}

SimplePeopleTracker::SimplePeopleTracker(ImageSource* img_in, 
					 char* pca_file,
					 ProfileSet* profiles_out,
					 FeatureDetector* features)
{
  if (profiles_out == NULL) profiles_out = new ProfileSet;
  profiles = profiles_out;
  previous_frame = NULL;
  FeatureDetector* tracking_force = NULL;
  bool use_background_image = FALSE;
  // do we need a full size background image for feature detection?
  if (strstr(detection_method, "FOREGROUND")) 
    use_background_image = TRUE;

  // construct motion detection object
  detect_motion = new MotionDetector(img_in, get_reference_frame(),
				     !use_background_image);

  detect_motion->process_next_frame();
  if (strstr(detection_method, "COLOUR") && 
      (img_in->get_current()->imgtype() != RGB32))
    {
      AmbErrorHandler::handle_warning("ignoring colour feature \
detection");
      detection_method = "FOREGROUND_EDGE";
    }

  // construct a feature detection object
  if (features != NULL) tracking_force = features;
  else  if (strcmp(detection_method, "COLOUR_FOREGROUND") == 0)
    {
      tracking_force = new 
	ColouredForegroundEdgeDetector(detect_motion->background->get_current()
				       ,img_in->get_current(),
				       feature_threshold,
				       feature_threshold);
    }
  else if (strcmp(detection_method,"MOVING_EDGE") == 0)
    {
      previous_frame = img_in->get_current()->copy();
      tracking_force = new
	MovingEdgeDetector(previous_frame, img_in->get_current(),
			   feature_threshold, STATISTICAL_SM);
    }
  else if (strcmp(detection_method,"SIMPLE_EDGE") == 0)
    {
      tracking_force = new
	SimpleEdgeDetector(img_in->get_current(), feature_threshold,
			   STATISTICAL_SM);
    }
  else if (strcmp(detection_method,"SOBEL_EDGE") == 0)
    {
      tracking_force = new
	SobelDetector(img_in->get_current(), feature_threshold,
		      STATISTICAL_SM);
    }
  else if (strcmp(detection_method,"FOREGROUND_EDGE") == 0)
    {
      tracking_force = new
	ForegroundEdgeDetector(detect_motion->background->get_current(),
			       img_in->get_current(),
			       feature_threshold, feature_threshold, 
			       MAXIMUM_SM);
    }
  else if (strcmp(detection_method,"NORMALISED_COLOUR_FOREGROUND") == 0)
    {
      tracking_force = new
	NormForeColourDetect(detect_motion->background->get_current(),
			     img_in->get_current(),
			     feature_threshold,
			     feature_threshold);
    }
  if (tracking_force == NULL)
    {
      AmbErrorHandler::handle_warning("GroundPeopleTracker:: \
Unknown feature detector\n Using FOREGROUND_EDGE");
       tracking_force = new
	 ForegroundEdgeDetector(detect_motion->background->get_current(),
				img_in->get_current(),
				feature_threshold, feature_threshold, 
				MAXIMUM_SM);
    }


 // construct an active model object
  if (pca_file == NULL) pca_file = the_pca_file;
  local_optimiser = new ActiveModel(pca_file);
  local_optimiser->setup();
  local_optimiser->set_feature_detector(tracking_force);
  local_optimiser->set_video_image(img_in->get_current());

  // construct a 2D occlusion handler
  occ_handler = 
    new OcclusionImage(img_in->get_current()->get_width(),
		       img_in->get_current()->get_height());


  local_optimiser->set_occlusion_handler(occ_handler);
}

  
void SimplePeopleTracker::predict_profiles()
{
  // update Kalman predictions
  TrackedProfile* curr_prf;
  if (local_optimiser->mean == NULL) local_optimiser->setup();
  occ_handler->clear(0);
  for (profiles->start(); profiles->current_ok(); profiles->forward())
    {      
      if (profiles->get_current().get_type() == BASE_PROFILE)
	{
	  TrackedProfile* tracked =
	    new TrackedProfile(profiles->get_current(),
			       ++(start_ref_no.value));
	  profiles->insert_before(profiles->current, tracked);
	  profiles->current  = profiles->current->prev;
	  profiles->destroy(profiles->current->next);
	}
      curr_prf = (TrackedProfile*) &profiles->get_current();
      local_optimiser->set_profile(curr_prf);
      if (curr_prf->pos_filter == NULL) 
	AmbErrorHandler::handle_error("in \
SimplePeopleTracker::predict_profiles, can't predict unitialised\
profile");
      else 
	{
	  curr_prf->update_shape_filters();
	  local_optimiser->apply_virtual_shape();
	  
	  DynamicKalmanTwoD* origin_filter =
	    (DynamicKalmanTwoD*) curr_prf->pos_filter;
	  
 	  origin_filter->
	    reset_noise_covariance(SQUARE(noise_vel_sd * curr_prf->height),
				   SQUARE(noise_vel_sd * curr_prf->height),
				   SQUARE(noise_acc_sd * curr_prf->height),
				   SQUARE(noise_acc_sd * curr_prf->height));

	  curr_prf->origin = origin_filter->get_prediction();
	  curr_prf->a_filter->set_polar_noise_cov(SQUARE(noise_scale_sd),
						  SQUARE(noise_theta_sd));
	  curr_prf->a_filter->update_state();

	}     
      local_optimiser->b_to_x();
      if ((++(curr_prf->time_alive)) >= SMALL_NO_FRAMES)
 	occ_handler->mark_rect(curr_prf);
    }   
#ifdef DEBUG
  if (!AmbErrorHandler::is_quiet_mode())
    {
      occ_handler->display();
    }
#endif
}


void SimplePeopleTracker::track_all()
{
  TrackedProfile* curr_prf;
  for (item(Profile)* curr = profiles->first; curr != NULL; )
    {
      curr_prf = (TrackedProfile*) curr->dat;
      local_optimiser->set_profile(curr_prf);
      realno fitness = local_optimiser->track_profile();

      realno track_error =
	sqrt(curr_prf->pos_filter->get_uncertainty());
      curr_prf->frame_error = track_error;
    
   
      DB_PRINTF("positional variance %f\n", track_error);
      DB_PRINTF("observed %d out of unoccluded %d out of %d",\
		local_optimiser->no_observed,\
		local_optimiser->no_looks,\
		local_optimiser->no_sampled);	    
      DB_PRINTF("fitness %f \n", fitness);

      curr->dat = curr_prf;

      bool is_ok = TRUE;
      ++(curr_prf->time_alive);
      if ((curr_prf->ref_no == 0) && 
          (curr_prf->time_alive >= SMALL_NO_FRAMES))
        curr_prf->ref_no = ++(start_ref_no.value);

      // if track is fairly new and the fit is poor then kill it
      if ((curr_prf->time_alive < SMALL_NO_FRAMES) && 
          (fitness < FITNESS_THRESHOLD)) is_ok = FALSE;
      // if the positional uncertainty has grown too large, kill it
      if (track_error > terminate_thresh) is_ok = FALSE;

      // if the fit is really bad and a significant part of the
      // contour is observable (unoccluded), kill it
      if ((fitness < VERY_POOR_FIT) &&
	  (local_optimiser->get_significance() > MIN_SIGNIFICANCE)) 
	is_ok = FALSE;

      if (!is_ok)
	{
	  DB_PRINTF("Killing object %d \n", curr_prf->ref_no);
	  DB_PRINTF( "fit = %f\n", track_error);
	  DB_PRINTF( "observed %d out of %d\n",\
		     local_optimiser->no_observed,\
		     local_optimiser->no_looks);
	  //curr = curr->next;
	  item(Profile)* tmp = curr->next;
	  profiles->destroy(curr);
	  curr = tmp;
	}
      
      else curr = curr->next;
    }
}

void SimplePeopleTracker::track_new_objects()
{

  if (local_optimiser->mean == NULL) local_optimiser->setup();
  for (item(Profile)* curr = profiles->first; curr != NULL;)
       
    {
      item(Profile)* next_node = curr->next;

      if (curr->dat->get_type() == BASE_PROFILE)
	{
	  TrackedProfile* tracked = 
	    new TrackedProfile(*curr->dat,start_ref_no.value+1);
	  local_optimiser->set_profile(tracked);
	  initialise_track(tracked); // get model hypothesis
	  local_optimiser->b_to_x();
	  realno fitness = local_optimiser->track_profile();
	  realno track_error = 
	    sqrt(tracked->pos_filter->get_uncertainty());

	  if ((track_error < terminate_thresh) && 
	      (fitness > FITNESS_THRESHOLD))
	    {
	      delete (curr->dat);
	      curr->dat = tracked;
	      (start_ref_no.value)++;
	    }
	  else 
	    {
	      delete (tracked);
	      profiles->destroy(curr);
	    }
	      
	}
      curr = next_node;
    }
}


void SimplePeopleTracker::initialise_track(TrackedProfile* curr_prf)
{
  realno init_scale =  curr_prf->height / 
    local_optimiser->mean->height;
  realno init_theta = 0.0;
  curr_prf->a_x = init_scale * cos(init_theta);
  curr_prf->a_y = init_scale * sin(init_theta);
  local_optimiser->set_profile(curr_prf);
  local_optimiser->setup_shape_filters();
  local_optimiser->b_to_x();
  curr_prf->get_size();

  curr_prf->pos_filter = 
    new DynamicKalmanTwoD(SQUARE(noise_vel_sd * curr_prf->height), 
		       SQUARE(noise_vel_sd * curr_prf->height),
		       SQUARE(noise_acc_sd * curr_prf->height),
		       SQUARE(noise_acc_sd * curr_prf->height),
		       curr_prf->origin, Point2(0,0),
		       SQUARE(init_pos_sd * curr_prf->height),
		       SQUARE(init_pos_sd * curr_prf->height),
		       SQUARE(init_vel_sd * curr_prf->height),
		       SQUARE(init_vel_sd * curr_prf->height), 1.0);
		       

  
  curr_prf->a_filter = new StaticKalmanTwoD(0, 0, 0, curr_prf->a_x, 
				      curr_prf->a_y, 0, 0, 0);
  curr_prf->a_filter->set_polar_noise_cov(SQUARE(init_scale_sd), 
					  SQUARE(init_theta_sd));
  curr_prf->a_filter->update_state();
  curr_prf->a_filter->set_polar_noise_cov(SQUARE(noise_scale_sd),
					  SQUARE(noise_theta_sd));

}

// tracking loop for each new image
void SimplePeopleTracker::MainLoop(int max_objects)
{
  ImageSource* vidsrc = detect_motion->input_frames;

  predict_profiles();
  track_all();
  detect_motion->process_current_frame();        
  bool detect_new = TRUE;

  if (profiles->no_items >= max_objects)  detect_new = FALSE;
  if (vidsrc != NULL)
    if ((vidsrc->get_frame_no() % TEMPORAL_SKIP) != 0)
      detect_new = FALSE;

  if (detect_new)
    {
      detect_motion->draw_old_objects(profiles);
      ProfileSet tmp_set;
      detect_motion->extract_new_objects(&tmp_set);
      tmp_set.keep_ok();
      profiles->concat(tmp_set);
      track_new_objects();
    }
  if (previous_frame != NULL)
    vidsrc->get_current()->copy(previous_frame);
  
}

	      

