/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/***************************************************************
 * C++ source
 *
 * File : 	TrackedObject.cc
 *
 * Module :	TrackedObject
 *
 * Author : 	A M Baumberg (CoMIR)
 *
 * Creation Date : Thu Aug  8 17:59:28 1996 
 *
 * Comments : 	
 *
 ***************************************************************/


/* include files */

#include "TrackedObject.h"
#include <Matrix2by2.h>
#include <utils.h>

#define GPtoIP GroundPlaneToImagePlane
#define IPtoGP ImagePlaneToGroundPlane

#ifndef ZSGN
#define ZSGN(a)		(((a)<0) ? -1 : (a)>0 ? 1 : 0)	
#endif
#ifndef ABS
#define ABS(a)		(((a)<0) ? -(a) : (a))
#endif

static const int MEMORY_LENGTH = 20;	// window size for averaging
					// velocities

static const realno myEPS = 1e-30;

static const realno SMALL_FRACTION = 0.01;
static const realno LARGE_FRACTION = 0.02;
static const realno EPS_TOLERANCE = 0.001;


TrackedObject::TrackedObject(Point2 p1, Point2 p2, Point2 p3, Point2 p4,
			     int the_id, char ch_id, ObjectType the_type)
{
  GroundPlaneBB[0] = p1;
  GroundPlaneBB[1] = p2;
  GroundPlaneBB[2] = p3;
  GroundPlaneBB[3] = p4;

  GroundPlaneOrigin = 0.25 * (p1 + p2 + p3 + p4);
  if (GPtoIP.get_data() == NULL)
    setup_default_projection();

  depth = GPtoIP.read(2,0) * GroundPlaneOrigin.x + 
    GPtoIP.read(2,1) * GroundPlaneOrigin.y + GPtoIP.read(2,2);

  is_stationary = FALSE;
  obj_id = the_id;
  id_char = ch_id;

  type = the_type;
  no_active_frames = 0;
  no_stationary_frames = 0;

  BB_velocity[0] = BB_velocity[1] = BB_velocity[2] = BB_velocity[3]
    = velocity = Point2(0,0);
}

TrackedObject::TrackedObject(Point2 pos, int the_id, char ch_id,
			     ObjectType the_type)
{
  GroundPlaneBB[0] = pos + Point2(-20,20);
  GroundPlaneBB[1] = pos + Point2(20,20);
  GroundPlaneBB[2] = pos + Point2(20,-20);
  GroundPlaneBB[3] = pos + Point2(-20,-20);

  GroundPlaneOrigin = pos;
  if (GPtoIP.get_data() == NULL)
    setup_default_projection();

  depth = GPtoIP.read(2,0) * GroundPlaneOrigin.x + 
    GPtoIP.read(2,1) * GroundPlaneOrigin.y + GPtoIP.read(2,2);

  is_stationary = FALSE;
  obj_id = the_id;
  id_char = ch_id;

  type = the_type;
  no_active_frames = 0;
  no_stationary_frames = 0;

  BB_velocity[0] = BB_velocity[1] = BB_velocity[2] = BB_velocity[3]
    = velocity = Point2(0,0);
}

  
bool TrackedObject::is_point_hidden(Point2& p)
{
  // does the GP bounding box occlude point p
  // (doesn't necessarily mean there's image plane occlusion)

  if (type == TrackedObject::PERSON)
    {
      // just return depth ordering
      Point2 zbasis(GPtoIP.read(2,0), GPtoIP.read(2,1));
      return ((zbasis * p) > (zbasis * GroundPlaneOrigin));
    }

  Point2 zbasis = Point2(GPtoIP.read(2,0), GPtoIP.read(2,1));
  Point2 camera_pos;
  if (get_GP_camera_pos(camera_pos))
    zbasis = p - camera_pos;

  realno p_z = zbasis * p - EPS_TOLERANCE; // depth + constant at p
  Point2 zline(zbasis.y, -zbasis.x);
  
  for (int i = 0; i < 4; i++)
    {
      int j = (i+1) % 4;
      
      realno sect_ij =  (zline * (p - GroundPlaneBB[i])) /
	(zline * (GroundPlaneBB[j] - GroundPlaneBB[i]));

      if ((sect_ij >= 0) && (sect_ij <= 1))
	{
	  Point2 sect = (sect_ij * GroundPlaneBB[j] + 
			 (1.0 - sect_ij) * GroundPlaneBB[i]);
	  // is the intersection point in from of p
	  if ((zbasis * sect) < p_z) return (TRUE);
	}
    }
	
  return (FALSE);
}
  
void TrackedObject::update(TrackedObject& prev_state)
{
  //if (id != prev_state.id) return;

  no_active_frames = prev_state.no_active_frames + 1;

  if (type == PERSON)  // people are never considered stationary
    return;

  int n = prev_state.no_active_frames;
  if (n >= MEMORY_LENGTH) n = MEMORY_LENGTH;
  
  velocity = 
    ((n * prev_state.velocity) + 
     (GroundPlaneOrigin - prev_state.GroundPlaneOrigin)) / (n+1.0);
  
  for (int i = 0; i < 4; i++)
    BB_velocity[i] =
      ((n * prev_state.BB_velocity[i]) + 
       (GroundPlaneBB[i] - prev_state.GroundPlaneBB[i])) / (n+1.0);
  
  // test if stationary

  realno sq_threshold = sq_size() * SMALL_FRACTION *
    SMALL_FRACTION;
  realno large_threshold = sq_size() * (LARGE_FRACTION * LARGE_FRACTION);
  
  if (((GroundPlaneOrigin - prev_state.GroundPlaneOrigin).length2() 
       < large_threshold) &&
      ((no_active_frames >= MEMORY_LENGTH) &&
       (BB_velocity[0].length2() < sq_threshold) &&
       (BB_velocity[1].length2() < sq_threshold) &&
       (BB_velocity[2].length2() < sq_threshold) &&
       (BB_velocity[3].length2() < sq_threshold)))
    {
      // the velocities have had time to average
      // AND they are all relatively small compared to the
      // object size => the CAR is stopped

      is_stationary = TRUE;
      no_stationary_frames = prev_state.no_stationary_frames + 1;
      DB_PRINTF("Car %d is stopped \n", obj_id);
    }
  else
    {
      is_stationary = FALSE;
      no_stationary_frames = 0;
    }
}

ostream& operator<<(ostream& strm_out, TrackedObject& obj)
{
  strm_out << "label " << obj.obj_id << 
    obj.GroundPlaneBB[0].x << " " << obj.GroundPlaneBB[0].y << " " <<
    obj.GroundPlaneBB[1].x << " " << obj.GroundPlaneBB[1].y << " " <<
    obj.GroundPlaneBB[2].x << " " << obj.GroundPlaneBB[2].y << " " <<
    obj.GroundPlaneBB[3].x << " " << obj.GroundPlaneBB[3].y << endl;
  return (strm_out);
}
 

istream& operator>>(istream& strm_in, TrackedObject& obj)
{
  char dummy[20];
  strm_in >> dummy >> obj.obj_id >>
    obj.GroundPlaneBB[0].x >> obj.GroundPlaneBB[0].y >>
    obj.GroundPlaneBB[1].x >> obj.GroundPlaneBB[1].y >>
    obj.GroundPlaneBB[2].x >> obj.GroundPlaneBB[2].y >>
    obj.GroundPlaneBB[3].x >> obj.GroundPlaneBB[3].y;
  return (strm_in);
}

    
bool TrackedObject::occludes(TrackedObject& obj2)
{
  if (this->type == TrackedObject::CAR)
    return (is_point_hidden(obj2.GroundPlaneOrigin));

  // otherwise
  if (obj2.type == TrackedObject::CAR)
    return (!obj2.is_point_hidden(GroundPlaneOrigin));

  // both objects are people

  return (this->depth < obj2.depth);

}

// encloses:
// return TRUE if the bounding box encloses the point
bool TrackedObject::encloses(Point2 pnt)
{
  if (this->type == TrackedObject::PERSON) return (FALSE);

  realno side01 = (pnt - GroundPlaneBB[0]) ^ 
    (GroundPlaneBB[1] - GroundPlaneBB[0]);
  realno side12 = (pnt - GroundPlaneBB[1]) ^ 
    (GroundPlaneBB[2] - GroundPlaneBB[1]);
  realno side23 = (pnt - GroundPlaneBB[2]) ^ 
    (GroundPlaneBB[3] - GroundPlaneBB[2]);
  realno side30 = (pnt - GroundPlaneBB[3]) ^ 
    (GroundPlaneBB[0] - GroundPlaneBB[3]);
  
  int sum_sgn = ZSGN(side01) + ZSGN(side12) + ZSGN(side23) +
    ZSGN(side30);
  if (ABS(sum_sgn) == 4) return (TRUE);

  return (FALSE);
}

Point2 TrackedObject::nearest_boundary_point(Point2 pnt)
{
  // return the nearest point on the bounding box lines
  Point2 vec01 = GroundPlaneBB[1] - GroundPlaneBB[0];
  Point2 vec12 = GroundPlaneBB[2] - GroundPlaneBB[1];
  Point2 vec23 = GroundPlaneBB[3] - GroundPlaneBB[2];
  Point2 vec30 = GroundPlaneBB[0] - GroundPlaneBB[3];

  Point2 pos01 = GroundPlaneBB[0] + 
    (((pnt - GroundPlaneBB[0]) * vec01) * (vec01 / vec01.length2()));
  Point2 pos12 = GroundPlaneBB[1] + 
    (((pnt - GroundPlaneBB[1]) * vec12) * (vec12 / vec12.length2()));
  Point2 pos23 = GroundPlaneBB[2] + 
    (((pnt - GroundPlaneBB[2]) * vec23) * (vec23 / vec23.length2()));
  Point2 pos30 = GroundPlaneBB[3] + 
    (((pnt - GroundPlaneBB[3]) * vec30) * (vec30 / vec30.length2()));

  Point2 tmp1, tmp2, res;
  if ((pnt - pos01).length2() < (pnt - pos12).length2())
    tmp1 = pos01;
  else tmp1 = pos12;
  if ((pnt - pos23).length2() < (pnt - pos30).length2())
    tmp2 = pos23;
  else tmp2 = pos30;
  if ((pnt - tmp1).length2() < (pnt - tmp2).length2())
    res = tmp1;
  else res = tmp2;

  return (res);
}

      
static inline bool line_intersect(Point2& p1, Point2& p2, 
				  Point2& q1, Point2& q2, Point2& res)
{
  // does line segment p1->p2 intersect line segment q1->q2
  Point2 vecp = p2 - p1;
  Point2 vecq = q2 - q1;

  Matrix2by2 M(-vecp.x, vecq.x, -vecp.y, vecq.y);
  if (fabs(M.determinant()) < 1e-30) return (FALSE);
  Point2 rvec = q2 - p2;
  Point2 scalars = M.inverse() * rvec;
  if (!((scalars.x >= -myEPS) && (scalars.x <= (1+myEPS)) &&
      (scalars.y >= -myEPS) && (scalars.y <= (1+myEPS))))
    return (FALSE);
  res = (scalars.x * p1) + (1.0 - scalars.x) * p2;
  return (TRUE);
}


static inline Point2 closest_point_to_line(Point2& l1, Point2& l2,
					   Point2& p)
{
  // return closest point to p on line l1->l2
  Point2 vec12 = l2 - l1;
  return (l1 + ((p - l1) * vec12) * (vec12 / vec12.length2()));
}

Point2 TrackedObject::constrain_movement(Point2 old_pos, Point2
					 new_pos)
{
  if (this->type == TrackedObject::PERSON) return (new_pos);
  for (int i = 0; i < 4; i++)
    {
      Point2 end1 = GroundPlaneBB[i];
      Point2 end2 = GroundPlaneBB[(i+1)%4];
      Point2 dummy;
      if (line_intersect(end1, end2, old_pos, new_pos, dummy))
	return (closest_point_to_line(end1, end2, new_pos)); 
    }
  return (new_pos);
}


