/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/***************************************************************
 * C++ source
 *
 * File : 	TrackedProfile.cc
 *
 * Module :	TrackedProfile
 *
 * Author : 	A M Baumberg (CoMIR)
 *
 * Creation Date : Fri Nov  3 15:20:19 1995 
 *
 * Comments : 	
 *
 ***************************************************************/


/* include files */

#include <GPfilter.h>
#include "TrackedProfile.h"
#include <base_pixel.h>
#include <RGBpixel.h>
#include <GrayPixel.h>
#include <AmbErrorHandler.h>

#define BLACK 0

TrackedProfile::TrackedProfile(TrackedProfile& original) :
Profile(original)
{
  filters = NULL; 
  time_alive = original.time_alive;
  no_shape_parameters = original.no_shape_parameters;
  if (original.filters != NULL)
    {
      filters = new FilterOneD*[no_shape_parameters];
      for (int i = 0; i < no_shape_parameters; i++)
	filters[i] = original.filters[i]->clone();
     
   }
  else filters = NULL;

  a_x = original.a_x;
  a_y = original.a_y;
  if (original.a_filter != NULL)
    {
      a_filter = new StaticKalmanTwoD();
      *a_filter = *original.a_filter;
    }
  else a_filter = NULL;

  if (original.s_filter != NULL)
    {
      s_filter = new StaticKalmanOneD();
      *s_filter = *original.s_filter;
    }
  else s_filter = NULL;

  if (original.th_filter != NULL)
    {
      th_filter = new StaticKalmanOneD();
      *th_filter = *original.th_filter;
    }
  else th_filter = NULL;

  if (original.pos_filter != NULL)
    {      
      pos_filter = original.pos_filter->clone();

    }
  else pos_filter = NULL;

  gp_filter = new GPfilter(*original.gp_filter, this);
  colour_info = original.colour_info;
}


void TrackedProfile::draw(bool flag)
{
  if (time_alive < 3)
    {
      int wdth = Profile::draw_linewidth;
      Profile::draw_linewidth = 3;
      if (flag) Profile::draw(ref_no, ref_no, ref_no);
      else Profile::draw(BLACK, BLACK, BLACK);
      Profile::draw_linewidth = wdth;
    }
  else
    {
      if (flag) Profile::draw(ref_no, ref_no, ref_no);
      else Profile::draw(BLACK, BLACK, BLACK);
    }
}


void TrackedProfile::get_enclosing_box(int& xmin, int& xmax,
				       int& ymin, int& ymax)
{
  get_box(xmin,xmax, ymin, ymax);
  int dx = 0.5 + sqrt(pos_filter->get_cov().val00);
  int dy = 0.5 + sqrt(pos_filter->get_cov().val11);
  
  int max_dx = 0.5 + (xmax - xmin) / 4;
  int max_dy = 0.5 + (ymax - ymin) / 4;

  if (dx > max_dx) dx = max_dx;
  if (dy > max_dy) dy = max_dy;

  xmin -= dx;
  xmax += dx;
  ymin -= dy;
  ymax += dy;

}  

void TrackedProfile::dump_reading_info(ostream& strm_out)
{
  strm_out << "label " << ref_no << endl; 
  strm_out << "wp_pos " << gp_filter->GP_position() << endl;
  strm_out << "wp_cov " << gp_filter->gp_pos_tracker->C00.val00 <<
    " " << gp_filter->gp_pos_tracker->C00.val01  <<
    " " <<  gp_filter->gp_pos_tracker->C00.val11  << endl;

  strm_out << "image_origin " << origin << endl;
  strm_out << "image_width " << width << endl;
  strm_out << "image_height " << height << endl;
  strm_out << "control_points ";
  for (int i = 0; i < NO_CONTROL_POINTS; i++)
    strm_out << spline[i] << " ";
  strm_out << endl;
}

void TrackedProfile::update_shape_filters()
{
  if ((this == NULL) || (filters == NULL))
    {
      AmbErrorHandler::handle_error("bad call to \
TrackedProfile::update_shape_filters");
      return;
    }
  for (int i = 0; i < no_shape_parameters; i++)
    filters[i]->update_state();

}

void TrackedProfile::output_to_stream(ostream& out_strm)
{
  // backwardly compatible stream output !
  if (ref_no != 0) out_strm << "label " << ref_no << endl; 
  out_strm << "origin " << origin << endl;
  out_strm << "width " << width << endl;
  out_strm << "height " << height << endl;

  if (direction != Point2(0,0))
    out_strm << "direction " << direction << endl;
  if (gp_filter != NULL)
    {
      out_strm << "wp_pos " << gp_filter->GP_position() << endl;
      out_strm << "wp_cov " << gp_filter->gp_pos_tracker->C00.val00 <<
	" " << gp_filter->gp_pos_tracker->C00.val01  <<
	" " <<  gp_filter->gp_pos_tracker->C00.val11  << endl;
    }
  out_strm << "shape_parameters";
  for (int i = 0; i < no_shape_parameters; i++)
    out_strm << "  " << filters[i]->get_state();
  out_strm << endl;

  if (Profile::output_points)
    {
      out_strm << "control_points ";
      for (int i = 0; i < NO_CONTROL_POINTS; i++)
	out_strm << spline[i] << " ";
    }
} 
