/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/***************************************************************
 * C++ source
 *
 * File : 	WorldOcclusionHandler.cc
 *
 * Module :	WorldOcclusionHandler
 *
 * Author : 	A M Baumberg (CoMIR)
 *
 * Creation Date : Fri Aug  9 15:59:13 1996 
 *
 * Comments : 	
 *
 ***************************************************************/


/* include files */

#include <GPfilter.h>
#include <TrackedProfile.h>

#include "WorldOcclusionHandler.h"
#include <EnvParameter.h>
#include <utils.h>
#include <ActiveModel.h>


static const unsigned char MUTUAL_OCCLUSION_CHAR = 255;

static const unsigned char FLAG_MOVING = 1;
static const unsigned char FLAG_FRONT = 2;
static const unsigned char FLAG_BEHIND = 0;
static const unsigned char FLAG_STATIC = 0;

bool WorldOcclusionHandler::quick_mode 
= AmbGlobals.register_bool("QUICK_OCCLUSION_REASONING",
			   TRUE, &WorldOcclusionHandler::quick_mode,
			   TRUE, "WorldOcclusionHandler",
			   "Draw people as rectangles for speed in \
occlusion reasoning");

	
WorldOcclusionHandler::WorldOcclusionHandler(ImageSource* input_images,
					     MotionDetector* mdetect, 
					     ProfileSet* curr_people,
					     ExternalData* cars)
{
  car_data = cars;

  video_images = input_images;
  movement_detector = mdetect;
  
  CameraView = backup_view = viz_image = NULL;
  tracked_people = NULL;
  curr_objects = new TrackedObjectList;
  prev_objects =  new TrackedObjectList;

  // stop_time is the expected time before a stationary 
  // object will be incorporated into the background
  stop_time = mdetect->run_length * mdetect->time_skip;
  stop_time *= 2;	// add safety margin

  sample_skip = mdetect->sample_skip;

  tracked_people = curr_people;
}


void WorldOcclusionHandler::clear_flags()
{
  for (int i = 0; i < MAX_OBJECT_ID; i++)
    flags[i] = FALSE;
}

void WorldOcclusionHandler::update_frame()
{
  prev_objects->destroy_all();
  delete (prev_objects);
  prev_objects = curr_objects;
  curr_objects = new TrackedObjectList;

  DB_PRINTF("Leeds frame %d \n", video_images->get_frame_no());

  
  int w = video_images->get_current()->get_width() / sample_skip;
  int h = video_images->get_current()->get_height() / sample_skip;

  if (CameraView == NULL)
    CameraView =  new Gray8Image(w,h);

  if (car_data != NULL)
    {
      car_data->get_next_frame();
      if (car_data->footprint_data != NULL)
	curr_objects->concat(car_data->footprint_data->duplicate());
      if (car_data->CamView != NULL)
	car_data->CamView->resample(sample_skip, sample_skip,
				    CameraView);
      else
	{      
	  CameraView->clear();
	  AmbErrorHandler::handle_warning("WorldOcclusionHandler using \
blank car image");
	}
    }
  else CameraView->clear();

  backup_view = CameraView->copy(backup_view);

}  
  
static void mask_flagged_objects(Image* mask, Image* all_img,
				 unsigned char* flags)
{
  unsigned char* pix1 = all_img->get_data();
  unsigned char* pix2 = mask->get_data();
  unsigned char* edat = mask->endofdata();
  while (pix2 < edat)
    {
      if (flags[*pix1++]) // mask this pixel ?
	*pix2 = 0;	  //  yes
      else *pix2 = 255;	  //  no
      pix2++;
    }
}  

void WorldOcclusionHandler::update_profiles()
{
  while (merge_tracks());

  ProfileSet& profiles = *tracked_people;
  ProfileSet LongProfiles;	// people that have been tracked for
				// more than 3 frames
  for (profiles.start(); profiles.current_ok(); profiles.forward())
    {
      if (profiles.get_current().get_type() == TRACKED_PROFILE)
	{
	  TrackedProfile* prf = (TrackedProfile*)
	    &profiles.get_current();
	  if ((prf->gp_filter != NULL) &&
	      (prf->gp_filter->gp_pos_tracker != NULL))
	    {
	      Point2& pos =
		prf->gp_filter->gp_pos_tracker->get_state();
	      curr_objects->constrain_position(pos);
	      char id_char = (prf->ref_no & 127) + 128;
	      curr_objects->add(new TrackedObject(pos, prf->ref_no, id_char));
	      if (prf->time_alive > 3)
		LongProfiles.add((Profile*) prf);
	    }
	}  
    }
  
  curr_objects->update_list(*prev_objects); 
  render_profiles(LongProfiles, quick_mode,  CameraView, sample_skip);
  LongProfiles.delete_all();	// delete list nodes
}

void WorldOcclusionHandler::setup_occlusion_map(TrackedProfile* currx)
{
  int curr_id = currx->ref_no;
  Point2 curr_pos = currx->gp_filter->GP_position();

  // setup flags for all objects that might occlude this object
  // flags[0] = FALSE;
  clear_flags();
  flags[MUTUAL_OCCLUSION_CHAR] = TRUE;

  for (curr_objects->start(); curr_objects->current_ok();
       curr_objects->forward())
    {
      TrackedObject& tmp_obj = curr_objects->get_current();
      flags[tmp_obj.id_char] = FALSE;
      if (tmp_obj.obj_id != curr_id)
	{
	  if (tmp_obj.type == TrackedObject::CAR)
	    {
	      // it's a car
	      if (tmp_obj.no_stationary_frames > stop_time)
		{
		  // it's part of the background
		  // and it occludes but doesn't enclose
		  if (tmp_obj.is_point_hidden(curr_pos))
		    flags[tmp_obj.id_char] = TRUE; 

		}
	      else
		{
		  // it's a moving silhouette
		  flags[tmp_obj.id_char] = TRUE;
		}
	    }
	  else
	    {
	      // it's another person
	      
	      // flag occlusion if object have been tracked
	      // for several frames
	      if (tmp_obj.no_active_frames > 3)
		flags[tmp_obj.id_char] = TRUE;	

	    }
	}
    }
}
      
void WorldOcclusionHandler::setup_motion_mask()
{


  ProfileSet LongProfiles;	// all profiles that have been alive
				// for > 3 frames
  ProfileSet ShortProfiles;	// ... the rest
  for (tracked_people->start(); tracked_people->current_ok();
       tracked_people->forward())
    {
      TrackedProfile* curr_prf = (TrackedProfile*)
	&tracked_people->get_current();
      if (curr_prf->time_alive > 3)
	LongProfiles.add((Profile*) curr_prf);
      else
	ShortProfiles.add((Profile*) curr_prf);
    }


  // copy over the backup CameraView
  backup_view->copy(CameraView);

  // build motion mask 

  // now use the backup image
  render_profiles(LongProfiles, TRUE, backup_view, sample_skip);
  render_profiles(ShortProfiles,TRUE, backup_view, sample_skip);


  int w = backup_view->get_width();
  int h = backup_view->get_height();
  
  if (movement_detector->tracked_mask == NULL)
    movement_detector->tracked_mask = new Gray8Image(w,h);
  
  // mask of already tracked motion to ignore
  Image* tmask = movement_detector->tracked_mask;
  

  clear_flags();

  // flag all the known motion
  // i.e. all PEOPLE and all cars that are not yet
  // incorporated into the background 

  for (curr_objects->start(); curr_objects->current_ok();
       curr_objects->forward())
    {
      TrackedObject& cobj = curr_objects->get_current();
      if (cobj.type == TrackedObject::CAR)
	{
	  if (cobj.no_stationary_frames < stop_time) 
	    flags[cobj.id_char] = TRUE;
	}
      else flags[cobj.id_char] = TRUE;
    }

  flags[MUTUAL_OCCLUSION_CHAR] = TRUE;

  mask_flagged_objects(tmask, backup_view, flags);

  // noew draw updated view for occlusion of new objects
  render_profiles(LongProfiles, quick_mode, CameraView, sample_skip);
  
  LongProfiles.delete_all();
  ShortProfiles.delete_all();


  return;

  // more accuracy -- use a mask in the background update
  // it's a bit slow for the moment

  if (movement_detector->background->motion_mask == NULL)
    movement_detector->background->set_motion_mask(new Gray8Image(w,h));

  // mask of pixels that should not be updated in background filter
  Image* mmask = movement_detector->background->motion_mask;

  clear_flags();

  // flag moving objects
  // all PEOPLE and all non-stationary cars

  for (curr_objects->start(); curr_objects->current_ok();
       curr_objects->forward())
    {
      TrackedObject& cobj = curr_objects->get_current();
      if (cobj.type == TrackedObject::CAR)
	{
	  if (cobj.is_stationary != TRUE) flags[cobj.id_char] = TRUE;
	}
      else flags[cobj.id_char] = TRUE;

    }
  flags[MUTUAL_OCCLUSION_CHAR] = TRUE;
  mask_flagged_objects(mmask, CameraView, flags);
  
  mmask->set_title("Background update mask");
  mmask->display();

}


static void draw_box_view(Image* view, int xmin, int xmax, int ymin,
			  int ymax, unsigned char* flags, int id)
{
  if (ymin < 0) ymin  = 0;
  if (ymax >= view->get_height()) ymax = view->get_height() - 1;
  if (xmin < 0) xmin = 0;
  if (xmax >= view->get_width()) xmax = view->get_width()  - 1;

  for (int y = ymin; y <= ymax; y++)
    {
      unsigned char *pnt1 = view->get_pixel(xmin, y);
      unsigned char *pnt2 = view->get_pixel(xmax, y);
      while (pnt1 <= pnt2)
	{
	  if ((flags[*pnt1] & FLAG_MOVING) != 0)
	    *pnt1 = MUTUAL_OCCLUSION_CHAR;
	  else if ((flags[*pnt1] & FLAG_FRONT) == 0)
	    *pnt1 = id;
	  pnt1++;
	}
    }
}


static void render_shape(Image* view, Image* pic,
			 unsigned char* flags,
			 int xoffset, int yoffset)
{
  for (int y = 0; y < pic->get_height(); y++)
    for (int x = 0; x < pic->get_width(); x++)
      {
	unsigned char* pnt1 = pic->get_pixel(x,y);
	if (*pnt1 != 0)
	  {
	    unsigned char* pnt2 = view->check_pixel(x+xoffset,
						    y+yoffset);
	    if (pnt2 != NULL)
	      {
		// two moving people mutually occlude
		if ((flags[*pnt2] & FLAG_MOVING) != 0)
		  *pnt2 = MUTUAL_OCCLUSION_CHAR;
		else if ((flags[*pnt2] & FLAG_FRONT) == 0)
		  *pnt2 = *pnt1;
	      }
	  }
      }
}

// render_profiles(profiles, use_boxes, theView, skip, viz_mode)
// profiles: the pedestrian outlines
// use_boxes: if (TRUE) then draw as rectangles
// theView: an image to draw into
// skip: spatial image resampling
// viz_mode: if (TRUE) then don't flag mutual occlusion
//		of two or more moving objects
void WorldOcclusionHandler::render_profiles(ProfileSet& profiles, 
					    bool use_boxes,
					    Image* theView,
					    int skip, bool viz_mode)
{  
  flags[MUTUAL_OCCLUSION_CHAR] = FLAG_MOVING + FLAG_FRONT;

  for (profiles.start(); profiles.current_ok(); profiles.forward())
    {
      if (profiles.get_current().get_type() == TRACKED_PROFILE)
	{
	  TrackedProfile* prf = (TrackedProfile*)
	    &profiles.get_current();
	  Point2 gp_pos = prf->gp_filter->GP_position();

	  int new_id = (prf->ref_no & 127) + 128;
	  int xmin,xmax, ymin, ymax;
	  prf->get_box(xmin, xmax, ymin, ymax);
	  xmin = (xmin / skip) - 1;
	  xmax = (xmax / skip) + 1;
	  ymin = (ymin / skip) - 1;
	  ymax = (ymax / skip) + 1;

	  for (curr_objects->start(); curr_objects->current_ok();
	       curr_objects->forward())
	    {
	      // flag every object that is behind the
	      // new person
	      TrackedObject& tobj = curr_objects->get_current();
	      if (tobj.is_point_hidden(gp_pos))
		flags[tobj.id_char] = FLAG_FRONT; 
	      else flags[tobj.id_char] = FLAG_BEHIND; 

	      if (!viz_mode)
		{
		  // flag objects as moving for mutual occlusion
		  if (tobj.type == TrackedObject::PERSON)
		    flags[tobj.id_char] += FLAG_MOVING;
		  else if (tobj.no_stationary_frames < stop_time)
		    flags[tobj.id_char] += FLAG_MOVING;
		}
	    }
	  flags[0] = FLAG_BEHIND;

	  // if resampling draw using rectangles
	  if ((use_boxes) || (skip != 1))
	    draw_box_view(theView, xmin, xmax, ymin, ymax,
			  flags, new_id);
	  else
	    {
	      int w1 = xmax - xmin + 11;
	      int h1 = ymax - ymin + 11;
	      Gray8Image* pic1 = new Gray8Image(w1,h1);
	      pic1->clear();
	      pic1->set_draw_colour(new_id);
	      prf->draw_in_image(pic1,0.5 * Point2(w1,h1));
	      pic1->set_line_width(1);
	      prf->draw_in_image(pic1,0.5 * Point2(w1,h1), FALSE);
	      render_shape(theView, pic1, flags, 
			   0.5 + prf->origin.x - (0.5 * w1),
			   0.5 + prf->origin.y - (0.5 * h1));
	      delete (pic1);
	    }	  
	}
    }
}


// merge tracks that are too similar
// return TRUE if two tracks were merged
bool WorldOcclusionHandler::merge_tracks()
{
  ProfileSet& profiles = *tracked_people;
  NagMatrix distance_table(profiles.no_items, profiles.no_items);
  Profile** object_table = new Profile*[profiles.no_items];

  int i = 0;
  int j = 0;
  for (profiles.start(); profiles.current_ok(); profiles.forward())
    {
      object_table[i] = &profiles.get_current();
      j = 0;
      for (item(Profile)* lp2 = profiles.first; lp2 != NULL;
	   lp2 = lp2->next)
	{
	  Profile* prf1 = &profiles.get_current();
	  Profile* prf2 = lp2->dat;
	  distance_table.set(i,j,1e30);
	  if (i != j)
	    {
	      if ((prf1->get_type() != TRACKED_PROFILE) ||
		  (prf2->get_type() != TRACKED_PROFILE)) continue;
	      
	      TrackedProfile* tprf1 = (TrackedProfile*) prf1;
	      TrackedProfile* tprf2 = (TrackedProfile*) prf2;
	      
	      DynamicKalmanTwoD* posf1 = tprf1->gp_filter->gp_pos_tracker;
	      DynamicKalmanTwoD* posf2 = tprf2->gp_filter->gp_pos_tracker;
	      
	      Point2 rel_pos = posf1->pos - posf2->pos;
	      Point2 rel_vel = posf1->velocity - posf2->velocity;
	      if ((tprf1->time_alive < 3) || (tprf2->time_alive < 3))
		rel_vel = Point2(0,0);
	      realno Mdist = rel_pos * (posf1->C00.inverse() * rel_pos) 
		+ rel_vel * (posf1->C11.inverse() * rel_vel);
	      distance_table.set(i,j, Mdist);
	    }
	  j++;
	}
      i++;
    }

  for (i = 0; i < profiles.no_items; i++)
    for (j = 0; j < profiles.no_items; j++)
      {
	if (distance_table.read(i,j) < (2.0 * 2.0))
	  {
	    TrackedProfile* tprf1 = (TrackedProfile*) object_table[i];
	    TrackedProfile* tprf2 = (TrackedProfile*) object_table[j];
	    if ((tprf1 == NULL) || (tprf2 == NULL))
	      continue;

	    DB_PRINTF("merging !\n");
	    int new_ref_no = tprf1->ref_no;
	    int new_time_alive = tprf1->time_alive;
	    if (tprf2->time_alive > new_time_alive)
	      {
		new_time_alive = tprf2->time_alive;
		new_ref_no = tprf2->ref_no;
	      }
	    if (tprf1->frame_error < tprf2->frame_error)
	      {
		profiles.destroy(profiles.find(tprf2));
		object_table[j] = NULL;
		tprf1->ref_no = new_ref_no;
		tprf2->time_alive = new_time_alive;
	      }
	    else
	      {
		profiles.destroy(profiles.find(tprf1));
		object_table[i] = NULL;
		tprf2->ref_no = new_ref_no;
		tprf2->time_alive = new_time_alive;
	      }
	  }
      }
  delete [] object_table;
  return (FALSE);
}

void WorldOcclusionHandler::constrain_position(TrackedProfile* currx)
{
  if (currx->gp_filter == NULL)
    {
      AmbErrorHandler::handle_warning("bad call to \
WorldOcclusionHandler::constrain_position");
      return;
    }

  Point2& pos = 
    currx->gp_filter->gp_pos_tracker->get_state();
  curr_objects->constrain_position(pos);
}

void WorldOcclusionHandler::constrain_motion(TrackedProfile* currx)
{
  // make sure you can't walk through cars
  
  // (i) find any cars that enclose the current position
  
  Point2& pos = currx->gp_filter->gp_pos_tracker->pos;
  Point2& old_pos = currx->gp_filter->gp_pos_tracker->old_pos;

  for (curr_objects->start(); curr_objects->current_ok();
       curr_objects->forward())
    if (curr_objects->get_current().encloses(pos))
      break;

  if (!curr_objects->current_ok()) return;

  // (ii) constrain movement

  AmbErrorHandler::handle_warning("WorldOcclusionHandler detected \
collision");

  pos = curr_objects->get_current().constrain_movement(old_pos, pos);
  currx->gp_filter->gp_pos_tracker->velocity = Point2(0,0);
}
	  
    
		    
bool WorldOcclusionHandler::is_off_image(int x, int y)
{
  return (video_images->get_current()->check_pixel(x,y) == NULL);
}


bool WorldOcclusionHandler::is_occlusion(int x, int y)
{
  if (this == NULL) return (FALSE);
  x /= sample_skip; y /= sample_skip;

  unsigned char* pnt = CameraView->check_pixel(x,y);
  if (pnt == NULL) return (FALSE);
  return (flags[*pnt]);
}

Image* WorldOcclusionHandler::setup_viz_image()
{
  ProfileSet LongProfiles;	// all profiles that have been alive
				// for a few frames
  for (tracked_people->start(); tracked_people->current_ok();
       tracked_people->forward())
    {
      TrackedProfile* curr_prf = (TrackedProfile*)
	&tracked_people->get_current();
      if (curr_prf->time_alive > 3)
	LongProfiles.add((Profile*) curr_prf);
    }

  if (viz_image == NULL)
    viz_image = new
      Gray8Image(video_images->get_current()->get_width(),
		 video_images->get_current()->get_height());

  if ((car_data != NULL) && (car_data->CamView != NULL))
    car_data->CamView->copy(viz_image);
  else viz_image->clear();

  render_profiles(LongProfiles, FALSE, viz_image, 1, TRUE);
  LongProfiles.delete_all();
  return (viz_image);
}

  
