/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/***************************************************************
 * C - C++ Header
 *
 * File : 	GlobalsManager.h
 *
 * Module :	GlobalsManager
 *
 * Author : 	A M Baumberg (CoMIR)
 *
 * Creation Date : Mon Oct 21 12:39:06 1996 
 *
 * Comments : 	keeps track of globals and initialises them
 *		using a parameter file or other interface
 *
 ***************************************************************/


#ifndef __GLOBALS_MANAGER_H__
#define __GLOBALS_MANAGER_H__

#include "boolean.h"
#include "realno.h"
#include <stream.h>

class GlobalsManager;

// a container for a single global
class GlobalVariable
{
protected:

  enum Type {REAL, INTEGER, STRING, BOOLEAN};
  Type type;

  // if fixed is TRUE then 
  // value should not be changed after initialisation
  bool fixed;	

  char *owner_class;	// the class that uses this global

public:

  char *tag;	// a unique identifier
  char *help_string;

  GlobalVariable(char* theTag, Type theType, bool isFixed,
		 char* theOwner, char* help)
  {
    tag = theTag; type = theType; fixed = isFixed;
    owner_class = theOwner; help_string = help;
  }

  virtual bool read_value(istream& ) {}
  virtual void output_value(ostream& ) {}
  virtual void full_printout(ostream& );

  virtual void* create_widget(void* parent)
  {}

  friend GlobalsManager;
};

class RealGlobalVariable : public GlobalVariable
{
protected:

  realno& value;
  bool constrained;	// is there a min or max
  realno min; // optional minimum allowed value
  realno max; // optional maximum allowed value;

public:

  RealGlobalVariable(char* theTag, bool isFixed,
		     char* theOwner, char* help, realno& theValue)
    : GlobalVariable(theTag, GlobalVariable::REAL, 
		     isFixed, theOwner, help),
      value(theValue)
  { constrained = FALSE; }

  RealGlobalVariable(char* theTag, bool isFixed,
		     char* theOwner, char* help, realno& theValue,
		     realno theMin, realno theMax)
    : GlobalVariable(theTag, GlobalVariable::REAL, isFixed, theOwner, help),
      value(theValue)
  { 
    min = theMin; max = theMax;
    constrained = TRUE;
  }

  bool read_value(istream& );
  realno& get_value() { return (value); }
  realno get_min() { return min; }
  realno get_max() { return max; }
  bool is_constrained() { return (constrained); }

  void output_value(ostream& );
  void full_printout(ostream& );
  void* create_widget(void* widget);

  friend GlobalsManager;

};

class IntGlobalVariable : public GlobalVariable
{
protected:

  int& value;
  bool constrained;	// is there a min or max
  int min; // optional minimum allowed value
  int max; // optional maximum allowed value;

public:

  IntGlobalVariable(char* theTag, bool isFixed,
		     char* theOwner, char* help, int& theValue)
    : GlobalVariable(theTag, GlobalVariable::INTEGER, 
		     isFixed, theOwner, help),
      value(theValue)
  { constrained = FALSE; }

  IntGlobalVariable(char* theTag, bool isFixed,
		     char* theOwner, char* help, int& theValue,
		     int theMin, int theMax)
    : GlobalVariable(theTag, GlobalVariable::INTEGER, 
		     isFixed, theOwner, help),
      value(theValue)
  { 
    min = theMin; max = theMax;
    constrained = TRUE;
  }

  bool read_value(istream& );
  void output_value(ostream& );
  void full_printout(ostream& );
  void* create_widget(void* widget);
  int& get_value() { return (value); }
  int get_min() { return (min); }
  int get_max() { return (max); }
  bool is_constrained() { return (constrained); }

  friend GlobalsManager;

};

class BoolGlobalVariable: public GlobalVariable
{
protected:

  bool& value;

public:

  BoolGlobalVariable(char* theTag, bool isFixed,
		     char* theOwner, char* help, bool& theValue)
    : GlobalVariable(theTag, GlobalVariable::BOOLEAN, 
		     isFixed, theOwner, help),
       value(theValue) {}

  bool read_value(istream& );
  bool& get_value() { return value; }

  void output_value(ostream& );
  void full_printout(ostream& );

  void* create_widget(void* parent);

  friend GlobalsManager;
};

class StringGlobalVariable: public GlobalVariable
{
protected:

  char*& value;
  bool is_filename;	// is it a file ? (for GUI)
  char** valid_strings; // if (non-NULL) contains array of possibles

public:

  StringGlobalVariable(char* theTag, bool isFixed,
		     char* theOwner, char* help, char*& theValue)
    : GlobalVariable(theTag, GlobalVariable::STRING, 
		     isFixed, theOwner, help),
       value(theValue) { is_filename = FALSE; }

  bool read_value(istream& );
  void output_value(ostream& );
  void full_printout(ostream& );
  void* create_widget(void* parent);
  char*& get_value() { return (value); }
  friend GlobalsManager;
};


class GlobalsManager
{
private:

  int add_global_widgets(void* parent, bool allow_fixed,
			 bool constrained, 
			 GlobalVariable::Type type,
			 char** owner_list = NULL,
			 char** tag_list = NULL);
			  
  void setup_flags();
  void *create_partial_interface(void*, bool, char** owners, char** tags );

public:
  
  GlobalVariable** table;
  int no_globals;
  int max_globals;
  bool *flags;


  ~GlobalsManager();

  // typical usage ...
  // extern int myGlobal;
  // int myGlobal = AmbGlobals.register_int(3, &myGlobal);
  // int& myGlobal2 = AmbGlobals.register_int(2);


  // registering a global returns a reference to the actual 
  // value (which may be changed later)

  realno& register_real(char* tag,
			realno default_value, realno* holder = NULL,
			bool fixed = TRUE,
			char* owner = NULL, char* help = NULL);

  realno& register_real(char* tag,
			realno default_value,
			realno min, realno max, realno* holder = NULL,
			bool fixed = TRUE,
			char* owner = NULL, char* help = NULL);

  int& register_int(char* tag,
		    int default_value, int* holder = NULL,
		    bool fixed = TRUE,
		    char* owner = NULL, char* help = NULL);

  int& register_int(char *tag,
		    int default_value,
		    int min, int max,  int* holder = NULL,
		    bool fixed = TRUE,
		    char* owner = NULL, char* help = NULL);


  bool& register_bool(char *tag,
		      bool default_value, bool* holder,
		      bool fixed = TRUE,
		      char* owner = NULL, char* help = NULL);
  
  
  char*& register_string(char* tag,
			 char *default_value, char** holder = NULL,
			 bool fixed = TRUE,
			 char* owner = NULL, char* help = NULL, 
			 char** valid_strings = NULL);
  
  char*& register_filename(char* tag,
			   char *default_value, char** holder = NULL,
			   char* owner = NULL, char* help = NULL);
  
  
  void set_value(char* tag, realno new_value);
  void set_value(char* tag, int new_value);
  void set_value(char* tag, char* new_value);

  void unmanage_global(char* tag);

  void register_global_variable(GlobalVariable* var);

  // read in global values from a file or stream
  void parse_parameter_stream(istream& input_stream);
  void parse_parameter_file(char* param_file);
  
  // save parameters in parseable format
  void dump_parameters(ostream& output_stream);
  void dump_parameters(char* param_file);
  
  // display all the info to stdout
  void output_info(); 


  // get_globals using a Motif User Interface 

  void create_interface(int argc, char** argv, char** owner_list =
			NULL, char** tag_list = NULL, bool
			use_constants = TRUE);

};

// the default globals for my code
extern GlobalsManager AmbGlobals;

#endif


