/*
This source file is the copyright property of the University of Leeds
('The University').

Permission to use, copy, modify, and distribute this source file for
educational, research, and not-for-profit purposes, without fee and
without a signed licensing agreement, is hereby granted, provided that
the above copyright notice, this paragraph and the following three
paragraphs appear in all copies, modifications, and distributions.

In no event shall The University be liable to any party for direct,
indirect, special, incidental or consequential damages, including lost
profits, arising out of the use of this software and its documentation.

The software is provided without warranty. The University has no
obligation to provide maintenance, support, updates, enhancements, or
modifications.

This software was written by Adam Baumberg, Vision Group,
School of Computer Studies, University of Leeds, UK. The code and use
thereof should be attributed to the author where appropriate
(including demonstrations which rely on it's use).
*/

/*
 *
 * list.h
 *
 * a general (2-way) linked list of some type
 * - defined using preprocessor directives to avoid
 * template problems
 *
 *
 */

#ifndef __LIST_H__
#define __LIST_H__

#include <stream.h>
#include <string.h>
#include "boolean.h"

#define name2(a,b)      _name2_aux(a,b)
#define _name2_aux(a,b)      a##b
#define list(type) name2(type,list)
#define item(type) name2(type,item)

//++
#define listdeclare(type)

class item(type)
{
 public:
  type* dat;
  item(type)* prev;
  item(type)* next;

  item(type) (type* d, item(type)* p = NULL, item(type)* n = NULL)
    { dat = d; prev = p; next = n; }

};

class list(type)
{

 public:
  item(type)* first;
  item(type)* last;
  int no_items;
  item(type)* current;

  list(type)() { first = last = current = NULL; no_items = 0; }


  list(type) duplicate();

  item(type)* operator[](int i);
  item(type)* find(type* item);

  int add(type* new_item);
  int add_a_copy(type* new_item);
  int add(type new_item) { return add_a_copy(&new_item);}
  type& add() { add(new type); return *(last->dat); }
  
  void concat(list(type)& l2);

  type* remove(item(type)* item);
  type* remove(type* item ) { return(remove (find(item))); }
  type* remove(int i) { return(remove ((*this)[i])); }
  type* pop() { return(remove(last)); }

  void destroy(item(type)* item);
  void destroy(type* item ) { destroy (find(item)); }
  void destroy(int i) { destroy ((*this)[i]); }

  void insert_before(item(type)* item, type* datum);
  void insert_at(int i, type* datum) { insert_before((*this)[i], datum); }
  void insert_before(type* item, type* datum)
    { insert_before(find(item), datum); }

  void delete_all();
  void destroy_all();

  
  void close();
  void open();

  item(type)* start() { return (current = first); }
  item(type)* forward() { return (current = current->next); }
  item(type)* back() { return (current = current->prev); }
  bool current_ok() { return (current != NULL); }

  type& get_current() { return *(current->dat); }
  type& get_first() { return *(first->dat); }
  type& get_last() { return *(last->dat); }

  type* to_vector();

};

ostream& operator<<(ostream&, list(type)&);
istream& operator>>(istream&, list(type)&);
//--


//++
#define listimplement(type)

list(type) list(type)::duplicate()
{
  list(type) list_copy;
  for (start();current_ok(); forward())
    list_copy.add(get_current());
  return (list_copy);
}

item(type)* list(type)::operator[](int i)
{
  if (i<0 || i >= no_items)
    {
      fprintf(stderr,  " item %d not in list \n", i);
	fflush(stderr);
      abort();
    }
  item(type)* current = first;
  while (i-- > 0  && current != NULL) current = current->next;
  return (current);
}

item(type)* list(type)::find(type* item_dat)
{
  item(type)* current = first;
  while (current != NULL && current->dat != item_dat) 
    current = current->next;
  return (current);
}

int list(type)::add(type* new_item_dat)
{
  if (no_items != 0)
    {
      item(type)*  new_item = new item(type) (new_item_dat, last, NULL);
      last->next = new_item;
      no_items++;
      last = new_item;
    }
  else
    {
      if (first != NULL)
	{
	  fprintf(stderr,  " error in list mechanism!! \n");
	fflush(stderr);
	  abort();
	}
      item(type)* new_item = new item(type) (new_item_dat, NULL, NULL);
      first = last = new_item;
      no_items = 1;
    }
  return (no_items);
}

int list(type)::add_a_copy(type* new_item_dat)
{
  type* copy_item_dat = new type;
  *copy_item_dat = *new_item_dat;
  return add(copy_item_dat);
}

void list(type)::concat(list(type)& l2)
{
  if (first == NULL)
    {
      first = l2.first;
      last = l2.last;
    }
  else
    {
      last->next = l2.first;
      if (l2.first != NULL)
	{
	  l2.first->prev = last;
	  last = l2.last;
	}
    }
  no_items += l2.no_items;

  l2.first = l2.last = NULL;
  l2.no_items = 0;
}

  
type* list(type)::remove(item(type)* item)
{
  if (item->prev != NULL) item->prev->next = item->next;
  if (item == first) first = item->next;
  if (item->next != NULL) item->next->prev = item->prev;
  if (item == last) last = item->prev;
  if ((item == last) && (item == first)) first = last = NULL;
  type* rdat = item->dat;
  delete item;
  no_items--;
  return (rdat);
}

void list(type)::destroy(item(type)* item)
{
  if (item->prev != NULL) item->prev->next = item->next;
  if (item == first) first = item->next;
  if (item->next != NULL) item->next->prev = item->prev;
  if (item == last) last = item->prev;
  if ((item == last) && (item == first)) first = last = NULL;
  delete (item->dat);
  delete item;
  no_items--;
}

void list(type)::insert_before(item(type)* item, type* datum)
{
  item(type)* new_item = new item(type) (datum, item->prev, item);
  if (item->prev != NULL) item->prev->next = new_item;
  if (item == first) first = new_item;
  item->prev = new_item;
  no_items++;
}

void list(type)::delete_all()
{
  while (no_items > 0) remove(last);
}

void list(type)::destroy_all()
{
  while (no_items > 0) destroy(last);
}

void list(type)::close()
{
  if (no_items > 0)
    {
      last->next = first;
      first->prev = last;
    }
}

void list(type)::open()
{
  if (no_items > 0)
    {
      last->next = first->prev = NULL;
    }
}

type* list(type)::to_vector()
{
  type* res = new type[no_items];
  item(type)* current = first;
  for (int j = 0; j < no_items; j++)
    {
      res[j] = *(current->dat);
      current = current->next;
    }
  return (res);
}

ostream& operator<<(ostream& out_strm, list(type)& l)
{
  item(type)* curr = l.first;
  out_strm << "list_length = " << l.no_items << "\n";
  for (int j = 0; j < l.no_items; j++)
    {
      out_strm << (*(curr->dat)) << endl;
      curr = curr->next;
    }
  return (out_strm);
}
      
istream& operator>>(istream& in_strm,list(type)& l)
{
  int nitems;
  char dummy [100];
  strcpy(dummy,"dummy");
  while (!(in_strm.eof()) && ((strlen(dummy) == 0) ||
	 (strncmp(dummy, "list_length",12) != 0))) in_strm >> dummy;
  if (!in_strm.eof())
    { 
      in_strm.ignore(100, '=');
      in_strm >> nitems;
      type* newdata;
      for (int j = 0; j < nitems; j++)
	{
	  newdata = new type();
	  in_strm >> (*newdata);
	  l.add(newdata);
	}
    }
  else 
    {
      l.first = l.last = NULL;
      l.no_items = 0;
    }
  return (in_strm);
}
//--

#endif


